c       Copyright (C) 2007 Patrice Koehl
c
c       This library is free software; you can redistribute it and/or
c       modify it under the terms of the GNU Lesser General Public
c       License as published by the Free Software Foundation; either
c       version 2.1 of the License, or (at your option) any later version.
c
c       This library is distributed in the hope that it will be useful,
c       but WITHOUT ANY WARRANTY; without even the implied warranty of
c       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
c       Lesser General Public License for more details.
c
c       You should have received a copy of the GNU Lesser General Public
c       License along with this library; if not, write to the Free Software
c       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
c
c ==========================================================================================
c ==========================================================================================
c	Cartesian_Grid_gpbl.f
c       (routines for discretization on structured grids, for 
c	GPBL solver, based on Michael Holst MG package)
c
c	Version 1:      7/4/07
c
c	Author: Patrice Koehl (in collaboration with Marc Delarue)
c
c	This file contains a series of subroutines for solving modified Poisson-Boltzmann
c	equations. It is inspired form Michael Holst MG code, which is available at:
c	http://www.fetk.org/codes/pmg/index.html
c
c
c ==========================================================================================
c ==========================================================================================
c 	Build_gpbl.f
c ==========================================================================================
c ==========================================================================================
c
c	This subroutine build the boundary arrays, and modify affine vectors.
c
c	if (ido=0)	do only fine level
c	if (ido=1)	do only coarse levels
c	if (ido=2)	do all levels
c	if (ido=3)	rebuild the second operator at the coarsest level
c
c       if(ijac=0)      build a 7-point stencil for fine level
c       if(ijac=1)      build a 27-point stencil for fine level
c
c	Note: the fine level must be built before any coarse levels
c
	subroutine build_gpbl(nx,ny,nz,nlevel,ipkey,ido,ijac,iz,
     1			ipc,rpc,
     2			cc,fc,xpos,ypos,zpos,gxcf,gycf,gzcf,ccf,fcf,
     4			grid_info)
c
c ==========================================================================================
c 	Declare all variables
c ==========================================================================================
c
c	Input variables
c ==========================================================================================
c
	integer	nx,ny,nz,nlevel,ido,ipkey,ijac
c
	integer	ipc(*),iz(50,*)
c
	real*8	rpc(*),cc(*),fc(*)
	real*8	ccf(*),fcf(*)
	real*8	gxcf(*),gycf(*),gzcf(*),xpos(*),ypos(*),zpos(*)
c
	integer*1	grid_info(*)
c
c	Local variables
c ==========================================================================================
c
	integer	nxx,nyy,nzz,nxold,nyold,nzold,numdia,key,k
	integer	level
c
c ==========================================================================================
c	Setup
c ==========================================================================================
c
	nxx	= nx
	nyy	= ny
	nzz	= nz
c
c ==========================================================================================
c	Build vectors on the finest level
c ==========================================================================================
c
	if(ido.eq.0.or.ido.eq.2) then
c
		level = 1
c
		call build_vect(nxx,nyy,nzz,ipkey,ido,ijac,numdia,
     1		ipc(iz(5,level)),rpc(iz(6,level)),
     2		cc(iz(1,level)),fc(iz(1,level)),xpos(iz(8,level)),
     3		ypos(iz(9,level)),zpos(iz(10,level)),gxcf(iz(2,level)),
     4		gycf(iz(3,level)),gzcf(iz(4,level)),ccf(iz(1,level)),
     5		fcf(iz(1,level)))
c
		iz(7,level+1) = iz(7,level) + numdia * nxx * nyy * nzz
c
	endif
c
c ==========================================================================================
c	Build the (nlevel-1) level operators
c ==========================================================================================
c
	if((ido.eq.1).or.(ido.eq.2).or.(ido.eq.3)) then
c
	    do 100 level = 2, nlevel
c
		nxold = nxx
		nyold = nyy
		nzold = nzz
c
		call make_coarse(1,nxold,nyold,nzold,nxx,nyy,nzz)
c
		if(ido.ne.3) then
c
			call gpbl_copy0(nxx,nyy,nzz,nxold,nyold,nzold,
     1			xpos(iz(8,level)),ypos(iz(9,level)),
     2			zpos(iz(10,level)),gxcf(iz(2,level)),
     3			gycf(iz(3,level)),gzcf(iz(4,level)),
     4			grid_info(iz(1,level)),ccf(iz(1,level)),
     6			fcf(iz(1,level)),
     7			xpos(iz(8,level-1)),ypos(iz(9,level-1)),
     8			zpos(iz(10,level-1)),gxcf(iz(2,level-1)),
     9			gycf(iz(3,level-1)),gzcf(iz(4,level-1)),
     S			grid_info(iz(1,level-1)),ccf(iz(1,level-1)),
     S			fcf(iz(1,level-1)))
c
			call build_vect(nxx,nyy,nzz,ipkey,ido,ijac,
     1			numdia,ipc(iz(5,level)),rpc(iz(6,level)),
     2			cc(iz(1,level)),fc(iz(1,level)),
     3			xpos(iz(8,level)),ypos(iz(9,level)),
     4			zpos(iz(10,level)),gxcf(iz(2,level)),
     5			gycf(iz(3,level)),gzcf(iz(4,level)),
     6			ccf(iz(1,level)),fcf(iz(1,level)))
c
		   	iz(7,level+1) = iz(7,level) + numdia*nxx*nyy*nzz
c
		endif
c
100	    continue
c
	endif
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c 	Build_vect.f
c ==========================================================================================
c ==========================================================================================
c
c	This subroutine creates the rhs and Helmholtz term of the PDE
c
	subroutine build_vect(nx,ny,nz,ipkey,ido,ijac,numdia,ipc,rpc,
     1		cc,fc,xf,yf,zf,gxcf,gycf,gzcf,ccf,fcf)
c
c ==========================================================================================
c 	Declare all variables
c ==========================================================================================
c
c	Input variables
c ==========================================================================================
c
	integer	nx,ny,nz,ipkey,numdia,ido,ijac
c
	integer	ipc(*)
c
	real*8	ccf(nx,ny,nz),fcf(nx,ny,nz)
	real*8	rpc(*)
	real*8	cc(nx,ny,nz),fc(nx,ny,nz)
	real*8	gxcf(ny,nz,*),gycf(nx,nz,*),gzcf(nx,ny,*)
	real*8	xf(*),yf(*),zf(*)
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k,ike,jke,kke
	integer	nxm1,nym1,nzm1
c
	real*8	hx,hy,hz
	real*8	hxm1,hym1,hzm1,coef_fc
	real*8	bc_cond_e,bc_cond_w,bc_cond_n,bc_cond_s
	real*8	bc_cond_u,bc_cond_d
	real*8	coef_oE,coef_oN,coef_uC
	real*8	coef_oEm1,coef_oNm1,coef_uCm1
	real*8	diag,a1cf,a2cf,a3cf
        real*8  eps_bulk
c
        common /bulk/   eps_bulk
c
c ==========================================================================================
c 	Setup
c ==========================================================================================
c
c
c	Save problem key
c
	ipc(10) = ipkey
c
c	Note number of nonzeros in this stencil
c
	ipc(12) = 1
	if(ido.eq.0) then
                if(ijac.eq.0) then
		        ipc(11) = 7
		        numdia  = 4
                else
		        ipc(11) = 27
		        numdia  = 14
                endif
	else
		ipc(11) = 27
		numdia = 14
	endif
c
c	Determine n and define number of mesh points
c
	nxm1 = nx -1
	nym1 = ny -1
	nzm1 = nz -1
c
c	Diagonal scale factor
c
	diag = 1.0d0
	a1cf = eps_bulk
	a2cf = eps_bulk
	a3cf = eps_bulk
c
c ==========================================================================================
c 	Build the operator
c ==========================================================================================
c
	do 300 k = 2,nz-1
c
	   hzm1 = zf(k) - zf(k-1)
	   hz   = zf(k+1) - zf(k)
c
	   do 200 j = 2,ny-1
c
		hym1 = yf(j) - yf(j-1)
		hy   = yf(j+1) - yf(j)
c
		do 100 i = 2,nx-1
c
			hxm1 = xf(i) - xf(i-1)
			hx   = xf(i+1) - xf(i)
c
c			Define coefficients
c                       =====================
c
			coef_oE  = diag *(hym1+hy)*(hzm1+hz)/(4.d0*hx)
			coef_oEm1= diag *(hym1+hy)*(hzm1+hz)/(4.d0*hxm1)
			coef_oN  = diag *(hxm1+hx)*(hzm1+hz)/(4.d0*hy)
			coef_oNm1= diag *(hxm1+hx)*(hzm1+hz)/(4.d0*hym1)
			coef_uC  = diag *(hxm1+hx)*(hym1+hy)/(4.d0*hz)
			coef_uCm1= diag *(hxm1+hx)*(hym1+hy)/(4.d0*hzm1)
			coef_fc  = diag *(hxm1+hx)*(hym1+hy)*(hzm1+hz)
     1					/8.d0
c
			fc(i,j,k) = coef_fc * fcf(i,j,k)
			cc(i,j,k) = coef_fc * ccf(i,j,k)
c
c			East neighbour
c                       =====================
c
			ike = min0(1,iabs(i-nxm1))
c
			bc_cond_e = (1-ike)*coef_oE*a1cf*
     1					gxcf(j,k,2)
			fc(i,j,k) = fc(i,j,k) + bc_cond_e
c
c			North neighbor
c                       =====================
c
			jke = min0(1,iabs(j-nym1))
			bc_cond_n = (1-jke)*coef_oN*a2cf*
     1					gycf(i,k,2)
			fc(i,j,k) = fc(i,j,k) + bc_cond_n
c
c			Up neighbour
c                       =====================
c
			kke = min0(1,iabs(k-nzm1))
			bc_cond_u = (1-kke)*coef_uC*a3cf*
     1					gzcf(i,j,2)
			fc(i,j,k) = fc(i,j,k) + bc_cond_u
c
c			West neighbour 
c                       =====================
c
			ike = min0(1,iabs(i-2))
			bc_cond_w = (1-ike)*coef_oEm1*a1cf*
     1					gxcf(j,k,1)
			fc(i,j,k) = fc(i,j,k) + bc_cond_w
c
			jke = min0(1,iabs(j-2))
			bc_cond_s = (1-jke)*coef_oNm1*a2cf*
     1					gycf(i,k,1)
			fc(i,j,k) = fc(i,j,k) + bc_cond_s
c
			kke = min0(1,iabs(k-2))
			bc_cond_d = (1-kke)*coef_uCm1*a3cf*
     1					gzcf(i,j,1)
			fc(i,j,k) = fc(i,j,k) + bc_cond_d
c
100		continue
c
200	   continue
c
300	continue
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c 	gpbl_copy0.f
c ==========================================================================================
c ==========================================================================================
c
c	This subroutine translates information from a fine grid to a coarser grid
c	It does a simple copy.
c	This is specialized for the GPBL equation
c
	subroutine gpbl_copy0(nxc,nyc,nzc,nxf,nyf,nzf,
     1			xc,yc,zc,gxc,gyc,gzc,a1c,ccc,fcc,
     2			xf,yf,zf,gxf,gyf,gzf,a1f,ccf,fcf)
c
c	Note:
c	Variables ending in "c" refer to the coarse grid, while variables ending in "f"
c	refer to the fine grid
c
c ==========================================================================================
c 	Declare all variables
c ==========================================================================================
c
c	Input variables
c ==========================================================================================
c
	integer	nxc,nyc,nzc,nxf,nyf,nzf
c
c	Coarse grid:
c
	integer*1	a1c(nxc,nyc,nzc)
c
	real*8	xc(nxc),yc(nyc),zc(nzc)
	real*8	gxc(nyc,nzc,*),gyc(nxc,nzc,*),gzc(nxc,nyc,*)
	real*8	ccc(nxc,nyc,nzc),fcc(nxc,nyc,nzc)
c
c	Fine grid:
c
	integer*1	a1f(nxf,nyf,nzf)
c
	real*8	xf(nxf),yf(nyf),zf(nzf)
	real*8	gxf(nyf,nzf,*),gyf(nxf,nzf,*),gzf(nxf,nyf,*)
	real*8	ccf(nxf,nyf,nzf),fcf(nxf,nyf,nzf)
c
c	Local variables
c ==========================================================================================
c
	integer	iadd,jadd,kadd,i,j,k,ii,jj,kk
c
c ==========================================================================================
c 	Compute coarse grid pde coefficients
c ==========================================================================================
c
c	How far to step in
c
	iadd	= (nxf-1)/(nxc-1)
	jadd	= (nyf-1)/(nyc-1)
	kadd	= (nzf-1)/(nzc-1)
c
	if(iadd.ne.2.or.jadd.ne.2.or.kadd.ne.2) then
		write(6,*) 'Problem with grid dimension in build_copy0'
		stop
	endif
c
	do 300 k = 1,nzc
c
		kk = 2*k -1
		zc(k) = zf(kk)
c
		do 200 j = 1,nyc
c
			jj = 2*j - 1
			yc(j) = yf(jj)
c
			do 100 i = 1,nxc
c
				ii = 2*i - 1
				xc(i) = xf(ii)
c
c ==========================================================================================
c	 Grid information    
c ==========================================================================================
c
                                call mvbits(a1f(ii,jj,kk),0,2,
     1                                  a1c(i,j,k),0)
                                call mvbits(a1f(min0(nxf,ii+1),jj,kk),
     1                                  0,2,a1c(i,j,k),2)
                                call mvbits(a1f(ii,min0(nyf,jj+1),kk),
     1                                  0,2,a1c(i,j,k),4)
                                call mvbits(a1f(ii,jj,min0(nzf,kk+1)),
     1                                  0,2,a1c(i,j,k),6)
c
100			continue
c
200		continue
c
300	continue
c
c ==========================================================================================
c	Copy Helmholtz term and source term (fixed charges): simple copy
c ==========================================================================================
c
	call extract(nxf,nyf,nzf,nxc,nyc,nzc,ccf,ccc)
	call extract(nxf,nyf,nzf,nxc,nyc,nzc,fcf,fcc)
c
c ==========================================================================================
c	Now copy boundary values
c ==========================================================================================
c
c	Planes i=1 and i=nx
c ==========================================================================================
c
	do 500 k = 1,nzc
c
		kk = 2*k-1
c
		do 400 j = 1,nyc
c
			jj = 2 * j - 1
c
			gxc(j,k,1) = gxf(jj,kk,1)
			gxc(j,k,2) = gxf(jj,kk,2)
c
400		continue
c
500	continue
c
c	Planes j=1 and j=ny
c ==========================================================================================
c
	do 700 k = 1,nzc
c
		kk = 2*k-1
c
		do 600 i = 1,nxc
c
			ii = 2 * i - 1
c
			gyc(i,k,1) = gyf(ii,kk,1)
			gyc(i,k,2) = gyf(ii,kk,2)
c
600		continue
c
700	continue
c
c	Planes k=1 and k=nz
c ==========================================================================================
c
	do 900 j = 1,nyc
c
		jj = 2*j-1
c
		do 800 i = 1,nxc
c
			ii = 2 * i - 1
c
			gzc(i,j,1) = gzf(ii,jj,1)
			gzc(i,j,2) = gzf(ii,jj,2)
c
800		continue
c
900	continue
c
c ==========================================================================================
c	Return to calling program
c ==========================================================================================
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c 	operatorJ_gpbl.f
c ==========================================================================================
c ==========================================================================================
c
c	This subroutine builds the Jacobian of the GPBL equation, on all grid levels
c
c	This version assumes that the Jacobian on the fine grid is based on a stencil
c	of size 7 (i.e. each row of the Jacobian has 7 non zero elements, of which
c	only 4 are stored as the matrix is symmetric). For any point on the mesh,
c	these 7 values corresponds to the point itself, and its N(orth), S(outh), E(ast), 
c	W(est), U(pper), L(ower) neighbours
c
	subroutine operatorJ_gpbl(nx,ny,nz,nlevel,ijac,iz,
     1			ipc,rpc,phi,jac,xpos,ypos,zpos,fc,cc,
     2			resid,grid_info,free_energy,iok,iflag,vect)
c
c ==========================================================================================
c 	Declare all variables
c ==========================================================================================
c
c	Input variables
c ==========================================================================================
c
	integer	nx,ny,nz,nlevel,ido,ipkey,iflag,ijac
        integer iok
c
	integer	ipc(*),iz(50,*)
c
        real*8  free_energy
	real*8	rpc(*)
	real*8	phi(*),jac(*)
	real*8	fc(*),cc(*),resid(*)
	real*8	xpos(*),ypos(*),zpos(*)
	real*8	vect(*)
c
	integer*1	grid_info(*)
c
c	Local variables
c ==========================================================================================
c
	integer	nxx,nyy,nzz,nxold,nyold,nzold,numdia,key,k
	integer	level
c
c ==========================================================================================
c	Setup
c ==========================================================================================
c
	nxx	= nx
	nyy	= ny
	nzz	= nz
c
c ==========================================================================================
c	Build Jacobian on the finest level
c ==========================================================================================
c
	level = 1
c
        if(ijac.eq.0) then
	        call buildJ_langevin7(phi(iz(1,level)),nxx,nyy,nzz,
     1	        jac(iz(7,level)),xpos(iz(8,level)),
     2	        ypos(iz(9,level)),zpos(iz(10,level)),
     3          grid_info(iz(1,level)),
     3	        fc(iz(1,level)),cc(iz(1,level)),resid,
     4          free_energy,iok,iflag,vect)
c
	        ipc(iz(5,level)+4) = 7
c
        else
	        call buildJ_langevin27(phi(iz(1,level)),nxx,nyy,nzz,
     1	        jac(iz(7,level)),xpos(iz(8,level)),
     2	        ypos(iz(9,level)),zpos(iz(10,level)),
     3          grid_info(iz(1,level)),
     3	        fc(iz(1,level)),cc(iz(1,level)),resid,
     4          free_energy,iok,iflag,vect)
c
	        ipc(iz(5,level)+4) = 27
c
        endif
c
c ==========================================================================================
c	Build the (nlevel-1) level Jacobians
c ==========================================================================================
c
	do 100 level = 2, nlevel
c
		nxold = nxx
		nyold = nyy
		nzold = nzz
c
		call make_coarse(1,nxold,nyold,nzold,nxx,nyy,nzz)
c
		if(level.eq.2.and.ijac.eq.0) then
			call build_G(nxold,nyold,nzold,nxx,nyy,nzz,7,
     1			jac(iz(7,level-1)),jac(iz(7,level)))
		else
			call build_G(nxold,nyold,nzold,nxx,nyy,nzz,27,
     1			jac(iz(7,level-1)),jac(iz(7,level)))
		endif
c
		ipc(iz(5,level)+4) = 27
c
100	continue
c
	return
	end
