c ==========================================================================================
c       Copyright (C) 2007 Patrice Koehl
c
c       This library is free software; you can redistribute it and/or
c       modify it under the terms of the GNU Lesser General Public
c       License as published by the Free Software Foundation; either
c       version 2.1 of the License, or (at your option) any later version.
c
c       This library is distributed in the hope that it will be useful,
c       but WITHOUT ANY WARRANTY; without even the implied warranty of
c       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
c       Lesser General Public License for more details.
c
c       You should have received a copy of the GNU Lesser General Public
c       License along with this library; if not, write to the Free Software
c       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
c
c ==========================================================================================
c ==========================================================================================
c	ion_density_pbe.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c       This subroutine generates the ion density map for PB equation
c       (it is the ion number density, expressed in mol/l)
c
c	Input:
c               U        : current electrostatic map
c		nx,ny,nz : size of the grid
c		grid     : info on regular grid and shifted grids
c
c	Output:
c               map     : ion density map
c               title   : title for file containing the map
c
        subroutine ion_density_pbe(u,nx,ny,nz,grid,map,title)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	include 'poisson.h'
c
	integer	nx,ny,nz
        integer i,j,k,i_ion
c
        real*8  exp_min,exp_max
        parameter (exp_min=-80.d0, exp_max = 80.d0)
c
        integer method,nlevel,nlin,equation,ndiscrete,ijac,nsmooth
        real*8  tol,coef,arg
c
        character       title*72
c
c	Information on grids
c ==========================================================================================
c
	integer*1       grid(nx,ny,nz)
c
c       electrostatic potential and ion map
c ==========================================================================================
c
        real*8  u(nx,ny,nz),map(nx,ny,nz)
c
c 	Parameter values
c ==========================================================================================
c
	integer	n_solv,n_ion,n_zz
c
	real*8  ion_conc(max_ion),ion_z(max_ion),ion_rad(max_ion)
	real*8  kappa(max_ion)
c
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
        common /pdesolver/      tol,method,nlevel,nlin,equation,
     1                          ndiscrete,ijac,nsmooth
c
c
c ==========================================================================================
c       Compute ion number density (in mol/l)
c ==========================================================================================
c
        do 400 k = 1,nz
                do 300 j = 1,ny
                        do 200 i = 1,nx
                                map(i,j,k) = 0.d0
                                if(ibits(grid(i,j,k),1,1).eq.0) goto 200
                                do 100 i_ion = 1,n_ion
                                        arg = -ion_z(i_ion)*u(i,j,k)
                                        if(arg.gt.exp_max) arg=exp_max
                                        if(arg.lt.exp_min) arg=exp_min
                                        coef = ion_conc(i_ion)
                                        if(equation.eq.2) then
                                                map(i,j,k)=map(i,j,k)+
     1                                          coef*exp(arg)
                                        elseif(equation.eq.1) then
                                                map(i,j,k)=map(i,j,k)+
     1                                          coef*(1.d0+arg)
                                        endif
100                             continue
200                     continue
300             continue
400     continue
c
        title='Ion number density (mol/l)'
c
        return
        end
c ==========================================================================================
c ==========================================================================================
c	charge_density_pbe.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c       This subroutine generates the ion density map for PB equation
c       (it is the ion number density, expressed in mol/l)
c
c	Input:
c               U        : current electrostatic map
c		nx,ny,nz : size of the grid
c		grid     : info on regular grid and shifted grids
c
c	Output:
c               map     : ion density map
c               title   : title for file containing the map
c
        subroutine charge_density_pbe(u,nx,ny,nz,grid,map,title)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	include 'poisson.h'
c
	integer	nx,ny,nz
        integer i,j,k,i_ion
c
        real*8  exp_min,exp_max
        parameter (exp_min=-80.d0, exp_max = 80.d0)
c
        integer method,nlevel,nlin,equation,ndiscrete,ijac,nsmooth
        real*8  tol,coef,arg
c
        character       title*72
c
c	Information on grids
c ==========================================================================================
c
	integer*1       grid(nx,ny,nz)
c
c       electrostatic potential and ion map
c ==========================================================================================
c
        real*8  u(nx,ny,nz),map(nx,ny,nz)
c
c 	Parameter values
c ==========================================================================================
c
	integer	n_solv,n_ion,n_zz
c
	real*8  ion_conc(max_ion),ion_z(max_ion),ion_rad(max_ion)
	real*8  kappa(max_ion)
c
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
        common /pdesolver/      tol,method,nlevel,nlin,equation,
     1                          ndiscrete,ijac,nsmooth
c
c
c ==========================================================================================
c       Compute charge density (in e.mol/l)
c ==========================================================================================
c
        do 400 k = 1,nz
                do 300 j = 1,ny
                        do 200 i = 1,nx
                                map(i,j,k) = 0.d0
                                if(ibits(grid(i,j,k),1,1).eq.0) goto 200
                                do 100 i_ion = 1,n_ion
                                        arg = -ion_z(i_ion)*u(i,j,k)
                                        if(arg.gt.exp_max) arg=exp_max
                                        if(arg.lt.exp_min) arg=exp_min
                                        coef = ion_conc(i_ion)*
     1                                          ion_z(i_ion)
                                        if(equation.eq.2) then
                                                map(i,j,k)=map(i,j,k)+
     1                                          coef*exp(arg)
                                        elseif(equation.eq.1) then
                                                map(i,j,k)=map(i,j,k)+
     1                                          coef*(1.d0+arg)
                                        endif
100                             continue
200                     continue
300             continue
400     continue
c
        title='Ion charge density (e.mol/l)'
c
        return
        end
c
c ==========================================================================================
c ==========================================================================================
c	ion_density_smpb.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c       This subroutine generates the ion density map for SMPB equation
c       (it is the ion number density, expressed in mol/l)
c
c	Input:
c               U        : current electrostatic map
c		nx,ny,nz : size of the grid
c		grid     : info on regular grid and shifted grids
c
c	Output:
c               map     : ion density map
c               title   : title for file containing the map
c
        subroutine ion_density_smpb(u,nx,ny,nz,grid,map,title)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	include 'poisson.h'
c
	integer	nx,ny,nz
        integer i,j,k,i_ion
c
        real*8  exp_min,exp_max
        parameter (exp_min=-80.d0, exp_max = 80.d0)
c
        real*8  coef,arg
c
        character       title*72
c
c	Information on grids
c ==========================================================================================
c
	integer*1       grid(nx,ny,nz)
c
c       electrostatic potential and ion map
c ==========================================================================================
c
        real*8  u(nx,ny,nz),map(nx,ny,nz)
c
c	information on ions
c===========================================================================================
c
	integer	n_ion,n_zz
c
	real*8	ratio,phi0,coef_Z0,coef_Z
	real*8	D,g,val1
	real*8	ion_conc(max_ion),ion_z(max_ion),kappa(max_ion)
	real*8	ion_rad(max_ion)
	real*8	ion_cvol(max_ion)
c
c ==========================================================================================
c == Common block (define ions)            =================================================
c ==========================================================================================
c
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
	common /smpb/		ion_cvol,ratio,phi0,coef_Z0
c
c
	do 400 k = 1,nz
		do 300 j = 1,ny
		    do 200 i = 1,nx
			map(i,j,k) = 0.d0
c
			if(ibits(grid(i,j,k),1,1).eq.0) goto 200
c
c		Add ion one (small size)
c
			arg = -ion_z(1)*u(i,j,k)
c
c			Make sure that the argument of the exponential function is not
c			too large (or too small), so that we do not have overflow (or
c			underflow)
c
			if(arg.gt.exp_max) arg = exp_max
			if(arg.lt.exp_min) arg = exp_min
c
			arg = exp(arg)
c
			val1 = (1.d0-phi0+ion_cvol(1)*arg)
			g = ion_conc(1)*arg
			d = val1
			coef_Z = coef_Z0/val1
			coef_Z = coef_Z**(ratio-1)
c
c		Add all other ions
c
			do 100 i_ion = 2,n_ion
c
				arg = -ion_z(i_ion)*u(i,j,k)
c
c				Make sure that the argument of the exponential function is not
c				too large (or too small), so that we do not have overflow (or
c				underflow)
c
				if(arg.gt.exp_max) arg=exp_max
				if(arg.lt.exp_min) arg=exp_min
c
				arg = exp(arg)
c
				g = g + coef_Z*ion_conc(i_ion)*arg
c
				d = d + coef_Z*ion_cvol(i_ion)*arg
c
100			continue
c
			map(i,j,k) = g/d
c
200		     continue
300		continue
400	continue
c
c
        title='Ion number density (mol/l)'
c
        return
        end
c
c ==========================================================================================
c ==========================================================================================
c	charge_density_smpb.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c       This subroutine generates the ion density map for SMPB equation
c       (it is the charge density, expressed in e.mol/l)
c
c	Input:
c               U        : current electrostatic map
c		nx,ny,nz : size of the grid
c		grid     : info on regular grid and shifted grids
c
c	Output:
c               map     : ion density map
c               title   : title for file containing the map
c
        subroutine charge_density_smpb(u,nx,ny,nz,grid,map,title)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	include 'poisson.h'
c
	integer	nx,ny,nz
        integer i,j,k,i_ion
c
        real*8  exp_min,exp_max
        parameter (exp_min=-80.d0, exp_max = 80.d0)
c
        real*8  coef,arg
c
        character       title*72
c
c	Information on grids
c ==========================================================================================
c
	integer*1       grid(nx,ny,nz)
c
c       electrostatic potential and ion map
c ==========================================================================================
c
        real*8  u(nx,ny,nz),map(nx,ny,nz)
c
c	information on ions
c===========================================================================================
c
	integer	n_ion,n_zz
c
	real*8	ratio,phi0,coef_Z0,coef_Z
	real*8	D,g,val1
	real*8	ion_conc(max_ion),ion_z(max_ion),kappa(max_ion)
	real*8	ion_rad(max_ion)
	real*8	ion_cvol(max_ion)
c
c ==========================================================================================
c == Common block (define ions)            =================================================
c ==========================================================================================
c
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
	common /smpb/		ion_cvol,ratio,phi0,coef_Z0
c
c
	do 400 k = 1,nz
		do 300 j = 1,ny
		    do 200 i = 1,nx
			map(i,j,k) = 0.d0
c
			if(ibits(grid(i,j,k),1,1).eq.0) goto 200
c
c		Add ion one (small size)
c
			arg = -ion_z(1)*u(i,j,k)
c
c			Make sure that the argument of the exponential function is not
c			too large (or too small), so that we do not have overflow (or
c			underflow)
c
			if(arg.gt.exp_max) arg = exp_max
			if(arg.lt.exp_min) arg = exp_min
c
			arg = exp(arg)
c
			val1 = (1.d0-phi0+ion_cvol(1)*arg)
			g = ion_conc(1)*ion_z(1)*arg
			d = val1
			coef_Z = coef_Z0/val1
			coef_Z = coef_Z**(ratio-1)
c
c		Add all other ions
c
			do 100 i_ion = 2,n_ion
c
				arg = -ion_z(i_ion)*u(i,j,k)
c
c				Make sure that the argument of the exponential function is not
c				too large (or too small), so that we do not have overflow (or
c				underflow)
c
				if(arg.gt.exp_max) arg=exp_max
				if(arg.lt.exp_min) arg=exp_min
c
				arg = exp(arg)
c
				g = g + coef_Z*ion_z(i_ion)*
     1                                  ion_conc(i_ion)*arg
c
				d = d + coef_Z*ion_cvol(i_ion)*arg
c
100			continue
c
			map(i,j,k) = g/d
c
200		     continue
300		continue
400	continue
c
c
        title='Ion charge density (e.mol/l)'
c
        return
        end
c
c ==========================================================================================
c ==========================================================================================
c	ion_density_gpbl.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c       This subroutine generates the ion density map for GPBL equation
c       (it is the ion number density, expressed in mol/l)
c
c	Input:
c               U        : current electrostatic map
c		nx,ny,nz : size of the grid
c		grid     : info on regular grid and shifted grids
c
c	Output:
c               map     : ion density map
c               title   : title for file containing the map
c
        subroutine ion_density_gpbl(u,x,y,z,nx,ny,nz,grid,map,title,
     1                          Enorm)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	include 'poisson.h'
c
	integer	nx,ny,nz
c
        real*8  Enorm(nx,ny)
c
        character       title*72
c
c	Information on grids
c ==========================================================================================
c
	real*8  x(nx),y(ny),z(nz)
c
	integer*1       grid(nx,ny,nz)
c
c       Electrostatic potential and map
c ==========================================================================================
c
        real*8  u(nx,ny,nz),map(nx,ny,nz)
c
c 	Parameter values
c ==========================================================================================
c
	integer	n_solv,n_ion,n_zz
c
        real*8  free_edens,free_charge,free_iondip,free_energy
	real*8	vol,p0_e,coef_l,dvol
	real*8	coef_Z0,phi0,ratio
c
	real*8  ion_cvol(max_ion),l_ion(max_ion)
	real*8	solv_conc(max_solv)
	real*8  solv_dipole(max_solv),solv_cvol(max_solv)
c
	real*8  ion_conc(max_ion),ion_z(max_ion),ion_rad(max_ion)
	real*8  kappa(max_ion)
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k,l,i_ion,info1,istep
c
        real*8  exp_min,exp_max
        parameter (exp_min = -80.d0, exp_max = 80.d0)
	real*8	sinhc_u,arg,u_val,F1_u,exp_mu
	real*8	d,g,dm1
	real*8	l_dip
c
c	Common blocks
c ==========================================================================================
c
	common /dipoles/	vol,p0_e,coef_l
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
	common /solvent/        solv_conc,solv_dipole,solv_cvol,n_solv
	common /smpb/           ion_cvol,ratio,phi0,coef_Z0
c
c ==========================================================================================
c	Define some parameters
c ==========================================================================================
c
	l_dip = solv_cvol(1)/(1-phi0)
	do 100 i = 1,n_ion
		l_ion(i) = ion_cvol(i)/(1-phi0)
100	continue
c
        call set_all_zeros(nx,ny,nz,map)
c
	do 500 k = 2,nz-1
c
           call plane_efield(u(1,1,k),u(1,1,k-1),
     1          u(1,1,k+1),nx,ny,x,y,z,k,Enorm)
c
	   do 400 j = 2,ny-1
c
		do 300 i = 2,nx-1
c
                        info1 = ibits(grid(i,j,k),1,1)
c
			if(info1.ne.0) then
c
                                u_val = p0_e*Enorm(i,j)
                                call compute_F1(u_val,F1_u,sinhc_u,
     1                          exp_mu)
c
				d = exp_mu
                                g = 0
				do 200 i_ion = 1,n_ion
					arg = -ion_z(i_ion)*u(i,j,k)
     1                                  -u_val
                                        if(arg.gt.exp_max) arg=exp_max
                                        if(arg.lt.exp_min) arg=exp_min
                                        arg = exp(arg)
					d=d+l_ion(i_ion)*arg
                                        g = g + ion_conc(i_ion)*arg
200				continue
				d = d + l_dip*sinhc_u
				dm1 = 1/d
c
                                map(i,j,k) =  g/d
c
			endif
c
300		continue
400	   continue
c
500	continue
c
c       Assume electrostatic potential and electric field at boundary is 0: 
c
        do 900 k = 1,nz,nz-1
                do 800 j = 1,ny
                        do 700 i = 1,nx
                                info1 = ibits(grid(i,j,k),1,1)
                                if(info1.ne.0) then
                                        sinhc_u = 1.d0
                                        d = 1 + l_dip*sinhc_u
                                        g = 0
                                        do 600 i_ion=1,n_ion
                                                d=d+l_ion(i_ion)
                                                g = g + 
     1                                          ion_conc(i_ion)*arg
600                                     continue
                                        dm1 = 1/d
                                        map(i,j,k)=g*dm1/(1-phi0)
                                endif
700                     continue
800             continue
900     continue
c
        do 1300 k = 2,nz-1
                do 1200 j = 1,ny
                        if(j.eq.1.or.j.eq.ny) then
                                istep = 1
                        else
                                istep = nx - 1
                        endif
                        do 1100 i = 1,nx,istep
                                info1 = ibits(grid(i,j,k),1,1)
                                if(info1.ne.0) then
                                        sinhc_u = 1.d0
                                        d = 1 + l_dip*sinhc_u
                                        g = 0
                                        do 1000 i_ion=1,n_ion
                                                d=d+l_ion(i_ion)
                                                g = g + 
     1                                          ion_conc(i_ion)*arg
1000                                    continue
                                        dm1 = 1/d
                                        map(i,j,k)=g*dm1/(1-phi0)
                                endif
1100                    continue
1200            continue
1300    continue
c
        title='Ion number density (mol/l)'
c
        return
        end
c ==========================================================================================
c ==========================================================================================
c	charge_density_gpbl.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c       This subroutine generates the charge density map for GPBL equation
c       (it is the electric charge density, expressed in e.mol/l)
c
c	Input:
c               U        : current electrostatic map
c		nx,ny,nz : size of the grid
c		grid     : info on regular grid and shifted grids
c
c	Output:
c               map     : ion density map
c               title   : title for file containing the map
c
        subroutine charge_density_gpbl(u,x,y,z,nx,ny,nz,grid,map,title,
     1                                  Enorm)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	include 'poisson.h'
c
	integer	nx,ny,nz
c
        real*8  Enorm(nx,ny)
c
        character       title*72
c
c       Electrostatic potential and map
c ==========================================================================================
c
        real*8  u(nx,ny,nz),map(nx,ny,nz)
c
c	Information on grids
c ==========================================================================================
c
	real*8  x(nx),y(ny),z(nz)
c
	integer*1       grid(nx,ny,nz)
c
c 	Parameter values
c ==========================================================================================
c
	integer	n_solv,n_ion,n_zz
c
        real*8  free_edens,free_charge,free_iondip,free_energy
	real*8	vol,p0_e,coef_l,dvol
	real*8	coef_Z0,phi0,ratio
c
	real*8  ion_cvol(max_ion),l_ion(max_ion)
	real*8	solv_conc(max_solv)
	real*8  solv_dipole(max_solv),solv_cvol(max_solv)
c
	real*8  ion_conc(max_ion),ion_z(max_ion),ion_rad(max_ion)
	real*8  kappa(max_ion)
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k,l,i_ion,info1,istep
c
        real*8  exp_min,exp_max
        parameter (exp_min = -80.d0, exp_max = 80.d0)
c
	real*8	sinhc_u,arg,u_val,F1_u,exp_mu
	real*8	d,g,dm1
	real*8	l_dip
c
c	Common blocks
c ==========================================================================================
c
	common /dipoles/	vol,p0_e,coef_l
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
	common /solvent/        solv_conc,solv_dipole,solv_cvol,n_solv
	common /smpb/           ion_cvol,ratio,phi0,coef_Z0
c
c ==========================================================================================
c	Define some parameters
c ==========================================================================================
c
	l_dip = solv_cvol(1)/(1-phi0)
	do 100 i = 1,n_ion
		l_ion(i) = ion_cvol(i)/(1-phi0)
100	continue
c
        call set_all_zeros(nx,ny,nz,map)
c
	do 500 k = 2,nz-1
c
           call plane_efield(u(1,1,k),u(1,1,k-1),
     1          u(1,1,k+1),nx,ny,x,y,z,k,Enorm)
c
	   do 400 j = 2,ny-1
c
		do 300 i = 2,nx-1
c
                        map(i,j,k) = 0
                        info1 = ibits(grid(i,j,k),1,1)
c
			if(info1.ne.0) then
c
                                u_val = p0_e*Enorm(i,j)
                                call compute_F1(u_val,F1_u,sinhc_u,
     1                          exp_mu)
c
				d = exp_mu
                                g = 0
				do 200 i_ion = 1,n_ion
					arg = -ion_z(i_ion)*u(i,j,k)
     1                                  -u_val
                                        if(arg.gt.exp_max) arg=exp_max
                                        if(arg.lt.exp_min) arg=exp_min
                                        arg = exp(arg)
					d=d+l_ion(i_ion)*arg
                                        g = g + ion_z(i_ion)*
     1                                          ion_conc(i_ion)*arg
200				continue
				d = d + l_dip*sinhc_u
				dm1 = 1/d
c
                                map(i,j,k) =  g*dm1/(1-phi0)
c
			endif
c
300		continue
400	   continue
c
500	continue
c
c       Assume electrostatic potential and electric field at boundary is 0: 
c
        do 900 k = 1,nz,nz-1
                do 800 j = 1,ny
                        do 700 i = 1,nx
                                info1 = ibits(grid(i,j,k),1,1)
                                if(info1.ne.0) then
                                        sinhc_u = 1.d0
                                        d = 1 + l_dip*sinhc_u
                                        g = 0
                                        do 600 i_ion=1,n_ion
                                                d=d+l_ion(i_ion)
                                                g = g + ion_z(i_ion)*
     1                                          ion_conc(i_ion)*arg
600                                     continue
                                        dm1 = 1/d
                                        map(i,j,k)=g*dm1/(1-phi0)
                                endif
700                     continue
800             continue
900     continue
c
        do 1300 k = 2,nz-1
                do 1200 j = 1,ny
                        if(j.eq.1.or.j.eq.ny) then
                                istep = 1
                        else
                                istep = nx - 1
                        endif
                        do 1100 i = 1,nx,istep
                                info1 = ibits(grid(i,j,k),1,1)
                                if(info1.ne.0) then
                                        sinhc_u = 1.d0
                                        d = 1 + l_dip*sinhc_u
                                        g = 0
                                        do 1000 i_ion=1,n_ion
                                                d=d+l_ion(i_ion)
                                                g = g + ion_z(i_ion)*
     1                                          ion_conc(i_ion)*arg
1000                                    continue
                                        dm1 = 1/d
                                        map(i,j,k)=g*dm1/(1-phi0)
                                endif
1100                    continue
1200            continue
1300    continue
c
        title='Ion charge density (e.mol/l)'
c
        return
        end
c ==========================================================================================
c ==========================================================================================
c	water_density_gpbl.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c       This subroutine generates the water density map for GPBL equation
c       (expressed in mol/l)
c
c	Input:
c               U        : current electrostatic map
c		nx,ny,nz : size of the grid
c		grid     : info on regular grid and shifted grids
c
c	Output:
c               map     : water density map
c               title   : title for file containing the map
c
        subroutine water_density_gpbl(u,x,y,z,nx,ny,nz,grid,map,title,
     1                                  Enorm,nwater)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	include 'poisson.h'
c
	integer	nx,ny,nz
c	
	real*8  nwater
        real*8  Enorm(nx,ny)
	real*8 dx,dy,dz,dr
c
        character       title*72
c
c       Electrostatic potential and map
c ==========================================================================================
c
        real*8  u(nx,ny,nz),map(nx,ny,nz)
c
c	Information on grids
c ==========================================================================================
c
	real*8 	x(nx),y(ny),z(nz)
c
	integer*1       grid(nx,ny,nz)
c
c 	Parameter values
c ==========================================================================================
c
	integer	n_solv,n_ion,n_zz
c
        real*8  free_edens,free_charge,free_iondip,free_energy
	real*8	vol,p0_e,coef_l,dvol
	real*8	coef_Z0,phi0,ratio
c
	real*8  ion_cvol(max_ion),l_ion(max_ion)
	real*8	solv_conc(max_solv)
	real*8  solv_dipole(max_solv),solv_cvol(max_solv)
c
	real*8  ion_conc(max_ion),ion_z(max_ion),ion_rad(max_ion)
	real*8  kappa(max_ion)
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k,l,i_ion,info1,info2,istep
c
	real*8	sinhc_u,arg,u_val,F1_u,exp_mu
	real*8	d,g,dm1
	real*8	l_dip
c
        real*8  exp_min,exp_max
        parameter (exp_min = -80.d0, exp_max = 80.d0)
c
c	Common blocks
c ==========================================================================================
c
	common /dipoles/	vol,p0_e,coef_l
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
	common /solvent/        solv_conc,solv_dipole,solv_cvol,n_solv
	common /smpb/           ion_cvol,ratio,phi0,coef_Z0
c
c ==========================================================================================
c	Define some parameters
c ==========================================================================================
c
	l_dip = solv_cvol(1)/(1-phi0)
	do 100 i = 1,n_ion
		l_ion(i) = ion_cvol(i)/(1-phi0)
100	continue
c
        call set_all_zeros(nx,ny,nz,map)
c
	do 500 k = 2,nz-1
c
           call plane_efield(u(1,1,k),u(1,1,k-1),
     1          u(1,1,k+1),nx,ny,x,y,z,k,Enorm)
c
	   do 400 j = 2,ny-1
c
		do 300 i = 2,nx-1
c
                        info1 = ibits(grid(i,j,k),0,1)
                        info2 = ibits(grid(i,j,k),1,1)
c
			if(info1.ne.0) then
c
                                u_val = p0_e*Enorm(i,j)
                                call compute_F1(u_val,F1_u,sinhc_u,
     1                          exp_mu)
c
				d = exp_mu
                                if(info2.ne.0) then
				   do 200 i_ion = 1,n_ion
					arg = -ion_z(i_ion)*u(i,j,k)
     1                                  -u_val
                                        if(arg.gt.exp_max) arg=exp_max
                                        if(arg.lt.exp_min) arg=exp_min
                                        arg = exp(arg)
					d=d+l_ion(i_ion)*arg
200				   continue
                                endif
				d = d + l_dip*sinhc_u
				dm1 = 1/d
c
                                map(i,j,k) =  sinhc_u*dm1
     1                                          /(1-phi0)
c
			endif
c
300		continue
400	   continue
c
500	continue
c
c       Assume electrostatic potential and electric field at boundary is 0: 
c
        do 900 k = 1,nz,nz-1
                do 800 j = 1,ny
                        do 700 i = 1,nx
                                info1 = ibits(grid(i,j,k),0,1)
                                info2 = ibits(grid(i,j,k),1,1)
                                if(info1.ne.0) then
                                        sinhc_u = 1.d0
                                        d = 1 + l_dip*sinhc_u
                                        if(info2.ne.0) then
                                                do 600 i_ion=1,n_ion
                                                        d=d+l_ion(i_ion)
600                                             continue
                                        endif
                                        dm1 = 1/d
c                                        map(i,j,k)=solv_conc(1)*sinhc_u
                                        map(i,j,k)=sinhc_u
     1                                  *dm1/(1-phi0)
                                endif
700                     continue
800             continue
900     continue
c
        do 1300 k = 2,nz-1
                do 1200 j = 1,ny
                        if(j.eq.1.or.j.eq.ny) then
                                istep = 1
                        else
                                istep = nx - 1
                        endif
                        do 1100 i = 1,nx,istep
                                info1 = ibits(grid(i,j,k),0,1)
                                info2 = ibits(grid(i,j,k),1,1)
                                if(info1.ne.0) then
                                        sinhc_u = 1.d0
                                        d = 1 + l_dip*sinhc_u
                                        if(info2.ne.0) then
                                                do 1000 i_ion=1,n_ion
                                                        d=d+l_ion(i_ion)
1000                                            continue
                                        endif
                                        dm1 = 1/d
c                                        map(i,j,k)=solv_conc(1)*sinhc_u
                                        map(i,j,k)=sinhc_u
     1                                  *dm1/(1-phi0)
                                endif
1100                    continue
1200            continue
1300    continue
                                
        title='Water density (mol/l)'
c	
c Now loop over the grid again to get the number of water molecules
c	
	dz = z(2)-z(1)
	dy = y(2)-y(1)
	dx = x(2)-x(1)
	dr = dx*dy*dz
	nwater = 0.d0
	do 1101 k = 1,nz
	  do 1102 j = 1,ny
	    do 1103 i = 1,nx
	      nwater = nwater + map(i,j,k)*dr
1103	    continue
1102	  continue
1101	continue
c	
	nwater = nwater - nx*ny*nz + (nx-1.d0)*(ny-1.d0)*(nz-1.d0)
c	
        return
        end
c ==========================================================================================
c ==========================================================================================
c	energy_density.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c       This subroutine generates the energy density on the map
c
c	Input:
c               U        : current electrostatic map
c		nx,ny,nz : size of the grid
c		grid     : info on regular grid and shifted grids
c
c	Output:
c               map     : ion density map
c               title   : title for file containing the map
c
        subroutine energy_density(u,x,y,z,nx,ny,nz,grid,map,title,
     1                          Enorm)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	integer	nx,ny,nz
        integer i,j,k
c
        real*8  eps,eps1,eps2
        real*8  Enorm(nx,ny)
c
        real*8  u(nx,ny,nz),map(nx,ny,nz)
c
        character       title*72
c
c	Information on grids
c ==========================================================================================
c
	real*8 	x(nx),y(ny),z(nz)
c
	integer*1       grid(nx,ny,nz)
c
        common /elec_const/     eps1,eps2
c
        call set_all_zeros(nx,ny,nz,map)
c
        do 300 k = 2,nz-1
                call plane_efield(u(1,1,k),u(1,1,k-1),
     1          u(1,1,k+1),nx,ny,x,y,z,k,Enorm)
c
                do 200 j = 2,ny-1
                        do 100 i = 2,nx-1
                                if(ibits(grid(i,j,k),0,1).eq.0) then
                                        eps = eps1
                                else
                                        eps = eps2
                                endif
                                map(i,j,k)=eps*Enorm(i,j)*Enorm(i,j)/2
100                     continue
200             continue
300     continue
c
        title='Electrostatics energy density ((kT/eA)^2)'
c
        return
        end
