c       Copyright (C) 2007 Patrice Koehl
c
c       This library is free software; you can redistribute it and/or
c       modify it under the terms of the GNU Lesser General Public
c       License as published by the Free Software Foundation; either
c       version 2.1 of the License, or (at your option) any later version.
c
c       This library is distributed in the hope that it will be useful,
c       but WITHOUT ANY WARRANTY; without even the implied warranty of
c       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
c       Lesser General Public License for more details.
c
c       You should have received a copy of the GNU Lesser General Public
c       License along with this library; if not, write to the Free Software
c       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
c
c ==========================================================================================
c ==========================================================================================
c	Langevin27.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This file contains a library of routines that are
c			useful for the Generalized PBL equation
c
c ==========================================================================================
c ==========================================================================================
c	BuildJ_langevin.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine is the driver to the subroutine that computes the 
c			Jacobian of the GPBL equation
c
c	Input:
c		phi      : the electrostatic potential map
c		nx,ny,nz : size of the grid
c		hx,hy,hz : coordinates on the grid
c
c	Output:
c		eps_X,eps_Y,eps_Z : the diagonal elements of the tensor in front of the Laplacian
c				    (for each grid point)
c
	subroutine buildJ_langevin27(phi,nx,ny,nz,jac,
     1			hx,hy,hz,grid_info,fc,cc,resid,
     2                  free_energy,iok,itype,Space)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
c
	integer	nx,ny,nz,iok,i1,i2,itype
c
        integer*1 grid_info(*)
c
        real*8  free_energy
	real*8	phi(*)
	real*8	fc(*),resid(*),cc(*)
	real*8	jac(nx*ny*nz,*)
	real*8	hx(*),hy(*),hz(*)
c
	real*8	Space(*)
c
        i1 = 1
        i2 = i1 + nx*ny
c
	call compute_J_langevin27(phi,nx,ny,nz,
     1			hx,hy,hz,grid_info,
     2			jac(1,1),jac(1,2),jac(1,3),jac(1,4),
     3                  jac(1,5),jac(1,6),jac(1,7),jac(1,8),jac(1,9),
     4                  jac(1,10),jac(1,11),jac(1,12),jac(1,13),
     5                  jac(1,14),
     6			fc,cc,resid,free_energy,iok,itype,Space(i1),
     7                  Space(i2))
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	Compute_J_langevin27.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine computes the position specific epsilon for
c			the GPBL equation
c
c	This is the standard version, that deals with any number of ions (up to 10)
c	and any valences for these ions
c
c	Input:
c		phi      : the electrostatic potential map
c		nx,ny,nz : size of the grid
c		x,y,z    : coordinates on the grid
c		grid_info: info on shifted grids
c
c	Output:
c		jac	: the Jacobian of the GPBL equation
c		resid	: residual vector
c
	subroutine compute_J_langevin27(phi,nx,ny,nz,
     1			x,y,z,grid_info,
     2			JoC,JoE,JoN,JuC,
     3                  JoNE,JoNW,JuE,JuW,JuN,JuS,JuNE,JuNW,JuSE,JuSW,
     4			fc,cc,resid,free_energy,iok,itype,Enorm,der)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	include 'poisson.h'
c
	integer	nx,ny,nz
	integer	info1,iok,itype
c
        real*8  free_energy
        real*8  free_edens,free_charge,free_iondip
	real*8	eps_X,eps_Y,eps_Z
        real*8  der_X(10),der_Y(10),der_Z(10)
	real*8	coef_oE,coef_oN,coef_uC
        real*8  phi_i,phi_j,phi_k
	real*8	hxm1,hym1,hzm1
	real*8	hx,hy,hz
        real*8  val4,val5,val6,val7,val8,val9,val10
        real*8  val11,val12,val13,val14
        real*8  xnorm1
c
	real*8	phi(nx,ny,nz)
	real*8	fc(nx,ny,nz),resid(nx,ny,nz)
	real*8	cc(nx,ny,nz)
c
c	Jacobian for GPBL equation
c ==========================================================================================
c
	real*8	JoC(nx,ny,nz),JoE(nx,ny,nz),JoN(nx,ny,nz)
	real*8	JuC(nx,ny,nz),JoNE(nx,ny,nz),JoNW(nx,ny,nz)
	real*8	JuE(nx,ny,nz),JuW(nx,ny,nz),JuN(nx,ny,nz)
	real*8	JuS(nx,ny,nz),JuNE(nx,ny,nz),JuNW(nx,ny,nz)
	real*8	JuSE(nx,ny,nz),JuSW(nx,ny,nz)
c
	real*8	Enorm(nx,ny)
        real*8  der(nx,ny,*)
c
c	Information on grids
c ==========================================================================================
c
	real*8  	x(nx),y(ny),z(nz)
c
	integer*1       grid_info(nx,ny,nz)
c
c 	Parameter values
c ==========================================================================================
c
	integer	n_solv,n_ion
c
	real*8	vol,p0_e,coef_l
	real*8	coef_Z0,phi0,ratio
	real*8  l_Bjerrum,pi,Temp,KB,electron,Ang,beta,Avogadro,eps0
	real*8  C_charge,C_prop,fac_edens
c
	real*8  ion_cvol(max_ion),l_ion(max_ion),k_ion(max_ion)
	real*8	solv_conc(max_solv)
	real*8  solv_dipole(max_solv),solv_cvol(max_solv)
c
	real*8  ion_conc(max_ion),ion_z(max_ion),ion_rad(max_ion)
	real*8  kappa(max_ion)
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k,l,i_ion,n_zz
        integer ike,jke,kke,ikem1,jkem1
c
	real*8	u_val,F1_u,sinhc_u,G1_u,val,exp_mu
	real*8	val1,val2,val3
	real*8	coef,phi_val,d,dm1,dm2,der_d
        real*8  zp_max,zn_max,exp_arg,arg_max
	real*8	der_d2,g,der_g
	real*8	arg,exp_max,exp_min
	real*8	l_dip,dvol
        real*8  eps1,eps2
c
	parameter       (exp_min = -60.d0, exp_max = 60.d0)
c
c	Common blocks
c ==========================================================================================
c
	common /dipoles/	vol,p0_e,coef_l
	common /solvent/        solv_conc,solv_dipole,solv_cvol,n_solv
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
	common /smpb/           ion_cvol,ratio,phi0,coef_Z0
	common /constants/ 	l_Bjerrum,pi,Temp,KB,electron,Ang,
     1                     	Avogadro,eps0,beta,C_charge,C_prop
        common /elec_const/     eps1,eps2
        common /ion_max/        zp_max,zn_max
 
c
c ==========================================================================================
c	Initialise
c ==========================================================================================
c
c	Make sure all arrays are initialized to 0
c ==========================================================================================
c
	call set_all_zeros(nx,ny,nz,resid)
c
	call set_all_zeros(nx,ny,nz,JoC)
	call set_all_zeros(nx,ny,nz,JoE)
	call set_all_zeros(nx,ny,nz,JoN)
	call set_all_zeros(nx,ny,nz,JuC)
	call set_all_zeros(nx,ny,nz,JoNE)
	call set_all_zeros(nx,ny,nz,JoNW)
	call set_all_zeros(nx,ny,nz,JuE)
	call set_all_zeros(nx,ny,nz,JuW)
	call set_all_zeros(nx,ny,nz,JuN)
	call set_all_zeros(nx,ny,nz,JuS)
	call set_all_zeros(nx,ny,nz,JuNE)
	call set_all_zeros(nx,ny,nz,JuNW)
	call set_all_zeros(nx,ny,nz,JuSE)
	call set_all_zeros(nx,ny,nz,JuSW)
c
c ==========================================================================================
c	Define some parameters
c ==========================================================================================
c
	coef = coef_l * solv_cvol(1)/(1-phi0)
	l_dip = solv_cvol(1)/(1-phi0)
c
	do 100 i = 1,n_ion
		l_ion(i) = ion_cvol(i)/(1-phi0)
		k_ion(i) = kappa(i)/(1-phi0)
100	continue
c
	fac_edens = -(1d-10/(beta*electron*electron))*eps0/2
c
c ==========================================================================================
c	Compute eps for the X-shifted map
c ==========================================================================================
c
c	Compute electric field on x-shifted grid
c
	do 600 k = 2,nz-1
c
	   hzm1 = z(k)-z(k-1)
	   hz = z(k+1)-z(k)
c
	   call plane_efield_X27(phi(1,1,k),phi(1,1,k-1),phi(1,1,k+1),
     1		nx,ny,x,y,z,k,Enorm,der)
c
	   do 500 j = 2,ny-1
c
		hym1 = y(j) - y(j-1)
		hy = y(j+1)-y(j)
c
		do 450 i = 1,nx-1
c
			hx = x(i+1)-x(i)
c
                        coef_oE  = (hym1+hy)*(hzm1+hz)/(4.d0*hx)
c
			info1 = ibits(grid_info(i,j,k),2,2)
c
			if(info1.eq.0) then
c
				eps_X = eps1
                                do 150 l = 1,10
                                        der_X(l) = 0
150                             continue
c
			else
c
				u_val = p0_e*Enorm(i,j)
				call compute_F1G1(u_val,F1_u,G1_u,
     1				sinhc_u,exp_mu)
c
				val = coef*F1_u
				phi_val = (phi(i,j,k)+phi(i+1,j,k))/2
c
				d = exp_mu
				der_d = 0
c
				if(info1.ne.1) then
c
				    do 200 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val -u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d+l_ion(i_ion)*arg
					der_d = der_d-0.5d0*ion_z(i_ion)
     1					*l_ion(i_ion)*arg
200				    continue
c
				endif
c
c                                der_d = 0
				d = d + l_dip*sinhc_u
                                dm1 = 1/d
c
				eps_X = eps2 + val*dm1
c
                                do 250 l = 1,7
                                  der_X(l)=coef*der(i,j,l)*G1_u*dm1
     1                                  *p0_e*p0_e
     2                            -val*dm1*dm1*l_dip*der(i,j,l)*F1_u
     3                                  *p0_e*p0_e
250                             continue
                                der_X(1)=der_X(1)-val*dm1*dm1*der_d
                                der_X(2)=der_X(2)-val*dm1*dm1*der_d
c
                                if(itype.eq.0) then
                                        do 275 l = 1,7
                                                der_X(l)=0.d0
275                                     continue
                                endif
c
			endif
c
                        ike = min0(1,iabs(i-nx+1))
                        jke = min0(1,iabs(j-ny+1))
                        kke = min0(1,iabs(k-nz+1))
                        ikem1 = min0(1,iabs(i-1))
c
                        phi_i = -coef_oE*(phi(i,j,k) - phi(i+1,j,k))
c
                        if(i.ne.1) then
c
                                JoC(i,j,k) = JoC(i,j,k) + coef_oE*eps_X-
     2                          der_X(1)*phi_i
c
                                JoE(i,j,k)=JoE(i,j,k)+ike*
     1                          (coef_oE*eps_X+ der_X(2)*phi_i)
c
                                JoN(i,j,k)=JoN(i,j,k) + jke*(
     1                          der_X(3)*phi_i)
c
                                JuC(i,j,k)=JuC(i,j,k) + kke*(
     1                          der_X(4)*phi_i)
c
                                JoNE(i,j,k)=JoNE(i,j,k) + ike*jke*(
     1                          der_X(5)*phi_i)
c
                                JuE(i,j,k)=JuE(i,j,k) + ike*kke*(
     1                          der_X(7)*phi_i)
c
                        endif
c
                        if(ike.ne.0) then
c
                                JoC(i+1,j,k)=JoC(i+1,j,k)+coef_oE*eps_X
     1                          +der_X(2)*phi_i
c
                                JoN(i+1,j,k)=JoN(i+1,j,k)-
     1                          jke*der_X(5)*phi_i
c
                                JuC(i+1,j,k)=JuC(i+1,j,k)-kke*
     1                          der_X(7)*phi_i
c
                                JoNW(i+1,j,k)=JoNW(i+1,j,k)-ikem1*jke*
     1                          der_X(3)*phi_i
c
                                JuW(i+1,j,k)=JuW(i+1,j,k)-ikem1*kke*
     1                          der_X(4)*phi_i
c
                        endif
c
                        if(i.ne.1) then
                                resid(i,j,k)=resid(i,j,k)-eps_X*phi_i
                        endif
c
                        if(ike.ne.0) then
                                resid(i+1,j,k)=resid(i+1,j,k)+
     1                          eps_X*phi_i
                        endif
c
450		continue
500	   continue
c
600	continue
c
c ==========================================================================================
c	Repeat for the Y-shifted map
c ==========================================================================================
c
c	Compute electric field on y-shifted grid
c
	do 1050 k = 2,nz-1
c
	   hzm1 = z(k)-z(k-1)
	   hz = z(k+1)-z(k)
c
	   call plane_efield_Y27(phi(1,1,k),phi(1,1,k-1),phi(1,1,k+1),
     1		nx,ny,x,y,z,k,Enorm,der)
c
	   do 1000 j = 1,ny-1
c
		hy = y(j+1)-y(j)
c
		do 950 i = 2,nx-1
c
                        hxm1 = x(i) - x(i-1)
			hx = x(i+1)-x(i)
c
                        coef_oN  = (hxm1+hx)*(hzm1+hz)/(4.d0*hy)
c
			info1 = ibits(grid_info(i,j,k),4,2)
c
			if(info1.eq.0) then
c
				eps_Y = eps1
                                do 650 l = 1,10
                                        der_Y(l) = 0
650                             continue
c
			else
c
				u_val = p0_e*Enorm(i,j)
				call compute_F1G1(u_val,F1_u,G1_u,
     1				sinhc_u,exp_mu)
c
				val = coef*F1_u
				phi_val = (phi(i,j,k)+phi(i,j+1,k))/2
c
				d = exp_mu
				der_d = 0
c
				if(info1.ne.1) then
c
				    do 700 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val-u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d+l_ion(i_ion)*arg
					der_d = der_d-0.5d0*ion_z(i_ion)
     1					*l_ion(i_ion)*arg
700				    continue
c
				endif
c
c                                der_d = 0
				d = d + l_dip*sinhc_u
                                dm1 = 1/d
c
				eps_Y = eps2 + val*dm1
c
                                do 750 l = 1,10
                                  der_Y(l)=coef*der(i,j,l)*G1_u*dm1
     1                                  *p0_e*p0_e
     2                             -val*dm1*dm1*l_dip*der(i,j,l)*F1_u
     3                                  *p0_e*p0_e
750                             continue
                                der_Y(1)=der_Y(1)-val*dm1*dm1*der_d
                                der_Y(3)=der_Y(3)-val*dm1*dm1*der_d
c
                                if(itype.eq.0) then
                                        do 775 l = 1,10
                                                der_Y(l)=0.d0
775                                     continue
                                endif
c
c
			endif
c
                        ike = min0(1,iabs(i-nx+1))
                        jke = min0(1,iabs(j-ny+1))
                        kke = min0(1,iabs(k-nz+1))
                        ikem1 = min0(1,iabs(i-2))
                        jkem1 = min0(1,iabs(j-1))
c
                        phi_j = -coef_oN*(phi(i,j,k) - phi(i,j+1,k))
c
                        if(j.ne.1) then
c
                                JoC(i,j,k) = JoC(i,j,k) +coef_oN*eps_Y -
     2                          der_Y(1)*phi_j
c
                                JoE(i,j,k)=JoE(i,j,k) + ike*
     1                          der_Y(2)*phi_j
c
                                JoN(i,j,k)=JoN(i,j,k) + jke*
     1                          (coef_oN*eps_Y+der_Y(3)*phi_j)
c
                                JuC(i,j,k)=JuC(i,j,k) + kke*
     1                          der_Y(4)*phi_j
c
                                JoNE(i,j,k)=JoNE(i,j,k) + ike*jke*
     1                          der_Y(5)*phi_j
c
                                JoNW(i,j,k)=JoNW(i,j,k) + ikem1*jke*
     1                          der_Y(6)*phi_j
c
                                JuN(i,j,k)=JuN(i,j,k) + jke*kke*
     1                          der_Y(9)*phi_j
c
                        endif
c
                        if(jke.ne.0) then
c
                                JoC(i,j+1,k)=JoC(i,j+1,k) +coef_oN*eps_Y
     2                          +der_Y(3)*phi_j
c
                                JoE(i,j+1,k)=JoE(i,j+1,k) - ike*
     1                          der_Y(5)*phi_j
c
                                JuC(i,j+1,k)=JuC(i,j+1,k) - kke*
     1                          der_Y(9)*phi_j
c
                                JuS(i,j+1,k)=JuS(i,j+1,k) - jkem1*kke*
     1                          der_Y(4)*phi_j
c
                        endif
c
                        if(j.ne.1) then
                                resid(i,j,k)=resid(i,j,k)-eps_Y*phi_j
                        endif
c
                        if(jke.ne.0) then
                                resid(i,j+1,k)=resid(i,j+1,k)+
     1                          eps_Y*phi_j
                        endif
c
950		continue
1000	   continue
c
1050	continue
c
c ==========================================================================================
c	Repeat for the Z-shifted map
c ==========================================================================================
c
c	Compute electric field on z-shifted grid
c
	do 1500 k = 1,nz-1
c
	   hz = z(k+1)-z(k)
c
	   call plane_efield_Z27(phi(1,1,k),phi(1,1,k+1),
     1		nx,ny,x,y,z,k,Enorm,der)
c
	   do 1450 j = 2,ny-1
c
		hym1 = y(j) - y(j-1)
		hy = y(j+1)-y(j)
c
		do 1400 i = 2,nx-1
c
                        hxm1 = x(i) - x(i-1)
			hx = x(i+1)-x(i)
c
                        coef_uC  = (hxm1+hx)*(hym1+hy)/(4.d0*hz)
c
			info1 = ibits(grid_info(i,j,k),6,2)
c
			if(info1.eq.0) then
c
				eps_Z = eps1
                                do 1100 l = 1,10
                                        der_Z(l) = 0
1100                            continue
c
			else
c
				u_val = p0_e*Enorm(i,j)
				call compute_F1G1(u_val,F1_u,G1_u,
     1				sinhc_u,exp_mu)
c
				val = coef*F1_u
				phi_val = (phi(i,j,k)+phi(i,j,k+1))/2
c
				d = exp_mu
				der_d = 0
c
				if(info1.ne.1) then
c
				    do 1150 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val-u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d+l_ion(i_ion)*arg
					der_d = der_d-0.5d0*ion_z(i_ion)
     1					*l_ion(i_ion)*arg
1150				   continue
c
				endif
c
c                                der_d = 0
				d = d + l_dip*sinhc_u
                                dm1 = 1/d
c
				eps_Z = eps2 + val*dm1
c
                                do 1200 l = 1,10
                                   der_Z(l)=coef*der(i,j,l)*G1_u*dm1
     1                                  *p0_e*p0_e
     2                             -val*dm1*dm1*l_dip*der(i,j,l)*F1_u
     3                                  *p0_e*p0_e
1200                            continue
                                der_Z(1)=der_Z(1)-val*dm1*dm1*der_d
                                der_Z(4)=der_Z(4)-val*dm1*dm1*der_d
c
                                if(itype.eq.0) then
                                        do 1275 l = 1,10
                                                der_Z(l)=0.d0
1275                                     continue
                                endif
			endif
c
                        ike = min0(1,iabs(i-nx+1))
                        jke = min0(1,iabs(j-ny+1))
                        kke = min0(1,iabs(k-nz+1))
                        ikem1 = min0(1,iabs(i-2))
                        jkem1 = min0(1,iabs(j-2))
c
                        phi_k = -coef_uC*(phi(i,j,k) - phi(i,j,k+1))
c
                        if(k.ne.1) then
c
                                JoC(i,j,k) = JoC(i,j,k)+coef_uC*eps_Z -
     2                          der_Z(1)*phi_k 
c
                                JoE(i,j,k)=JoE(i,j,k) + ike*
     1                          der_Z(2)*phi_k
c
                                JoN(i,j,k)=JoN(i,j,k) + jke*
     1                          der_Z(3)*phi_k
c
                                JuC(i,j,k)=JuC(i,j,k)+kke*(coef_uC*eps_Z
     1                          +der_Z(4)*phi_k)
c
                                JuE(i,j,k)=JuE(i,j,k) + ike*kke*
     1                          der_Z(7)*phi_k
c
                                JuW(i,j,k)=JuW(i,j,k) + ikem1*kke*
     1                          der_Z(8)*phi_k
c
                                JuN(i,j,k)=JuN(i,j,k) + jke*kke*
     1                          der_Z(9)*phi_k 
c
                                JuS(i,j,k)=JuS(i,j,k) + jkem1*kke*
     1                          der_Z(10)*phi_k
c
                        endif
c
                        if(kke.ne.0) then
c
                                JoC(i,j,k+1)=JoC(i,j,k+1)+coef_uC*eps_Z
     2                          +der_Z(4)*phi_k 
c
                                JoE(i,j,k+1)=JoE(i,j,k+1) - ike*
     1                          der_Z(7)*phi_k
c
                                JoN(i,j,k+1)=JoN(i,j,k+1) - jke*
     1                          der_Z(9)*phi_k
c
                        endif
c
                        if(k.ne.1) then
                                resid(i,j,k)=resid(i,j,k)-eps_Z*phi_k
                        endif
c
                        if(kke.ne.0) then
                                resid(i,j,k+1)=resid(i,j,k+1)+
     1                          eps_Z*phi_k
                        endif
c
1400		continue
1450	   continue
c
1500	continue
c
c ========================================================================================
c Add ions, if any
c ========================================================================================
c
c Case1: do not compute free energy
c
c
        if(iok.eq.1) then
c
	        if(n_ion.eq.0) goto 2200
c
	        do 2100 k = 2,nz-1
c
                    hz = 1.d0/(z(k+1)-z(k-1))
	            call plane_efield27(phi(1,1,k),phi(1,1,k-1),
     1			phi(1,1,k+1),nx,ny,x,y,z,k,Enorm,der)
c
	            do 2000 j = 2,ny-1
c
	   	        hy = 1.d0/(y(j+1)-y(j-1))
c
		        do 1900 i = 2,nx-1
c
	   		    hx = 1.d0/(x(i+1)-x(i-1))
c
                            info1 = ibits(grid_info(i,j,k),1,1)
			    if(info1.ne.0) then
c
				u_val = p0_e*Enorm(i,j)
c
				call compute_F1G1(u_val,F1_u,G1_u,
     1					sinhc_u,exp_mu)
c
				phi_val = phi(i,j,k)
c
				d = exp_mu
				der_d = 0
				der_g = 0
				g = 0
				do 1800 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val-u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d+l_ion(i_ion)*arg
					g=g+k_ion(i_ion)*arg
					der_d = der_d-ion_z(i_ion)
     1					*l_ion(i_ion)*arg
					der_g = der_g -ion_z(i_ion)
     1					*k_ion(i_ion)*arg
1800				continue
				d = d + l_dip*sinhc_u
                                dm1 = 1/d
                                dm2=dm1*dm1
c
                                do 1850 l=1,4
                                   der_X(l)=-l_dip*g*der(i,j,l)*F1_u*dm2
     1                                  *p0_e*p0_e
1850                            continue
                                if(itype.eq.0) then
                                        do 1875 l = 1,4
                                                der_X(l)=0.d0
1875                                     continue
                                endif
c
                                der_X(1)=der_X(1)+der_g*dm1-der_d*g*dm2
c
				resid(i,j,k)=resid(i,j,k)+cc(i,j,k)*g
     1				*dm1
c
				JoC(i,j,k) = JoC(i,j,k) +cc(i,j,k)*
     1				der_X(1)
c
                                ike = min0(1,iabs(i-nx+1))
                                jke = min0(1,iabs(j-ny+1))
                                kke = min0(1,iabs(k-nz+1))
                                JoE(i,j,k) = JoE(i,j,k)+ike*cc(i,j,k)
     1                          *der_X(2)
                                JoN(i,j,k) = JoN(i,j,k)+jke*cc(i,j,k)
     1                          *der_X(3)
                                JuC(i,j,k) = JuC(i,j,k)+kke*cc(i,j,k)
     1                          *der_X(4)
c
			endif
c
1900                continue
2000	        continue
c
2100	    continue
c
2200        continue
c
        else
c
                free_edens = 0
                free_charge = 0
                free_iondip = 0
c
	        do 2600 k = 2,nz-1
c
                    hz = 1.d0/(z(k+1)-z(k-1))
	            call plane_efield27(phi(1,1,k),phi(1,1,k-1),
     1			phi(1,1,k+1),nx,ny,x,y,z,k,Enorm,der)
c
	            do 2500 j = 2,ny-1
c
	   	        hy = 1.d0/(y(j+1)-y(j-1))
c
		        do 2400 i = 2,nx-1
c
	   		    hx = 1.d0/(x(i+1)-x(i-1))
c
                            dvol = hx*hy*hz/8
			    phi_val = phi(i,j,k)
c
                            free_edens = free_edens + Enorm(i,j)
     1                          *Enorm(i,j)*dvol
c
                            free_charge = free_charge+phi_val*fc(i,j,k)
c
			    info1 = ibits(grid_info(i,j,k),0,2)
c
			    if(info1.ne.0) then
c
				u_val = p0_e*Enorm(i,j)
c
				call compute_F1G1(u_val,F1_u,G1_u,
     1					sinhc_u,exp_mu)
c
				d = exp_mu
				der_d = 0
				g = 0
				der_g = 0
                                if(info1.ne.1) then
				    do 2300 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val-u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d+l_ion(i_ion)*arg
					g=g+k_ion(i_ion)*arg
					der_d = der_d-ion_z(i_ion)
     1					*l_ion(i_ion)*arg
					der_g = der_g -ion_z(i_ion)
     1					*k_ion(i_ion)*arg
2300				    continue
                                endif
				d = d + l_dip*sinhc_u
                                dm1 = 1/d
                                dm2 = dm1*dm1
c
				resid(i,j,k)=resid(i,j,k)+cc(i,j,k)*g
     1				*dm1
c
                                do 2350 l=1,4
                                   der_X(l)=-l_dip*g*der(i,j,l)*F1_u*dm2
     1                                  *p0_e*p0_e
2350                            continue
c
                                if(itype.eq.0) then
                                        do 2375 l = 1,4
                                                der_X(l)=0.d0
2375                                     continue
                                endif
                                der_X(1)=der_X(1)+der_g*dm1-der_d*g*dm2
c
				JoC(i,j,k) = JoC(i,j,k) +cc(i,j,k)*
     1				der_X(1)
c
                                ike = min0(1,iabs(i-nx+1))
                                jke = min0(1,iabs(j-ny+1))
                                kke = min0(1,iabs(k-nz+1))
                                JoE(i,j,k) = JoE(i,j,k)-ike*cc(i,j,k)
     1                          *der_X(2)
                                JoN(i,j,k) = JoN(i,j,k)-jke*cc(i,j,k)
     1                          *der_X(3)
                                JuC(i,j,k) = JuC(i,j,k)-kke*cc(i,j,k)
     1                          *der_X(4)
c
			endif
c
2400                continue
2500	        continue
c
2600	    continue
c
            free_edens = -fac_edens*free_edens
            free_charge = free_charge/C_charge
            free_iondip = free_iondip/vol
c
            free_energy = free_edens + free_charge + free_iondip
c
        endif
c
	do 2900 k = 2,nz-1
		do 2800 j = 2,ny-1
			do 2700 i = 2,nx-1
				resid(i,j,k) = -resid(i,j,k)+fc(i,j,k)
2700                    continue
2800            continue
2900    continue
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	Plane_Efield27.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine computes the Electric field on one plane
c			of the grid
c
c	Input:
c		phi      : the electrostatic potential map
c		nx,ny,nz : current size of the grid
c		iplane	 : plane level on which electric field is computed
c
	subroutine plane_efield27(phik,phikm1,phikp1,nx,ny,hx,hy,hz,
     1                  iplane,Enorm,der)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
c	Input/Output variables
c ==========================================================================================
c
	integer	nx,ny
	integer	iplane
c
	real*8	phik(nx,ny),phikm1(nx,ny),phikp1(nx,ny)
	real*8	Enorm(nx,ny)
        real*8  der(nx,ny,*)
c
c	Information on grid
c ==========================================================================================
c
	real*8  hx(*),hy(*),hz(*)
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k
c
	real*8	dx,dy,dz,val
c
        do 50 j = 1,ny
                do 25 i = 1,nx
                        der(i,j,1) = 0.d0
                        der(i,j,2) = 0.d0
                        der(i,j,3) = 0.d0
                        der(i,j,4) = 0.d0
25              continue
50      continue
c
	k = iplane
c
c ==========================================================================================
c	Compute e-field in x direction
c ==========================================================================================
c
c	Efield_x(i,j,k) = (u(i+1/2,j,k) - u(i-1/2,j,k))/dx
c
c	where dx = (hx(j+1)-hx(j-1))/2
c
c	Using interpolation:
c
c	u(i+1/2,j,k) = (u(i+1,j,k)+u(i,j,k))/2
c	and
c	u(i-1/2,j,k) = (u(i-1,j,k)+u(i,j,k))/2
c
c	Then:
c
c	Efield_x(i,j,k) = (u(i+1,j,k) - u(i-1,j,k))/(2*dx)
c
	do 200 j = 2,ny-1
		do 100 i = 2,nx-1
			dx = 1.d0/(hx(i+1)-hx(i-1))
			val = (phik(i+1,j)-phik(i-1,j))*dx
			Enorm(i,j) = val*val
                        der(i,j,2) = dx*val
100		continue 
200	continue
c
c ==========================================================================================
c	Compute e-field in y direction
c ==========================================================================================
c
c	Efield_y(i,j,k) = (u(i,j+1/2,k) - u(i,j-1/2,k))/dy
c
c	where dy = (hy(j+1)-hy(j-1))/2
c
c	Using interpolation:
c
c	u(i,j+1/2,k) = (u(i,j+1,k)+u(i,j,k))/2
c	and
c	u(i,j-1/2,k) = (u(i,j-1,k)+u(i,j,k))/2
c
c	Then:
c
c	Efield_y(i,j,k) = (u(i,j+1,k) - u(i,j-1,k))/(2*dy)
c
	do 500 j = 2,ny-1
		dy = 1.d0/(hy(j+1)-hy(j-1))
		do 400 i = 2,nx-1
			val = (phik(i,j+1)-phik(i,j-1))*dy
			Enorm(i,j)=Enorm(i,j) + val*val
                        der(i,j,3) = dy*val
400		continue
500	continue
c
c ==========================================================================================
c	Compute e-field in z direction
c ==========================================================================================
c
c	Efield_z(i,j,k) = (u(i,j,k+1/2) - u(i,j,k-1/2))/dz
c
c	where dz = (hz(k+1)-hz(k-1))/2
c
c	Using interpolation:
c
c	u(i,j,k+1/2) = (u(i,j,k+1)+u(i,j,k))/2
c	and
c	u(i,j,k-1/2) = (u(i,j,k-1)+u(i,j,k))/2
c
c	Then:
c
c	Efield_z(i,j,k) = (u(i,j,k+1) - u(i,j,k-1))/(2*dz)
c
	dz = 1.d0/(hz(k+1)-hz(k-1))
	do 900 j = 2,ny-1
		do 800 i = 2,nx-1
			val = (phikp1(i,j)-phikm1(i,j))*dz
			Enorm(i,j)=Enorm(i,j) + val*val
                        der(i,j,4) = dz*val
800		continue
900	continue
c
c ==========================================================================================
c	We currently have: ||E||^2; get ||E||
c ==========================================================================================
c
	do 1100 j = 2,ny-1
		do 1000 i = 2,nx-1
			Enorm(i,j) = sqrt(Enorm(i,j))
1000		continue
1100	continue
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c ==========================================================================================
c	Plane_Efield_X27.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine computes the Electric field on one plane
c			of the grid
c
c	Input:
c		phi      : the electrostatic potential map
c		nx,ny,nz : current size of the grid
c		iplane	 : plane level on which electric field is computed
c
	subroutine plane_efield_X27(phik,phikm1,phikp1,nx,ny,hx,hy,
     1                  hz,iplane,Enorm,der)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
c	Input/Output variables
c ==========================================================================================
c
	integer	nx,ny
	integer	iplane
c
	real*8	phik(nx,ny),phikm1(nx,ny),phikp1(nx,ny)
	real*8	Enorm(nx,ny)
        real*8  der(nx,ny,*)
c
c	Information on grid
c ==========================================================================================
c
	real*8  hx(*),hy(*),hz(*)
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k
c
	real*8	dx,dy,dz,val
c
        do 50 j = 1,ny
                do 25 i = 1,nx
                        der(i,j,1) = 0.d0
                        der(i,j,2) = 0.d0
                        der(i,j,3) = 0.d0
                        der(i,j,4) = 0.d0
                        der(i,j,5) = 0.d0
                        der(i,j,6) = 0.d0
                        der(i,j,7) = 0.d0
25              continue
50      continue
c
c
	k = iplane
c
c ==========================================================================================
c	Compute e-field in x direction
c ==========================================================================================
c
c	We have shifted the grid in the X-direction: the Electric field in this direction
c	at index i is therefore simply phi(i+1,j,k) - phi(i,j,k), up to i = nx-1 
c
	do 200 j = 2,ny-1
		do 100 i = 1,nx-1
			dx = 1.d0/(hx(i+1)-hx(i))
			val = (phik(i+1,j)-phik(i,j))*dx
			Enorm(i,j) = val*val
                        der(i,j,1) = -dx*val
                        der(i,j,2) = dx*val
100		continue 
200	continue
c
c ==========================================================================================
c	Compute e-field in y direction
c ==========================================================================================
c
c	The grid is now positioned at half interval along the x-axis. To compute the electric
c	field along y at position (i,j,k) (shifted grid), we need to compute:
c
c	j=2, ny-1:
c	Efield_y(i,j,k) = (u(i+1/2,j+1/2,k) - u(i+1/2,j-1/2,k))/dy
c
c	where dy = (hy(j+1)-hy(j-1))/2
c
c	j = 1:
c	Efield_y(i,j,k) = (u(i+1/2,1+1/2,k) - u(i+1/2,1,k))/dy1
c
c	where dy1 = (hy(2)-hy(1))/2
c
c	Note that we do not need this values for j = ny
c
c	Using bilinear interpolation:
c
c	u(i+1/2,j+1/2,k) = (u(i,j,k) + u(i+1,j,k) + u(i,j+1,k) + u(i+1,j+1,k))/4
c	u(i+1/2,j-1/2,k) = (u(i,j,k) + u(i+1,j,k) + u(i,j-1,k) + u(i+1,j-1,k))/4
c	u(i+1/2,j,k)     = (u(i,j,k) + u(i+1,j,k))/2
c
c	Therefore:
c
c	For j = 2, ny-1
c	Efield_y(i,j,k) = (u(i,j+1,k)+u(i+1,j+1,k)-u(i,j-1,k)-u(i+1,j-1,k))/(4*dy)
c
c	For j = 1
c	Efield_y(i,j,k) = (u(i,j+1,k)+u(i+1,j+1,k)-u(i,j,k)-u(i+1,j,k))/(4*dy1)
c
	do 500 j = 2,ny-1
		dy = 1.d0/(2*(hy(j+1)-hy(j-1)))
		do 400 i = 1,nx-1
			val = (phik(i,j+1)+phik(i+1,j+1)-phik(i,j-1)
     1				-phik(i+1,j-1))*dy
			Enorm(i,j)=Enorm(i,j) + val*val
                        der(i,j,3) = dy*val
                        der(i,j,5) = dy*val
400		continue
500	continue
c
c ==========================================================================================
c	Compute e-field in z direction
c ==========================================================================================
c
c	The grid is positioned at half interval along the x-axis. To compute the electric
c	field along z at position (i,j,k) (shifted grid), we need to compute:
c
c	For k = 2,nz-1
c	Efield_z(i,j,k) = (u(i+1/2,j,k+1/2) - u(i+1/2,j,k-1/2))/dz
c
c	where dz = (hz(k+1)-hz(k-1))/2
c
c	and for k = 1
c	Efield_z(i,j,k) = (u(i+1/2,j,1+1/2) - u(i+1/2,j,1))/dz1
c
c	where dz1 = (hz(2)-hz(1))/2
c
c	Note that we do not need this values for k = nz
c
c	Using bilinear interpolation:
c
c	u(i+1/2,j,k+1/2) = (u(i,j,k) + u(i+1,j,k) + u(i,j,k+1) + u(i+1,j,k+1))/4
c	u(i+1/2,j,k-1/2) = (u(i,j,k) + u(i+1,j,k) + u(i,j,k-1) + u(i+1,j,k-1))/4
c	u(i+1/2,j,k)     = (u(i,j,k) + u(i+1,j,k))/2
c
c	Therefore:
c
c	For k=iplane = 2, nz-1
c	Efield_z(i,j,k) = (u(i,j,k+1)+u(i+1,j,k+1)-u(i,j,k-1)-u(i+1,j,k-1))/(4*dz)
c
c	For k = 1
c	Efield_z(i,j,k) = (u(i,j,k+1)+u(i+1,j,k+1)-u(i,j,k)-u(i+1,j,k))/(4*dz1)
c
	dz = 1.d0/(2.d0*(hz(k+1)-hz(k-1)))
	do 900 j = 2,ny-1
		do 800 i = 1,nx-1
			val = (phikp1(i,j)+phikp1(i+1,j)-
     1			phikm1(i,j)-phikm1(i+1,j))*dz
			Enorm(i,j)=Enorm(i,j) + val*val
                        der(i,j,4) = val*dz
                        der(i,j,7) = val*dz
800		continue
900	continue
c
c ==========================================================================================
c	We currently have: ||E||^2; get ||E||
c ==========================================================================================
c
	do 1100 j = 2,ny-1
		do 1000 i = 1,nx-1
			Enorm(i,j) = sqrt(Enorm(i,j))
1000		continue
1100	continue
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	Plane_Efield_Y27.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine computes the Electric field on one plane
c			of the grid
c
c	Input:
c		phi      : the electrostatic potential map
c		nx,ny,nz : current size of the grid
c		iplane	 : plane level on which electric field is computed
c
	subroutine plane_efield_Y27(phik,phikm1,phikp1,nx,ny,hx,hy,hz,
     1                  iplane,Enorm,der)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
c	Input/Output variables
c ==========================================================================================
c
	integer	nx,ny
	integer	iplane
c
	real*8	phik(nx,ny),phikm1(nx,ny),phikp1(nx,ny)
	real*8	Enorm(nx,ny)
        real*8  der(nx,ny,*)
c
c	Information on grid
c ==========================================================================================
c
	real*8  hx(*),hy(*),hz(*)
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k
c
	real*8	dx,dy,dym1,dz,val
c
	k = iplane
c
        do 50 j = 1,ny
                do 25 i = 1,nx
                        der(i,j,1) = 0.d0
                        der(i,j,2) = 0.d0
                        der(i,j,3) = 0.d0
                        der(i,j,4) = 0.d0
                        der(i,j,5) = 0.d0
                        der(i,j,6) = 0.d0
                        der(i,j,7) = 0.d0
                        der(i,j,8) = 0.d0
                        der(i,j,9) = 0.d0
                        der(i,j,10) = 0.d0
25              continue
50      continue
c
c ==========================================================================================
c	Compute e-field in y direction
c ==========================================================================================
c
c	We have shifted the grid in the Y-direction: the Electric field in this direction
c	at index j is therefore simply phi(i,j+1,k) - phi(i,j,k), up to j = ny-1 
c
	do 200 j = 1,ny-1
		dy = 1.d0/(hy(j+1)-hy(j))
		do 100 i = 2,nx-1
			val = (phik(i,j+1)-phik(i,j))*dy
			Enorm(i,j) = val*val
                        der(i,j,1) = -val*dy
                        der(i,j,3) = val*dy
100		continue 
200	continue
c
c ==========================================================================================
c	Compute e-field in x direction
c ==========================================================================================
c
c	The grid is now positioned at half interval along the y-axis. To compute the electric
c	field along x at position (i,j,k) (shifted grid), we need to compute:
c
c	For x = 2, nx-1:
c	Efield_x(i,j,k) = (u(i+1/2,j+1/2,k) - u(i-1/2,j+1/2,k))/dx
c
c	where dx = (hx(i+1)-hx(i-1))/2
c
c	For x = 1:
c	Efield_x(1,j,k) = (u(1+1/2,j+1/2,k) - u(1,j+1/2,k))/dx1
c
c	where dx1 = (hx(2)-hx(1))/2
c
c	Note that we do not need this values for i = nx
c
c	Using bilinear interpolation:
c
c	u(i+1/2,j+1/2,k) = (u(i,j,k) + u(i+1,j,k) + u(i,j+1,k) + u(i+1,j+1,k))/4
c	u(i-1/2,j+1/2,k) = (u(i-1,j,k) + u(i,j+1,k) + u(i,j,k) + u(i-1,j+1,k))/4
c	u(i,j+1/2,k)     = (u(i,j,k) + u(i,j+1,k))/2
c
c	Therefore:
c
c	For i = 2, nx-1
c	Efield_x(i,j,k) = (u(i+1,j,k)+u(i+1,j+1,k)-u(i-1,j,k)-u(i-1,j+1,k))/(4*dx)
c
c	For i = 1
c	Efield_x(i,j,k) = (u(i+1,j,k)+u(i+1,j+1,k)-u(i,j,k)-u(i,j+1,k))/(4*dx1)
c
	do 500 j = 1,ny-1
		do 400 i = 2,nx-1
			dx = 1.d0/(2*(hx(i+1)-hx(i-1)))
			val = (phik(i+1,j)+phik(i+1,j+1)-phik(i-1,j)
     1				-phik(i-1,j+1))*dx
			Enorm(i,j)=Enorm(i,j) + val*val
                        der(i,j,2) = val*dx
                        der(i,j,5) = val*dx
                        der(i,j,6) = -val*dx
400		continue
500	continue
c
c ==========================================================================================
c	Compute e-field in z direction
c ==========================================================================================
c
c	The grid is positioned at half interval along the y-axis. To compute the electric
c	field along z at position (i,j,k) (shifted grid), we need to compute:
c
c	For k = 2,nz-1:
c	Efield_z(i,j,k) = (u(i,j+1/2,k+1/2) - u(i,j+1/2,k-1/2))/dz
c
c	with dz = (hz(k+1)-hz(k-1))/2
c
c	and for k = 1:
c	Efield_z(i,j,1) = (u(i,j+1/2,1+1/2) - u(i,j+1/2,1))/dz1
c
c	with dz1 = (hz(2)-hz(1))/2
c
c	Note that we do not need this values for k = nz
c
c	Using bilinear interpolation:
c
c	u(i,j+1/2,k+1/2) = (u(i,j,k) + u(i,j+1,k) + u(i,j,k+1) + u(i,j+1,k+1))/4
c	u(i,j+1/2,k-1/2) = (u(i,j,k) + u(i,j+1,k) + u(i,j,k-1) + u(i,j+1,k-1))/4
c	u(i,j+1/2,k)     = (u(i,j,k) + u(i,j+1,k))/2
c
c	Therefore:
c
c	For k=iplane = 2, nz-1
c	Efield_z(i,j,k) = (u(i,j,k+1)+u(i,j+1,k+1)-u(i,j,k-1)-u(i,j+1,k-1))/(4*dz)
c
c	For k = 1
c	Efield_z(i,j,k) = (u(i,j,k+1)+u(i,j+1,k+1)-u(i,j,k)-u(i,j+1,k))/(4*dz1)
c
	dz = 1.d0/(2*(hz(k+1)-hz(k-1)))
	do 900 j = 1,ny-1
		do 800 i = 2,nx-1
			val = (phikp1(i,j)+phikp1(i,j+1)-
     1			phikm1(i,j)-phikm1(i,j+1))*dz
			Enorm(i,j)=Enorm(i,j) + val*val
                        der(i,j,4) = val*dz
                        der(i,j,9) = val*dz
800		continue
900	continue
c
c ==========================================================================================
c	We currently have: ||E||^2; get ||E||
c ==========================================================================================
c
	do 1100 j = 1,ny-1
		do 1000 i = 2,nx-1
			Enorm(i,j) = sqrt(Enorm(i,j))
1000		continue
1100	continue
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	Plane_Efield_Z27.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine computes the Electric field on one plane
c			of the grid
c
c	Input:
c		phi      : the electrostatic potential map
c		nx,ny,nz : current size of the grid
c		iplane	 : plane level on which electric field is computed
c
	subroutine plane_efield_Z27(phik,phikp1,nx,ny,hx,hy,hz,
     1                  iplane,Enorm,der)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
c	Input/Output variables
c ==========================================================================================
c
	integer	nx,ny
	integer	iplane
c
	real*8	phik(nx,ny),phikp1(nx,ny)
	real*8	Enorm(nx,ny)
        real*8  der(nx,ny,*)
c
c	Information on grid
c ==========================================================================================
c
	real*8  hx(*),hy(*),hz(*)
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k
c
	real*8	dx,dy,dz,dzm1,val
c
	k = iplane
c
        do 50 j = 1,ny
                do 25 i = 1,nx
                        der(i,j,1) = 0.d0
                        der(i,j,2) = 0.d0
                        der(i,j,3) = 0.d0
                        der(i,j,4) = 0.d0
                        der(i,j,5) = 0.d0
                        der(i,j,6) = 0.d0
                        der(i,j,7) = 0.d0
                        der(i,j,8) = 0.d0
                        der(i,j,9) = 0.d0
                        der(i,j,10) = 0.d0
25              continue
50      continue
c
c ==========================================================================================
c	Compute e-field in z direction
c ==========================================================================================
c
c	We have shifted the grid in the Z-direction: the Electric field in this direction
c	at index (i,j) is therefore simply phi(i,j,k+1) - phi(i,j,k), up to k = nz-1 
c
	dz = 1.d0/(hz(k+1)-hz(k))
	do 200 j = 2,ny-1
		do 100 i = 2,nx-1
			val = (phikp1(i,j)-phik(i,j))*dz
			Enorm(i,j) = val*val
                        der(i,j,1) = -val*dz
                        der(i,j,4) = val*dz
100		continue 
200	continue
c
c ==========================================================================================
c	Compute e-field in x direction
c ==========================================================================================
c
c	The grid is now positioned at half interval along the z-axis. To compute the electric
c	field along x at position (i,j,k) (shifted grid), we need to compute:
c
c	x = 2,nx-1:
c	Efield_x(i,j,k) = (u(i+1/2,j,k+1/2) - u(i-1/2,j,k+1/2))/dx
c
c	where dx = (hx(i+1)-hx(i-1))/2
c
c	x = 1:
c	Efield_x(1,j,k) = (u(1+1/2,j,k+1/2) - u(1,j,k+1/2))/dx
c
c	where dx1 = (hx(2)-hx(1))/2
c
c	Note that we do not need this values for i = nx
c
c	Using bilinear interpolation:
c
c	u(i+1/2,j,k+1/2) = (u(i,j,k) + u(i+1,j,k) + u(i,j,k+1) + u(i+1,j,k+1))/4
c	u(i-1/2,j,k+1/2) = (u(i-1,j,k) + u(i,j,k+1) + u(i,j,k) + u(i-1,j,k+1))/4
c	u(i,j,k+1/2)     = (u(i,j,k) + u(i,j,k+1))/2
c
c	Therefore:
c
c	For i = 2, nx-1
c	Efield_x(i,j,k) = (u(i+1,j,k)+u(i+1,j,k+1)-u(i-1,j,k)-u(i-1,j,k+1))/(4*dx)
c
c	For i = 1
c	Efield_x(i,j,k) = (u(i+1,j,k)+u(i+1,j,k+1)-u(i,j,k)-u(i,j,k+1))/(4*dx1)
c
	do 500 j = 2,ny-1
		do 400 i = 2,nx-1
			dx = 1.d0/(2*(hx(i+1)-hx(i-1)))
			val = (phik(i+1,j)+phikp1(i+1,j)-phik(i-1,j)
     1				-phikp1(i-1,j))*dx
			Enorm(i,j)=Enorm(i,j) + val*val
                        der(i,j,2) = val*dx
                        der(i,j,7) = val*dx
                        der(i,j,8) = -val*dx
400		continue
500	continue
c
c ==========================================================================================
c	Compute e-field in y direction
c ==========================================================================================
c
c	The grid is positioned at half interval along the y-axis. To compute the electric
c	field along z at position (i,j,k) (shifted grid), we need to compute:
c
c	For y = 2,ny-1:
c	Efield_y(i,j,k) = (u(i,j+1/2,k+1/2) - u(i,j-1/2,k+1/2))/dy
c
c	where dy = (hy(j+1)-hy(j-1))/2
c
c	For y = 1:
c	Efield_y(i,1,k) = (u(i,1+1/2,k+1/2) - u(i,1,k+1/2))/dy1
c
c	where dy1 = (hy(2)-hy(1))/2
c
c	Note that we do not need this values for k = nz
c
c	Using bilinear interpolation:
c
c	u(i,j+1/2,k+1/2) = (u(i,j,k) + u(i,j+1,k) + u(i,j,k+1) + u(i,j+1,k+1))/4
c	u(i,j-1/2,k+1/2) = (u(i,j,k) + u(i,j,k+1) + u(i,j-1,k) + u(i,j-1,k+1))/4
c	u(i,j,k+1/2)     = (u(i,j,k) + u(i,j,k+1))/2
c
c	Therefore:
c
c	For j= 2, ny-1
c	Efield_z(i,j,k) = (u(i,j+1,k)+u(i,j+1,k+1)-u(i,j-1,k)-u(i,j-1,k+1))/(4*dy)
c
c	For j = 1:
c	Efield_z(i,j,k) = (u(i,j+1,k+1)+u(i,j+1,k+1)-u(i,j,k)-u(i,j,k+1))/(4*dy1)
c
	do 900 j = 2,ny-1
		dy=1.d0/(2*(hy(j+1)-hy(j-1)))
		do 800 i = 2,nx-1
			val = (phik(i,j+1)+phikp1(i,j+1)-
     1			phik(i,j-1)-phikp1(i,j-1))*dy
			Enorm(i,j)=Enorm(i,j) + val*val
                        der(i,j,3) = val*dy
                        der(i,j,9) = val*dy
                        der(i,j,10) = -val*dy
800		continue
900	continue
c
c ==========================================================================================
c	We currently have: ||E||^2; get ||E||
c ==========================================================================================
c
	do 1100 j = 2,ny-1
		do 1000 i = 2,nx-1
			Enorm(i,j) = sqrt(Enorm(i,j))
1000		continue
1100	continue
c
	return
	end
