c       Copyright (C) 2007 Patrice Koehl
c
c       This library is free software; you can redistribute it and/or
c       modify it under the terms of the GNU Lesser General Public
c       License as published by the Free Software Foundation; either
c       version 2.1 of the License, or (at your option) any later version.
c
c       This library is distributed in the hope that it will be useful,
c       but WITHOUT ANY WARRANTY; without even the implied warranty of
c       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
c       Lesser General Public License for more details.
c
c       You should have received a copy of the GNU Lesser General Public
c       License along with this library; if not, write to the Free Software
c       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
c
c ==========================================================================================
c ==========================================================================================
c	Langevin_diel.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine is the driver to the routine that
c			computes the residual of the GPBL equation
c
c	Input:
c		phi      : the electrostatic potential map
c		nx,ny,nz : size of the grid
c		x,y,z    : coordinates on the grid
c		grid_info: infor on regular grid and shifted grids
c		fc	 : source term (fixed charges)
c		cc	 : Helmholtz term
c		Enorm  	 : storage array
c
c	Output:
c		resid	 : residual vector
c               epsX,epsY,epsZ : shifted dielectric maps
c
	subroutine langevin_diel(phi,nx,ny,nz,x,y,z,grid_info,
     2			Enorm,fc,cc,resid,epsX,epsY,epsZ)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	include 'poisson.h'
c
	integer	nx,ny,nz
        integer info1
c
	real*8	eps_X,eps_Y,eps_Z
	real*8	coef_oE,coef_oN,coef_uC
	real*8	coef_oEh,coef_oNh,coef_uCh
	real*8	hxm1,hym1,hzm1
	real*8	hx,hy,hz
	real*8	dvol
c
	real*8	phi(nx,ny,nz)
	real*8	fc(nx,ny,nz),resid(nx,ny,nz)
	real*8	cc(nx,ny,nz)
        real*8  epsX(nx,ny,nz),epsY(nx,ny,nz)
        real*8  epsZ(nx,ny,nz)
c
	real*8	Enorm(nx,ny)
c
c	Information on grids
c ==========================================================================================
c
	real*8  	x(nx),y(ny),z(nz)
c
	integer*1       grid_info(nx,ny,nz)
c
c 	Parameter values
c ==========================================================================================
c
	integer	n_solv,n_ion,n_zz
c
	real*8	vol,p0_e,coef_l
	real*8	coef_Z0,phi0,ratio
c
	real*8  ion_cvol(max_ion),l_ion(max_ion),k_ion(max_ion)
	real*8	solv_conc(max_solv)
	real*8  solv_dipole(max_solv),solv_cvol(max_solv)
c
	real*8  ion_conc(max_ion),ion_z(max_ion),ion_rad(max_ion)
	real*8  kappa(max_ion)
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k,l,i_ion
c
	real*8	u_val,F1_u,sinhc_u,val,exp_mu
	real*8	val1,val2,val3
	real*8	coef,phi_val,d,g,dm1
	real*8	arg,exp_max,exp_min
	real*8	l_dip
        real*8  eps1,eps2
c
	parameter       (exp_min = -80.d0, exp_max = 80.d0)
c
c	Common blocks
c ==========================================================================================
c
	common /dipoles/	vol,p0_e,coef_l
	common /solvent/        solv_conc,solv_dipole,solv_cvol,n_solv
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
	common /smpb/           ion_cvol,ratio,phi0,coef_Z0
        common /elec_const/     eps1,eps2
c
c ==========================================================================================
c	Initialise
c ==========================================================================================
c
c	Make sure all arrays are initialized to 0
c ==========================================================================================
c
	call set_all_zeros(nx,ny,nz,resid)
	call set_all_zeros(nx,ny,nz,epsX)
	call set_all_zeros(nx,ny,nz,epsY)
	call set_all_zeros(nx,ny,nz,epsZ)
c
c ==========================================================================================
c	Define some parameters
c ==========================================================================================
c
c
	coef = coef_l * solv_cvol(1)/(1-phi0)
	l_dip = solv_cvol(1)/(1-phi0)
        if(l_dip.eq.0) phi0 = 0.d0
	do 100 i = 1,n_ion
		l_ion(i) = ion_cvol(i)/(1-phi0)
                if(l_dip.eq.0) l_ion(i) = 0.d0
		k_ion(i) = kappa(i)/(1-phi0)
100	continue
c
c ==========================================================================================
c	Compute eps for the X-shifted map
c ==========================================================================================
c
c	Compute electric field on x-shifted grid
c
	do 600 k = 2,nz-1
c
	   hzm1 = z(k)-z(k-1)
	   hz = z(k+1)-z(k)
c
	   call plane_efield_X(phi(1,1,k),phi(1,1,k-1),phi(1,1,k+1),
     1		nx,ny,x,y,z,k,Enorm)
c
	   do 500 j = 2,ny-1
c
		hym1 = y(j) - y(j-1)
		hy = y(j+1)-y(j)
c
		do 400 i = 1,nx-1
c
			hx = x(i+1)-x(i)
c
                       	coef_oE  = (hym1+hy)*(hzm1+hz)/(4.d0*hx)
			coef_oEh = coef_oE*(phi(i,j,k)-phi(i+1,j,k))
c
			info1 = ibits(grid_info(i,j,k),2,2)
c
			if(info1.eq.0) then
c
                                eps_X = eps1
c
			else
c
				u_val = p0_e*Enorm(i,j)
c
				call compute_F1(u_val,F1_u,
     1					sinhc_u,exp_mu)
				val = coef*F1_u
c
				phi_val = (phi(i,j,k)+phi(i+1,j,k))/2
c
				d = exp_mu
c
				if(info1.ne.1) then
c
				    do 200 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val-u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d+l_ion(i_ion)*arg
200				    continue
c
				endif
c
				d = d + l_dip*sinhc_u
				dm1 = 1/d
c
				eps_X = eps2 + val*dm1
c
			endif
c
                        epsX(i,j,k) = eps_X
			if(i.ne.1) then
				resid(i,j,k)=resid(i,j,k)+coef_oEh*eps_X
			endif
			if(i.ne.nx-1) then
				resid(i+1,j,k)=resid(i+1,j,k)
     1     				-coef_oEh*eps_X
			endif
c
400		continue
500	   continue
c
600	continue
c
c ==========================================================================================
c	Define eps for the Y-shifted map
c ==========================================================================================
c
c	Compute electric field on Y-shifted grid
c
	do 1100 k = 2,nz-1
c
	   hzm1 = z(k)-z(k-1)
	   hz = z(k+1)-z(k)
c
	   call plane_efield_Y(phi(1,1,k),phi(1,1,k-1),phi(1,1,k+1),
     1		nx,ny,x,y,z,k,Enorm)
c
	   do 1000 j = 1,ny-1
c
		hy = y(j+1)-y(j)
c
		do 900 i = 2,nx-1
c
			hxm1 = x(i)-x(i-1)
			hx = x(i+1)-x(i)
c
                       	coef_oN  = (hxm1+hx)*(hzm1+hz)/(4.d0*hy)
			coef_oNh = coef_oN*(phi(i,j,k)-phi(i,j+1,k))
c
			info1 = ibits(grid_info(i,j,k),4,2)
c
			if(info1.eq.0) then
c
				eps_Y = eps1
c
			else
c
				u_val = p0_e*Enorm(i,j)
c
				call compute_F1(u_val,F1_u,
     1				sinhc_u,exp_mu)
				val = coef*F1_u
c 
				phi_val = (phi(i,j,k)+
     1				phi(i,j+1,k))/2
c
				d = exp_mu
c
				if(info1.ne.1) then
c
				    do 700 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val-u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d+l_ion(i_ion)*arg
700				    continue
c
				endif
c
				d = d + l_dip*sinhc_u
				dm1 = 1/d
				eps_Y = eps2 + val*dm1
c
			endif
c
                        epsY(i,j,k) = eps_Y
			if(j.ne.1) then
				resid(i,j,k)=resid(i,j,k)+coef_oNh*eps_Y
			endif
			if(j.ne.ny-1) then
				resid(i,j+1,k)=resid(i,j+1,k)
     1				-coef_oNh*eps_Y
			endif
c
900		continue
1000	   continue
c
1100	continue
c
c ==========================================================================================
c	Define eps for the Z-shifted map
c ==========================================================================================
c
c	Compute electric field on Z-shifted grid
c
	do 1700 k = 1,nz-1
c
	   hz = z(k+1)-z(k)
c
	   call plane_efield_Z(phi(1,1,k),phi(1,1,k+1),
     1		nx,ny,x,y,z,k,Enorm)
c
	   do 1600 j = 2,ny-1
c
		hy = y(j+1)-y(j)
		hym1 = y(j)-y(j-1)
c
		do 1500 i = 2,nx-1
c
			hxm1 = x(i)-x(i-1)
			hx = x(i+1)-x(i)
c
                       	coef_uC  = (hxm1+hx)*(hym1+hy)/(4.d0*hz)
			coef_uCh = coef_uC*(phi(i,j,k)-phi(i,j,k+1))
c
			info1 = ibits(grid_info(i,j,k),6,2)
c
			if(info1.eq.0) then
c
				eps_Z = eps1
c
			else
c
				u_val = p0_e*Enorm(i,j)
c
				call compute_F1(u_val,F1_u,
     1				sinhc_u,exp_mu)
				val = coef*F1_u
c
				phi_val = (phi(i,j,k)+
     1				phi(i,j,k+1))/2
c
				d = exp_mu
c
				if(info1.ne.1) then
c
				    do 1200 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val-u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d + l_ion(i_ion)*arg
1200				    continue
c
				endif
c
				d = d + l_dip*sinhc_u
				dm1 = 1/d
c
				eps_Z = eps2 + val*dm1
c
			endif
c
                        epsZ(i,j,k) = eps_Z
			if(k.ne.1) then
				resid(i,j,k)=resid(i,j,k)+coef_uCh*eps_Z
			endif
			if(k.ne.nz-1) then
				resid(i,j,k+1)=resid(i,j,k+1)
     1				-coef_uCh*eps_Z
			endif
c
1500		continue
1600	   continue
c
1700	continue
c
	if(n_ion.eq.0) goto 2200
c
	do 2100 k = 2,nz-1
c
	   hz = z(k+1)-z(k-1)
c
	   call plane_efield(phi(1,1,k),phi(1,1,k-1),
     1		phi(1,1,k+1),nx,ny,x,y,z,k,Enorm)
c
	   do 2000 j = 2,ny-1
c
                hy = y(j+1)-y(j-1)
c
		do 1900 i = 2,nx-1
c
                        hx = x(i+1)-x(i-1)
c
			if(cc(i,j,k).ne.0) then
c
                                dvol = hx*hy*hz/8.d0
c
				u_val = p0_e*Enorm(i,j)
c
				call compute_F1(u_val,F1_u,
     1					sinhc_u,exp_mu)
c
				phi_val = phi(i,j,k)
c
				d = exp_mu
                                g = 0
				do 1800 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val-u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d+l_ion(i_ion)*arg
					g=g+k_ion(i_ion)*arg
1800				continue
				d = d + l_dip*sinhc_u
				dm1 = 1/d
c
				resid(i,j,k)=resid(i,j,k)+cc(i,j,k)*
     1						g*dm1*dvol
c
			endif
c
1900		continue
2000	   continue
c
2100	continue
c
2200	continue
c
	do 2500 k = 2,nz-1
	        hz = z(k+1)-z(k-1)
		do 2400 j = 2,ny-1
	                hy = y(j+1)-y(j-1)
			do 2300 i = 2,nx-1
	                        hx = x(i+1)-x(i-1)
                                d = hx*hy*hz/8.d0
				resid(i,j,k) = -resid(i,j,k)+fc(i,j,k)*d
2300			continue
2400		continue
2500	continue
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	Langevin_new_eps.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine is the driver to the routine that
c			computes the residual of the GPBL equation
c
c	Input:
c		phi      : the electrostatic potential map
c		nx,ny,nz : size of the grid
c		x,y,z    : coordinates on the grid
c		grid_info: infor on regular grid and shifted grids
c		cc	 : Helmholtz term
c		Enorm  	 : storage array
c
c	Output:
c               epsX,epsY,epsZ : shifted dielectric maps
c		cc	       : updated Helmholtz term
c
	subroutine langevin_new_eps(phi,nx,ny,nz,x,y,z,grid_info,
     2			Enorm,epsX,epsY,epsZ)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	include 'poisson.h'
c
	integer	nx,ny,nz
        integer info1,flag_psi
c
	real*8	eps_X,eps_Y,eps_Z
c
	real*8	phi(nx,ny,nz)
	real*8	psi(nx,ny,nz)
        real*8  epsX(nx,ny,nz),epsY(nx,ny,nz)
        real*8  epsZ(nx,ny,nz)
c
	real*8	Enorm(nx,ny)
c
	pointer (ptr_psi,psi)
c
c	Information on grids
c ==========================================================================================
c
	real*8  	x(nx),y(ny),z(nz)
c
	integer*1       grid_info(nx,ny,nz)
c
c 	Parameter values
c ==========================================================================================
c
	integer	n_solv,n_ion,n_zz
c
	real*8	vol,p0_e,coef_l
	real*8	coef_Z0,phi0,ratio
	real*8	psi_val,psi_coef
	real*8	psi_bulk,v0,b
c
	real*8  ion_cvol(max_ion),l_ion(max_ion),k_ion(max_ion)
	real*8	solv_conc(max_solv)
	real*8  solv_dipole(max_solv),solv_cvol(max_solv)
c
	real*8  ion_conc(max_ion),ion_z(max_ion),ion_rad(max_ion)
	real*8  kappa(max_ion)
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k,l,i_ion
c
	real*8	u_val,F1_u,sinhc_u,val,exp_mu
	real*8	val1,val2,val3
	real*8	coef,phi_val,d,g,dm1
	real*8	arg,exp_max,exp_min
	real*8	l_dip
        real*8  eps1,eps2
c
	parameter       (exp_min = -80.d0, exp_max = 80.d0)
c
c	Common blocks
c ==========================================================================================
c
	common /dipoles/	vol,p0_e,coef_l
	common /solvent/        solv_conc,solv_dipole,solv_cvol,n_solv
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
	common /smpb/           ion_cvol,ratio,phi0,coef_Z0
        common /elec_const/     eps1,eps2
	common /psi_pot/	ptr_psi,flag_psi
	common /yukawa/		v0,b,psi_bulk
c
c ==========================================================================================
c	Initialise
c ==========================================================================================
c
c
	coef = coef_l * solv_cvol(1)/(1-phi0)
	l_dip = solv_cvol(1)/(1-phi0)
        if(l_dip.eq.0) phi0 = 0.d0
	do 100 i = 1,n_ion
		l_ion(i) = ion_cvol(i)/(1-phi0)
                if(l_dip.eq.0) l_ion(i) = 0.d0
		k_ion(i) = kappa(i)/(1-phi0)
100	continue
c
c	Make sure all arrays are initialized to 0
c ==========================================================================================
c
	call set_all_zeros(nx,ny,nz,epsX)
	call set_all_zeros(nx,ny,nz,epsY)
	call set_all_zeros(nx,ny,nz,epsZ)
c
c ==========================================================================================
c	Compute eps for the X-shifted map
c ==========================================================================================
c
c	Compute electric field on x-shifted grid
c
	do 600 k = 2,nz-1
c
	   call plane_efield_X(phi(1,1,k),phi(1,1,k-1),phi(1,1,k+1),
     1		nx,ny,x,y,z,k,Enorm)
c
	   do 500 j = 2,ny-1
c
		do 400 i = 1,nx-1
c
			info1 = ibits(grid_info(i,j,k),2,2)
c
			if(info1.eq.0) then
c
                                eps_X = eps1
c
			else
c
				if(flag_psi.eq.1) then
					psi_val = (psi(i,j,k)+
     1					psi(i+1,j,k))/2
					arg = -psi_val+psi_bulk
					if(arg.gt.exp_max) arg=exp_max
					if(arg.lt.exp_min) arg=exp_min
					psi_coef = exp(arg)
				else
					psi_coef = 1
				endif
c
				u_val = p0_e*Enorm(i,j)
c
				call compute_F1(u_val,F1_u,
     1					sinhc_u,exp_mu)
				val = coef*F1_u*psi_coef
c
				phi_val = (phi(i,j,k)+phi(i+1,j,k))/2
c
				d = exp_mu
c
				if(info1.ne.1) then
c
				    do 200 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val-u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d+l_ion(i_ion)*arg
200				    continue
c
				endif
c
				d = d + l_dip*sinhc_u*psi_coef
				dm1 = 1/d
c
				eps_X = eps2 + val*dm1
c
			endif
c
                        epsX(i,j,k) = eps_X
c
400		continue
500	   continue
c
600	continue
c
c ==========================================================================================
c	Define eps for the Y-shifted map
c ==========================================================================================
c
c	Compute electric field on Y-shifted grid
c
	do 1100 k = 2,nz-1
c
	   call plane_efield_Y(phi(1,1,k),phi(1,1,k-1),phi(1,1,k+1),
     1		nx,ny,x,y,z,k,Enorm)
c
	   do 1000 j = 1,ny-1
c
		do 900 i = 2,nx-1
c
			info1 = ibits(grid_info(i,j,k),4,2)
c
			if(info1.eq.0) then
c
				eps_Y = eps1
c
			else
c
				if(flag_psi.eq.1) then
					psi_val = (psi(i,j,k)+
     1					psi(i,j+1,k))/2
					arg = -psi_val+psi_bulk
					if(arg.gt.exp_max) arg=exp_max
					if(arg.lt.exp_min) arg=exp_min
					psi_coef = exp(arg)
				else
					psi_coef = 1
				endif
c
				u_val = p0_e*Enorm(i,j)
c
				call compute_F1(u_val,F1_u,
     1				sinhc_u,exp_mu)
				val = coef*F1_u*psi_coef
c 
				phi_val = (phi(i,j,k)+
     1				phi(i,j+1,k))/2
c
				d = exp_mu
c
				if(info1.ne.1) then
c
				    do 700 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val-u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d+l_ion(i_ion)*arg
700				    continue
c
				endif
c
				d = d + l_dip*sinhc_u*psi_coef
				dm1 = 1/d
				eps_Y = eps2 + val*dm1
c
			endif
c
                        epsY(i,j,k) = eps_Y
c
900		continue
1000	   continue
c
1100	continue
c
c ==========================================================================================
c	Define eps for the Z-shifted map
c ==========================================================================================
c
c	Compute electric field on Z-shifted grid
c
	do 1700 k = 1,nz-1
c
	   call plane_efield_Z(phi(1,1,k),phi(1,1,k+1),
     1		nx,ny,x,y,z,k,Enorm)
c
	   do 1600 j = 2,ny-1
c
		do 1500 i = 2,nx-1
c
			info1 = ibits(grid_info(i,j,k),6,2)
c
			if(info1.eq.0) then
c
				eps_Z = eps1
c
			else
c
				if(flag_psi.eq.1) then
					psi_val = (psi(i,j,k)+
     1					psi(i,j,k+1))/2
					arg = -psi_val+psi_bulk
					if(arg.gt.exp_max) arg=exp_max
					if(arg.lt.exp_min) arg=exp_min
					psi_coef = exp(arg)
				else
					psi_coef = 1
				endif
c
				u_val = p0_e*Enorm(i,j)
				u_val = p0_e*Enorm(i,j)
c
				call compute_F1(u_val,F1_u,
     1				sinhc_u,exp_mu)
				val = coef*F1_u*psi_coef
c
				phi_val = (phi(i,j,k)+
     1				phi(i,j,k+1))/2
c
				d = exp_mu
c
				if(info1.ne.1) then
c
				    do 1200 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val-u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d + l_ion(i_ion)*arg
1200				    continue
c
				endif
c
				d = d + l_dip*sinhc_u*psi_coef
				dm1 = 1/d
c
				eps_Z = eps2 + val*dm1
c
			endif
c
                        epsZ(i,j,k) = eps_Z
c
1500		continue
1600	   continue
c
1700	continue
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	Langevin_resid.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine is the driver to the routine that
c			computes the residual of the GPBL equation
c
c	Input:
c		phi      : the electrostatic potential map
c		nx,ny,nz : size of the grid
c		x,y,z    : coordinates on the grid
c		fc	 : source term (fixed charges)
c		cc	 : Helmholtz term
c		Enorm  	 : storage array
c               epsX,epsY,epsZ : shifted dielectric maps
c
c	Output:
c		resid	 : residual vector
c
	subroutine langevin_resid(phi,nx,ny,nz,x,y,z,
     2			Enorm,fc,cc,resid,epsX,epsY,epsZ,
     3                  free_ene,flag_free,grid_info)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	include 'poisson.h'
c
	integer	nx,ny,nz
        integer info1,flag_psi,flag_free
c
	real*8	eps_X,eps_Y,eps_Z
	real*8	coef_oE,coef_oN,coef_uC
	real*8	coef_oEh,coef_oNh,coef_uCh
	real*8	hxm1,hym1,hzm1
	real*8	hx,hy,hz
	real*8	free_edens,free_charge,free_iondip,free_ene,dvol
        real*8  fac_edens
	real*8	psi_bulk,psi_val,psi_coef,v0,b
c
	real*8	phi(nx,ny,nz)
	real*8	fc(nx,ny,nz),resid(nx,ny,nz)
	real*8	cc(nx,ny,nz)
        real*8  epsX(nx,ny,nz),epsY(nx,ny,nz)
        real*8  epsZ(nx,ny,nz)
c
	real*8	psi(nx,ny,nz)
c
	real*8	Enorm(nx,ny)
c
	pointer (ptr_psi,psi)
c
c	Information on grids
c ==========================================================================================
c
	real*8  	x(nx),y(ny),z(nz)
c
	integer*1       grid_info(nx,ny,nz)
c
c 	Parameter values
c ==========================================================================================
c
	integer	n_solv,n_ion,n_zz
c
	real*8	vol,p0_e,coef_l
	real*8	coef_Z0,phi0,ratio
c
	real*8  ion_cvol(max_ion),l_ion(max_ion),k_ion(max_ion)
	real*8	solv_conc(max_solv)
	real*8  solv_dipole(max_solv),solv_cvol(max_solv)
c
	real*8  ion_conc(max_ion),ion_z(max_ion),ion_rad(max_ion)
	real*8  kappa(max_ion)
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k,l,i_ion
c
	real*8	u_val,F1_u,sinhc_u,val,exp_mu
	real*8	val1,val2,val3
	real*8	coef,phi_val,d,g,dm1
	real*8	arg,exp_max,exp_min
	real*8	l_dip
        real*8  eps1,eps2
        real*8  l_Bjerrum,pi,Temp,KB,electron,Ang,Avogadro
        real*8  eps0,beta,C_charge,C_prop
c
	parameter       (exp_min = -80.d0, exp_max = 80.d0)
c
c	Common blocks
c ==========================================================================================
c
	common /dipoles/	vol,p0_e,coef_l
	common /solvent/        solv_conc,solv_dipole,solv_cvol,n_solv
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
	common /smpb/           ion_cvol,ratio,phi0,coef_Z0
        common /elec_const/     eps1,eps2
	common /psi_pot/	ptr_psi,flag_psi
	common /yukawa/		v0,b,psi_bulk
        common /constants/      l_Bjerrum,pi,Temp,KB,electron,Ang,
     1                          Avogadro,eps0,beta,C_charge,C_prop
c
c ==========================================================================================
c	Initialise
c ==========================================================================================
c
        fac_edens = -(1d-10/(beta*electron*electron))*eps0/2
c
c	Make sure all arrays are initialized to 0
c ==========================================================================================
c
	call set_all_zeros(nx,ny,nz,resid)
c
c ==========================================================================================
c	Define some parameters
c ==========================================================================================
c
	coef = coef_l * solv_cvol(1)/(1-phi0)
	l_dip = solv_cvol(1)/(1-phi0)
        if(l_dip.eq.0) phi0 = 0.d0
	do 100 i = 1,n_ion
		l_ion(i) = ion_cvol(i)/(1-phi0)
                if(l_dip.eq.0) l_ion(i) = 0.d0
		k_ion(i) = kappa(i)/(1-phi0)
100	continue
c
c ==========================================================================================
c	X-shifted map
c ==========================================================================================
c
c	Compute electric field on x-shifted grid
c
	do 600 k = 2,nz-1
c
	   hzm1 = z(k)-z(k-1)
	   hz = z(k+1)-z(k)
c
	   do 500 j = 2,ny-1
c
		hym1 = y(j) - y(j-1)
		hy = y(j+1)-y(j)
c
		do 400 i = 1,nx-1
c
			hx = x(i+1)-x(i)
c
                       	coef_oE  = (hym1+hy)*(hzm1+hz)/(4.d0*hx)
			coef_oEh = coef_oE*(phi(i,j,k)-phi(i+1,j,k))
c
                        eps_X = epsX(i,j,k)
c
			if(i.ne.1) then
				resid(i,j,k)=resid(i,j,k)+coef_oEh*eps_X
			endif
			if(i.ne.nx-1) then
				resid(i+1,j,k)=resid(i+1,j,k)
     1				-coef_oEh*eps_X
			endif
c
400		continue
500	   continue
c
600	continue
c
c ==========================================================================================
c	Y-shifted map
c ==========================================================================================
c
c	Compute electric field on Y-shifted grid
c
	do 1100 k = 2,nz-1
c
	   hzm1 = z(k)-z(k-1)
	   hz = z(k+1)-z(k)
c
	   do 1000 j = 1,ny-1
c
		hy = y(j+1)-y(j)
c
		do 900 i = 2,nx-1
c
			hxm1 = x(i)-x(i-1)
			hx = x(i+1)-x(i)
c
                       	coef_oN  = (hxm1+hx)*(hzm1+hz)/(4.d0*hy)
			coef_oNh = coef_oN*(phi(i,j,k)-phi(i,j+1,k))
c
                        eps_Y = epsY(i,j,k)
			if(j.ne.1) then
				resid(i,j,k)=resid(i,j,k)+coef_oNh*eps_Y
			endif
			if(j.ne.ny-1) then
				resid(i,j+1,k)=resid(i,j+1,k)
     1				-coef_oNh*eps_Y
			endif
c
900		continue
1000	   continue
c
1100	continue
c
c ==========================================================================================
c	Z-shifted map
c ==========================================================================================
c
c	Compute electric field on Z-shifted grid
c
	do 1700 k = 1,nz-1
c
	   hz = z(k+1)-z(k)
c
	   do 1600 j = 2,ny-1
c
		hy = y(j+1)-y(j)
		hym1 = y(j)-y(j-1)
c
		do 1500 i = 2,nx-1
c
			hxm1 = x(i)-x(i-1)
			hx = x(i+1)-x(i)
c
                       	coef_uC  = (hxm1+hx)*(hym1+hy)/(4.d0*hz)
			coef_uCh = coef_uC*(phi(i,j,k)-phi(i,j,k+1))
c
                        eps_Z = epsZ(i,j,k)
			if(k.ne.1) then
				resid(i,j,k)=resid(i,j,k)+coef_uCh*eps_Z
			endif
			if(k.ne.nz-1) then
				resid(i,j,k+1)=resid(i,j,k+1)
     1				-coef_uCh*eps_Z
			endif
c
1500		continue
1600	   continue
c
1700	continue
c
c       If we do not need the free energy, just check ions
c
        if(flag_free.eq.0) then
c
	        if(n_ion.eq.0) goto 2200
c
	        do 2100 k = 2,nz-1
c
	           hz = z(k+1)-z(k-1)
c
	           call plane_efield(phi(1,1,k),phi(1,1,k-1),
     1		   phi(1,1,k+1),nx,ny,x,y,z,k,Enorm)
c
	           do 2000 j = 2,ny-1
c
                        hy = y(j+1)-y(j-1)
c
		        do 1900 i = 2,nx-1
c
                            hx = x(i+1)-x(i-1)
c
			    if(cc(i,j,k).ne.0) then
c
				if(flag_psi.eq.1) then
					psi_val = psi(i,j,k)
					arg = -psi_val+psi_bulk
					if(arg.gt.exp_max) arg=exp_max
					if(arg.lt.exp_min) arg=exp_min
					psi_coef = exp(arg)
				else
					psi_coef = 1
				endif
c
                                dvol = hx*hy*hz/8.d0
c
				u_val = p0_e*Enorm(i,j)
c
				call compute_F1(u_val,F1_u,
     1					sinhc_u,exp_mu)
c
				phi_val = phi(i,j,k)
c
				d = exp_mu
                                g = 0
				do 1800 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val-u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d+l_ion(i_ion)*arg
					g=g+k_ion(i_ion)*arg
1800				continue
				d = d + l_dip*sinhc_u*psi_coef
				dm1 = 1/d
c
				resid(i,j,k)=resid(i,j,k)+cc(i,j,k)*
     1						g*dm1*dvol
				cc(i,j,k)=cc(i,j,k)*exp_mu*dm1/(1-phi0)
c
			    endif
c
1900		        continue
2000	           continue
c
2100	        continue
c
2200	        continue
c
        else
c
                free_edens  = 0
                free_charge = 0
                free_iondip = 0
c
	        do 2600 k = 2,nz-1
c
	           hz = z(k+1)-z(k-1)
c
	           call plane_efield(phi(1,1,k),phi(1,1,k-1),
     1		   phi(1,1,k+1),nx,ny,x,y,z,k,Enorm)
c
	           do 2500 j = 2,ny-1
c
                        hy = y(j+1)-y(j-1)
c
		        do 2400 i = 2,nx-1
c
                            hx = x(i+1)-x(i-1)
c
                            dvol = hx*hy*hz/8.d0
                            phi_val = phi(i,j,k)
c
                            free_edens = free_edens + Enorm(i,j)
     1                          * Enorm(i,j)*dvol
c
                            free_charge = free_charge+phi_val*fc(i,j,k)
     1                          *dvol
c
                            info1 = ibits(grid_info(i,j,k),0,2)
c
                            if(info1.ne.0) then
c
				if(flag_psi.eq.1) then
					psi_val = psi(i,j,k)
					arg = -psi_val+psi_bulk
					if(arg.gt.exp_max) arg=exp_max
					if(arg.lt.exp_min) arg=exp_min
					psi_coef = exp(arg)
				else
					psi_coef = 1
				endif
c
				u_val = p0_e*Enorm(i,j)
c
				call compute_F1(u_val,F1_u,
     1					sinhc_u,exp_mu)
c
				d = exp_mu
                                g = 0
				do 2300 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val-u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d+l_ion(i_ion)*arg
					g=g+k_ion(i_ion)*arg
2300				continue
				d = d + l_dip*sinhc_u*psi_coef
				dm1 = 1/d
c
				resid(i,j,k)=resid(i,j,k)+cc(i,j,k)*
     1						g*dm1*dvol
				cc(i,j,k)=cc(i,j,k)*exp_mu*dm1/(1-phi0)
c
                                free_iondip=free_iondip +
     1                                  log(d/exp_mu)*dvol
c
			    endif
c
2400		        continue
2500	           continue
c
2600	        continue
c
                free_edens = - fac_edens*free_edens
                free_charge = free_charge
                free_iondip = -free_iondip/vol
c
                free_ene = free_edens + free_charge + free_iondip
c
        endif
c
	do 2900 k = 2,nz-1
	        hz = z(k+1)-z(k-1)
		do 2800 j = 2,ny-1
	                hy = y(j+1)-y(j-1)
			do 2700 i = 2,nx-1
	                        hx = x(i+1)-x(i-1)
                                d = hx*hy*hz/8.d0
				resid(i,j,k) = -resid(i,j,k)+fc(i,j,k)*d
2700			continue
2800		continue
2900	continue
c
	return
	end
