c       Copyright (C) 2007 Patrice Koehl
c
c       This library is free software; you can redistribute it and/or
c       modify it under the terms of the GNU Lesser General Public
c       License as published by the Free Software Foundation; either
c       version 2.1 of the License, or (at your option) any later version.
c
c       This library is distributed in the hope that it will be useful,
c       but WITHOUT ANY WARRANTY; without even the implied warranty of
c       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
c       Lesser General Public License for more details.
c
c       You should have received a copy of the GNU Lesser General Public
c       License along with this library; if not, write to the Free Software
c       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
c
c ==========================================================================================
c ==========================================================================================
c	Langevin_tools.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This file contains a library of routines that are
c			useful for the Generalized PBL equation
c
c ==========================================================================================
c ==========================================================================================
c	langevin_scal.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine computes the residual of the discretized
c			GPBL equation at the specified vertex on the mesh
c
c	This is the standard version, that deals with any number of ions (up to 10)
c	and any valences for these ions
c
c	Input:
c		phi      : the electrostatic potential map
c		nx,ny,nz : size of the grid
c		x,y,z    : coordinates on the grid
c		grid_info: info on shifted grids
c               i,j,k    : position on grid considered
c
c	Output:
c               zNN     : residual at position i,j,k
c               DzNN    : Jacobian(1) at position i,j,k
c
	subroutine langevin_scal(i,j,k,phi,nx,ny,nz,
     1			x,y,z,grid_info,fc,cc,zNN,DzNN)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	include 'poisson.h'
c
	integer	nx,ny,nz
        integer i,j,k
	integer	info1
c
	real*8	eps_X,eps_Y,eps_Z
	real*8	coef_oE,coef_oN,coef_uC
	real*8	coef_oEm,coef_oNm,coef_uCm
	real*8	hxm1,hym1,hzm1
	real*8	hx,hy,hz
c
        real*8  zNN,DzNN
c
        real*8  eps_Xp,eps_Xm,eps_Yp,eps_Ym,eps_Zp,eps_Zm
        real*8  der_Xp,der_Xm,der_Yp,der_Ym,der_Zp,der_Zm
        real*8  Eip,Eim,Ejp,Ejm,Ekp,Ekm,Ec
        real*8  DerEip,DerEim,DerEjp,DerEjm,DerEkp,DerEkm
        real*8  phi_i,phi_im1,phi_j,phi_jm1,phi_k,phi_km1
c
        real*8  fc,cc
	real*8	phi(nx,ny,nz)
c
c	Information on grids
c ==========================================================================================
c
	real*8  	x(nx),y(ny),z(nz)
c
	integer*1       grid_info(nx,ny,nz)
c
c 	Parameter values
c ==========================================================================================
c
	integer	n_solv,n_ion,n_zz
c
	real*8	vol,p0_e,p0_e2,coef_l
	real*8	coef_Z0,phi0,ratio
	real*8  l_Bjerrum,pi,Temp,KB,electron,Ang,beta,Avogadro,eps0
	real*8  C_charge,C_prop,fac_edens
c
	real*8  ion_cvol(max_ion),l_ion(max_ion),k_ion(max_ion)
	real*8	solv_conc(max_solv)
	real*8  solv_dipole(max_solv),solv_cvol(max_solv)
c
	real*8  ion_conc(max_ion),ion_z(max_ion),ion_rad(max_ion)
	real*8  kappa(max_ion)
c
c	Local variables
c ==========================================================================================
c
	integer l,i_ion
c
	real*8	u_val,F1_u,G1_u,sinhc_u,val,exp_mu,exp_arg
	real*8	val1,val2,val3
	real*8	coef,phi_val,d,dm1,der_d
	real*8	der_d2,g,der_g
	real*8	arg,exp_max,exp_min
	real*8	l_dip,dvol
        real*8  eps1,eps2
	real*8	zp_max,zn_max,arg_max
c
	parameter       (exp_min = -80.d0, exp_max = 80.d0)
c
c	Common blocks
c ==========================================================================================
c
	common /dipoles/	vol,p0_e,coef_l
	common /solvent/        solv_conc,solv_dipole,solv_cvol,n_solv
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
	common /smpb/           ion_cvol,ratio,phi0,coef_Z0
	common /constants/ 	l_Bjerrum,pi,Temp,KB,electron,Ang,
     1                     	Avogadro,eps0,beta,C_charge,C_prop
        common /elec_const/     eps1,eps2
 
c
c ==========================================================================================
c	Initialise
c ==========================================================================================
c
        zNN  = 0
        DzNN = 0
c
c ==========================================================================================
c	Define some parameters
c ==========================================================================================
c
        p0_e2 = p0_e*p0_e
c
	coef = coef_l * solv_cvol(1)/(1-phi0)
	l_dip = solv_cvol(1)/(1-phi0)
	do 100 i_ion = 1,n_ion
		l_ion(i_ion) = ion_cvol(i_ion)/(1-phi0)
		k_ion(i_ion) = kappa(i_ion)/(1-phi0)
100	continue
c
	fac_edens = -(1d-10/(beta*electron*electron))*eps0/2
c
c ==========================================================================================
c	Define "neighborhood" of point (i,j,k)
c ==========================================================================================
c
        hzm1 = z(k)-z(k-1)
        hz = z(k+1)-z(k)
c
	hym1 = y(j) - y(j-1)
	hy = y(j+1)-y(j)
c
	hx = x(i+1)-x(i)
	hxm1 = x(i)-x(i-1)
c
       	coef_oE  = (hym1+hy)*(hzm1+hz)/(4.d0*hx)
       	coef_oEm  = (hym1+hy)*(hzm1+hz)/(4.d0*hxm1)
c
       	coef_oN  = (hxm1+hx)*(hzm1+hz)/(4.d0*hy)
       	coef_oNm  = (hxm1+hx)*(hzm1+hz)/(4.d0*hym1)
c
       	coef_uC  = (hxm1+hx)*(hym1+hy)/(4.d0*hz)
       	coef_uCm  = (hxm1+hx)*(hym1+hy)/(4.d0*hzm1)
c
c ==========================================================================================
c	Compute eps for the X-shifted map
c ==========================================================================================
c
c
c	Compute electric field on x-shifted grid
c
        call efield_X(phi(1,1,k),phi(1,1,k-1),phi(1,1,k+1),
     1		nx,ny,x,y,z,i,j,k,Eip,Eim,DerEip,DerEim)
c
	info1 = ibits(grid_info(i,j,k),2,2)
c
	if(info1.eq.0) then
c
		eps_Xp = eps1
                der_Xp = 0
c
	else
c
		u_val = p0_e*Eip
c
		call compute_F1G1(u_val,F1_u,G1_u,sinhc_u,exp_mu)
c
		val = coef*F1_u
c
		phi_val = (phi(i,j,k)+phi(i+1,j,k))/2
c
		d = exp_mu
		der_d = 0
c
		if(info1.ne.1) then
c
	                do 200 i_ion = 1,n_ion
	                        arg=-ion_z(i_ion)*phi_val -u_val
				if(arg.gt.exp_max) arg = exp_max
				if(arg.lt.exp_min) arg = exp_min
				arg = exp(arg)
				d=d+l_ion(i_ion)*arg
				der_d = der_d-0.5d0*ion_z(i_ion)
     1					*l_ion(i_ion)*arg
200			continue
c
		endif
c
		d = d + l_dip*sinhc_u
		dm1 = 1/d
c
		eps_Xp = eps2 + val*dm1
c
                Der_Xp=coef*dm1*p0_e2*DerEip*(G1_u-dm1*l_dip*F1_u*F1_u)
     1                  - val*dm1*dm1*der_d
c
	endif
c
	info1 = ibits(grid_info(i-1,j,k),2,2)
c
	if(info1.eq.0) then
c
		eps_Xm = eps1
                der_Xm = 0
c
	else
c
		u_val = p0_e*Eim
c
		call compute_F1G1(u_val,F1_u,G1_u,sinhc_u,exp_mu)
		val = coef*F1_u
c
		phi_val = (phi(i,j,k)+phi(i-1,j,k))/2
c
		d = exp_mu
		der_d = 0
c
		if(info1.ne.1) then
c
	                do 300 i_ion = 1,n_ion
	                        arg=-ion_z(i_ion)*phi_val -u_val
				if(arg.gt.exp_max) arg = exp_max
				if(arg.lt.exp_min) arg = exp_min
				arg = exp(arg)
				d=d+l_ion(i_ion)*arg
				der_d = der_d-0.5d0*ion_z(i_ion)
     1					*l_ion(i_ion)*arg
300			continue
c
		endif
c
		d = d + l_dip*sinhc_u
		dm1 = 1/d
c
		eps_Xm = eps2 + val*dm1
c
                Der_Xm=coef*dm1*p0_e2*DerEim*(G1_u-dm1*l_dip*F1_u*F1_u)
     1                  - val*dm1*dm1*der_d
c
	endif
c
        phi_i = phi(i,j,k) - phi(i+1,j,k)
        phi_im1=phi(i,j,k) - phi(i-1,j,k)
c
        DzNN = coef_oE*eps_Xp + coef_oEm*eps_Xm
     1          + coef_oE*der_Xp*phi_i
     2          + coef_oEm*der_Xm*phi_im1
c
        zNN = coef_oE*eps_Xp*phi_i
     1          + coef_oEm*eps_Xm*phi_im1
c
c ==========================================================================================
c	Define eps for the Y-shifted map
c ==========================================================================================
c
        call efield_Y(phi(1,1,k),phi(1,1,k-1),phi(1,1,k+1),
     1		nx,ny,x,y,z,i,j,k,Ejp,Ejm,DerEjp,DerEjm)
c
	info1 = ibits(grid_info(i,j,k),4,2)
c
	if(info1.eq.0) then
c
	        eps_Yp = eps1
                Der_Yp = 0
c
	else
c
		u_val = p0_e*Ejp
c
		call compute_F1G1(u_val,F1_u,G1_u,
     1				sinhc_u,exp_mu)
		val = coef*F1_u
c 
		phi_val = (phi(i,j,k)+phi(i,j+1,k))/2
c
		d = exp_mu
		der_d = 0
c
		if(info1.ne.1) then
c
                        do 400 i_ion = 1,n_ion
				arg = -ion_z(i_ion)*phi_val -u_val
				if(arg.gt.exp_max) arg = exp_max
				if(arg.lt.exp_min) arg = exp_min
				arg = exp(arg)
				d=d+l_ion(i_ion)*arg
				der_d = der_d-0.5d0*ion_z(i_ion)
     1					*l_ion(i_ion)*arg
400		        continue
c
		endif
c
		d = d + l_dip*sinhc_u
		dm1 = 1/d
c
		eps_Yp = eps2 + val*dm1
c
                Der_Yp=coef*dm1*p0_e2*DerEjp*(G1_u-dm1*l_dip*F1_u*F1_u)
     1                  - val*dm1*dm1*der_d
c
        endif
c
	info1 = ibits(grid_info(i,j-1,k),4,2)
c
	if(info1.eq.0) then
c
	        eps_Ym = eps1
                Der_Ym = 0
c
	else
c
		u_val = p0_e*Ejm
c
		call compute_F1G1(u_val,F1_u,G1_u,
     1				sinhc_u,exp_mu)
		val = coef*F1_u
c 
		phi_val = (phi(i,j,k)+phi(i,j-1,k))/2
c
		d = exp_mu
		der_d = 0
c
		if(info1.ne.1) then
c
                        do 500 i_ion = 1,n_ion
				arg = -ion_z(i_ion)*phi_val-u_val
				if(arg.gt.exp_max) arg = exp_max
				if(arg.lt.exp_min) arg = exp_min
				arg = exp(arg)
				d=d+l_ion(i_ion)*arg
				der_d = der_d-0.5d0*ion_z(i_ion)
     1					*l_ion(i_ion)*arg
500		        continue
c
		endif
c
		d = d + l_dip*sinhc_u
		dm1 = 1/d
c
		eps_Ym = eps2 + val*dm1
c
                Der_Ym=coef*dm1*p0_e2*DerEjm*(G1_u-dm1*l_dip*F1_u*F1_u)
     1                  - val*dm1*dm1*der_d
c
        endif
c
        phi_j = phi(i,j,k) - phi(i,j+1,k)
        phi_jm1=phi(i,j,k) - phi(i,j-1,k)
c
        DzNN = DzNN + coef_oN*eps_Yp + coef_oNm*eps_Ym
     1          + coef_oN*der_Yp*phi_j
     2          + coef_oNm*der_Ym*phi_jm1
c
        zNN = zNN + coef_oN*eps_Yp*phi_j
     1          + coef_oNm*eps_Ym*phi_jm1
c
c ==========================================================================================
c	Define eps for the Z-shifted map
c ==========================================================================================
c
	call efield_Z(phi(1,1,k),phi(1,1,k-1),phi(1,1,k+1),
     1		nx,ny,x,y,z,i,j,k,Ekp,Ekm,DerEkp,DerEkm)
c
	info1 = ibits(grid_info(i,j,k),6,2)
c
	if(info1.eq.0) then
c
		eps_Zp = eps1
                Der_Zp = 0
c
	else
c
		u_val = p0_e*Ekp
c
		call compute_F1G1(u_val,F1_u,G1_u,sinhc_u,exp_mu)
		val = coef*F1_u
c
		phi_val = (phi(i,j,k)+phi(i,j,k+1))/2
c
		d = exp_mu
		der_d = 0
c
		if(info1.ne.1) then
c
                        do 600 i_ion = 1,n_ion
				arg = -ion_z(i_ion)*phi_val -u_val
				if(arg.gt.exp_max) arg = exp_max
				if(arg.lt.exp_min) arg = exp_min
				arg = exp(arg)
				d=d + l_ion(i_ion)*arg
				der_d = der_d-0.5d0*ion_z(i_ion)
     1				*l_ion(i_ion)*arg
600		        continue
c
		endif
c
		d = d + l_dip*sinhc_u
		dm1 = 1/d
c
		eps_Zp = eps2 + val*dm1
c
                Der_Zp=coef*dm1*p0_e2*DerEkp*(G1_u-dm1*l_dip*F1_u*F1_u)
     1                  - val*dm1*dm1*der_d
c
        endif
c
	info1 = ibits(grid_info(i,j,k-1),6,2)
c
	if(info1.eq.0) then
c
		eps_Zm = eps1
                Der_Zm = 0
c
	else
c
		u_val = p0_e*Ekm
c
		call compute_F1G1(u_val,F1_u,G1_u,sinhc_u,exp_mu)
		val = coef*F1_u
c
		phi_val = (phi(i,j,k)+phi(i,j,k-1))/2
c
		d = exp_mu
		der_d = 0
c
		if(info1.ne.1) then
c
                        do 700 i_ion = 1,n_ion
				arg = -ion_z(i_ion)*phi_val -u_val
				if(arg.gt.exp_max) arg = exp_max
				if(arg.lt.exp_min) arg = exp_min
				arg = exp(arg)
				d=d + l_ion(i_ion)*arg
				der_d = der_d-0.5d0*ion_z(i_ion)
     1				*l_ion(i_ion)*arg
700		        continue
c
		endif
c
		d = d + l_dip*sinhc_u
		dm1 = 1/d
c
		eps_Zm = eps2 + val*dm1
c
                Der_Zm=coef*dm1*p0_e2*DerEkm*(G1_u-dm1*l_dip*F1_u*F1_u)
     1                  - val*dm1*dm1*der_d
c
        endif
c
        phi_k = phi(i,j,k) - phi(i,j,k+1)
        phi_km1=phi(i,j,k) - phi(i,j,k-1)
c
        DzNN = DzNN + coef_uC*eps_Zp + coef_uCm*eps_Zm
     1          + coef_uC*der_Zp*phi_k
     2          + coef_uCm*der_Zm*phi_km1
c
        zNN = zNN + coef_uC*eps_Zp*phi_k
     1          + coef_uCm*eps_Zm*phi_km1
c
c ==========================================================================================
c	Add ion, if any
c ==========================================================================================
c
	if(n_ion.eq.0) goto 1000
c
	call efield(phi(1,1,k),phi(1,1,k-1),
     1		phi(1,1,k+1),nx,ny,x,y,z,i,j,k,Ec)
c
        if(cc.ne.0) then
c
		u_val = p0_e*Ec
c
		call compute_F1(u_val,F1_u,sinhc_u,exp_mu)
c
		phi_val = phi(i,j,k)
c
		d = exp_mu
		der_d = 0
		der_g = 0
		g = 0
		do 800 i_ion = 1,n_ion
			arg=-ion_z(i_ion)*phi_val -u_val+arg_max
			if(arg.gt.exp_max) arg = exp_max
			if(arg.lt.exp_min) arg = exp_min
			arg = exp(arg)
			d=d+l_ion(i_ion)*arg
			g=g+k_ion(i_ion)*arg
			der_d = der_d-ion_z(i_ion)
     1				*l_ion(i_ion)*arg
			der_g = der_g -ion_z(i_ion)
     1				*k_ion(i_ion)*arg
800		continue
		d = d + l_dip*sinhc_u
		dm1 = 1/d
c
                zNN = zNN + cc*g*dm1
                DzNN = DzNN + cc*dm1*(der_g-g*der_d*dm1)
c
	endif
c
1000	continue
c
        zNN = zNN - fc
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	Compute_F1G1.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine computes the function F1(u) = Langevin(u)*sinhc(u)
c			as well as G1(u) = F1'(u)/u
c
	subroutine compute_F1G1(u_val,F1_u,G1_u,sinhc_u,exp_mu)
c
	real*8	u_val,u2,F1_u,sinhc_u,G1_u
	real*8	tol
	real*8	sinh_u,cosh_u,exp_2mu,exp_mu
	real*8	fast_exp
c
	parameter	(tol=1.d-04)
c
	if(u_val.le.tol) then
		F1_u = 1.d0/3.d0
                G1_u = 1.d0/15.d0
		sinhc_u = 1.d0
		exp_mu = 1.d0
        else
		exp_mu = exp(-u_val)
		exp_2mu = exp_mu*exp_mu/2.d0
		sinh_u = 0.5d0 - exp_2mu
		cosh_u = 0.5d0 + exp_2mu
		sinhc_u = sinh_u/u_val
		u2 = u_val*u_val
		F1_u = (cosh_u-sinhc_u)/u2
                G1_u = (sinhc_u - 3*F1_u)/u2
	endif
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	matrix_gpbl.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine is the driver to the routine that
c			computes the residual of the GPBL equation
c
c	Input:
c		phi      : the electrostatic potential map
c		nx,ny,nz : size of the grid
c		x,y,z    : coordinates on the grid
c		grid_info: infor on regular grid and shifted grids
c
c	This routine considers two cases; z:z electrolytes, and more general
c	systems
c
c	Output:
c		resid	: residual vector
c
	subroutine matrix_gpbl(phi,nx,ny,nz,x,y,z,grid_info,
     2			Enorm,cc,resid)
c
	integer	nx,ny,nz
	integer	n_ion,n_zz
c
        integer*1 grid_info(*)
c
	real*8	phi(*),Enorm(*),cc(*),resid(*)
        real*8  x(*),y(*),z(*)
c
	call matrix_langevin(phi,nx,ny,nz,x,y,z,grid_info,
     1		Enorm,cc,resid)
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	matrix_langevin.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine is the driver to the routine that
c			computes the residual of the GPBL equation
c
c	Input:
c		phi      : the electrostatic potential map
c		nx,ny,nz : size of the grid
c		x,y,z    : coordinates on the grid
c		grid_info: infor on regular grid and shifted grids
c
c	This is the generic routine, for any number of ions (up to 10), with
c	any valence
c
	subroutine matrix_langevin(phi,nx,ny,nz,x,y,z,grid_info,
     2			Enorm,cc,resid)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	include 'poisson.h'
c
	integer	nx,ny,nz
        integer info1
c
	real*8	eps_X,eps_Y,eps_Z
	real*8	coef_oE,coef_oN,coef_uC
	real*8	coef_oEh,coef_oNh,coef_uCh
	real*8	hxm1,hym1,hzm1
	real*8	hx,hy,hz
	real*8	free_edens,free_charge,free_iondip,free_energy,dvol
c
	real*8	phi(nx,ny,nz)
	real*8	resid(nx,ny,nz)
	real*8	cc(nx,ny,nz)
c
	real*8	Enorm(nx,ny)
c
c	Information on grids
c ==========================================================================================
c
	real*8  	x(nx),y(ny),z(nz)
c
	integer*1       grid_info(nx,ny,nz)
c
c 	Parameter values
c ==========================================================================================
c
	integer	n_solv,n_ion,n_zz
c
	real*8	vol,p0_e,coef_l
	real*8	coef_Z0,phi0,ratio
c
	real*8  ion_cvol(max_ion),l_ion(max_ion),k_ion(max_ion)
	real*8	solv_conc(max_solv)
	real*8  solv_dipole(max_solv),solv_cvol(max_solv)
c
	real*8  ion_conc(max_ion),ion_z(max_ion),ion_rad(max_ion)
	real*8  kappa(max_ion)
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k,l,i_ion
c
	real*8	u_val,F1_u,sinhc_u,val,exp_mu
	real*8	val1,val2,val3
	real*8	coef,phi_val,d,g,dm1
	real*8	arg,exp_max,exp_min
	real*8	l_dip
        real*8  eps1,eps2
c
	parameter       (exp_min = -80.d0, exp_max = 80.d0)
c
c	Common blocks
c ==========================================================================================
c
	common /dipoles/	vol,p0_e,coef_l
	common /solvent/        solv_conc,solv_dipole,solv_cvol,n_solv
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
	common /smpb/           ion_cvol,ratio,phi0,coef_Z0
        common /elec_const/     eps1,eps2
c
c ==========================================================================================
c	Initialise
c ==========================================================================================
c
c	Make sure all arrays are initialized to 0
c ==========================================================================================
c
	call set_all_zeros(nx,ny,nz,resid)
c
c ==========================================================================================
c	Define some parameters
c ==========================================================================================
c
	coef = coef_l * solv_cvol(1)/(1-phi0)
	l_dip = solv_cvol(1)/(1-phi0)
	do 100 i = 1,n_ion
		l_ion(i) = ion_cvol(i)/(1-phi0)
		k_ion(i) = kappa(i)/(1-phi0)
100	continue
c
c ==========================================================================================
c	Compute eps for the X-shifted map
c ==========================================================================================
c
c	Compute electric field on x-shifted grid
c
	do 600 k = 2,nz-1
c
	   hzm1 = z(k)-z(k-1)
	   hz = z(k+1)-z(k)
c
	   call plane_efield_X(phi(1,1,k),phi(1,1,k-1),phi(1,1,k+1),
     1		nx,ny,x,y,z,k,Enorm)
c
	   do 500 j = 2,ny-1
c
		hym1 = y(j) - y(j-1)
		hy = y(j+1)-y(j)
c
		do 400 i = 1,nx-1
c
			hx = x(i+1)-x(i)
c
                       	coef_oE  = (hym1+hy)*(hzm1+hz)/(4.d0*hx)
			coef_oEh = coef_oE*(phi(i,j,k)-phi(i+1,j,k))
c
			info1 = ibits(grid_info(i,j,k),2,2)
c
			if(info1.eq.0) then
c
                                eps_X = eps1
c
			else
c
				u_val = p0_e*Enorm(i,j)
c
				call compute_F1(u_val,F1_u,
     1					sinhc_u,exp_mu)
				val = coef*F1_u
c
				phi_val = (phi(i,j,k)+phi(i+1,j,k))/2
c
				d = exp_mu
c
				if(info1.ne.1) then
c
				    do 200 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val -u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d+l_ion(i_ion)*arg
200				    continue
c
				endif
c
				d = d + l_dip*sinhc_u
				dm1 = 1/d
c
				eps_X = eps2 + val*dm1
c
			endif
c
			if(i.ne.1) then
				resid(i,j,k)=resid(i,j,k)+coef_oEh*eps_X
			endif
			if(i.ne.nx-1) then
				resid(i+1,j,k)=resid(i+1,j,k)
     1				-coef_oEh*eps_X
			endif
c
400		continue
500	   continue
c
600	continue
c
c ==========================================================================================
c	Define eps for the Y-shifted map
c ==========================================================================================
c
c	Compute electric field on Y-shifted grid
c
	do 1100 k = 2,nz-1
c
	   hzm1 = z(k)-z(k-1)
	   hz = z(k+1)-z(k)
c
	   call plane_efield_Y(phi(1,1,k),phi(1,1,k-1),phi(1,1,k+1),
     1		nx,ny,x,y,z,k,Enorm)
c
	   do 1000 j = 1,ny-1
c
		hy = y(j+1)-y(j)
c
		do 900 i = 2,nx-1
c
			hxm1 = x(i)-x(i-1)
			hx = x(i+1)-x(i)
c
                       	coef_oN  = (hxm1+hx)*(hzm1+hz)/(4.d0*hy)
			coef_oNh = coef_oN*(phi(i,j,k)-phi(i,j+1,k))
c
			info1 = ibits(grid_info(i,j,k),4,2)
c
			if(info1.eq.0) then
c
				eps_Y = eps1
c
			else
c
				u_val = p0_e*Enorm(i,j)
c
				call compute_F1(u_val,F1_u,
     1				sinhc_u,exp_mu)
				val = coef*F1_u
c 
				phi_val = (phi(i,j,k)+
     1				phi(i,j+1,k))/2
c
				d = exp_mu
c
				if(info1.ne.1) then
c
				    do 700 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val-u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d+l_ion(i_ion)*arg
700				    continue
c
				endif
c
				d = d + l_dip*sinhc_u
				dm1 = 1/d
				eps_Y = eps2 + val*dm1
c
			endif
c
			if(j.ne.1) then
				resid(i,j,k)=resid(i,j,k)+coef_oNh*eps_Y
			endif
			if(j.ne.ny-1) then
				resid(i,j+1,k)=resid(i,j+1,k)
     1				-coef_oNh*eps_Y
			endif
c
900		continue
1000	   continue
c
1100	continue
c
c ==========================================================================================
c	Define eps for the Z-shifted map
c ==========================================================================================
c
c	Compute electric field on Z-shifted grid
c
	do 1700 k = 1,nz-1
c
	   hz = z(k+1)-z(k)
c
	   call plane_efield_Z(phi(1,1,k),phi(1,1,k+1),
     1		nx,ny,x,y,z,k,Enorm)
c
	   do 1600 j = 2,ny-1
c
		hy = y(j+1)-y(j)
		hym1 = y(j)-y(j-1)
c
		do 1500 i = 2,nx-1
c
			hxm1 = x(i)-x(i-1)
			hx = x(i+1)-x(i)
c
                       	coef_uC  = (hxm1+hx)*(hym1+hy)/(4.d0*hz)
			coef_uCh = coef_uC*(phi(i,j,k)-phi(i,j,k+1))
c
			info1 = ibits(grid_info(i,j,k),6,2)
c
			if(info1.eq.0) then
c
				eps_Z = eps1
c
			else
c
				u_val = p0_e*Enorm(i,j)
c
				call compute_F1(u_val,F1_u,
     1				sinhc_u,exp_mu)
				val = coef*F1_u
c
				phi_val = (phi(i,j,k)+
     1				phi(i,j,k+1))/2
c
				d = exp_mu
c
				if(info1.ne.1) then
c
				    do 1200 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val-u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d + l_ion(i_ion)*arg
1200				    continue
c
				endif
c
				d = d + l_dip*sinhc_u
				dm1 = 1/d
c
				eps_Z = eps2 + val*dm1
c
			endif
c
			if(k.ne.1) then
				resid(i,j,k)=resid(i,j,k)+coef_uCh*eps_Z
			endif
			if(k.ne.nz-1) then
				resid(i,j,k+1)=resid(i,j,k+1)
     1				-coef_uCh*eps_Z
			endif
c
1500		continue
1600	   continue
c
1700	continue
c
	if(n_ion.eq.0) goto 2200
c
	do 2100 k = 2,nz-1
c
	   call plane_efield(phi(1,1,k),phi(1,1,k-1),
     1		phi(1,1,k+1),nx,ny,x,y,z,k,Enorm)
c
	   do 2000 j = 2,ny-1
c
		do 1900 i = 2,nx-1
c
			if(cc(i,j,k).ne.0) then
c
				u_val = p0_e*Enorm(i,j)
c
				call compute_F1(u_val,F1_u,
     1					sinhc_u,exp_mu)
c
				phi_val = phi(i,j,k)
c
				d = exp_mu
                                g = 0
				do 1800 i_ion = 1,n_ion
					arg=-ion_z(i_ion)*phi_val-u_val
					if(arg.gt.exp_max) arg = exp_max
					if(arg.lt.exp_min) arg = exp_min
					arg = exp(arg)
					d=d+l_ion(i_ion)*arg
					g=g+k_ion(i_ion)*arg
1800				continue
				d = d + l_dip*sinhc_u
				dm1 = 1/d
c
				resid(i,j,k)=resid(i,j,k)+cc(i,j,k)*
     1						g*dm1
c
			endif
c
1900		continue
2000	   continue
c
2100	continue
c
2200	continue
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	Efield.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine computes the Electric field at one
c                       specific position of the grid
c
c	Input:
c		phi      : the electrostatic potential map
c		nx,ny,nz : current size of the grid
c		i,j,k	 : position considered
c
c       See plane_efield (in Langevin.f) for details
c
	subroutine efield(phik,phikm1,phikp1,nx,ny,hx,hy,hz,
     1                  i,j,k,Enorm)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	include 'poisson.h'
c
c	Input/Output variables
c ==========================================================================================
c
	integer	nx,ny
	integer	i,j,k
c
	real*8	phik(nx,ny),phikm1(nx,ny),phikp1(nx,ny)
	real*8	Enorm
c
c	Information on grid
c ==========================================================================================
c
	real*8  hx(*),hy(*),hz(*)
c
c	Local variables
c ==========================================================================================
c
	real*8	dx,dy,dz,val
c
c ==========================================================================================
c	Compute e-field in x direction
c ==========================================================================================
c
	dx = 1.d0/(hx(i+1)-hx(i-1))
	val = (phik(i+1,j)-phik(i-1,j))*dx
	Enorm = val*val
c
c ==========================================================================================
c	Compute e-field in y direction
c ==========================================================================================
c
	dy = 1.d0/(hy(j+1)-hy(j-1))
	val = (phik(i,j+1)-phik(i,j-1))*dy
	Enorm=Enorm + val*val
c
c ==========================================================================================
c	Compute e-field in z direction
c ==========================================================================================
c
	dz = 1.d0/(hz(k+1)-hz(k-1))
	val = (phikp1(i,j)-phikm1(i,j))*dz
	Enorm=Enorm + val*val
c
c ==========================================================================================
c	We currently have: ||E||^2; get ||E||
c ==========================================================================================
c
        Enorm = sqrt(Enorm)
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c ==========================================================================================
c	Efield_X.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine computes the Electric field in the
c                       neighborhood of a point (i,j,k) on the grid
c                       (at position (i-1/2,j,k) and (i+1/2,j,k)
c
c       See plnae_efield_X (in Langevin.f) for details
c
c	Input:
c		phi      : the electrostatic potential map
c		nx,ny,nz : current size of the grid
c               i,j,k    : position considered
c
	subroutine efield_X(phik,phikm1,phikp1,nx,ny,hx,hy,
     1                  hz,i,j,k,Eip,Eim,DerEip,DerEim)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	include 'poisson.h'
c
c	Input/Output variables
c ==========================================================================================
c
	integer	nx,ny
        integer i,j,k
c
        real*8  Eip,Eim,DerEip,DerEim
	real*8	phik(nx,ny),phikm1(nx,ny),phikp1(nx,ny)
c
c	Information on grid
c ==========================================================================================
c
	real*8  hx(*),hy(*),hz(*)
c
c	Local variables
c ==========================================================================================
c
	real*8	dx,dy,dz,val
c
c ==========================================================================================
c	Compute e-field in x direction
c ==========================================================================================
c
c	We have shifted the grid in the X-direction: the Electric field in this direction
c	at index i is therefore simply phi(i+1,j,k) - phi(i,j,k), up to i = nx-1 
c
	dx = 1.d0/(hx(i+1)-hx(i))
	val = (phik(i+1,j)-phik(i,j))*dx
	Eip = val*val
        DerEip = -dx*val
	dx = 1.d0/(hx(i)-hx(i-1))
	val = (phik(i,j)-phik(i-1,j))*dx
	Eim = val*val
        DerEim = dx*val
c
c ==========================================================================================
c	Compute e-field in y direction
c ==========================================================================================
c
	dy = 1.d0/(2*(hy(j+1)-hy(j-1)))
	val = (phik(i,j+1)+phik(i+1,j+1)-phik(i,j-1)
     1			-phik(i+1,j-1))*dy
	Eip=Eip + val*val
	val = (phik(i-1,j+1)+phik(i,j+1)-phik(i-1,j-1)
     1			-phik(i,j-1))*dy
	Eim=Eim + val*val
c
c ==========================================================================================
c	Compute e-field in z direction
c ==========================================================================================
c
	dz = 1.d0/(2.d0*(hz(k+1)-hz(k-1)))
	val = (phikp1(i,j)+phikp1(i+1,j)-
     1		phikm1(i,j)-phikm1(i+1,j))*dz
	Eip=Eip + val*val
	val = (phikp1(i-1,j)+phikp1(i,j)-
     1		phikm1(i-1,j)-phikm1(i,j))*dz
	Eim=Eim + val*val
c
c ==========================================================================================
c	We currently have: ||E||^2; get ||E||
c ==========================================================================================
c
        Eip = sqrt(Eip)
        Eim = sqrt(Eim)
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	Efield_Y.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine computes the Electric field in the
c                       neighborhood of grid point i,j,k (in fact, a position
c                       (i,j-1/2,k) and (i,j+1/2,k)
c
c	Input:
c		phi      : the electrostatic potential map
c		nx,ny,nz : current size of the grid
c               i,j,k    : position considered
c
	subroutine efield_Y(phik,phikm1,phikp1,nx,ny,hx,hy,hz,
     1                  i,j,k,Ejp,Ejm,DerEjp,DerEjm)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
c	Input/Output variables
c ==========================================================================================
c
	integer	nx,ny
        integer i,j,k
c
        real*8  Ejp,Ejm,DerEjp,DerEjm
	real*8	phik(nx,ny),phikm1(nx,ny),phikp1(nx,ny)
c
c	Information on grid
c ==========================================================================================
c
	real*8  hx(*),hy(*),hz(*)
c
c	Local variables
c ==========================================================================================
c
	real*8	dx,dy,dz,val
c
c ==========================================================================================
c	Compute e-field in y direction
c ==========================================================================================
c
c	We have shifted the grid in the Y-direction: the Electric field in this direction
c	at index j is therefore simply phi(i,j+1,k) - phi(i,j,k), up to j = ny-1 
c
	dy = 1.d0/(hy(j+1)-hy(j))
	val = (phik(i,j+1)-phik(i,j))*dy
        Ejp = val*val
        DerEjp = -val*dy
	dy = 1.d0/(hy(j)-hy(j-1))
	val = (phik(i,j)-phik(i,j-1))*dy
        Ejm = val*val
        DerEjm = val*dy
c
c ==========================================================================================
c	Compute e-field in x direction
c ==========================================================================================
c
	dx = 1.d0/(2*(hx(i+1)-hx(i-1)))
	val = (phik(i+1,j)+phik(i+1,j+1)-phik(i-1,j)
     1			-phik(i-1,j+1))*dx
	Ejp=Ejp + val*val
	val = (phik(i+1,j-1)+phik(i+1,j)-phik(i-1,j-1)
     1			-phik(i-1,j))*dx
	Ejm=Ejm + val*val
c
c ==========================================================================================
c	Compute e-field in z direction
c ==========================================================================================
c
	dz = 1.d0/(2*(hz(k+1)-hz(k-1)))
	val = (phikp1(i,j)+phikp1(i,j+1)-
     1		phikm1(i,j)-phikm1(i,j+1))*dz
	Ejp=Ejp + val*val
	val = (phikp1(i,j-1)+phikp1(i,j)-
     1		phikm1(i,j-1)-phikm1(i,j))*dz
	Ejm=Ejm + val*val
c
c ==========================================================================================
c	We currently have: ||E||^2; get ||E||
c ==========================================================================================
c
        Ejp = sqrt(Ejp)
        Ejm = sqrt(Ejm)
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	Efield_Z.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      9/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	This subroutine computes the Electric field in the
c                       neighborhood of point (i,j,k) on the grid
c                       (at position (i,j,k-1/2) and (i,j,k+1/2))
c
c	Input:
c		phi      : the electrostatic potential map
c		nx,ny,nz : current size of the grid
c               i,j,k    : position considered
c
	subroutine efield_Z(phik,phikm1,phikp1,nx,ny,hx,hy,hz,
     1                  i,j,k,Ekp,Ekm,DerEkp,DerEkm)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	include 'poisson.h'
c
c	Input/Output variables
c ==========================================================================================
c
	integer	nx,ny
	integer	i,j,k
c
        real*8  Ekp,Ekm,DerEkp,DerEkm
	real*8	phik(nx,ny),phikp1(nx,ny),phikm1(nx,ny)
c
c	Information on grid
c ==========================================================================================
c
	real*8  hx(*),hy(*),hz(*)
c
c	Local variables
c ==========================================================================================
c
	real*8	dx,dy,dz,val
c
c ==========================================================================================
c	Compute e-field in z direction
c ==========================================================================================
c
c	We have shifted the grid in the Z-direction: the Electric field in this direction
c	at index (i,j) is therefore simply phi(i,j,k+1) - phi(i,j,k), up to k = nz-1 
c
	dz = 1.d0/(hz(k+1)-hz(k))
	val = (phikp1(i,j)-phik(i,j))*dz
	Ekp = val*val
        DerEkp = -val*dz
	dz = 1.d0/(hz(k)-hz(k-1))
	val = (phik(i,j)-phikm1(i,j))*dz
	Ekm = val*val
        DerEkm = val*dz
c
c ==========================================================================================
c	Compute e-field in x direction
c ==========================================================================================
c
	dx = 1.d0/(2*(hx(i+1)-hx(i-1)))
	val = (phik(i+1,j)+phikp1(i+1,j)-phik(i-1,j)
     1			-phikp1(i-1,j))*dx
	Ekp=Ekp + val*val
	val = (phikm1(i+1,j)+phik(i+1,j)-phikm1(i-1,j)
     1			-phik(i-1,j))*dx
	Ekm=Ekm + val*val
c
c ==========================================================================================
c	Compute e-field in y direction
c ==========================================================================================
c
	dy=1.d0/(2*(hy(j+1)-hy(j-1)))
	val = (phik(i,j+1)+phikp1(i,j+1)-
     1		phik(i,j-1)-phikp1(i,j-1))*dy
	Ekp=Ekp + val*val
	val = (phikm1(i,j+1)+phik(i,j+1)-
     1		phikm1(i,j-1)-phik(i,j-1))*dy
	Ekm=Ekm + val*val
c
c ==========================================================================================
c	We currently have: ||E||^2; get ||E||
c ==========================================================================================
c
        Ekp = sqrt(Ekp)
        Ekm = sqrt(Ekm)
c
	return
	end
