c       Copyright (C) 2007 Patrice Koehl
c
c       This library is free software; you can redistribute it and/or
c       modify it under the terms of the GNU Lesser General Public
c       License as published by the Free Software Foundation; either
c       version 2.1 of the License, or (at your option) any later version.
c
c       This library is distributed in the hope that it will be useful,
c       but WITHOUT ANY WARRANTY; without even the implied warranty of
c       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
c       Lesser General Public License for more details.
c
c       You should have received a copy of the GNU Lesser General Public
c       License along with this library; if not, write to the Free Software
c       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
c
c ==========================================================================================
c ==========================================================================================
c	NlinSmoothing.f
c
c	Version 1:      7/4/07
c
c	Author: Patrice Koehl (in collaboration with Marc Delarue)
c
c	This file contains a set of routines for non linear smoothing
c
c ==========================================================================================
c ==========================================================================================
c
c
c ==========================================================================================
c ==========================================================================================
c	nonlin_smoothing.f
c
c	This is the driver routine for the smoothing procedure. We only implement the fast
c	diagonal iterative method
c ==========================================================================================
c ==========================================================================================
c
	subroutine nonlin_smoothing(nx,ny,nz,ipc,rpc,ac,cc,fc,x,w1,w2,r,
     1		itmax,iters,errtol,omega,iresid,iadjoint,meth)
c
c ==========================================================================================
c	Declaration
c ==========================================================================================
c
	integer	nx,ny,nz,itmax,iters,iresid,iadjoint,meth
c
	integer	ipc(*)
c
	real*8	omega,errtol
	real*8	rpc(*),ac(nx*ny*nz,*),cc(nx,ny,nz),fc(nx,ny,nz)
	real*8	x(nx,ny,nz),w1(nx,ny,nz),w2(nx,ny,nz)
	real*8	r(nx,ny,nz)
c
c ==========================================================================================
c	Call the appropriate smoothing routine
c ==========================================================================================
c
	if(meth.eq.1) then
		call NGaussSeidel_RB(nx,ny,nz,ipc,rpc,ac,cc,fc,x,
     1		w1,w2,r,itmax,iters,errtol,omega,iresid,iadjoint)
	elseif(meth.eq.2) then
		call NWJacobi(nx,ny,nz,ipc,rpc,ac,cc,fc,x,
     1		w1,w2,r,itmax,iters,errtol,omega,iresid,iadjoint)
	elseif(meth.eq.3) then
		call NSOR(nx,ny,nz,ipc,rpc,ac,cc,fc,x,
     1		w1,w2,r,itmax,iters,errtol,omega,iresid,iadjoint)
	elseif(meth.eq.4) then
		write(6,*) 'Nonlin CG for smoothing not yet implemented'
		stop
	elseif(meth.eq.5) then
		call Nrichardson(nx,ny,nz,ipc,rpc,ac,cc,fc,x,
     1		w1,w2,r,itmax,iters,errtol,omega,iresid,iadjoint)
	else
		write(6,*) 'Bad smoothing method'
		stop
	endif
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	NGaussSeidel_RB.f
c
c	This is the driver routine for the Non-linear Gauss Seidel red-black smoothing procedure. 
c ==========================================================================================
c ==========================================================================================
c
	subroutine NGaussSeidel_RB(nx,ny,nz,ipc,rpc,ac,cc,fc,x,w1,w2,r,
     1		itmax,iters,errtol,omega,iresid,iadjoint)
c
c ==========================================================================================
c	Declaration
c ==========================================================================================
c
	integer	nx,ny,nz,itmax,iters,iresid,iadjoint
	integer	numdia
c
	integer	ipc(*)
c
	real*8	omega,errtol
	real*8	rpc(*),ac(nx*ny*nz,*),cc(nx,ny,nz),fc(nx,ny,nz)
	real*8	x(nx,ny,nz),w1(nx,ny,nz),w2(nx,ny,nz)
	real*8	r(nx,ny,nz)
c
c ==========================================================================================
c	Call the proper routine based on Stencil size
c ==========================================================================================
c
	numdia = ipc(11)
c
	if(numdia.eq.7) then
		call NGaussSeidel7(nx,ny,nz,ipc,rpc,ac(1,1),cc,fc,
     1		ac(1,2),ac(1,3),ac(1,4),
     2		x,w1,w2,r,
     3		itmax,iters,errtol,omega,iresid,iadjoint)
	elseif(numdia.eq.27) then
		call NGaussSeidel27(nx,ny,nz,ipc,rpc,ac(1,1),cc,fc,
     1		ac(1,2),ac(1,3),ac(1,4),ac(1,5),ac(1,6),ac(1,7),ac(1,8),
     1		ac(1,9),ac(1,10),ac(1,11),ac(1,12),ac(1,13),ac(1,14),
     2		x,w1,w2,r,
     3		itmax,iters,errtol,omega,iresid,iadjoint)
	else
		write(6,*) 'In GaussSeidel_RB : bad stencil type...'
	endif
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	NGaussSeidel7.f
c
c	7-diagonal non linear Gauss Seidel routine
c ==========================================================================================
c ==========================================================================================
c
	subroutine NGaussSeidel7(nx,ny,nz,ipc,rpc,
     1			oC,cc,fc,oE,oN,uC,
     2			x,w1,w2,r,itmax,iters,errtol,omega,iresid,
     3			iadjoint)
c
c ==========================================================================================
c	Declaration
c ==========================================================================================
c
	integer	nx,ny,nz,itmax,iters,iresid,iadjoint
	integer	i,j,k,i1,i2,j1,j2,k1,k2,istep
	integer	nitmax,niters,ifail_tol,ipkey
c
	integer	ipc(*)
c
	real*8	omega,errtol
	real*8	c_scal,dc_scal
	real*8	UU,AA,BB,FF,zNN,DzNN,zntol,zndenom,znerror
	real*8	change
        real*8  xnorm1
c
	real*8	rpc(*)
	real*8	oE(nx,ny,nz),oN(nx,ny,nz),uC(nx,ny,nz)
	real*8	fc(nx,ny,nz),oC(nx,ny,nz),cc(nx,ny,nz)
	real*8	x(nx,ny,nz),w1(nx,ny,nz),w2(nx,ny,nz)
	real*8	r(nx,ny,nz)
c
	nitmax	  = 10
	zntol     = 1.d-5
	ifail_tol = 0
	ipkey	  = ipc(10)
c
c ==========================================================================================
c	Do Gauss Seidel iterations itmax times
c ==========================================================================================
c
	i1 = (1-iadjoint) * 2 + iadjoint	*(nx-1)
	i2 = iadjoint     * 2 + (1-iadjoint)	*(nx-1)
	j1 = (1-iadjoint) * 2 + iadjoint	*(ny-1)
	j2 = iadjoint     * 2 + (1-iadjoint)	*(ny-1)
	k1 = (1-iadjoint) * 2 + iadjoint	*(nz-1)
	k2 = iadjoint     * 2 + (1-iadjoint)	*(nz-1)
c
	istep = iadjoint*(-1) + (1-iadjoint)*(1)
c
	do 700 iters = 1,itmax
c
		do 600 k=k1,k2,istep
		   do 500 j = j1,j2,istep
			do 400 i = i1,i2,istep
c
c				For this unknown, use a 1D Newton method
c
				UU = x(i,j,k)
				AA = oC(i,j,k)
				BB = cc(i,j,k)
				FF = fc(i,j,k) 
     1				+ oN(i,j,k)	* x(i,j+1,k)
     2				+ oN(i,j-1,k)	* x(i,j-1,k)
     3				+ oE(i,j,k)	* x(i+1,j,k)
     4				+ oE(i-1,j,k)	* x(i-1,j,k)
     5				+ uC(i,j,k)	* x(i,j,k+1)
     6				+ uC(i,j,k-1)	* x(i,j,k-1)
c
c				Evaluate residual of 1D system
c
				zNN = AA*UU + c_scal(BB,UU,ipkey)-FF
c
c				Setup for Newton method
c
				niters = 0
				zndenom = zNN
				if(zndenom.eq.0.d0) zndenom = 1.d0
				znerror = zNN / zndenom
c
				if(znerror.lt.zntol) goto 300
				if(niters.gt.nitmax) goto 200
c
100				continue
c
					niters = niters + 1
c
					DzNN = AA + dc_scal(BB,UU,ipkey)
c
					change = - zNN / DzNN
					change = sign(min(abs(change),
     1					10.d0),change)
c
					UU = UU + change
c
					zNN=AA*UU+c_scal(BB,UU,ipkey)-FF
c
					znerror = zNN/zndenom
c
					if(znerror.lt.zntol) goto 300
					if(niters.gt.nitmax) goto 200
c
				goto 100
c
200				continue
				ifail_tol = ifail_tol + 1
c
300				continue
c
				x(i,j,k) = omega*UU + (1-omega)*x(i,j,k)
c
400			continue
c
500		    continue
c
600		continue
c
700	continue
c
c	If needed, return the new residual
c
	if(iresid.eq.1) then
		call residual7(nx,ny,nz,ipc,rpc,oC,cc,fc,oE,oN,uC,
     1                  x,r,w1)
	endif
c
c	if(ifail_tol.gt.0) then
c		write(6,*) 
c     1          'Nonlinear GSRB: 1D Newton tol. failures: ',ifail_tol
c	endif
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	NGaussSeidel27.f
c
c	27-diagonal non linear Gauss Seidel routine
c ==========================================================================================
c ==========================================================================================
c
	subroutine NGaussSeidel27(nx,ny,nz,ipc,rpc,
     1			oC,cc,fc,oE,oN,uC,
     2			oNE,oNW,uE,uW,uN,uS,uNE,uNW,uSE,uSW,
     2			x,w1,w2,r,itmax,iters,errtol,omega,iresid,
     3			iadjoint)
c
c ==========================================================================================
c	Declaration
c ==========================================================================================
c
	integer	nx,ny,nz,itmax,iters,iresid,iadjoint
	integer	i,j,k,i1,i2,j1,j2,k1,k2,istep
	integer	nitmax,niters,ifail_tol,ipkey
c
	integer	ipc(*)
c
	real*8	omega,errtol
	real*8	UU,AA,BB,FF,zNN,DzNN,zntol,zndenom,znerror
	real*8	change
	real*8	c_scal,dc_scal
	real*8	xnorm1
c
	real*8	tmp0,tmpU,tmpD
	real*8	rpc(*)
	real*8	oE(nx,ny,nz),oN(nx,ny,nz),uC(nx,ny,nz)
	real*8	oNE(nx,ny,nz),oNW(nx,ny,nz),uE(nx,ny,nz)
	real*8	uW(nx,ny,nz),uN(nx,ny,nz),uS(nx,ny,nz)
	real*8	uNE(nx,ny,nz),uNW(nx,ny,nz),uSE(nx,ny,nz)
	real*8	uSW(nx,ny,nz)
	real*8	fc(nx,ny,nz),oC(nx,ny,nz),cc(nx,ny,nz)
	real*8	x(nx,ny,nz),w1(nx,ny,nz),w2(nx,ny,nz)
	real*8	r(nx,ny,nz)
c
	real*8	val1,val2,val3,eps
c
	nitmax	  = 10
	zntol     = 1.d-5
	ifail_tol = 0
	ipkey	  = ipc(10)
c
	eps=1.d-4
c
c ==========================================================================================
c	Do Gauss Seidel iterations itmax times
c ==========================================================================================
c
	i1 = (1-iadjoint) * 2 + iadjoint	*(nx-1)
	i2 = iadjoint     * 2 + (1-iadjoint)	*(nx-1)
	j1 = (1-iadjoint) * 2 + iadjoint	*(ny-1)
	j2 = iadjoint     * 2 + (1-iadjoint)	*(ny-1)
	k1 = (1-iadjoint) * 2 + iadjoint	*(nz-1)
	k2 = iadjoint     * 2 + (1-iadjoint)	*(nz-1)
c
	istep = iadjoint*(-1) + (1-iadjoint)*(1)
c
	call residual27(nx,ny,nz,ipc,rpc,oC,cc,fc,
     1		oE,oN,uC,oNE,oNW,uE,uW,uN,uS,uNE,uNW,uSE,uSW,
     3		x,r,w1)
c
	do 700 iters = 1,itmax
c
		do 600 k=k1,k2,istep
		   do 500 j = j1,j2,istep
			do 500 i = i1,i2,istep
c
				UU = x(i,j,k)
				AA = oC(i,j,k)
				BB = cc(i,j,k)
c
				tmp0 =
     1				+ oN(i,j,k)	* x(i,j+1,k)
     2				+ oN(i,j-1,k)	* x(i,j-1,k)
     3				+ oE(i,j,k)	* x(i+1,j,k)
     4				+ oE(i-1,j,k)	* x(i-1,j,k)
     3				+ oNE(i,j,k)	* x(i+1,j+1,k)
     4				+ oNE(i-1,j-1,k)* x(i-1,j-1,k)
     3				+ oNW(i,j,k)	* x(i-1,j+1,k)
     4				+ oNW(i+1,j-1,k)* x(i+1,j-1,k)
c
				tmpU =
     1				+ uC(i,j,k)	* x(i,j,k+1)
     2				+ uN(i,j,k)	* x(i,j+1,k+1)
     3				+ uS(i,j,k)	* x(i,j-1,k+1)
     4				+ uE(i,j,k)	* x(i+1,j,k+1)
     3				+ uW(i,j,k)	* x(i-1,j,k+1)
     4				+ uNE(i,j,k)	* x(i+1,j+1,k+1)
     3				+ uNW(i,j,k)	* x(i-1,j+1,k+1)
     4				+ uSE(i,j,k)	* x(i+1,j-1,k+1)
     4				+ uSW(i,j,k)	* x(i-1,j-1,k+1)
c
				tmpD =
     1				+ uC(i,j,k-1)	* x(i,j,k-1)
     2				+ uS(i,j+1,k-1)	* x(i,j+1,k-1)
     3				+ uN(i,j-1,k-1)	* x(i,j-1,k-1)
     4				+ uW(i+1,j,k-1)	* x(i+1,j,k-1)
     3				+ uE(i-1,j,k-1)	* x(i-1,j,k-1)
     4				+ uSW(i+1,j+1,k-1)* x(i+1,j+1,k-1)
     3				+ uSE(i-1,j+1,k-1)* x(i-1,j+1,k-1)
     4				+ uNW(i+1,j-1,k-1)* x(i+1,j-1,k-1)
     4				+ uNE(i-1,j-1,k-1)* x(i-1,j-1,k-1)
c
				FF = fc(i,j,k) + tmp0 + tmpU + tmpD
c
c				Evaluate residual of 1D system
c
				zNN = AA*UU+c_scal(BB,UU,ipkey)-FF
c
c				Setup for Newton method
c
				niters = 0
				zndenom = zNN
				if(zndenom.eq.0.d0) zndenom = 1.d0
				znerror = zNN / zndenom
c
				if(znerror.lt.zntol) goto 300
				if(niters.gt.nitmax) goto 200
c
100				continue
c
					niters = niters + 1
c
					DzNN = AA + dc_scal(BB,UU,ipkey)
c
					change = - zNN / DzNN
c
					UU = UU + change
c
					zNN=AA*UU+c_scal(BB,UU,ipkey)-FF
c
					znerror = zNN/zndenom
c
					if(znerror.lt.zntol) goto 300
					if(niters.gt.nitmax) goto 200
c
				goto 100
c
200				continue
				ifail_tol = ifail_tol + 1
c
300				continue
c
				x(i,j,k) = UU
c
c
400			continue
500		    continue
600		continue
c
700	continue
c
c	If needed, return the new residual
c
	if(iresid.eq.1) then
		call residual27(nx,ny,nz,ipc,rpc,oC,cc,fc,
     1			oE,oN,uC,oNE,oNW,uE,uW,uN,uS,uNE,uNW,uSE,uSW,
     3			x,r,w1)
	endif
c
c	if(ifail_tol.gt.0) then
c		write(6,*) 
c     1          'Nonlinear GSRB: 1D Newton tol. failures: ',ifail_tol
c		write(6,*) 'nx,ny,nz:',nx,ny,nz
c	endif
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	conjgrad_nlin.f
c
c	This routine solves the system of equation Ax=b using conjugate gradient
c ==========================================================================================
c ==========================================================================================
c
	subroutine conjgrad_nlin(nx,ny,nz,ipc,rpc,ac,cc,fc,x,p,ap,r,
     1		zk,zkp1,tmp,
     2		itmax,iters,errtol,omega,iresid,iadjoint)
c
c ==========================================================================================
c	Declaration
c ==========================================================================================
c
	integer	nx,ny,nz,itmax,iters,iresid,iadjoint
	integer	i,j,k
c
	integer	ipc(*)
c
	real*8	omega,errtol,rsnrm,pAp,denom,xnorm2
	real*8	rhok1,rhok2,alpha,beta,dotvect
        real*8  dot,betam1
c
	real*8	rpc(*),ac(nx,ny,nz,*),fc(nx,ny,nz),cc(nx,ny,nz)
	real*8	x(nx,ny,nz),r(nx,ny,nz),p(nx,ny,nz),ap(nx,ny,nz)
	real*8	zk(nx,ny,nz),zkp1(nx,ny,nz),tmp(nx,ny,nz)
c
c ==========================================================================================
c	Setup
c ==========================================================================================
c
	iters = 0
	if(iters.ge.itmax.and.iresid.eq.0) return
c
	call residual(nx,ny,nz,ipc,rpc,ac,cc,fc,x,r,zk)
c
	denom = xnorm2(nx,ny,nz,r)
c
	if(denom.eq.0.d0) return
	if(iters.ge.itmax) return
c
c ==========================================================================================
c	ConjGrad loop
c ==========================================================================================
c
100	continue
c
c		Compute current stopping criteria
c
		rhok2 = dotvect(nx,ny,nz,r,r)
		rsnrm = dsqrt(rhok2)
c
		if(rsnrm/denom.le.errtol) goto 200
		if(iters.ge.itmax) goto 200
c
c		Form new direction vector
c
		if(iters.eq.0) then
			call vectcopy(nx,ny,nz,r,p)
		else
			beta = rhok2/rhok1
                        betam1 = 1.d0/beta
			call vectsmult(nx,ny,nz,betam1,r,p)
			call scalevect(nx,ny,nz,beta,p)
		endif
c
c		Find alpha that minimizes error
c
		call vectcopy(nx,ny,nz,tmp)
		call linesearch(nx,ny,nz,alpha,
     2			ipc,rpc,ac,cc,fc,p,x,tmp,ap,zk,zkp1)
c
c		Save rhok2
c
		rhok1 = rhok2
c
c		Update solution in direction p of length alpha
c
		call vectsmult(nx,ny,nz,alpha,p,x)
c
c		Update residual
c
		call vectsmult(nx,ny,nz,(-alpha),ap,r)
		call vectsmult(nx,ny,nz,(1.d0),zk,r)
		call vectsmult(nx,ny,nz,(-1.d0),zkp1,r)
c
		iters = iters + 1
c
		goto 100
c
200	continue
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	NWJacobi.f
c
c	This is the driver routine for the non linear Weighted Jacobi smoothing procedure. 
c ==========================================================================================
c ==========================================================================================
c
	subroutine NWJacobi(nx,ny,nz,ipc,rpc,ac,cc,fc,x,w1,w2,r,
     1		itmax,iters,errtol,omega,iresid,iadjoint)
c
c ==========================================================================================
c	Declaration
c ==========================================================================================
c
	integer	nx,ny,nz,itmax,iters,iresid,iadjoint
	integer	numdia
c
	integer	ipc(*)
c
	real*8	omega,errtol
	real*8	rpc(*),ac(nx*ny*nz,*),cc(nx,ny,nz),fc(nx,ny,nz)
	real*8	x(nx,ny,nz),w1(nx,ny,nz),w2(nx,ny,nz)
	real*8	r(nx,ny,nz)
c
c ==========================================================================================
c	Call the proper routine based on Stencil size
c ==========================================================================================
c
	numdia = ipc(11)
c
	if(numdia.eq.7) then
		call NWJacobi7(nx,ny,nz,ipc,rpc,ac(1,1),cc,fc,
     1		ac(1,2),ac(1,3),ac(1,4),
     2		x,w1,w2,r,
     3		itmax,iters,errtol,omega,iresid,iadjoint)
	elseif(numdia.eq.27) then
		call NWJacobi27(nx,ny,nz,ipc,rpc,ac(1,1),cc,fc,
     1		ac(1,2),ac(1,3),ac(1,4),ac(1,5),ac(1,6),ac(1,7),ac(1,8),
     1		ac(1,9),ac(1,10),ac(1,11),ac(1,12),ac(1,13),ac(1,14),
     2		x,w1,w2,r,
     3		itmax,iters,errtol,omega,iresid,iadjoint)
	else
		write(6,*) 'In GaussSeidel_RB : bad stencil type...'
	endif
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	NWJacobi7.f
c
c	7-diagonal Non linear Weighted Jacobi routine
c ==========================================================================================
c ==========================================================================================
c
	subroutine NWJacobi7(nx,ny,nz,ipc,rpc,
     1			oC,cc,fc,oE,oN,uC,
     2			x,w1,w2,r,itmax,iters,errtol,omega,iresid,
     3			iadjoint)
c
c ==========================================================================================
c	Declaration
c ==========================================================================================
c
	integer	nx,ny,nz,itmax,iters,iresid,iadjoint
	integer	i,j,k
	integer	nitmax,niters,ifail_tol,ipkey
c
	integer	ipc(*)
c
	real*8	omega,errtol,fac
c
	real*8	UU,AA,BB,FF,zNN,DzNN,zntol,zndenom,znerror,change
	real*8	c_scal,dc_scal
	real*8	rpc(*)
	real*8	oE(nx,ny,nz),oN(nx,ny,nz),uC(nx,ny,nz)
	real*8	fc(nx,ny,nz),oC(nx,ny,nz),cc(nx,ny,nz)
	real*8	x(nx,ny,nz),w1(nx,ny,nz),w2(nx,ny,nz)
	real*8	r(nx,ny,nz)
c
	nitmax	  = 10
	zntol     = 1.d-5
	ifail_tol = 0
	ipkey	  = ipc(10)
c
c ==========================================================================================
c	Do Jacobi iterations itmax times
c ==========================================================================================
c
	do 700 iters = 1,itmax
c
		do 600 k=2,nz-1
		   do 500 j = 2,ny-1
			do 400 i = 2,nx-1
c
				UU = x(i,j,k)
				AA = oC(i,j,k)
				BB = cc(i,j,k)
				FF = fc(i,j,k)
     1				+ oN(i,j,k)	* x(i,j+1,k)
     2				+ oN(i,j-1,k)	* x(i,j-1,k)
     3				+ oE(i,j,k)	* x(i+1,j,k)
     4				+ oE(i-1,j,k)	* x(i-1,j,k)
     5				+ uC(i,j,k)	* x(i,j,k+1)
     6				+ uC(i,j,k-1)	* x(i,j,k-1)
c
c				Evaluate residual of 1D system
c
				zNN = AA*UU + c_scal(BB,UU,ipkey)-FF
c
c				Setup for Newton method
c
				niters = 0
				zndenom = zNN
				if(zndenom.eq.0.d0) zndenom = 1.d0
				znerror = zNN / zndenom
c
				if(znerror.lt.zntol) goto 300
				if(niters.gt.nitmax) goto 200
c
100				continue
c
					niters = niters + 1
c
					DzNN = AA + dc_scal(BB,UU,ipkey)
c
					change = - zNN / DzNN
c
					UU = UU + change
c
					zNN=AA*UU+c_scal(BB,UU,ipkey)-FF
c
					znerror = zNN/zndenom
c
					if(znerror.lt.zntol) goto 300
					if(niters.gt.nitmax) goto 200
c
				goto 100
c
200				continue
				ifail_tol = ifail_tol + 1
c
300				continue
c
				w1(i,j,k) = UU
c
400			continue
500		   continue
600		continue
c
700	continue
c
        fac = 1.d0 - omega
c
        do 1000 k = 2,nz-1
                do 900 j = 2,ny-1
                        do 800 i = 2,nx-1
                                x(i,j,k)=omega*w1(i,j,k)+fac*x(i,j,k)
800                     continue
900             continue
1000    continue
c
c	If needed, return the new residual
c
	if(iresid.eq.1) then
		call residual7(nx,ny,nz,ipc,rpc,oC,cc,fc,oE,oN,uC,
     1                  x,r,w1)
	endif
c
	if(ifail_tol.gt.0) then
		write(6,*) 
     1          'Nonlinear GSRB: 1D Newton tol. failures: ',ifail_tol
	endif
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	WJacobi27.f
c
c	27-diagonal Weighted Jacobi routine
c ==========================================================================================
c ==========================================================================================
c
	subroutine NWJacobi27(nx,ny,nz,ipc,rpc,
     1			oC,cc,fc,oE,oN,uC,
     2			oNE,oNW,uE,uW,uN,uS,uNE,uNW,uSE,uSW,
     2			x,w1,w2,r,itmax,iters,errtol,omega,iresid,
     3			iadjoint)
c
c ==========================================================================================
c	Declaration
c ==========================================================================================
c
	integer	nx,ny,nz,itmax,iters,iresid,iadjoint
	integer	i,j,k
	integer	nitmax,niters,ifail_tol,ipkey
c
	integer	ipc(*)
c
	real*8	omega,errtol,fac
c
	real*8	tmp0,tmpU,tmpD
	real*8	c_scal,dc_scal
	real*8	UU,AA,BB,FF,zNN,DzNN,zntol,zndenom,znerror,change
	real*8	rpc(*)
	real*8	oE(nx,ny,nz),oN(nx,ny,nz),uC(nx,ny,nz)
	real*8	oNE(nx,ny,nz),oNW(nx,ny,nz),uE(nx,ny,nz)
	real*8	uW(nx,ny,nz),uN(nx,ny,nz),uS(nx,ny,nz)
	real*8	uNE(nx,ny,nz),uNW(nx,ny,nz),uSE(nx,ny,nz)
	real*8	uSW(nx,ny,nz)
	real*8	fc(nx,ny,nz),oC(nx,ny,nz),cc(nx,ny,nz)
	real*8	x(nx,ny,nz),w1(nx,ny,nz),w2(nx,ny,nz)
	real*8	r(nx,ny,nz)
c
	nitmax	  = 10
	zntol     = 1.d-5
	ifail_tol = 0
	ipkey	  = ipc(10)
c
c ==========================================================================================
c	Do Gauss Seidel iterations itmax times
c ==========================================================================================
c
	do 700 iters = 1,itmax
c
		do 600 k=2,nz-1
		   do 500 j = 2,ny-1
			do 400 i = 2,nx-1
c
				UU = x(i,j,k)
				AA = oC(i,j,k)
				BB = cc(i,j,k)
c
				tmp0 =
     1				+ oN(i,j,k)	* x(i,j+1,k)
     2				+ oN(i,j-1,k)	* x(i,j-1,k)
     3				+ oE(i,j,k)	* x(i+1,j,k)
     4				+ oE(i-1,j,k)	* x(i-1,j,k)
     3				+ oNE(i,j,k)	* x(i+1,j+1,k)
     4				+ oNE(i-1,j-1,k)* x(i-1,j-1,k)
     3				+ oNW(i,j,k)	* x(i-1,j+1,k)
     4				+ oNW(i+1,j-1,k)* x(i+1,j-1,k)
c
				tmpU =
     1				+ uC(i,j,k)	* x(i,j,k+1)
     2				+ uN(i,j,k)	* x(i,j+1,k+1)
     3				+ uS(i,j,k)	* x(i,j-1,k+1)
     4				+ uE(i,j,k)	* x(i+1,j,k+1)
     3				+ uW(i,j,k)	* x(i-1,j,k+1)
     4				+ uNE(i,j,k)	* x(i+1,j+1,k+1)
     3				+ uNW(i,j,k)	* x(i-1,j+1,k+1)
     4				+ uSE(i,j,k)	* x(i+1,j-1,k+1)
     4				+ uSW(i,j,k)	* x(i-1,j-1,k+1)
c
				tmpD =
     1				+ uC(i,j,k-1)	* x(i,j,k-1)
     2				+ uS(i,j+1,k-1)	* x(i,j+1,k-1)
     3				+ uN(i,j-1,k-1)	* x(i,j-1,k-1)
     4				+ uW(i+1,j,k-1)	* x(i+1,j,k-1)
     3				+ uE(i-1,j,k-1)	* x(i-1,j,k-1)
     4				+ uSW(i+1,j+1,k-1)* x(i+1,j+1,k-1)
     3				+ uSE(i-1,j+1,k-1)* x(i-1,j+1,k-1)
     4				+ uNW(i+1,j-1,k-1)* x(i+1,j-1,k-1)
     4				+ uNE(i-1,j-1,k-1)* x(i-1,j-1,k-1)
c
				FF = fc(i,j,k) + tmp0 + tmpU + tmpD
c
c				Evaluate residual of 1D system
c
				zNN = AA*UU + c_scal(BB,UU,ipkey)-FF
c
c				Setup for Newton method
c
				niters = 0
				zndenom = zNN
				if(zndenom.eq.0.d0) zndenom = 1.d0
				znerror = zNN / zndenom
c
				if(znerror.lt.zntol) goto 300
				if(niters.gt.nitmax) goto 200
c
100				continue
c
					niters = niters + 1
c
					DzNN = AA + dc_scal(BB,UU,ipkey)
c
					change = - zNN / DzNN
c
					UU = UU + change
c
					zNN=AA*UU+c_scal(BB,UU,ipkey)-FF
c
					znerror = zNN/zndenom
c
					if(znerror.lt.zntol) goto 300
					if(niters.gt.nitmax) goto 200
c
				goto 100
c
200				continue
				ifail_tol = ifail_tol + 1
c
300				continue
c
				w1(i,j,k) = UU
c
c
400			continue
500		    continue
600		continue
c
700	continue
c
        fac = 1.d0 - omega
c
        do 1000 k = 2,nz-1
                do 900 j = 2,ny-1
                        do 800 i = 2,nx-1
                                x(i,j,k)=omega*w1(i,j,k)+fac*x(i,j,k)
800                     continue
900             continue
1000     continue
c
c	If needed, return the new residual
c
	if(iresid.eq.1) then
		call residual27(nx,ny,nz,ipc,rpc,oC,cc,fc,
     1			oE,oN,uC,oNE,oNW,uE,uW,uN,uS,uNE,uNW,uSE,uSW,
     3			x,r,w1)
	endif
c
	if(ifail_tol.gt.0) then
		write(6,*) 
     1          'Nonlinear GSRB: 1D Newton tol. failures: ',ifail_tol
	endif
c
        return
        end
c
c ==========================================================================================
c ==========================================================================================
c	NSOR.f
c
c	This is the driver routine for the Non linear SOR procedure.
c ==========================================================================================
c ==========================================================================================
c
	subroutine nsor(nx,ny,nz,ipc,rpc,ac,cc,fc,x,w1,w2,r,
     1		itmax,iters,errtol,omega,iresid,iadjoint)
c
c ==========================================================================================
c	Declaration
c ==========================================================================================
c
	integer	nx,ny,nz,itmax,iters,iresid,iadjoint
	integer	numdia
c
	integer	ipc(*)
c
	real*8	omega,errtol
	real*8	rpc(*),ac(nx*ny*nz,*),cc(nx,ny,nz),fc(nx,ny,nz)
	real*8	x(nx,ny,nz),w1(nx,ny,nz),w2(nx,ny,nz)
	real*8	r(nx,ny,nz)
c
c ==========================================================================================
c	Call the proper routine based on Stencil size
c ==========================================================================================
c
	numdia = ipc(11)
c
	if(numdia.eq.7) then
		call nsor7(nx,ny,nz,ipc,rpc,ac(1,1),cc,fc,
     1		ac(1,2),ac(1,3),ac(1,4),
     2		x,w1,w2,r,
     3		itmax,iters,errtol,omega,iresid,iadjoint)
	elseif(numdia.eq.27) then
		call nsor27(nx,ny,nz,ipc,rpc,ac(1,1),cc,fc,
     1		ac(1,2),ac(1,3),ac(1,4),ac(1,5),ac(1,6),ac(1,7),ac(1,8),
     1		ac(1,9),ac(1,10),ac(1,11),ac(1,12),ac(1,13),ac(1,14),
     2		x,w1,w2,r,
     3		itmax,iters,errtol,omega,iresid,iadjoint)
	else
		write(6,*) 'In GaussSeidel_RB : bad stencil type...'
	endif
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	nsor7.f
c
c	7-diagonal SOR routine
c ==========================================================================================
c ==========================================================================================
c
	subroutine nsor7(nx,ny,nz,ipc,rpc,
     1			oC,cc,fc,oE,oN,uC,
     2			x,w1,w2,r,itmax,iters,errtol,omega,iresid,
     3			iadjoint)
c
c ==========================================================================================
c purpose:
c
c       fast 7 diagonal nonlinear red/black sor routine.
c
c       note that the interior grid points from 2,...,nx-1, etc.
c       we then begin coloring with a red point, and do a point red/black
c       coloring.
c
c       the red points are:
c
c       if ((j even) and (k even)) then
c               begin row at first point, i=2 (i even), or ioff = 0
c       else if ((j odd) and (k even)) then
c               begin row at second point, i=3 (i odd), or ioff = 1
c       else if ((j even) and (k odd)) then
c               begin row at second point, i=3 (i odd), or ioff = 1
c       else if ((j odd) and (k odd)) then
c               begin row at first point, i=2 (i even), or ioff = 0
c       endif
c       then: begin row at:  i=2+ioff
c
c       the appropriate ioff function for the red points is then:
c               ioff = dabs(mod(j,2)-mod(k,2))
c
c       the appropriate ioff function for the black points is then:
c               ioff = 1 - dabs(mod(j,2)-mod(k,2))
c
c       alternatively, the red points are:
c
c       those whose indices add up to an even number.
c       to see this, consider that all surrounding points are
c       only one index different, hence the sum will differ by one.
c       thus, if a given point has an even sum, then the surrounding
c       points will have an odd sum.
c
c       the black points are:
c
c       those whose indices add up to an odd number.
c
c       algorithm parameters:
c       =====================
c
c       U0      = initial guess
c       Llin    = matrix representing linear part of equations
c       F       = vector representing rhs of equations
c       nx      = number of unknowns
c       itmax   = number of iterations to perform
c       c_scal  = function representing nonlinear part of equations
c       dc_scal = function representing derivative of nonlinear
c                 part of equations
c       U       = final result
c
c       algorithm idea:
c       ===============
c
c       the ith nonlinear equation is represented as:
c
c       N(UU) = AA * UU + c_scal(UU) - FF
c
c       which is solved for UU using newton method, where:
c
c       AA = ith diagonal entry of Llin
c       FF = F(i), minus the nondiag of ith eqn of Llin times current U
c       UU = the unknown, initialized to U(i)
c
c ==========================================================================================
c	Declaration
c ==========================================================================================
c
	integer	nx,ny,nz,itmax,iters,iresid,iadjoint
	integer	i,j,k,i1,i2,j1,j2,k1,k2,istep,ioff
        integer nitmax,niters,ifail_tol,ipkey
c
	integer	ipc(*)
c
	real*8	omega,errtol,fac
c
        real*8  UU,AA,BB,FF,zNN,DzNN,zntol,zndenom,znerror
        real*8  c_scal,dc_scal,change
c
	real*8	rpc(*)
	real*8	oE(nx,ny,nz),oN(nx,ny,nz),uC(nx,ny,nz)
	real*8	fc(nx,ny,nz),oC(nx,ny,nz),cc(nx,ny,nz)
	real*8	x(nx,ny,nz),w1(nx,ny,nz),w2(nx,ny,nz)
	real*8	r(nx,ny,nz)
c
c ==========================================================================================
c	Nonlinear iteration tolerance and itmax
c ==========================================================================================
c
        nitmax          = 10
        zntol           = 1.d-5
        ifail_tol       = 0
        ipkey           = ipc(10)
c
        fac     = 1.0d0 - omega
c
	do 1300 iters = 1,itmax
c
		do 600 k=2,nz-1
c
		   do 500 j = 2,ny-1
c
                        ioff = mod((j+k+2),2)
c
			do 400 i = 2+ioff,nx-1,2
c
c                               For this unknown, do a 1D Newton method
c
c                               Determine the 1D equation
c
                                UU = x(i,j,k)
                                AA = oC(i,j,k)
                                BB = cc(i,j,k)
c
				FF = fc(i,j,k) 
     1				+ oN(i,j,k)	* x(i,j+1,k)
     2				+ oN(i,j-1,k)	* x(i,j-1,k)
     3				+ oE(i,j,k)	* x(i+1,j,k)
     4				+ oE(i-1,j,k)	* x(i-1,j,k)
     5				+ uC(i,j,k)	* x(i,j,k+1)
     6				+ uC(i,j,k-1)	* x(i,j,k-1)
c
c                               Evaluate residual of 1D system
c
                                zNN = AA * UU + c_scal(BB,UU,ipkey) - FF
c
c                               Setup for Newton method
c
                                niters = 0
                                zndenom = zNN
                                if(zndenom.eq.0.0d0) zndenom = 1.0d0
                                znerror = zNN / zndenom
c
c                               the 1D Newton method
c
                                if(znerror.lt.zntol) goto 300
                                if(niters.gt.nitmax) goto 200
c
100                             continue
c
                                        niters = niters + 1
c
c                                       Construct Jacobian matrix of NN
c
                                        DzNN = AA + dc_scal(BB,UU,ipkey)
c
c                                       Solve the linear system
c
                                        change = - zNN / DzNN
					change=sign(min(abs(change),
     1					5.d0),change)
c
c                                       Update the solution
c
                                        UU = UU + change
c
c                                       Evaluate residual of 1D system
c
                                        zNN=AA*UU+c_scal(BB,UU,ipkey)-FF
c
c                                       Compute error
c
                                        znerror = zNN / zndenom
c
c                                       tolerance and itmax check
c
                                        if(znerror.lt.zntol) goto 300
                                        if(niters.gt.nitmax) goto 200
c
c                                       Loop
c
                                goto 100
c
c                               tolerance not reached
c
200                             continue
c
                                ifail_tol = ifail_tol + 1
c
c                               tolerance reached
c
300                             continue
c
c                               Newton method complete: update solution
c
                                x(i,j,k) = x(i,j,k)*fac + omega*UU
c
400                     continue
c
500                 continue
c
600             continue
c
c               Do the black points
c
		do 1200 k=2,nz-1
c
		   do 1100 j = 2,ny-1
c
                        ioff = 1-mod((j+k+2),2)
c
			do 1000 i = 2+ioff,nx-1,2
c
c                               For this unknown, do a 1D Newton method
c
c                               Determine the 1D equation
c
                                UU = x(i,j,k)
                                AA = oC(i,j,k)
                                BB = cc(i,j,k)
c
				FF = fc(i,j,k) 
     1				+ oN(i,j,k)	* x(i,j+1,k)
     2				+ oN(i,j-1,k)	* x(i,j-1,k)
     3				+ oE(i,j,k)	* x(i+1,j,k)
     4				+ oE(i-1,j,k)	* x(i-1,j,k)
     5				+ uC(i,j,k)	* x(i,j,k+1)
     6				+ uC(i,j,k-1)	* x(i,j,k-1)
c
c                               Evaluate residual of 1D system
c
                                zNN = AA * UU + c_scal(BB,UU,ipkey) - FF
c
c                               Setup for Newton method
c
                                niters = 0
                                zndenom = zNN
                                if(zndenom.eq.0.0d0) zndenom = 1.0d0
                                znerror = zNN / zndenom
c
c                               the 1D Newton method
c
                                if(znerror.lt.zntol) goto 900
                                if(niters.gt.nitmax) goto 800
c
700                             continue
c
                                        niters = niters + 1
c
c                                       Construct Jacobian matrix of NN
c
                                        DzNN = AA + dc_scal(BB,UU,ipkey)
c
c                                       Solve the linear system
c
                                        change = - zNN / DzNN
					change=sign(min(abs(change),
     1					5.d0),change)
c
c                                       Update the solution
c
                                        UU = UU + change
c
c                                       Evaluate residual of 1D system
c
                                        zNN=AA*UU+c_scal(BB,UU,ipkey)-FF
c
c                                       Compute error
c
                                        znerror = zNN / zndenom
c
c                                       tolerance and itmax check
c
                                        if(znerror.lt.zntol) goto 900
                                        if(niters.gt.nitmax) goto 800
c
c                                       Loop
c
                                goto 700
c
c                               tolerance not reached
c
800                             continue
c
                                ifail_tol = ifail_tol + 1
c
c                               tolerance reached
c
900                             continue
c
c                               Newton method complete: update solution
c
                                x(i,j,k) = x(i,j,k)*fac + omega*UU
c
1000                     continue
c
1100                 continue
c
1200             continue
c

1300	continue
c
c	If needed, return the new residual
c
	if(iresid.eq.1) then
		call residual7(nx,ny,nz,ipc,rpc,oC,cc,fc,oE,oN,uC,
     1          x,r,w1)
	endif
c
c        if(ifail_tol.gt.0) then
c                write(6,*)'NSOR: 1D Newton tolerance failures: ',
c     1			ifail_tol
c        endif
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	nsor27.f
c
c	27-diagonal non-linear SOR routine
c ==========================================================================================
c ==========================================================================================
c
	subroutine nsor27(nx,ny,nz,ipc,rpc,
     1			oC,cc,fc,oE,oN,uC,
     2			oNE,oNW,uE,uW,uN,uS,uNE,uNW,uSE,uSW,
     2			x,w1,w2,r,itmax,iters,errtol,omega,iresid,
     3			iadjoint)
c
c ==========================================================================================
c       purpose:
c
c       27 diagonal nonlinear sor routine.
c
c       algorithm parameters:
c       ---------------------
c
c       U0      = initial guess
c       Llin    = matrix representing linear part of equations
c       F       = vector representing rhs of equations
c       nx      = number of unknowns
c       itmax   = number of iterations to perform
c       c_scal  = function representing nonlinear part of equations
c       dc_scal = function representing derivative of nonlinear
c                 part of equations
c       U       = final result
c
c       algorithm idea:
c       ---------------
c
c       the ith nonlinear equation is represented as:
c
c       N(UU) = AA * UU + c_scal(UU) - FF
c
c       which is solved for UU using Newton method, where:
c
c       AA = ith diagonal entry of Llin
c       FF = F(i), minus the nondiag of ith eqn of Llin times current U
c       UU = the unknown, initialized to U(i)
c
c ==========================================================================================
c	Declaration
c ==========================================================================================
c
	integer	nx,ny,nz,itmax,iters,iresid,iadjoint
	integer	i,j,k,i1,i2,j1,j2,k1,k2,istep
        integer nitmax,niters,ifail_tol,ipkey
c
	integer	ipc(*)
c
	real*8	omega,errtol,fac
        real*8  xnorm1
c
        real*8  c_scal,dc_scal
        real*8  UU,AA,BB,FF,zNN,DzNN,zntol,zndenom,znerror
        real*8  change
	real*8	tmp0,tmpU,tmpD
	real*8	rpc(*)
	real*8	oE(nx,ny,nz),oN(nx,ny,nz),uC(nx,ny,nz)
	real*8	oNE(nx,ny,nz),oNW(nx,ny,nz),uE(nx,ny,nz)
	real*8	uW(nx,ny,nz),uN(nx,ny,nz),uS(nx,ny,nz)
	real*8	uNE(nx,ny,nz),uNW(nx,ny,nz),uSE(nx,ny,nz)
	real*8	uSW(nx,ny,nz)
	real*8	fc(nx,ny,nz),oC(nx,ny,nz),cc(nx,ny,nz)
	real*8	x(nx,ny,nz),w1(nx,ny,nz),w2(nx,ny,nz)
	real*8	r(nx,ny,nz)
c
c       Non linear iteration tolerance and itmax
c
        nitmax  = 10
        zntol   = 1.d-5
        ifail_tol = 0
        ipkey = ipc(10)
c
        fac = 1.d0 - omega
c
	do 700 iters = 1,itmax
c
		do 600 k=2,nz-1
		   do 500 j = 2,ny-1
			do 400 i = 2,nx-1
c
                                UU = x(i,j,k)
                                AA = oC(i,j,k)
                                BB = cc(i,j,k)
c
				tmp0 =
     1				+ oN(i,j,k)	* x(i,j+1,k)
     2				+ oN(i,j-1,k)	* x(i,j-1,k)
     3				+ oE(i,j,k)	* x(i+1,j,k)
     4				+ oE(i-1,j,k)	* x(i-1,j,k)
     3				+ oNE(i,j,k)	* x(i+1,j+1,k)
     4				+ oNE(i-1,j-1,k)* x(i-1,j-1,k)
     3				+ oNW(i,j,k)	* x(i-1,j+1,k)
     4				+ oNW(i+1,j-1,k)* x(i+1,j-1,k)
c
				tmpU =
     1				+ uC(i,j,k)	* x(i,j,k+1)
     2				+ uN(i,j,k)	* x(i,j+1,k+1)
     3				+ uS(i,j,k)	* x(i,j-1,k+1)
     4				+ uE(i,j,k)	* x(i+1,j,k+1)
     3				+ uW(i,j,k)	* x(i-1,j,k+1)
     4				+ uNE(i,j,k)	* x(i+1,j+1,k+1)
     3				+ uNW(i,j,k)	* x(i-1,j+1,k+1)
     4				+ uSE(i,j,k)	* x(i+1,j-1,k+1)
     4				+ uSW(i,j,k)	* x(i-1,j-1,k+1)
c
				tmpD =
     1				+ uC(i,j,k-1)	* x(i,j,k-1)
     2				+ uS(i,j+1,k-1)	* x(i,j+1,k-1)
     3				+ uN(i,j-1,k-1)	* x(i,j-1,k-1)
     4				+ uW(i+1,j,k-1)	* x(i+1,j,k-1)
     3				+ uE(i-1,j,k-1)	* x(i-1,j,k-1)
     4				+ uSW(i+1,j+1,k-1)* x(i+1,j+1,k-1)
     3				+ uSE(i-1,j+1,k-1)* x(i-1,j+1,k-1)
     4				+ uNW(i+1,j-1,k-1)* x(i+1,j-1,k-1)
     4				+ uNE(i-1,j-1,k-1)* x(i-1,j-1,k-1)
c
                                FF = fc(i,j,k) + tmp0 + tmpU + tmpD
c
c                               Evaluate residual of 1D system
c
                                zNN = AA*UU + c_scal(BB,UU,ipkey)-FF
c
c                               Setup for Newton method
c
                                niters = 0
                                zndenom = zNN
                                if(zndenom . eq. 0.0d0) zndenom = 1.0d0
c
                                znerror = zNN / zndenom
c
c                               the 1D Newton method
c
                                if(znerror . lt.zntol) goto 300
c
                                if(niters.gt.nitmax) goto 200
c
100                             continue
c
                                        niters = niters + 1
c
c                                       Construct Jacobian matrix of NN
c
                                        DzNN = AA + dc_scal(BB,UU,ipkey)
c
c                                       Solve the linear system
c
                                        change = - zNN / DzNN
					change=sign(min(abs(change),
     1					5.d0),change)
c
c                                       Update the solution
c
                                        UU = UU + change
c
c                                       Evaluate residual of 1D system
c
                                        zNN=AA*UU+c_scal(BB,UU,ipkey)-FF
c
c                                       Compute error
c
                                        znerror = zNN / zndenom
c
c                                       tolerance and itmax check
c
                                        if(znerror.lt.zntol) goto 300
                                        if(niters.gt.nitmax) goto 200
c
                                goto 100
c
c                               tolerance not reached
c
200                             continue
c
                                ifail_tol = ifail_tol + 1
c
c                               tolerance reached
c
300                             continue
c
c                               Newton method complete -- update solution value
c
                                x(i,j,k)=x(i,j,k)*fac + omega*UU
c
400                     continue
500                 continue
600             continue
c
700     continue
c
c	If needed, return the new residual
c
	if(iresid.eq.1) then
		call residual27(nx,ny,nz,ipc,rpc,oC,cc,fc,
     1			oE,oN,uC,oNE,oNW,uE,uW,uN,uS,uNE,uNW,uSE,uSW,
     3			x,r,w1)
	endif
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	NRichardson.f
c
c	This is the driver routine for the nonlinear Richardson smoothing procedure. 
c ==========================================================================================
c ==========================================================================================
c
	subroutine Nrichardson(nx,ny,nz,ipc,rpc,ac,cc,fc,x,w1,w2,r,
     1		itmax,iters,errtol,omega,iresid,iadjoint)
c
c ==========================================================================================
c	Declaration
c ==========================================================================================
c
	integer	nx,ny,nz,itmax,iters,iresid,iadjoint
	integer	numdia
c
	integer	ipc(*)
c
	real*8	omega,errtol
	real*8	rpc(*),ac(nx*ny*nz,*),cc(nx,ny,nz),fc(nx,ny,nz)
	real*8	x(nx,ny,nz),w1(nx,ny,nz),w2(nx,ny,nz)
	real*8	r(nx,ny,nz)
c
c ==========================================================================================
c	Call the proper routine based on Stencil size
c ==========================================================================================
c
	numdia = ipc(11)
c
	if(numdia.eq.7) then
		call nrich7(nx,ny,nz,ipc,rpc,ac(1,1),cc,fc,
     1		ac(1,2),ac(1,3),ac(1,4),
     2		x,w1,w2,r,
     3		itmax,iters,errtol,omega,iresid,iadjoint)
	elseif(numdia.eq.27) then
		call nrich27(nx,ny,nz,ipc,rpc,ac(1,1),cc,fc,
     1		ac(1,2),ac(1,3),ac(1,4),ac(1,5),ac(1,6),ac(1,7),ac(1,8),
     1		ac(1,9),ac(1,10),ac(1,11),ac(1,12),ac(1,13),ac(1,14),
     2		x,w1,w2,r,
     3		itmax,iters,errtol,omega,iresid,iadjoint)
	else
		write(6,*) 'In GaussSeidel_RB : bad stencil type...'
	endif
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	nrich7.f
c
c	7-diagonal Weighted Jacobi routine
c ==========================================================================================
c ==========================================================================================
c
	subroutine nrich7(nx,ny,nz,ipc,rpc,
     1			oC,cc,fc,oE,oN,uC,
     2			x,w1,w2,r,itmax,iters,errtol,omega,iresid,
     3			iadjoint)
c
c ==========================================================================================
c	Declaration
c ==========================================================================================
c
	integer	nx,ny,nz,itmax,iters,iresid,iadjoint
	integer	i,j,k,i1,i2,j1,j2,k1,k2,istep
c
	integer	ipc(*)
c
	real*8	omega,errtol,fac
c
	real*8	rpc(*)
	real*8	oE(nx,ny,nz),oN(nx,ny,nz),uC(nx,ny,nz)
	real*8	fc(nx,ny,nz),oC(nx,ny,nz),cc(nx,ny,nz)
	real*8	x(nx,ny,nz),w1(nx,ny,nz),w2(nx,ny,nz)
	real*8	r(nx,ny,nz)
c
c ==========================================================================================
c	Do Richardson iterations itmax times
c ==========================================================================================
c
	do 400 iters = 1,itmax
c
		call residual7(nx,ny,nz,ipc,rpc,oC,cc,fc,oE,oN,uC,
     1                  x,r,w1)
c
		do 300 k=2,nz-1
		   do 200 j = 2,ny-1
			do 100 i = 2,nx-1
c
                                x(i,j,k)=x(i,j,k)+omega*r(i,j,k)
c
100			continue
200		   continue
300		continue
c
400	continue
c
	if(iresid.eq.1) then
		call residual7(nx,ny,nz,ipc,rpc,oC,cc,fc,oE,oN,uC,
     1                  x,r,w1)
	endif
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	NRich27.f
c
c	27-diagonal nonlinear Richardson routine
c ==========================================================================================
c ==========================================================================================
c
	subroutine nrich27(nx,ny,nz,ipc,rpc,
     1			oC,cc,fc,oE,oN,uC,
     2			oNE,oNW,uE,uW,uN,uS,uNE,uNW,uSE,uSW,
     2			x,w1,w2,r,itmax,iters,errtol,omega,iresid,
     3			iadjoint)
c
c ==========================================================================================
c	Declaration
c ==========================================================================================
c
	integer	nx,ny,nz,itmax,iters,iresid,iadjoint
	integer	i,j,k,i1,i2,j1,j2,k1,k2,istep
c
	integer	ipc(*)
c
	real*8	omega,errtol,fac
c
	real*8	tmp0,tmpU,tmpD
	real*8	rpc(*)
	real*8	oE(nx,ny,nz),oN(nx,ny,nz),uC(nx,ny,nz)
	real*8	oNE(nx,ny,nz),oNW(nx,ny,nz),uE(nx,ny,nz)
	real*8	uW(nx,ny,nz),uN(nx,ny,nz),uS(nx,ny,nz)
	real*8	uNE(nx,ny,nz),uNW(nx,ny,nz),uSE(nx,ny,nz)
	real*8	uSW(nx,ny,nz)
	real*8	fc(nx,ny,nz),oC(nx,ny,nz),cc(nx,ny,nz)
	real*8	x(nx,ny,nz),w1(nx,ny,nz),w2(nx,ny,nz)
	real*8	r(nx,ny,nz)
c
c ==========================================================================================
c	Do Richardson iterations itmax times
c ==========================================================================================
c
	do 400 iters = 1,itmax
c
		call residual27(nx,ny,nz,ipc,rpc,oC,cc,fc,
     1			oE,oN,uC,oNE,oNW,uE,uW,uN,uS,uNE,uNW,uSE,uSW,
     3			x,r,w1)
c
		do 300 k=2,nz-1
		   do 200 j = 2,ny-1
			do 100 i = 2,nx-1
                                x(i,j,k)=x(i,j,k)+omega*r(i,j,k)
100			continue
200		    continue
300		continue
c
400	continue
c
	if(iresid.eq.1) then
		call residual27(nx,ny,nz,ipc,rpc,oC,cc,fc,
     1			oE,oN,uC,oNE,oNW,uE,uW,uN,uS,uNE,uNW,uSE,uSW,
     3			x,r,w1)
	endif
c
        return
        end
