c       Copyright (C) 2007 Patrice Koehl
c
c       This library is free software; you can redistribute it and/or
c       modify it under the terms of the GNU Lesser General Public
c       License as published by the Free Software Foundation; either
c       version 2.1 of the License, or (at your option) any later version.
c
c       This library is distributed in the hope that it will be useful,
c       but WITHOUT ANY WARRANTY; without even the implied warranty of
c       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
c       Lesser General Public License for more details.
c
c       You should have received a copy of the GNU Lesser General Public
c       License along with this library; if not, write to the Free Software
c       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
c
c ==========================================================================================
c ==========================================================================================
c	NlinSmoothing_gpbl.f
c
c	Version 1:      7/4/07
c
c	Author: Patrice Koehl (in collaboration with Marc Delarue)
c
c	This file contains a set of routines for non linear smoothing
c
c ==========================================================================================
c ==========================================================================================
c
c
c ==========================================================================================
c ==========================================================================================
c	nonlin_smoothing_gpbl.f
c
c	This is the driver routine for the smoothing procedure. We only implement the fast
c	diagonal iterative method
c ==========================================================================================
c ==========================================================================================
c
	subroutine nonlin_smoothing_gpbl(nx,ny,nz,ipc,rpc,cc,fc,x,
     1		w1,w2,r,itmax,iters,errtol,omega,iresid,iadjoint,meth,
     2		grid_info,xpos,ypos,zpos,vect)
c
c ==========================================================================================
c	Declaration
c ==========================================================================================
c
	integer	nx,ny,nz,itmax,iters,iresid,iadjoint,meth
c
	integer	ipc(*)
c
	integer*1	grid_info(*)
c
	real*8	omega,errtol
	real*8	rpc(*),cc(nx,ny,nz),fc(nx,ny,nz)
	real*8	x(nx,ny,nz),w1(nx,ny,nz),w2(nx,ny,nz)
	real*8	r(nx,ny,nz),xpos(nx),ypos(ny),zpos(nz)
	real*8	vect(*)
c
c ==========================================================================================
c	Call the appropriate smoothing routine
c ==========================================================================================
c
	if(meth.eq.1) then
		call NGaussSeidel_RB_gpbl(nx,ny,nz,ipc,rpc,vect,cc,fc,x,
     1		w1,w2,r,itmax,iters,errtol,omega,iresid,iadjoint,
     2		grid_info,xpos,ypos,zpos)
	elseif(meth.eq.2) then
		call NWJacobi_gpbl(nx,ny,nz,ipc,rpc,vect,cc,fc,x,
     1		w1,w2,r,itmax,iters,errtol,omega,iresid,iadjoint,
     2		grid_info,xpos,ypos,zpos)
	elseif(meth.eq.3) then
		call NSOR_gpbl(nx,ny,nz,ipc,rpc,vect,cc,fc,x,
     1		w1,w2,r,itmax,iters,errtol,omega,iresid,iadjoint,
     2		grid_info,xpos,ypos,zpos)
	elseif(meth.eq.4) then
		call conjgrad_nlin_gpbl(nx,ny,nz,ipc,rpc,vect,cc,fc,x,
     1		w1,w2,r,itmax,iters,errtol,omega,iresid,iadjoint,
     2		grid_info,xpos,ypos,zpos)
	elseif(meth.eq.5) then
		call Nrichardson_gpbl(nx,ny,nz,ipc,rpc,vect,cc,fc,x,
     1		w1,w2,r,itmax,iters,errtol,omega,iresid,iadjoint,
     2		grid_info,xpos,ypos,zpos)
	else
		write(6,*) 'Bad smoothing method'
		stop
	endif
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	NGaussSeidel_RB_gpbl.f
c
c	This is the driver routine for the Non-linear Gauss Seidel red-black smoothing procedure. 
c ==========================================================================================
c ==========================================================================================
c
	subroutine NGaussSeidel_RB_gpbl(nx,ny,nz,ipc,rpc,vect,cc,fc,x,
     1		w1,w2,r,itmax,iters,errtol,omega,iresid,iadjoint,
     2		grid_info,xpos,ypos,zpos)
c
c ==========================================================================================
c	Declaration
c ==========================================================================================
c
	integer	nx,ny,nz,itmax,iters,iresid,iadjoint
	integer	i,j,k,i1,i2,j1,j2,k1,k2,istep
	integer	nitmax,niters,ifail_tol,ipkey
	integer	ii,jj,kk
c
	integer	ipc(*)
c
	integer*1	grid_info(*)
c
	real*8	omega,fac,errtol,epstol
	real*8	FF,BB,zNN,DzNN,zntol,zndenom,znerror
	real*8	change,keep
        real*8  xnorm1
c
	real*8	rpc(*)
	real*8	fc(nx,ny,nz),cc(nx,ny,nz)
	real*8	x(nx,ny,nz),w1(nx,ny,nz),w2(nx,ny,nz)
	real*8	r(nx,ny,nz)
	real*8	xpos(*),ypos(*),zpos(*),vect(*)
c
	nitmax	  = 10
	zntol     = 1.d-5
        epstol    = 1.d-10
	ifail_tol = 0
	ipkey	  = ipc(10)
c
c ==========================================================================================
c	Do Gauss Seidel iterations itmax times
c ==========================================================================================
c
	i1 = (1-iadjoint) * 2 + iadjoint	*(nx-1)
	i2 = iadjoint     * 2 + (1-iadjoint)	*(nx-1)
	j1 = (1-iadjoint) * 2 + iadjoint	*(ny-1)
	j2 = iadjoint     * 2 + (1-iadjoint)	*(ny-1)
	k1 = (1-iadjoint) * 2 + iadjoint	*(nz-1)
	k2 = iadjoint     * 2 + (1-iadjoint)	*(nz-1)
c
	istep = iadjoint*(-1) + (1-iadjoint)*(1)
c
	call vectcopy(nx,ny,nz,x,w1)
        call set_bound_zero(nx,ny,nz,w1)
c
	do 700 iters = 1,itmax
c
		do 600 k=k1,k2,istep
		   do 500 j = j1,j2,istep
			do 400 i = i1,i2,istep
c
c				For this unknown, use a 1D Newton method
c
				FF = fc(i,j,k)
				BB = cc(i,j,k)
c
				keep = w1(i,j,k)
c
				call langevin_scal(i,j,k,w1,nx,ny,nz,
     1				xpos,ypos,zpos,grid_info,FF,BB,zNN,DzNN)
c
c				Setup for Newton method
c
				niters = 0
				zndenom = zNN
				if(zndenom.eq.0.d0) zndenom = 1.d0
c
				znerror = zNN / zndenom
c
				if(znerror.lt.zntol.or.abs(zNN).
     1				lt.epstol) goto 300
c
100				continue
c
					niters = niters + 1
c
					change = - zNN / DzNN
					change=sign(min(abs(change),
     1					5.d0),change)
c
					w1(i,j,k) = w1(i,j,k) + change
c
					call langevin_scal(i,j,k,w1,nx,
     1					ny,nz,xpos,ypos,zpos,grid_info,
     2					FF,BB,zNN,DzNN)
c
					znerror = zNN/zndenom
c
					if(znerror.lt.zntol) goto 300
					if(niters.gt.nitmax) goto 200
c
				goto 100
c
200				continue
				ifail_tol = ifail_tol + 1
c
300				continue
c
				w1(i,j,k)=(1-omega)*keep+omega*w1(i,j,k)
c
400			continue
c
500		    continue
c
600		continue
c
700	continue
c
	call vectcopy(nx,ny,nz,w1,x)
c
c	If needed, return the new residual
c
	if(iresid.eq.1) then
		call residue_gpbl(x,nz,ny,nz,xpos,ypos,zpos,grid_info,
     1		vect,fc,cc,r)
	endif
c
c	if(ifail_tol.gt.0) then
c		write(6,*) 
c     1          'Nonlinear GSRB: 1D Newton tol. failures: ',ifail_tol
c		stop
c	endif
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	NWJacobi_gpbl.f
c
c	This is the driver routine for the non linear Weighted Jacobi smoothing procedure. 
c ==========================================================================================
c ==========================================================================================
c
	subroutine NWJacobi_gpbl(nx,ny,nz,ipc,rpc,vect,cc,fc,x,w1,w2,r,
     1		itmax,iters,errtol,omega,iresid,iadjoint,
     2		grid_info,xpos,ypos,zpos)
c
c ==========================================================================================
c	Declaration
c ==========================================================================================
c
	integer	nx,ny,nz,itmax,iters,iresid,iadjoint
	integer	i,j,k
	integer	nitmax,niters,ifail_tol,ipkey
c
	integer	ipc(*)
c
	integer*1	grid_info(*)
c
	real*8	omega,errtol,fac
c
	real*8	BB,FF,zNN,DzNN,zntol,zndenom,znerror,change
	real*8	rpc(*)
	real*8	fc(nx,ny,nz),cc(nx,ny,nz)
	real*8	x(nx,ny,nz),w1(nx,ny,nz),w2(nx,ny,nz)
	real*8	r(nx,ny,nz)
	real*8	xpos(*),ypos(*),zpos(*),vect(*)
c
	nitmax	  = 10
	zntol     = 1.d-5
	ifail_tol = 0
	ipkey	  = ipc(10)
c
c ==========================================================================================
c	Do Jacobi iterations itmax times
c ==========================================================================================
c
	call vectcopy(nx,ny,nz,x,w1)
c
	do 700 iters = 1,itmax
c
		do 600 k=2,nz-1
		   do 500 j = 2,ny-1
			do 400 i = 2,nx-1
c
				FF = fc(i,j,k)
				BB = cc(i,j,k)
c
				call langevin_scal(i,j,k,w1,nx,ny,nz,
     1				xpos,ypos,zpos,grid_info,FF,BB,zNN,DzNN)
c
c				Setup for Newton method
c
				niters = 0
				zndenom = zNN
				if(zndenom.eq.0.d0) zndenom = 1.d0
				znerror = zNN / zndenom
c
				if(znerror.lt.zntol) goto 300
				if(niters.gt.nitmax) goto 200
c
100				continue
c
					niters = niters + 1
c
					change = - zNN / DzNN
					change=sign(min(abs(change),
     1					5.d0),change)
c
					w1(i,j,k) = w1(i,j,k) + change
c
					call langevin_scal(i,j,k,w1,nx,
     1					ny,nz,xpos,ypos,zpos,grid_info,
     2					FF,BB,zNN,DzNN)
c
					znerror = zNN/zndenom
c
					if(znerror.lt.zntol) goto 300
					if(niters.gt.nitmax) goto 200
c
				goto 100
c
200				continue
				ifail_tol = ifail_tol + 1
c
300				continue
c
400			continue
c
500		    continue
c
600		continue
c
700	continue
c
        fac = 1.d0 - omega
c
        do 1000 k = 2,nz-1
                do 900 j = 2,ny-1
                        do 800 i = 2,nx-1
                                x(i,j,k)=omega*w1(i,j,k)+fac*x(i,j,k)
800                     continue
900             continue
1000    continue
c
c	If needed, return the new residual
c
	if(iresid.eq.1) then
		call residue_gpbl(x,nx,ny,nz,xpos,ypos,zpos,grid_info,
     1		vect,fc,cc,r)
	endif
c
c	if(ifail_tol.gt.0) then
c		write(6,*) 
c     1          'Nonlinear GSRB: 1D Newton tol. failures: ',ifail_tol
c	endif
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	NSOR_gpbl.f
c
c	This is the driver routine for the Non linear SOR procedure.
c ==========================================================================================
c ==========================================================================================
c
	subroutine nsor_gpbl(nx,ny,nz,ipc,rpc,vect,cc,fc,x,w1,w2,r,
     1		itmax,iters,errtol,omega,iresid,iadjoint,
     2          grid_info,xpos,ypos,zpos)
c
c
c       fast diagonal nonlinear red/black sor routine.
c
c       note that the interior grid points from 2,...,nx-1, etc.
c       we then begin coloring with a red point, and do a point red/black
c       coloring.
c
c       the red points are:
c
c       if ((j even) and (k even)) then
c               begin row at first point, i=2 (i even), or ioff = 0
c       else if ((j odd) and (k even)) then
c               begin row at second point, i=3 (i odd), or ioff = 1
c       else if ((j even) and (k odd)) then
c               begin row at second point, i=3 (i odd), or ioff = 1
c       else if ((j odd) and (k odd)) then
c               begin row at first point, i=2 (i even), or ioff = 0
c       endif
c       then: begin row at:  i=2+ioff
c
c       the appropriate ioff function for the red points is then:
c               ioff = dabs(mod(j,2)-mod(k,2))
c
c       the appropriate ioff function for the black points is then:
c               ioff = 1 - dabs(mod(j,2)-mod(k,2))
c
c       alternatively, the red points are:
c
c       those whose indices add up to an even number.
c       to see this, consider that all surrounding points are
c       only one index different, hence the sum will differ by one.
c       thus, if a given point has an even sum, then the surrounding
c       points will have an odd sum.
c
c       the black points are:
c
c       those whose indices add up to an odd number.
c
c       algorithm parameters:
c       =====================
c
c       U0      = initial guess
c       Llin    = matrix representing linear part of equations
c       F       = vector representing rhs of equations
c       nx      = number of unknowns
c       itmax   = number of iterations to perform
c       U       = final result
c
c       algorithm idea:
c       ===============
c
c       the ith nonlinear equation is represented as:
c
c       N(UU) = langevin_scal(UU) - FF
c
c       which is solved for UU using newton method, where:
c
c       FF = F(i)
c       UU = the unknown, initialized to U(i)
c
c ==========================================================================================
c	Declaration
c ==========================================================================================
c
	integer	nx,ny,nz,itmax,iters,iresid,iadjoint
	integer	i,j,k,istep,ioff
        integer nitmax,niters,ifail_tol,ipkey
c
	integer	ipc(*)
c
	real*8	omega,errtol,fac
c
        real*8  FF,BB,zNN,DzNN,zntol,zndenom,znerror
        real*8  change
c
	real*8	rpc(*)
	real*8	fc(nx,ny,nz),cc(nx,ny,nz)
	real*8	x(nx,ny,nz),w1(nx,ny,nz),w2(nx,ny,nz)
	real*8	r(nx,ny,nz)
        real*8  xpos(*),ypos(*),zpos(*),vect(*)
c
        integer*1 grid_info(*)
c
c ==========================================================================================
c	Nonlinear iteration tolerance and itmax
c ==========================================================================================
c
        nitmax          = 10
        zntol           = 1.d-5
        ifail_tol       = 0
        ipkey           = ipc(10)
c
        fac     = 1.0d0 - omega
c
        call vectcopy(nx,ny,nz,x,w1)
        call set_bound_zero(nx,ny,nz,w1)
c
	do 1300 iters = 1,itmax
c
		do 600 k=2,nz-1
c
		   do 500 j = 2,ny-1
c
                        ioff = mod((j+k+2),2)
c
			do 400 i = 2+ioff,nx-1,2
c
c                               For this unknown, do a 1D Newton method
c
c                               Determine the 1D equation
c                               compute the residual, and 1D Jacobian
c
                                FF = fc(i,j,k)
                                BB = cc(i,j,k)
c
		                call langevin_scal(i,j,k,
     1                          w1,nx,ny,nz,
     2                          xpos,ypos,zpos,grid_info,
     3                          FF,BB,zNN,DzNN)
c
c                               Setup for Newton method
c
                                niters = 0
                                zndenom = zNN
                                if(zndenom.eq.0.0d0) zndenom = 1.0d0
                                znerror = zNN / zndenom
c
c                               the 1D Newton method
c
                                if(znerror.lt.zntol) goto 300
                                if(niters.gt.nitmax) goto 200
c
100                             continue
c
                                        niters = niters + 1
c
c                                       Solve the linear system
c
                                        change = - zNN / DzNN
					change=sign(min(abs(change),
     1					5.d0),change)
c
c                                       Update the solution
c
                                        w1(i,j,k) = w1(i,j,k) + change
c
c                                       Evaluate residual of 1D system
c
		                        call langevin_scal(i,j,k,
     1                                  w1,nx,ny,nz,
     2                                  xpos,ypos,zpos,grid_info,
     3                                  FF,BB,zNN,DzNN)
c
c                                       Compute error
c
                                        znerror = zNN / zndenom
c
c                                       tolerance and itmax check
c
                                        if(znerror.lt.zntol) goto 300
                                        if(niters.gt.nitmax) goto 200
c
c                                       Loop
c
                                goto 100
c
c                               tolerance not reached
c
200                             continue
c
                                ifail_tol = ifail_tol + 1
c
c                               tolerance reached
c
300                             continue
c
c                               Newton method complete: update solution
c
                                w1(i,j,k)=x(i,j,k)*fac+omega*w1(i,j,k)
c
400                     continue
c
500                 continue
c
600             continue
c
c               Do the black points
c
		do 1200 k=2,nz-1
c
		   do 1100 j = 2,ny-1
c
                        ioff = 1-mod((j+k+2),2)
c
			do 1000 i = 2+ioff,nx-1,2
c
c                               For this unknown, do a 1D Newton method
c
c                               Determine the 1D equation
c                               compute the residual, and 1D Jacobian
c
				FF = fc(i,j,k)
				BB = cc(i,j,k)
c
		                call langevin_scal(i,j,k,
     1                          w1,nx,ny,nz,
     1                          xpos,ypos,zpos,grid_info,
     2                          FF,BB,zNN,DzNN)
c
c                               Setup for Newton method
c
                                niters = 0
                                zndenom = zNN
                                if(zndenom.eq.0.0d0) zndenom = 1.0d0
                                znerror = zNN / zndenom
c
c                               the 1D Newton method
c
                                if(znerror.lt.zntol) goto 900
                                if(niters.gt.nitmax) goto 800
c
700                             continue
c
                                        niters = niters + 1
c
c                                       Solve the linear system
c
                                        change = - zNN / DzNN
					change=sign(min(abs(change),
     1					5.d0),change)
c
c                                       Update the solution
c
                                        w1(i,j,k)=w1(i,j,k)+change
c
c                                       Evaluate residual of 1D system
c
		                        call langevin_scal(i,j,k,
     1                                  w1,nx,ny,nz,
     2                                  xpos,ypos,zpos,grid_info,
     3                                  FF,BB,zNN,DzNN)
c
c                                       Compute error
c
                                        znerror = zNN / zndenom
c
c                                       tolerance and itmax check
c
                                        if(znerror.lt.zntol) 
     1						goto 900
                                        if(niters.gt.nitmax) goto 800
c
c                                       Loop
c
                                goto 700
c
c                               tolerance not reached
c
800                             continue
c
                                ifail_tol = ifail_tol + 1
c
c                               tolerance reached
c
900                             continue
c
c                               Newton method complete: update solution
c
                                w1(i,j,k)=x(i,j,k)*fac+omega*w1(i,j,k)
c
1000                     continue
c
1100                 continue
c
1200             continue
c
                call vectcopy(nx,ny,nz,w1,x)
                call set_bound_zero(nx,ny,nz,x)
c
1300	continue
c
c	If needed, return the new residual
c
	if(iresid.eq.1) then
		call residue_gpbl(x,nx,ny,nz,
     1          xpos,ypos,zpos,grid_info,vect,fc,cc,r)
	endif
c
c        if(ifail_tol.gt.0) then
c                write(6,*)'NSOR: 1D Newton tolerance failures: ',
c     1			ifail_tol
c        endif
c
	return
	end
c
c ==========================================================================================
c	conjgrad_nlin_gpbl.f
c
c	This routine solves the system of equation Ax=b using conjugate gradient
c ==========================================================================================
c ==========================================================================================
c
	subroutine conjgrad_nlin_gpbl(nx,ny,nz,ipc,rpc,vect,cc,fc,x,p,
     2		ap,r,itmax,iters,errtol,omega,iresid,iadjoint,
     3          grid_info,xpos,ypos,zpos)
c
c ==========================================================================================
c	Declaration
c ==========================================================================================
c
	integer	nx,ny,nz,itmax,iters,iresid,iadjoint
	integer	i,j,k
c
	integer	ipc(*)
c
        integer*1 grid_info(*)
c
	real*8	omega,errtol,rsnrm,pAp,denom,xnorm2
	real*8	rhok1,rhok2,alpha,beta,dotvect
        real*8  dot,betam1
c
        real*8  xpos(*),ypos(*),zpos(*),vect(*)
	real*8	rpc(*),fc(nx,ny,nz),cc(nx,ny,nz)
	real*8	x(nx,ny,nz),r(nx,ny,nz),p(nx,ny,nz),ap(nx,ny,nz)
c
c ==========================================================================================
c	Setup
c ==========================================================================================
c
	iters = 0
	if(iters.ge.itmax.and.iresid.eq.0) return
c
	call residue_gpbl(x,nx,ny,nz,xpos,ypos,zpos,grid_info,fc,cc,r)
c
	denom = xnorm2(nx,ny,nz,r)
c
	if(denom.eq.0.d0) return
	if(iters.ge.itmax) return
c
c ==========================================================================================
c	ConjGrad loop
c ==========================================================================================
c
100	continue
c
c		Compute current stopping criteria
c
		rhok2 = dotvect(nx,ny,nz,r,r)
		rsnrm = dsqrt(rhok2)
c
		if(rsnrm/denom.le.errtol) goto 200
		if(iters.ge.itmax) goto 200
c
c		Form new direction vector
c
		if(iters.eq.0) then
			call vectcopy(nx,ny,nz,r,p)
		else
			beta = rhok2/rhok1
                        betam1 = 1.d0/beta
			call vectsmult(nx,ny,nz,betam1,r,p)
			call scalevect(nx,ny,nz,beta,p)
		endif
c
c		Find alpha that minimizes error
c
		call linesearch_gpbl(nx,ny,nz,alpha,
     2			ipc,rpc,cc,fc,p,x,r,ap,
     3                  grid_info,xpos,ypos,zpos,vect)
c
c		Save rhok2
c
		rhok1 = rhok2
c
c		Update solution in direction p of length alpha
c
		call vectsmult(nx,ny,nz,alpha,p,x)
c
c		Update residual
c
	        call residue_gpbl(x,nx,ny,nz,xpos,ypos,zpos,
     1                 grid_info,fc,cc,r)
c
		iters = iters + 1
c
		goto 100
c
200	continue
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	NRichardson_gpbl.f
c
c	This is the driver routine for the nonlinear Richardson smoothing procedure. 
c ==========================================================================================
c ==========================================================================================
c
	subroutine Nrichardson_gpbl(nx,ny,nz,ipc,rpc,vect,cc,fc,x,
     1		w1,w2,r,itmax,iters,errtol,omega,iresid,iadjoint,
     2		grid_info,xpos,ypos,zpos)
c
c ==========================================================================================
c	Declaration
c ==========================================================================================
c
	integer	nx,ny,nz,itmax,iters,iresid,iadjoint
	integer	i,j,k,i1,i2,j1,j2,k1,k2,istep
c
	integer	ipc(*)
c
	integer*1	grid_info(*)
c
	real*8	omega,errtol,fac
c
	real*8	rpc(*)
	real*8	fc(nx,ny,nz),cc(nx,ny,nz)
	real*8	x(nx,ny,nz),w1(nx,ny,nz),w2(nx,ny,nz)
	real*8	r(nx,ny,nz)
	real*8	xpos(*),ypos(*),zpos(*),vect(*)
c
c ==========================================================================================
c	Do Richardson iterations itmax times
c ==========================================================================================
c
	do 400 iters = 1,itmax
c
		call residue_gpbl(x,nx,ny,nz,xpos,ypos,zpos,grid_info,
     1		vect,fc,cc,r)
c
                call vectsmult(nx,ny,nz,omega,r,x)
c
400	continue
c
	if(iresid.eq.1) then
		call residue_gpbl(x,nx,ny,nz,xpos,ypos,zpos,grid_info,
     1		vect,fc,cc,r)
	endif
c
	return
	end
