c ==========================================================================================
c ==========================================================================================
c	C_vec.f
c       ((non linear) function in PDE, based on Michael Holst MG package)
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      8/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c
c	Purpose:	Defines the non linear term (vector version)
c
c	Input parameters:
c
c		coef:	defines region in which ion is considered
c		u:	dimensionless potential (in)
c		ipkey:  define type of equation
c
c	Output:
c
c		c_nl: value of non linear term in PBE
c
	subroutine c_vec(coef,u,c_nl,nx,ny,nz,ipkey)
c
	include 'poisson.h'
c
c ==========================================================================================
c == Define all variables                  =================================================
c ==========================================================================================
c
c	Local variables
c===========================================================================================
c
	integer	i,j,k,i_ion
c
	real*8	argument,argument0,coef2,coef20
	real*8	u_val,l_dip,exp_mu,F1_u,sinhc_u
	real*8	exp_min,exp_max
	parameter	(exp_min = -80.d0, exp_max = 80.d0)
c
c	Parameters (in and out)
c===========================================================================================
c
	integer	ipkey
	integer nx,ny,nz
c
	real*8	coef(nx,ny,nz),u(nx,ny,nz),c_nl(nx,ny,nz)
c
c	information on ions
c===========================================================================================
c
	integer	n_ion,n_zz,n_solv
c
	real*8	ratio,phi0,coef_Z0,coef_Z
	real*8	D,g,val1
	real*8	ion_conc(max_ion),ion_z(max_ion),kappa(max_ion)
	real*8	ion_rad(max_ion)
	real*8	ion_cvol(max_ion)
	real*8	solv_conc(max_solv),solv_dipole(max_solv)
	real*8	solv_cvol(max_solv)
c
c	Coulomb field (if charge free method is used)
c===========================================================================================
c
	integer	flag
	integer	flag_dip,flag_psi
c
	real*8	value
	real*8	psi_coef,psi_val,psi_bulk,v0,b
	real*8	dip(nx,ny,nz)
	real*8	psi(nx,ny,nz)
c
	pointer	(ptr_dip,dip)
	pointer (ptr_psi,psi)
c
c ==========================================================================================
c == Common block (define ions)            =================================================
c ==========================================================================================
c
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
	common /solvent/	solv_conc,solv_dipole,solv_cvol,n_solv
	common /smpb/		ion_cvol,ratio,phi0,coef_Z0
	common /water_sc/	ptr_dip,flag_dip
	common /psi_pot/	ptr_psi,flag_psi
	common /yukawa/		v0,b,psi_bulk
c
c ==========================================================================================
c == Compute non linear term               =================================================
c ==========================================================================================
c
	do 300 k = 1,nz
		do 200 j = 1,ny
			do 100 i = 1,nx
				c_nl(i,j,k) = 0.d0
100			continue
200		continue
300	continue
c
c	If solving Poisson equation, return
c
	if(ipkey.eq.0) return
c
c	Loop over all ions
c
c	If solving linearized Poisson Boltzmann, or non linear Poisson Boltzmann:
c
	if(ipkey.eq.1) then
c
	    do 700 i_ion = 1,n_ion
c
		coef20 = kappa(i_ion)
c
		argument0 = -ion_z(i_ion)
c
		do 600 k = 1,nz
		  do 500 j = 1,ny
		    do 400 i = 1,nx
c
			value = u(i,j,k)
			coef2 = coef20 * coef(i,j,k)
			argument = argument0*value
c
c			Make sure that the argument of the exponential function is not
c			too large (or too small), so that we do not have overflow (or
c			underflow)
c
			if(argument.gt.exp_max) argument = exp_max
			if(argument.lt.exp_min) argument = exp_min
c
			c_nl(i,j,k) = c_nl(i,j,k) + coef2*(1+argument)
c
400		    continue
500		  continue
600		continue
c
700	    continue
c
	elseif(ipkey.eq.2) then
c
	    do 1100 i_ion = 1,n_ion
c
		coef20 = kappa(i_ion)
c
		argument0 = -ion_z(i_ion)
c
		do 1000 k = 1,nz
		  do 900 j = 1,ny
		    do 800 i = 1,nx
c
			value = u(i,j,k)
c
			coef2 = coef20 * coef(i,j,k)
			argument = argument0*value
c
c			Make sure that the argument of the exponential function is not
c			too large (or too small), so that we do not have overflow (or
c			underflow)
c
			if(argument.gt.exp_max) argument = exp_max
			if(argument.lt.exp_min) argument = exp_min
c
			c_nl(i,j,k) = c_nl(i,j,k)+coef2*exp(argument)
c
800		    continue
900		  continue
1000		continue
c
1100	    continue
c
	elseif(ipkey.eq.3) then
c
	    do 1500 k = 1,nz
		do 1400 j = 1,ny
		   do 1300 i = 1,nx
c
			if(coef(i,j,k).eq.0) then
				c_nl(i,j,k) = 0.d0
				goto 1300
			endif
c
c		Add ion one (small size)
c
			value = u(i,j,k)
c
			argument = -ion_z(1)*value
c
c			Make sure that the argument of the exponential function is not
c			too large (or too small), so that we do not have overflow (or
c			underflow)
c
			if(argument.gt.exp_max) argument = exp_max
			if(argument.lt.exp_min) argument = exp_min
c
			argument = exp(argument)
c
			val1 = (1.d0-phi0+ion_cvol(1)*argument)
			if(flag_dip.eq.1) val1 = val1 + dip(i,j,k)
			g = kappa(1)*argument
			d = val1
			coef_Z = coef_Z0/val1
			coef_Z = coef_Z**(ratio-1)
c
c		Add all other ions
c
			do 1200 i_ion = 2,n_ion
c
				argument = -ion_z(i_ion)*value
c
c				Make sure that the argument of the exponential function is not
c				too large (or too small), so that we do not have overflow (or
c				underflow)
c
				if(argument.gt.exp_max) argument=exp_max
				if(argument.lt.exp_min) argument=exp_min
c
				argument = exp(argument)
c
				g = g + coef_Z*kappa(i_ion)*argument
c
				d = d + coef_Z*ion_cvol(i_ion)*argument
c
1200			continue
c
			c_nl(i,j,k) = coef(i,j,k)*g/d
c
1300		    continue
1400		continue
1500	    continue
c
	elseif(ipkey.eq.4) then
c
c		We set c_nl to zero, as it is included in the Jacobian
c
		call set_all_zeros(nx,ny,nz,c_nl)
c
	elseif(ipkey.eq.5) then
c
	    do 1900 k = 1,nz
		do 1800 j = 1,ny
		   do 1700 i = 1,nx
c
			if(coef(i,j,k).eq.0) then
				c_nl(i,j,k) = 0.d0
				goto 1700
			endif
c
			value = u(i,j,k)
c
			if(flag_dip.eq.1) then
				u_val = dip(i,j,k)
				call compute_F1(u_val,F1_u,
     1					sinhc_u,exp_mu)
				l_dip = solv_cvol(1)*sinhc_u
				if(flag_psi.eq.1) then
					psi_val = -psi(i,j,k)+psi_bulk
					if(psi_val.gt.exp_max) 
     1						psi_val=exp_max
					if(psi_val.lt.exp_min) 
     1						psi_val=exp_min
					psi_coef=exp(psi_val)
					l_dip=l_dip*psi_coef
				endif
			else
				u_val = 0
				exp_mu = 1
				l_dip = 0
			endif
c
			g = 0
			d = (1.d0 - phi0)*exp_mu + l_dip
c
c		Add all ions
c
			do 1600 i_ion = 1,n_ion
c
				argument = -ion_z(i_ion)*value -u_val
c
c				Make sure that the argument of the exponential function is not
c				too large (or too small), so that we do not have overflow (or
c				underflow)
c
				if(argument.gt.exp_max) argument=exp_max
				if(argument.lt.exp_min) argument=exp_min
c
				argument = exp(argument)
c
				g = g + kappa(i_ion)*argument
c
				d = d + ion_cvol(i_ion)*argument
c
1600			continue
c
			c_nl(i,j,k) = coef(i,j,k)*g/d
c
1700		    continue
1800		continue
1900	    continue
	endif
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	DC_vec.f
c       (derivative of (non linear) term  in PBE, based on Michael Holst MG package)
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      8/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c
c	Purpose:	Defines the derivative of the non linear term (vector version)
c
c	Input parameters:
c
c		coef:	defines region in which ion is considered
c		u:	dimensionless potential (in)
c		ipkey:  not used!
c
c	Output:
c
c		dc_nl: value of non linear term in PBE
c
	subroutine dc_vec(coef,u,dc_nl,nx,ny,nz,ipkey)
c
	include 'poisson.h'
c
c ==========================================================================================
c == Define all variables                  =================================================
c ==========================================================================================
c
c	Local variables
c===========================================================================================
c
	integer	i,j,k,i_ion
c
	real*8	argument,coef2,argument0,coef20
	real*8	u_val,l_dip,exp_mu,F1_u,sinhc_u
	real*8	exp_min,exp_max
	parameter	(exp_min = -80.d0, exp_max = 80.d0)
c
c	Parameters (in and out)
c===========================================================================================
c
	integer	ipkey
	integer	nx,ny,nz
c
	real*8	coef(nx,ny,nz),u(nx,ny,nz),dc_nl(nx,ny,nz)
c
c	information on ions
c===========================================================================================
c
	integer	n_ion,n_zz,n_solv
c
	real*8	D,dprime,g,gprime,val,val1,dval1
	real*8	ratio,phi0,coef_Z0,coef_Z,coef_Z_km1,dcoef_Z_km1
	real*8	ion_conc(max_ion),ion_z(max_ion),kappa(max_ion)
	real*8	ion_rad(max_ion)
	real*8	ion_cvol(max_ion)
	real*8	solv_conc(max_solv),solv_dipole(max_solv)
	real*8	solv_cvol(max_solv)
c
c	Coulomb field (if charge free method is used)
c===========================================================================================
c
	integer	flag,flag_dip
	integer	flag_psi
c
	real*8	value
	real*8	psi_bulk,psi_val,psi_coef,v0,b
	real*8	dip(nx,ny,nz)
	real*8	psi(nx,ny,nz)
c
	pointer (ptr_dip,dip)
	pointer (ptr_psi,psi)
c
c ==========================================================================================
c == Common block (define ions)            =================================================
c ==========================================================================================
c
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
	common /solvent/	solv_conc,solv_dipole,solv_cvol,n_solv
	common /smpb/		ion_cvol,ratio,phi0,coef_Z0
	common /water_sc/	ptr_dip,flag_dip
	common /psi_pot/	ptr_psi,flag_psi
	common /yukawa/		v0,b,psi_bulk
c
c ==========================================================================================
c == Compute non linear term               =================================================
c ==========================================================================================
c
	do 300 k = 1,nz
	    do 200 j = 1,ny
		do 100 i = 1,nx
			dc_nl(i,j,k) = 0
100		continue
200	    continue
300	continue
c
	if(ipkey.eq.0) return
c
	if(ipkey.eq.1) then
c
	    do 700 i_ion = 1,n_ion
c
		coef20 = kappa(i_ion)
c
		argument0 = -ion_z(i_ion)
c
		do 600 k = 1,nz
		  do 500 j = 1,ny
		    do 400 i = 1,nx
c
			coef2 = coef20 * coef(i,j,k)
c
			dc_nl(i,j,k) = dc_nl(i,j,k) + coef2*argument0
c
400		    continue
500		  continue
600		continue
c
700	    continue
c
	elseif(ipkey.eq.2) then
c
	    do 1100 i_ion = 1,n_ion
c
		coef20 = -ion_z(i_ion)*kappa(i_ion)
c
		argument0 = -ion_z(i_ion)
c
		do 1000 k = 1,nz
		  do 900 j = 1,ny
		    do 800 i = 1,nx
c
			value = u(i,j,k)
			coef2 = coef20 * coef(i,j,k)
			argument = argument0*value
c
c			Make sure that the argument of the exponential function is not
c			too large (or too small), so that we do not have overflow (or
c			underflow)
c
			if(argument.gt.exp_max) argument = exp_max
			if(argument.lt.exp_min) argument = exp_min
c
			dc_nl(i,j,k) = dc_nl(i,j,k)+coef2*exp(argument)
c
800		    continue
900		  continue
1000		continue
c
1100	    continue
c
	elseif(ipkey.eq.3) then
c
	    do 1500 k = 1,nz
		do 1400 j = 1,ny
		   do 1300 i = 1,nx
c
			if(coef(i,j,k).eq.0) then
				dc_nl(i,j,k) = 0.d0
				goto 1300
			endif
c
c		Add ion one (small size)
c
			value = u(i,j,k)
c
			argument = -ion_z(1)*value
c
c			Make sure that the argument of the exponential function is not
c			too large (or too small), so that we do not have overflow (or
c			underflow)
c
			if(argument.gt.exp_max) argument = exp_max
			if(argument.lt.exp_min) argument = exp_min
c
			argument = exp(argument)
c
			val1 = (1.d0-phi0+ion_cvol(1)*argument)
			if(flag_dip.eq.1) val1 = val1 + dip(i,j,k)
c
			dval1 = -ion_z(1)*ion_cvol(1)*argument
			g = kappa(1)*argument
			d = val1
			dprime = dval1
			gprime = -ion_z(1)*g
			coef_Z = coef_Z0/val1
			coef_Z_km1 = coef_Z**(ratio-1)
			dcoef_Z_km1 = -(ratio-1)*coef_Z_km1*dval1/val1
c
c		Add all other ions
c
			do 1200 i_ion = 2,n_ion
c
				argument = -ion_z(i_ion)*value
c
c				Make sure that the argument of the exponential function is not
c				too large (or too small), so that we do not have overflow (or
c				underflow)
c
				if(argument.gt.exp_max) argument=exp_max
				if(argument.lt.exp_min) argument=exp_min
c
				argument = exp(argument)
c
				val = coef_Z_km1*kappa(i_ion)*argument
				val1=coef_Z_km1*ion_cvol(i_ion)*argument
c
				g = g + val
c
				gprime = gprime + dcoef_Z_km1*
     1				kappa(i_ion)*argument -val*ion_z(i_ion)
c
				d = d + val1
				dprime = dprime - val1*ion_z(i_ion)
     1				+ dcoef_Z_km1*ion_cvol(i_ion)*argument
c
1200			continue
c
			dc_nl(i,j,k) = coef(i,j,k)*(gprime*d-g*dprime)
     1					/(d*d)
c
1300		    continue
1400		continue
1500	    continue
c
	elseif(ipkey.eq.4) then
c
c		We set dc_nl to zero, as it is included in the Jacobian
c
		call set_all_zeros(nx,ny,nz,dc_nl)
c
	elseif(ipkey.eq.5) then
c
	    do 1900 k = 1,nz
		do 1800 j = 1,ny
		   do 1700 i = 1,nx
c
			if(coef(i,j,k).eq.0) then
				dc_nl(i,j,k) = 0.d0
				goto 1700
			endif
c
c		Add ion one (small size)
c
			value = u(i,j,k)
c
			if(flag_dip.eq.1) then
				u_val = dip(i,j,k)
				call compute_F1(u_val,F1_u,
     1					sinhc_u,exp_mu)
				l_dip = solv_cvol(1)*sinhc_u
				if(flag_psi.eq.1) then
					psi_val = -psi(i,j,k)+psi_bulk
					if(psi_val.gt.exp_max) 
     1						psi_val=exp_max
					if(psi_val.lt.exp_min) 
     1						psi_val=exp_min
					psi_coef=exp(psi_val)
					l_dip=l_dip*psi_coef
				endif
			else
				u_val = 0
				exp_mu = 1
				l_dip = 0
			endif
c
			g = 0
			d = (1.d0 - phi0)*exp_mu + l_dip
			gprime = 0
			dprime = 0
c
c		Add all ions
c
			do 1600 i_ion = 1,n_ion
c
				argument = -ion_z(i_ion)*value -u_val
c
c				Make sure that the argument of the exponential function is not
c				too large (or too small), so that we do not have overflow (or
c				underflow)
c
				if(argument.gt.exp_max) argument=exp_max
				if(argument.lt.exp_min) argument=exp_min
c
				argument = exp(argument)
c
				val = kappa(i_ion)*argument
				val1 = ion_cvol(i_ion)*argument
c
				g = g + val
				gprime = gprime -val*ion_z(i_ion)
c
				d = d + val1
				dprime = dprime -val1*ion_z(i_ion)
c
1600			continue
c
			dc_nl(i,j,k) = coef(i,j,k)*(gprime*d-g*dprime)
     1					/(d*d)
c
1700		    continue
1800		continue
1900	    continue
	endif
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	C_scal.f
c       ((non linear) function in PDE, based on Michael Holst MG package)
c	(This is the scalar version: only computes at one position in the grid)
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      8/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c
c	Purpose:	Defines the non linear term (vector version)
c
c	Input parameters:
c
c		coef:	defines region in which ion is considered
c		u:	dimensionless potential (in)
c		ipkey:  define type of equation
c
c	Output:
c
c		c_scal: value of non linear term in PBE
c
	function c_scal(coef,u,ipkey)
c
	include 'poisson.h'
c
c ==========================================================================================
c == Define all variables                  =================================================
c ==========================================================================================
c
c	Local variables
c===========================================================================================
c
	integer	i,j,k,i_ion
c
	real*8	argument,argument0,coef2,coef20
	real*8	exp_min,exp_max
	parameter	(exp_min = -80.d0, exp_max = 80.d0)
c
c	Parameters (in and out)
c===========================================================================================
c
	integer	ipkey
c
	real*8	coef,u,c_scal
c
c	information on ions
c===========================================================================================
c
	integer	n_ion,n_zz
c
	real*8	ratio,phi0,coef_Z0,coef_Z
        real*8  zp_max,zn_max,phi,exp_phi
	real*8	D,g,val1
	real*8	ion_conc(max_ion),ion_z(max_ion),kappa(max_ion)
	real*8	ion_rad(max_ion)
	real*8	ion_cvol(max_ion)
c
c ==========================================================================================
c == Common block (define ions)            =================================================
c ==========================================================================================
c
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
        common /ion_max/        zp_max,zn_max
	common /smpb/		ion_cvol,ratio,phi0,coef_Z0
c
c
c ==========================================================================================
c == Compute non linear term               =================================================
c ==========================================================================================
c
	c_scal = 0.d0
c
c	If solving Poisson equation, return
c
	if(ipkey.eq.0) return
c
c	Loop over all ions
c
c	If solving linearized Poisson Boltzmann, or non linear Poisson Boltzmann:
c
	if(ipkey.eq.1) then
c
	    do 700 i_ion = 1,n_ion
c
		coef20 = kappa(i_ion)
c
		argument0 = -ion_z(i_ion)
c
		coef2 = coef20 * coef
		argument = argument0*u
c
c		Make sure that the argument of the exponential function is not
c		too large (or too small), so that we do not have overflow (or
c		underflow)
c
		if(argument.gt.exp_max) argument = exp_max
		if(argument.lt.exp_min) argument = exp_min
c
		c_scal = c_scal + coef2*(1+argument)
c
700	    continue
c
	elseif(ipkey.eq.2) then
c
	    do 1100 i_ion = 1,n_ion
c
		coef20 = kappa(i_ion)
c
		argument0 = -ion_z(i_ion)
c
		coef2 = coef20 * coef
		argument = argument0*u
c
c		Make sure that the argument of the exponential function is not
c		too large (or too small), so that we do not have overflow (or
c		underflow)
c
		if(argument.gt.exp_max) argument = exp_max
		if(argument.lt.exp_min) argument = exp_min
c
		c_scal = c_scal+coef2*exp(argument)
c
1100	    continue
c
	elseif(ipkey.eq.3) then
c
		if(coef.eq.0) then
			c_scal = 0.d0
			return
		endif
c
c		Add ion one (small size)
c
                if(ion_z(1)*u.gt.0) then
c
		        argument = -ion_z(1)*u
c
		        if(argument.lt.exp_min) argument = exp_min
c
		        argument = exp(argument)
c
		        val1 = (1.d0-phi0+ion_cvol(1)*argument)
		        coef_Z = coef_Z0/val1
		        coef_Z = coef_Z**(ratio-1)
c
                else
		        argument = ion_z(1)*u
c
		        if(argument.lt.exp_min) argument = exp_min
c
		        argument = exp(argument)
c
		        val1 = (1.d0-phi0)*argument+ion_cvol(1)
		        coef_Z = coef_Z0*argument/val1
		        coef_Z = coef_Z**(ratio-1)
                endif
c
                if(u.gt.0) then
                        phi = zn_max*u
                else
                        phi = zp_max*u
                endif
                exp_phi = exp(phi)
c
                argument = -ion_z(1)*u + phi
		if(argument.lt.exp_min) argument = exp_min
                argument = exp(argument)
		g = kappa(1)*argument
		d = (1.d0-phi0)*exp_phi + ion_cvol(1)*argument
c
c		Add all other ions
c
		do 1200 i_ion = 2,n_ion
c
			argument = -ion_z(i_ion)*u +phi
c
c			Make sure that the argument of the exponential function is not
c			too large (or too small), so that we do not have overflow (or
c			underflow)
c
			if(argument.lt.exp_min) argument=exp_min
c
			argument = exp(argument)
c
			g = g + coef_Z*kappa(i_ion)*argument
c
			d = d + coef_Z*ion_cvol(i_ion)*argument
c
1200		continue
c
		c_scal = coef*g/d
c
	elseif(ipkey.eq.4) then
c
c		We set c_scal to zero, as it is included in the Jacobian
c
		c_scal = 0
c
	endif
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	DC_scal.f
c       (derivative of (non linear) term  in PBE, based on Michael Holst MG package)
c	(This is the scalar version)
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      8/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c
c	Purpose:	Defines the derivative of the non linear term (vector version)
c
c	Input parameters:
c
c		coef:	defines region in which ion is considered
c		u:	dimensionless potential (in)
c		ipkey:  not used!
c
c	Output:
c
c		dc_scal: value of non linear term in PBE
c
	function dc_scal(coef,u,ipkey)
c
	include 'poisson.h'
c
c ==========================================================================================
c == Define all variables                  =================================================
c ==========================================================================================
c
c	Local variables
c===========================================================================================
c
	integer	i,j,k,i_ion
c
	real*8	argument,coef2,argument0,coef20
	real*8	exp_min,exp_max
	parameter	(exp_min = -80.d0, exp_max = 80.d0)
c
c	Parameters (in and out)
c===========================================================================================
c
	integer	ipkey
c
	real*8	coef,u,dc_scal
c
c	information on ions
c===========================================================================================
c
	integer	n_ion,n_zz
c
	real*8	D,dprime,g,gprime,val,val1,dval1
        real*8  zn_max,zp_max,phi,exp_phi
	real*8	ratio,phi0,coef_Z0,coef_Z,coef_Z_km1,dcoef_Z_km1
	real*8	ion_conc(max_ion),ion_z(max_ion),kappa(max_ion)
	real*8	ion_rad(max_ion)
	real*8	ion_cvol(max_ion)
c
c ==========================================================================================
c == Common block (define ions)            =================================================
c ==========================================================================================
c
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
        common /ion_max/        zp_max,zn_max
	common /smpb/		ion_cvol,ratio,phi0,coef_Z0
c
c
c ==========================================================================================
c == Compute non linear term               =================================================
c ==========================================================================================
c
	dc_scal = 0
c
	if(ipkey.eq.0) return
c
	if(ipkey.eq.1) then
c
	    do 700 i_ion = 1,n_ion
c
		coef20 = kappa(i_ion)
c
		argument0 = -ion_z(i_ion)
c
		coef2 = coef20 * coef
c
		dc_scal = dc_scal + coef2*argument0
c
700	    continue
c
	elseif(ipkey.eq.2) then
c
	    do 1100 i_ion = 1,n_ion
c
		coef20 = -ion_z(i_ion)*kappa(i_ion)
c
		argument0 = -ion_z(i_ion)
c
		coef2 = coef20 * coef
		argument = argument0*u
c
c		Make sure that the argument of the exponential function is not
c		too large (or too small), so that we do not have overflow (or
c		underflow)
c
		if(argument.gt.exp_max) argument = exp_max
		if(argument.lt.exp_min) argument = exp_min
c
		dc_scal = dc_scal+coef2*exp(argument)
c
1100	    continue
c
	elseif(ipkey.eq.3) then
c
		if(coef.eq.0) then
			dc_scal = 0.d0
			return
		endif
c
c		Add ion one (small size)
c
                if(ion_z(1)*u.gt.0) then
c
		        argument = -ion_z(1)*u
c
		        if(argument.lt.exp_min) argument = exp_min
c
		        argument = exp(argument)
c
		        val1 = (1.d0-phi0+ion_cvol(1)*argument)
		        dval1 = -ion_z(1)*ion_cvol(1)*argument
		        coef_Z = coef_Z0/val1
		        coef_Z_km1 = coef_Z**(ratio-1)
		        dcoef_Z_km1 = -(ratio-1)*coef_Z_km1*dval1/val1
c
                else
c
		        argument = ion_z(1)*u
c
		        if(argument.lt.exp_min) argument = exp_min
c
		        argument = exp(argument)
c
		        val1 = (1.d0-phi0)*argument+ion_cvol(1)
		        dval1 = -ion_z(1)*ion_cvol(1)
		        coef_Z = coef_Z0*argument/val1
		        coef_Z_km1 = coef_Z**(ratio-1)
		        dcoef_Z_km1 = -(ratio-1)*coef_Z_km1*dval1/val1
c
                endif
c
                if(u.gt.0) then
                        phi = zn_max*u
                else
                        phi = zp_max*u
                endif
                exp_phi = exp(phi)
                argument = -ion_z(1)*u + phi
		if(argument.lt.exp_min) argument = exp_min
		argument = exp(argument)
c
		g = kappa(1)*argument
		d = (1.d0-phi0)*exp_phi +ion_cvol(1)*argument
		dprime = -ion_z(1)*ion_cvol(1)*argument
		gprime = -ion_z(1)*g
c
c		Add all other ions
c
		do 1200 i_ion = 2,n_ion
c
			argument = -ion_z(i_ion)*u + phi
c
c			Make sure that the argument of the exponential function is not
c			too large (or too small), so that we do not have overflow (or
c			underflow)
c
			if(argument.lt.exp_min) argument=exp_min
c
			argument = exp(argument)
c
			val = coef_Z_km1*kappa(i_ion)*argument
			val1=coef_Z_km1*ion_cvol(i_ion)*argument
c
			g = g + val
c
			gprime = gprime + dcoef_Z_km1*
     1			kappa(i_ion)*argument -val*ion_z(i_ion)
c
			d = d + val1
			dprime = dprime - val1*ion_z(i_ion)
     1			+ dcoef_Z_km1*ion_cvol(i_ion)*argument
c
1200		continue
c
		dc_scal = coef*(gprime*d-g*dprime)
     1					/(d*d)
c
	elseif(ipkey.eq.4) then
c
c		We set dc_scal to zero, as it is included in the Jacobian
c
		dc_scal = 0.d0
c
	endif
c
	return
	end
