c       Copyright (C) 2007 Patrice Koehl
c
c       This library is free software; you can redistribute it and/or
c       modify it under the terms of the GNU Lesser General Public
c       License as published by the Free Software Foundation; either
c       version 2.1 of the License, or (at your option) any later version.
c
c       This library is distributed in the hope that it will be useful,
c       but WITHOUT ANY WARRANTY; without even the implied warranty of
c       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
c       Lesser General Public License for more details.
c
c       You should have received a copy of the GNU Lesser General Public
c       License along with this library; if not, write to the Free Software
c       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
c
c ==========================================================================================
c ==========================================================================================
c	PDE_init.f
c       (PDE solver, based on Michael Holst MG package)
c ==========================================================================================
c ==========================================================================================
c
        subroutine PDE_init(iparam,rparam,nx,ny,nz,hx,hy,hz,
     1                  xf,yf,zf,itm,rwork_size,iwork_size,info)
c
	include 'poisson.h'
c
c ==========================================================================================
c == Define all variables                        ===========================================
c ==========================================================================================
c	
c	Local variables
c ==========================================================================================
c
	integer	i,itm,info
c
	integer	flag_method,nlevel,nlin,equation,ndiscrete,ijac,nsmooth
        integer rwork_size,iwork_size
c
c	Variables for Michael Holst MG package
c ==========================================================================================
c
        integer nx,ny,nz
	integer	iparam(100)
	integer	nlev,nu1,nu2,mgkey,itmax,istop,iinfo
	integer	irite,ipkey,ipcon,nonlin,mgprol,mgcoar,mgdisc
	integer	mgsmoo,iperf
	integer	mgsolv
c
	real*8	errtol,tol
	real*8	omegal,omegan
c
	real*8	rparam(100)
c
	real*8	xf(nx,5),yf(ny,5),zf(nz,5)
        real*8  hx(*),hy(*),hz(*)
c
c	Common blocks
c ==========================================================================================
c
	common /pdesolver/	tol,flag_method,nlevel,nlin,equation,
     1                          ndiscrete,ijac,nsmooth
c
c ==========================================================================================
c == Set all parameters                          ===========================================
c ==========================================================================================
c	
c	Parameters for solver:
c ==========================================================================================
c
c	Parameters from input parameter file
c
	errtol	= tol
c
c	Hard coded parameters:
c
	itmax = itm		! maximum number of iterations
	istop = 1		! stopping criteria
	iinfo = 1		! runtime status messages:
				! 0=none, 1=some, 2=many, 3=more)
	irite = 8		! probably useless: logical unit number for output file
	ipkey = equation	! Define the Helmholtz term in the equation:
                                ! zero (0), linear PB (1), non linear PB (2), Size modified PB (3)
				! Generalized PB Langevin (4)
c	key   = 0		! print output to ascii file (0=no, 1=yes)
	iperf = 0		! analysis, based on true solution (0=no)
c
        nonlin = nlin           ! 0 (linear), 1 (nonlinear)
	mgkey  = 0		! multigrid method: 0=variable v-cycle, 1=nested operation
	nu1    = 2	        ! number of pre-smoothings
	nu2    = 2	        ! number of post-smoothings
	mgsmoo = nsmooth	! type of smoothing (1=gsrb, 2=wjac, 3=sor, 4=cghs, 5=Richardson)
	mgprol = 0		! prolongation (0=trilinear)
	mgcoar = ndiscrete	! coarse equations (0=stan, 1=harm, 2=Galerkin)
	mgsolv = 0		! coarse equation solver(0=cghs, 1 = nanded linpack)
	mgdisc = 0		! discretization (0=fv, 1=fe)
	ipcon  = 3		! preconditioner flag (0=diag, 1=iccg, 2=iccgdw, 3=miccgdw, 4=none)
c
	nlev   = nlevel  	! Number of mesh levels to use (max 6 for 65 grid points)
	omegal = 0.9d0		! linear relaxation parameter
	omegan = 0.80d0		! non linear relaxation parameter
c
c	Pack iparam and rparam, to match with MG code
c ==========================================================================================
c
	do 100 i = 1,100
		iparam(i) = 0
		rparam(i) = 0
100	continue
c
	iparam(1)	= rwork_size
	iparam(2)	= iwork_size
	iparam(3)	= nx
	iparam(4)	= ny
	iparam(5)	= nz
	iparam(6)	= nlev
	iparam(7)	= nu1
	iparam(8)	= nu2
	iparam(9)	= mgkey
	iparam(10)	= itmax
	iparam(11)	= istop
	iparam(12)	= iinfo
	iparam(13)	= irite
	iparam(14)	= ipkey
	iparam(15)	= ipcon
	iparam(16)	= nonlin
	iparam(17)	= mgprol
	iparam(18)	= mgcoar
	iparam(19)	= mgdisc
	iparam(20)	= mgsmoo
	iparam(21)	= mgsolv
	iparam(22)	= iperf
        iparam(23)      = info
c
	rparam(1)	= errtol
	rparam(3)	= 0.d0
	rparam(4)	= hx(nx)
	rparam(5)	= 0.d0
	rparam(6)	= hy(ny)
	rparam(7)	= 0.d0
	rparam(8)	= hz(nz)
	rparam(9)	= omegal
	rparam(10)	= omegan
c
c ==========================================================================================
c == fill in arrays that define PDE              ===========================================
c ==========================================================================================
c
c	Start by defining coordinates of grid points
c ==========================================================================================
c
	do 200 i = 1,nx
		xf(i,1) = hx(i)
200	continue
c
	do 300 i = 1,ny
		yf(i,1) = hy(i)
300	continue
c
	do 400 i = 1,nz
		zf(i,1) = hz(i)
400	continue
c
        return
        end
c
c ==========================================================================================
c ==========================================================================================
c	PDE_wrapper.f
c       (PDE solver, based on Michael Holst MG package)
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      7/6/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	provide a wrapper to Michael Holst PDE solvers
c
c	This is a general wrapper for Michael Holst Fortran package MG that solves
c	elliptic equation using a large collection of routines
c
c       it can solve:
c                       The Laplace equation
c                       The Poisson equation
c                       The linearized Poisson Boltzmann equation
c                       The Non linear Poisson Boltzmann equation
c                       The Dipolar Poisson Boltzmann Langevin equation
c
c	At this stage, we use the version based on finite differences. In the future,
c	we want to use finite elements
c
	subroutine solve_pde(grid_shift,epsilon_X,epsilon_Y,
     1		epsilon_Z,scalar_map,crg_map,gxcf,gycf,gzcf,
     2          xf,yf,zf,nx,ny,nz,phi_map,
     3          rwork,iwork,rparam,iparam)
c
	include 'poisson.h'
c
c ==========================================================================================
c == Define all variables                        ===========================================
c ==========================================================================================
c	
c	Local variables
c ==========================================================================================
c
	integer	i
c
	integer	flag_method,nlevel,nlin,equation,ndiscrete,ijac,nsmooth
c
c	Variables for Michael Holst MG package
c ==========================================================================================
c
        integer nx,ny,nz
	integer	iparam(100),iwork(*)
	integer	nlev,nu1,nu2,mgkey,itmax,istop,iinfo
	integer	irite,ipkey,ipcon,nonlin,mgprol,mgcoar,mgdisc
	integer	mgsmoo,iperf
	integer	mgsolv
c
	real*8	errtol,tol
	real*8	omegal,omegan
c
	real*8	rwork(*)
	real*8	rparam(100)
c
	real*8	xf(nx,5),yf(ny,5),zf(nz,5)
	real*8	gxcf(ny,nz,5),gycf(nx,nz,5)
	real*8	gzcf(nx,ny,5)
c
c	Information on grid
c ==========================================================================================
c
	integer*1 grid_shift(*)
c
c	Information on molecule
c ==========================================================================================
c
	integer	natom,ncharge,iform
c
	real*8	atom_coord(*)
	real*8	atom_radius(*),atom_charge(*)
c
	character atom_label(*)*30
	character atom_label2(*)*12
c
c	Information on maps over the grid:
c ==========================================================================================
c
c	phi_map:	electrostatics potential
c	crg_map:	charge map
c
	real*8	phi_map(*)
	real*8	crg_map(*)
	real*8	scalar_map(*)
c
c	epsilon_X defines the permittivity on the edges along the X axis
c	epsilon_Y defines the permittivity on the edges along the Y axis
c	epsilon_Z defines the permittivity on the edges along the Z axis
c
	real*8	epsilon_X(*)
	real*8	epsilon_Y(*)
	real*8	epsilon_Z(*)
c
c       Pointers
c ==========================================================================================
c
        pointer (ptr_coord,atom_coord)
        pointer (ptr_radius,atom_radius)
        pointer (ptr_charge,atom_charge)
        pointer (ptr_label,atom_label)
        pointer (ptr_label2,atom_label2)
c
c	Common blocks
c ==========================================================================================
c
        common /charges/        ptr_coord,ptr_charge,ptr_radius,
     1                          ptr_label,ptr_label2,iform,
     2                          natom,ncharge
	common /pdesolver/	tol,flag_method,nlevel,nlin,equation,
     1                          ndiscrete,ijac,nsmooth
c
c	We are solving either the Poisson Boltzmann equation, or
c	the size modified PB equation (SMPB), or the Generalized
c	Poisson Boltzmann Langevin (GPBL) equation.
c
c	the PB equation is:
c
c	-Delta (epsilon(r) Delta(u(r))) + kappa(r) f(u(r)) = C Sum_i zi d(r-ri)
c
c	Where:
c
c		u is the dimensionless electrostatic potential
c		(i.e. u(r) = electron * phi(r)/(kB T) )
c
c		Delta means Laplacian

c		Epsilon is the permittivity:
c			epsilon(r) = eps1 	inside the solute
c			epsilon(r) = eps2	inside the solvent
c
c		For each ion type:
c		kappa is the Debye-Huckel coefficient that depends on the ionic
c		concentration C :
c		kappa = Ck *(298/T) * ion_conc * ion_z
c			Ck = 4.2434514035 in CGS unit
c			ion_conc is the ionic concentration in mol/l
c			ion_z is the valence
c
c		f(u) is:
c			f(u(r)) = - exp(-ion_z u(r))
c
c			(note that for the linearized Poisson Boltzmann equation
c			f(u(r)) = -1 + ion_z u(r)
c
c		C Sum_i zi d(r-r_i) is the charge source. The sum extends over
c		all charges i, placed at position i, of fractional value zi.
c		C = 7046.528885 (298/T) in CGS unit
c
c ==========================================================================================
c == Now solve the equation                      ===========================================
c ==========================================================================================
c
c
c	Test all methods implemented by Michael Holst!!
c
	if(flag_method.eq.0) then
c
c		Conjugate Gradient, Multi Grid for linear equation
c
                write(6,*) 'Conjugate Gradient, MultiGrid solver'
                write(6,*) ' '
		if(nonlin.eq.1) then
			write(6,*) ' '
			write(6,*) 'CGMG is only for linear equation !!'
			write(6,*) 'Please change method'
			write(6,*) ' '
			stop
		endif
c
		call CGMG_Solver(iparam,rparam,iwork,rwork,phi_map,
     1		xf,yf,zf,gxcf,gycf,gzcf,epsilon_X,epsilon_Y,epsilon_Z,
     2		scalar_map,crg_map)
c
	elseif(flag_method.eq.1) then
c
                if(equation.le.3.or.equation.eq.5) then
c
                        if(iparam(23).eq.1) then
                        write(6,*) 'Truncated Newton, Stencil Size : 7'
                        write(6,*) ' '
                        endif
c
		        call Newton_Solver_pbe(iparam,rparam,iwork,
     1                  rwork,phi_map,xf,yf,zf,gxcf,gycf,gzcf,epsilon_X,
     2                  epsilon_Y,epsilon_Z,scalar_map,crg_map)
c
                else
c
                        if(iparam(23).eq.1) then
                        if(ijac.eq.0) then
                          write(6,*) 'Inexact Newton, Stencil Size : 7'
                        else
                          write(6,*) 'Inexact Newton, Stencil Size : 27'
                        endif
                        write(6,*) ' '
                        endif
c
		        call Newton_Solver_gpbl(ijac,iparam,rparam,
     1                  iwork,rwork,phi_map,xf,yf,zf,gxcf,gycf,gzcf,
     2                  epsilon_X,epsilon_Y,epsilon_Z,scalar_map,
     3                  crg_map,grid_shift)
c
                endif
c
	elseif(flag_method.eq.2) then
c
                if(iparam(23).eq.1) then
                        write(6,*) 'Direct, iterative Solver'
                        write(6,*) ' '
                endif
c
                if(equation.le.3) then
c
		        call Direct_Solver_pbe(iparam,rparam,iwork,
     1                  rwork,phi_map,xf,yf,zf,gxcf,gycf,gzcf,
     2                  epsilon_X,epsilon_Y,epsilon_Z,scalar_map,
     3                  crg_map)
c
                else
c
		        call Direct_Solver_gpbl(iparam,rparam,iwork,
     1                  rwork,phi_map,xf,yf,zf,gxcf,gycf,gzcf,
     2                  epsilon_X,epsilon_Y,epsilon_Z,scalar_map,
     3                  crg_map,grid_shift)
c
                endif
c
	elseif(flag_method.eq.3) then
c
                if(iparam(23).eq.1) then
                        write(6,*) 'Full Approximation Solver (FAS)'
                        write(6,*) ' '
                endif
c
                if(equation.le.3) then
c
		        call FAS_Solver_pbe(iparam,rparam,iwork,rwork,
     1                  phi_map,xf,yf,zf,gxcf,gycf,gzcf,epsilon_X,
     2                  epsilon_Y,epsilon_Z,scalar_map,crg_map)
c
                else
c
		        call FAS_Solver_gpbl(iparam,rparam,iwork,rwork,
     1                  phi_map,xf,yf,zf,gxcf,gycf,gzcf,epsilon_X,
     2                  epsilon_Y,epsilon_Z,scalar_map,crg_map,
     3                  grid_shift)
c
                endif
c
	else
		write(6,*) ' '
		write(6,*) 'Bad method key ...'
		write(6,*) ' '
c
	endif
c
        return
        end
