c ==========================================================================================
c ==========================================================================================
c	Charge_distrib.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	7/05/07
c	Author:		Patrice Koehl 
c			in collaboration with Marc Delarue
c
c	This subroutine distributes the charges onto grid points.
c	Currently, there are two methods implemented:
c		- trilinear (the charge is distributed onto the 8 vertices
c			     defining the cuboid it is in)
c		- spherical (the charge is distributed onto all vertices
c		  that are inside the VdW sphere of the atom
c
c	Input:
c		flag_distrib:	flag defining the method of distribution
c					0 : trilinear
c					1 : spherical, uniform
c					2 : antialiasing (not yet implemented)
c                                       4 : Uniform on surface of molecule
c                                           (only for spherical molecules)
c
	subroutine charge_distrib(crg_map,grid_partition,hx,hy,hz,
     1                          nx,ny,nz)
c
	include 'poisson.h'
c
c ==========================================================================================
c	Define all variables
c ==========================================================================================
c
c	Local variables
c ==========================================================================================
c
	integer	i,j,k
	integer	nx,ny,nz
c
	integer	ipoint,istepx,istepy,istepz,itot
	integer	ix,iy,iz,kx,ky,kz
	integer	Nvert_in,n_in
c
        integer*1       flag(8)
        integer*1       grid_partition(nx,ny,nz)
c
	real*8	val,valx,valy,valz,rad,rad2,crg,r_max
	real*8	diffx,diffy,diffz
c
	integer	posx(*),posy(*),posz(*)
c
c	Information on grid
c ==========================================================================================
c
c
	real*8	dx,dy,dz
	real*8	hx(nx),hy(ny),hz(nz)
c
c	Information on molecule
c ==========================================================================================
c
	integer	natom,ncharge,iform
c
	real*8	atom_coord(*)
	real*8	atom_radius(*),atom_charge(*)
c
	character atom_label(*)*30
	character atom_label2(*)*12
c
c	Input parameters
c ==========================================================================================
c
	integer	surf_type,flag_pymol,flag_box,flag_distrib,flag_eps
	integer	grid_type,bnd_type
	integer	ion_zone
c
	real*8	R_probe,R_ion
c
c	Output
c ==========================================================================================
c
c	crg_map:		charge map
c
	real*8	crg_map(nx,ny,nz)
c
c	Constants
c ==========================================================================================
c
	real*8	l_Bjerrum,pi,Temp,KB,electron,Ang,beta,Avogadro,eps0
	real*8	C_charge,C_prop,dvol
c
c       Uniform spheres
c ==========================================================================================
c
        integer npoint_sphere, nsphere
        integer distidx(8)
        real*8  radP,Point(3),distval(8)
        real*8  allspheres(360,120)
c
c       Pointers       
c ==========================================================================================
c
        pointer (ptr_posx,posx)
        pointer (ptr_posy,posy)
        pointer (ptr_posz,posz)
c
        pointer (ptr_coord,atom_coord)
        pointer (ptr_radius,atom_radius)
        pointer (ptr_charge,atom_charge)
        pointer (ptr_label,atom_label)
        pointer (ptr_label2,atom_label2)
c
        common /charges/   ptr_coord,ptr_charge,ptr_radius,ptr_label,
     1                     ptr_label2,iform,natom,ncharge
	common /constants/ l_Bjerrum,pi,Temp,KB,electron,Ang,Avogadro,
     1			   eps0,beta,C_charge,C_prop
	common /input_par/ R_probe,R_ion,flag_pymol,surf_type,
     1			   grid_type,ion_zone,flag_box,
     2			   flag_distrib,flag_eps,bnd_type
        common /spheres/   allspheres
        common /point_sphere/ npoint_sphere
c
1	format('from pymol.cgo import *')
2	format('from pymol import cmd',/)
3	format('ChargeSurf = [',/,
     1  '   COLOR, 0.0, 0.0, 1.0,',/,/)
4	format('   SPHERE,',f8.3,',',f8.3,',',f8.3,',',f8.3,',')
5	format('   END',/,'   ]',/,/)
6	format('cmd.load_cgo(ChargeSurf,''ChargeSurf'')')
c
c ==========================================================================================
c	Initialise charge distribution to 0 everywhere
c ==========================================================================================
c
	do 300 k = 1,nz
		do 200 j = 1,ny
			do 100 i = 1,nx
				crg_map(i,j,k) = 0.d0
100			continue
200		continue
300	continue
c
	if(flag_distrib.eq.0) then
c
c ==========================================================================================
c	Method 1: trilinear
c ==========================================================================================
c
	     do 400 ipoint = 1,natom
c
                if(atom_charge(ipoint).eq.0) goto 400
c
c	   	Find grid cell to which it belongs (coordinate of the LFD vertex,
c		see below for definition of vertices)
c
c
		call locate_cell(atom_coord(3*(ipoint-1)+1),hx,ix,
     1				 nx)
		call locate_cell(atom_coord(3*(ipoint-1)+2),hy,iy,
     1				 ny)
		call locate_cell(atom_coord(3*(ipoint-1)+3),hz,iz,
     1				 nz)
c
                call flag_inside(nx,ny,nz,grid_partition,ix,iy,iz,
     1                  flag,itot)
c
		dx = hx(ix+1)-hx(ix)
		dy = hy(iy+1)-hy(iy)
		dz = hz(iz+1)-hz(iz)
c
                dvol = dx*dy*dz
c
c		Get distance from point to YZ, XZ and XY planes of the grid cell
c
c
		diffx = (atom_coord(3*(ipoint-1)+1)-hx(ix))/dx
		diffy = (atom_coord(3*(ipoint-1)+2)-hy(iy))/dy
		diffz = (atom_coord(3*(ipoint-1)+3)-hz(iz))/dz
c
c		Need to scale charge: factor (4*pi*(electron)^2 / (kB *Temp)
c		in front of source term (=C_charge), plus divide by cuboid
c		volume (dvol) to make it a Dirac function
c
c               We also exclude all grid points that fall outside the molecule,
c               in which case the remaining points get more charge.
c
		crg = (atom_charge(ipoint)*C_charge/dvol)*(8.d0/itot)
c
c		Now interpolate charge
c
		call trilinear_distrib(crg_map,ix,iy,iz,diffx,diffy,
     1		diffz,flag,crg,nx,ny,nz)
c
400	    continue
c
	elseif(flag_distrib.eq.1) then
c
c ==========================================================================================
c	Method 2: Uniform spherical
c		  For each charge, find list of vertices in the grid that are within the
c		  vdW sphere of the corresponding atom; if there are more than 8 vertices,
c		  distribute charge uniformly on these vertices; if there are less than
c		  8 vertices, use trilinear interpolation on the 8 vertices of the grid cell
c		  that contains the charge
c		  This method is adapter brom Bruccoleri, J. Comp. Chem., 14, 1417-1422 (1993)
c
c ==========================================================================================
c
c               First make sure arrays are big enough
c
            r_max = 0
            do 450 ipoint = 1,natom
                if(atom_radius(ipoint).gt.r_max) then
                        r_max = atom_radius(ipoint)
                endif
450         continue
            dx = hx(2)-hx(1)
            n_in = nint(2*r_max/dx) + 1
            n_in = n_in**3
            ptr_posx = malloc(4*n_in)
            ptr_posy = malloc(4*n_in)
            ptr_posz = malloc(4*n_in)
c
	    do 900 ipoint = 1,natom
c
                if(atom_charge(ipoint).eq.0) goto 900
c
c	   	Find grid cell to which it belongs (coordinate of the LFD vertex,
c		see below for definition of vertices)
c
c
		call locate_cell(atom_coord(3*(ipoint-1)+1),hx,ix,
     1				 nx)
		call locate_cell(atom_coord(3*(ipoint-1)+2),hy,iy,
     1				 ny)
		call locate_cell(atom_coord(3*(ipoint-1)+3),hz,iz,
     1				 nz)
c
		dx = hx(ix+1)-hx(ix)
		dy = hy(iy+1)-hy(iy)
		dz = hz(iz+1)-hz(iz)
c
                dvol = dx*dy*dz
c
		rad = atom_radius(ipoint)
		crg = atom_charge(ipoint)*C_charge/dvol
		rad2 = rad*rad
c
c	   	Number of vertices to check in each dimension:
c
		call defstep(rad,hx,ix,istepx,nx)
		call defstep(rad,hy,iy,istepy,ny)
		call defstep(rad,hz,iz,istepz,nz)
c
c		Now loop over all these vertices
c
		Nvert_in = 0
c
		do 700 k = 1,2*istepz + 1
c
		    kz = iz - istepz + k -1
		    valz=(atom_coord(3*ipoint)-hz(kz))**2
c
		    do 600 j = 1,2*istepy + 1
c
			ky = iy - istepy + j -1 
		   	valy=(atom_coord(3*ipoint-1)-hy(ky))**2
c
			do 500 i = 1,2*istepx + 1
c
				kx = ix - istepx + i -1
				valx=(atom_coord(3*ipoint-2)-hx(kx))**2
c
c				if vertex inside ball ipoint, store
c
				if(valx+valy+valz.lt.rad2.and.
     1                          ibits(grid_partition(kx,ky,kz),0,1)
     2                          .eq.0) then
					Nvert_in = Nvert_in + 1
					posx(Nvert_in)=kx
					posy(Nvert_in)=ky
					posz(Nvert_in)=kz
				endif
c
500			continue
600		    continue
700		continue
c
		if(Nvert_in.gt.8) then
c
c			Distribute charge uniformly
c
			val = crg/dble(Nvert_in)
			do 800 i = 1,Nvert_in
			   crg_map(posx(i),posy(i),posz(i))=
     1			   crg_map(posx(i),posy(i),posz(i))+val
800			continue
c
		else
c
c			Do trilinear interpolation
c
c			Get distance from point to YZ, XZ and XY planes of the grid cell
c
			dx = hx(ix+1)-hx(ix)
			dy = hy(iy+1)-hy(iy)
			dz = hz(iz+1)-hz(iz)
c
			diffx = (atom_coord(3*(ipoint-1)+1)-hx(ix))/dx
			diffy = (atom_coord(3*(ipoint-1)+2)-hy(iy))/dy
			diffz = (atom_coord(3*(ipoint-1)+3)-hz(iz))/dz
c
                        call flag_inside(nx,ny,nz,grid_partition,ix,iy,
     1                  iz,flag,itot)
c
                        crg = crg * 8.d0/itot
c
			call trilinear_distrib(crg_map,ix,iy,iz,diffx,
     1			diffy,diffz,flag,crg,nx,ny,nz)
c
		endif
c
900	    continue
c
            call free(ptr_posx)
            call free(ptr_posy)
            call free(ptr_posz)
c
        elseif(flag_distrib.eq.2) then
c
                write(6,*) 'Antialiasing not yet implemented!'
                stop
c
	elseif(flag_distrib.eq.4) then
c
c ==========================================================================================
c	Method 4: Set charges uniformly on the surface of each atom
c ==========================================================================================
c
             open(unit=15,file='ChargeSurf.py',status='unknown')
             write(15,1)
             write(15,2)
             write(15,3)
             radP = 0.05
c
	     do 1500 ipoint = 1,natom
c
                npoint_sphere = 4*nint(pi*atom_radius(ipoint)*
     1                  atom_radius(ipoint)*3)
                if(npoint_sphere.lt.20.or.npoint_sphere.gt.120) then
                        write(6,*) 'Cannot handle npoint_sphere =',
     1                  npoint_sphere
                        stop
                endif
c
                if(atom_charge(ipoint).eq.0) goto 1500
c
c		Need to scale charge: factor (4*pi*(electron)^2 / (kB *Temp)
c		in front of source term (=C_charge), plus divide by cuboid
c		volume (dvol) to make it a Dirac function
c
		crg = (atom_charge(ipoint)*C_charge/dvol)
c
c               Loop over all points on surface of sphere
c
                crg = crg/npoint_sphere
c
                do 1400 i = 1,npoint_sphere
c
                        do 1000 j = 1,3
                                Point(j)=atom_radius(ipoint)*
     1                          allspheres(3*(i-1)+j,npoint_sphere)
1000                    continue
                        write(15,4) Point(1),Point(2),Point(3),radP
                        
c	   	Find grid cell to which Point belongs (coordinate of the LFD vertex,
c		see below for definition of vertices)
c
c
		        call locate_cell(Point(1),hx,ix,nx)
		        call locate_cell(Point(2),hy,iy,ny)
		        call locate_cell(Point(3),hz,iz,nz)
c
                        call flag_inside(nx,ny,nz,grid_partition,
     1                  ix,iy,iz,flag,itot)
c
		        dx = hx(ix+1)-hx(ix)
		        dy = hy(iy+1)-hy(iy)
		        dz = hz(iz+1)-hz(iz)
c
c		Get distance from point to YZ, XZ and XY planes of the grid cell
c
		        diffx = (atom_coord(3*(ipoint-1)+1)-hx(ix))/dx
		        diffy = (atom_coord(3*(ipoint-1)+2)-hy(iy))/dy
		        diffz = (atom_coord(3*(ipoint-1)+3)-hz(iz))/dz
c
                        distval(1) = diffx**2 + diffy**2 + diffz**2
                        distval(2) = (1-diffx)**2 + diffy**2 + diffz**2
                        distval(3) = diffx**2 + (1-diffy)**2 + diffz**2
                        distval(4) = (1-diffx)**2+(1-diffy)**2+diffz**2
                        distval(5) = diffx**2 + diffy**2 + (1-diffz)**2
                        distval(6) = (1-diffx)**2+diffy**2+ (1-diffz)**2
                        distval(7) = diffx**2+(1-diffy)**2 +(1-diffz)**2
                        distval(8) = (1-diffx)**2+(1-diffy)**2+
     1                                  (1-diffz)**2
c
                        call hpsort_key(distval,distidx,8)
c
                        do 1100 j = 1,8
                                if(flag(distidx(j)).ne.0) goto 1200
1100                    continue
c
                        write(6,*) 'Problem with point number: ',i
c
1200                    continue
c
                        if(distidx(j).eq.1) then
                         crg_map(ix,iy,iz)=crg_map(ix,iy,iz)+crg
                        elseif(distidx(j).eq.2) then
                         crg_map(ix+1,iy,iz)=crg_map(ix+1,iy,iz)
     1                                  +crg
                        elseif(distidx(j).eq.3) then
                         crg_map(ix,iy+1,iz)=crg_map(ix,iy+1,iz)
     1                                  +crg
                        elseif(distidx(j).eq.4) then
                         crg_map(ix+1,iy+1,iz)=crg_map(ix+1,iy+1,iz)
     1                                  +crg
                        elseif(distidx(j).eq.5) then
                         crg_map(ix,iy,iz+1)=crg_map(ix,iy,iz+1)
     1                                  +crg
                        elseif(distidx(j).eq.6) then
                         crg_map(ix+1,iy,iz+1)=crg_map(ix+1,iy,iz+1)
     1                                  +crg
                        elseif(distidx(j).eq.7) then
                         crg_map(ix,iy+1,iz+1)=crg_map(ix,iy+1,iz+1)
     1                                  +crg
                        elseif(distidx(j).eq.8) then
                         crg_map(ix+1,iy+1,iz+1)=crg_map(ix+1,iy+1,iz+1)
     1                                  +crg
                        endif
c
1400              continue
c
1500            continue
c
                write(15,5)
                write(15,6)
                close(unit=15)
c
	endif
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	trilinear_distrib.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	7/05/07
c	Author:		Patrice Koehl 
c			in collaboration with Marc Delarue
c
c	This subroutine distributes a charge on the 8 corners of a grid cell
c
c	The grid cell has 6 faces: L(eft) R(ight), F(ront) B(ack),
c	and U(p) D(own).
c	The eight vertices can be defined as :
c	LFD, LFU, RFD, RFU, LBD, LBU, RBD, RBU.
c
c	Input:
c
c		ix,iy,iz:	position of the LFB corner of the cube
c
c		diffx,diffy,diffz : distance of point to the YZ, XZ and XY plane of the
c				    grid cell, respectively
c
c               flag              : array of size 8; each element is a flag
c                                   set to 1 if corresponding corner of cube
c                                   is inside the molecule surface, 0 otherwise
c		crg		  : quantity to be distributed
c		crg_map		  : map to be updated		  
c
	subroutine trilinear_distrib(crg_map,ix,iy,iz,diffx,diffy,
     1			diffz,flag,crg,nx,ny,nz)
c
c ==========================================================================================
c = Define variables                          
c ==========================================================================================
c
	integer	ix,iy,iz
	integer	nx,ny,nz
c
        integer*1       flag(8)
c
	real*8	diffx,diffy,diffz,One_diffx,One_diffy,One_diffz
	real*8	crg
c
	real*8	crg_map(nx,ny,nz)
c
	One_diffx = 1.d0-diffx
	One_diffy = 1.d0-diffy
	One_diffz = 1.d0-diffz
c
c ==========================================================================================
c = Interpolate value                         
c ==========================================================================================
c
c	Charge on LFD (point (ix,iy,iz))
c
	crg_map(ix,iy,iz) = crg_map(ix,iy,iz)+
     1		flag(1)*One_diffx*One_diffy*One_diffz*crg
c
c	Charge on RFD (point (ix+1,iy,iz)
c
	crg_map(ix+1,iy,iz) = crg_map(ix+1,iy,iz)+
     1		flag(2)*diffx*One_diffy*One_diffz*crg
c
c	Charge on LBD (point (ix,iy+1,iz)
c
	crg_map(ix,iy+1,iz) = crg_map(ix,iy+1,iz)+
     1		flag(3)*One_diffx*diffy*One_diffz*crg
c
c	Charge on RBD (point (ix+1,iy+1,iz)
c
	crg_map(ix+1,iy+1,iz) = crg_map(ix+1,iy+1,iz)+
     1		flag(4)*diffx*diffy*One_diffz*crg
c
c	Charge on LFU (point (ix,iy,iz+1)
c
	crg_map(ix,iy,iz+1) = crg_map(ix,iy,iz+1)+
     1		flag(5)*One_diffx*One_diffy*diffz*crg
c
c	Charge on RFU (point (ix+1,iy,iz+1)
c
	crg_map(ix+1,iy,iz+1)=crg_map(ix+1,iy,iz+1)+
     1		flag(6)*diffx*One_diffy*diffz*crg
c
c	Charge on LBU (point (ix,iy+1,iz+1)
c
	crg_map(ix,iy+1,iz+1) = crg_map(ix,iy+1,iz+1)+
     1		flag(7)*One_diffx*diffy*diffz*crg
c
c	Charge on RBU (point (ix+1,iy+1,iz+1)
c
	crg_map(ix+1,iy+1,iz+1)=crg_map(ix+1,iy+1,iz+1)+
     1		flag(8)*diffx*diffy*diffz*crg
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	Flag_inside.f
c ==========================================================================================
c ==========================================================================================
c
c       Flag_inside tests if the grid points at the vertices of a cube are
c       inside the surface of the molecule, or outside; this is needed for
c       projecting the charges on the grid: we do not want to put a charge in
c       the solvent!
c
c       The array flag is organised as follows:
c
c               Position 1:     i,j,k (LFD)
c               Position 2:     i+1,j,k (RFD)
c               Position 3:     i,j+1,k (LBD)
c               Position 4:     i+1,j+1,k (RBD)
c               Position 5:     i,j,k+1 (LFU)
c               Position 2:     i+1,j,k+1 (RFU)
c               Position 3:     i,j+1,k+1 (LBU)
c               Position 4:     i+1,j+1,k+1 (RBU)
c
c       Input:
c               nx,ny,nz:       size of the grid
c               i,j,k:          position of LFD vertex of the cube
c
c       Output:
c               flag            the flag array
c               itot            total number of points with flag = 1
c
        subroutine flag_inside(nx,ny,nz,grid_partition,i,j,k,
     1                  flag,itot)
c
c ==========================================================================================
c       Declare all variables
c ==========================================================================================
c
        integer nx,ny,nz
        integer i,j,k,l,itot
c
        integer*1       flag(8)
        integer*1       grid_partition(nx,ny,nz)
c
        do 100 l = 1,8
                flag(l) = 1
100     continue
c
        if(ibits(grid_partition(i,j,k),0,1).ne.0) flag(1) = 0
        if(ibits(grid_partition(i+1,j,k),0,1).ne.0) flag(2) = 0
        if(ibits(grid_partition(i,j+1,k),0,1).ne.0) flag(3) = 0
        if(ibits(grid_partition(i+1,j+1,k),0,1).ne.0) flag(4) = 0
        if(ibits(grid_partition(i,j,k+1),0,1).ne.0) flag(5) = 0
        if(ibits(grid_partition(i+1,j,k+1),0,1).ne.0) flag(6) = 0
        if(ibits(grid_partition(i,j+1,k+1),0,1).ne.0) flag(7) = 0
        if(ibits(grid_partition(i+1,j+1,k+1),0,1).ne.0) flag(8) = 0
c
        itot = 0
        do 200 l = 1,8
                itot = itot + flag(l)
200     continue
c
        return
        end
c
c ==========================================================================================
c ==========================================================================================
c	set_epsilon.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      7/13/07
c	Version 2:	8/28/07: uses GenSurf to generate epsilon map
c
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Set epsilon values on the grid: 
c	The surface of the protein separates two regions: the interior,
c	which is then set with epsilon = epsilon1, and the exterior where
c	the solvent is, set with epsilon = epsilon2
c
c	Input:
c		epsilon 1	: "interior" epsilon
c		epsilon 2	: "exterior" epsilon
c		grid_shift    	: geometry of the molecule on the grid
c		flag_eps    	: flag for computing epsilon:
c					0 : epsilon set everywhere to 1
c					1 : define epsilon on the middle of 
c					    an edge according to the position
c					    of the two adjacent grid points
c					2 : perform harmonic averaging, 9 points
c					3 : perform harmonic averaging, 15 points
c					    (not implemented)
c
c	Information available through common block:
c
c		NpointSurf	: number of points sampling the accessible surface
c				  area
c		Pointsurf	: coordinates of these points
c
c	Output:
c		epsilon_X	: epsilon values on the grid (half point on along X axis
c		epsilon_Y	: epsilon values on the grid (half point on along Y axis
c		epsilon_Z	: epsilon values on the grid (half point on along Z axis
c
c
	subroutine set_epsilon(epsilon_X,epsilon_Y,epsilon_Z,grid_shift,
     1			hx,hy,hz,nx,ny,nz)
c
c ==========================================================================================
c == Define all variables                        ===========================================
c ==========================================================================================
c	
c	Local variables
c ==========================================================================================
c
	integer	i,j,k
	integer	ipos,nval
	integer	nx,ny,nz
c
	real*8	ep1,val
        real*8  dx,dy,dz
c
	real*8	eps_val(3)
        real*8  mol_min(3),mol_max(3)
        real*8  test(3),rad
c
c	Output
c ==========================================================================================
c
	real*8	epsilon_X(nx,ny,nz)
	real*8	epsilon_Y(nx,ny,nz)
	real*8	epsilon_Z(nx,ny,nz)
c
c	Working array
c ==========================================================================================
c
	real*8	work_X(*)
	real*8	work_Y(*)
	real*8	work_Z(*)
c
c	Dielectric information
c ==========================================================================================
c
	real*8	eps1,eps2
c
c	Information on grid
c ==========================================================================================
c
	real*8	hx(nx),hy(ny),hz(nz)
        real*8  hz_shift(*)
c
c	Information on shifted grids
c ==========================================================================================
c
	integer*1 grid_shift(nx,ny,nz)
c
c       Information on z-slab (if any)
c ==========================================================================================
c
        integer flag_slab
        real*8  z0,zh,eps_slab,dist_test
c
c	Parameter values
c ==========================================================================================
c
	integer	surf_type,flag_pymol,flag_box,flag_distrib,flag_eps
	integer	grid_type,bnd_type
	integer	ion_zone
c
	real*8	R_probe,R_ion
c
c       Pointers
c ==========================================================================================
c
        pointer (ptr_hz_shift,hz_shift)
        pointer (ptr_Work_X,Work_X)
        pointer (ptr_Work_Y,Work_Y)
        pointer (ptr_Work_Z,Work_Z)
c
c ==========================================================================================
c == Define all common blocks                    ===========================================
c ==========================================================================================
c	
	common /elec_const/ 	eps1,eps2
	common /input_par/	R_probe,R_ion,flag_pymol,surf_type,
     1				grid_type,ion_zone,flag_box,
     2				flag_distrib,flag_eps,bnd_type
        common /slab/           z0,zh,eps_slab,flag_slab
        common /mol_size/       mol_min,mol_max
c
c ==========================================================================================
c == Set local memory needed                     ===========================================
c ==========================================================================================
c
        ptr_hz_shift=malloc(8*nz)
c
	eps_val(1) = eps1
	eps_val(2) = eps2
	eps_val(3) = eps2
c
c	Set initial epsilon map based on molecule boundaries
c
	if(flag_eps.eq.0) then
c
		if(flag_eps.eq.0) then
			ep1 = 1.d0
		else
			ep1 = eps1
		endif
		do 300 k = 1,nz
		    do 200 j = 1,ny
			do 100 i = 1,nx
				epsilon_X(i,j,k) = ep1
				epsilon_Y(i,j,k) = ep1
				epsilon_Z(i,j,k) = ep1
100			continue
200		    continue
300		continue
c
	elseif(flag_eps.eq.1) then
c
c		X-shifted map
c
		do 800 k = 1,nz
		   do 700 j = 1,ny
			do 600 i = 1,nx
				ipos=ibits(grid_shift(i,j,k),2,1)+1
				epsilon_X(i,j,k) = eps_val(ipos)
600			continue
700		   continue
800		continue
c
c               Add slab, if present
c
                if(flag_slab.eq.1) then
c
                        call EpsilonSlab(z0,zh,epsilon_X,
     1                  eps_slab,grid_shift,2,hz,nx,ny,nz)
c
                endif
c
c		Repeat for Y-shifted map
c
		do 1300 k = 1,nz
		   do 1200 j = 1,ny
			do 1100 i = 1,nx
				ipos=ibits(grid_shift(i,j,k),4,1)+1
				epsilon_Y(i,j,k) = eps_val(ipos)
1100			continue
1200		   continue
1300		continue
c
c               Add slab, if present
c
                if(flag_slab.eq.1) then
c
                        call EpsilonSlab(z0,zh,epsilon_Y,
     1                  eps_slab,grid_shift,4,hz,nx,ny,nz)
c
                endif
c
c		Repeat for Z-shifted map
c
		do 1700 k = 1,nz
		   do 1600 j = 1,ny
			do 1500 i = 1,nx
				ipos=ibits(grid_shift(i,j,k),6,1)+1
				epsilon_Z(i,j,k) = eps_val(ipos)
1500			continue
1600		   continue
1700		continue
c
c               Add slab, if present
c
                if(flag_slab.eq.1) then
c
	                do 1800 k = 1,nz-1
		                hz_shift(k) = (hz(k)+hz(k+1))/2
1800                    continue
                        hz_shift(nz) = hz(nz)
                        call EpsilonSlab(z0,zh,epsilon_Z,
     1                  eps_slab,grid_shift,6,hz_shift,nx,ny,nz)
c
                endif
c
	elseif(flag_eps.eq.2) then
c
c		First set epsilon value based on local neighborhood,
c		and store in work arrays
c
                ptr_work_X = malloc(8*nx*ny*nz)
                ptr_work_Y = malloc(8*nx*ny*nz)
                ptr_work_Z = malloc(8*nx*ny*nz)
c
                call eps_harmo(nx,ny,nz,hz,hz_shift,grid_shift,eps_val,
     1          z0,zh,eps_slab,work_X,work_Y,work_Z,epsilon_X,epsilon_Y,
     2          epsilon_z)
c
                call free(ptr_Work_X)
                call free(ptr_Work_Y)
                call free(ptr_Work_Z)
c
	endif
c
c ==========================================================================================
c == free local memory                           ===========================================
c ==========================================================================================
c
        call free(ptr_hz_shift)
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	eps_harmo.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      8/13/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c       Perform harmonic averaging when computing eps map
c
        subroutine eps_harmo(nx,ny,nz,hz,hz_shift,grid_shift,eps_val,
     1          flag_slab,z0,zh,eps_slab,work_X,work_Y,work_Z,
     2          epsilon_X,epsilon_Y,epsilon_z)
c
c       Input:
c               nx,ny,nz        : size of the grid
c               hz              : Z-position of grid vertices
c               grid_shift      : information of position of molecule on the
c                                 grid
c               eps_val         : values of epsilon in different regions
c               flag_slab       : 0 if no slab, 1 if a slab is considered
c               z0,zh           : parameters for slab (if present)
c               eps_slab        : epsilon value inside slab
c               work_X,work_Y
c               work_Z          : storage vectors
c
c       Output:
c               epsilon_X       : epsilon map shifted along X
c               epsilon_Y       : epsilon map shifted along Y
c               epsilon_Z       : epsilon map shifted along Z
c
c ==========================================================================================
c == Define all variables                        ===========================================
c ==========================================================================================
c	
c	Local variables
c ==========================================================================================
c
	integer	i,j,k,ipos,nval
	integer	nx,ny,nz
c
        real*8  val
        real*8  hz(nz),hz_shift(nz)
c
c	Output
c ==========================================================================================
c
	real*8	epsilon_X(nx,ny,nz)
	real*8	epsilon_Y(nx,ny,nz)
	real*8	epsilon_Z(nx,ny,nz)
c
c	Working array
c ==========================================================================================
c
	real*8	work_X(nx,ny,nz)
	real*8	work_Y(nx,ny,nz)
	real*8	work_Z(nx,ny,nz)
c
c	Dielectric information
c ==========================================================================================
c
	real*8	eps_val(3)
c
c	Information on shifted grids
c ==========================================================================================
c
	integer*1 grid_shift(nx,ny,nz)
c
c       Information on z-slab (if any)
c ==========================================================================================
c
        integer flag_slab
        real*8  z0,zh,eps_slab,dist_test
c
c ==========================================================================================
c == Set epsilon maps                            ===========================================
c ==========================================================================================
c
	do 300 k = 1,nz
	   do 200 j = 1,ny
		do 100 i = 1,nx
			ipos=ibits(grid_shift(i,j,k),2,1)+1
			work_X(i,j,k) = eps_val(ipos)
100		continue
200	   continue
300	continue
c
c       Add slab, if present
c
        if(flag_slab.eq.1) then
c
                call EpsilonSlab(z0,zh,work_X,
     1          eps_slab,grid_shift,2,hz,nx,ny,nz)
c
        endif
c
c	Repeat for Y-shifted map
c
	do 800 k = 1,nz
	   do 700 j = 1,ny
		do 600 i = 1,nx
			ipos=ibits(grid_shift(i,j,k),4,1)+1
			work_Y(i,j,k) = eps_val(ipos)
600		continue
700	   continue
800	continue
c
c       Add slab, if present
c
        if(flag_slab.eq.1) then
c
                 call EpsilonSlab(z0,zh,work_Y,
     1           eps_slab,grid_shift,4,hz,nx,ny,nz)
c
        endif
c
c	Repeat for Z-shifted map
c
	do 1100 k = 1,nz
	   do 1000 j = 1,ny
		do 900 i = 1,nx
			ipos=ibits(grid_shift(i,j,k),6,1)+1
			work_Z(i,j,k) = eps_val(ipos)
900		continue
1000	   continue
1100	continue
c
c       Add slab, if present
c
        if(flag_slab.eq.1) then
c
                do 1200 k = 1,nz-1
	                hz_shift(k) = (hz(k)+hz(k+1))/2
1200            continue
                hz_shift(nz) = hz(nz)
                call EpsilonSlab(z0,zh,work_Z,
     1          eps_slab,grid_shift,6,hz_shift,nx,ny,nz)
c
        endif
c
c	Now perform harmonic averaging
c	(see Bruccoleri et al, J. Comp. Chem., 18, 268 (1997)
c
	do 1600 k = 1,nz
		do 1500 j = 1,ny
			do 1400 i = 1,nx-1
				nval=1
				val=1.d0/work_X(i,j,k)
				if(j.ne.1) then
				   nval=nval+2
				   val=val+1.d0/work_Y(i,j-1,k)
     1					+1.d0/work_Y(i+1,j-1,k)
				endif
				if(j.ne.ny) then
				   nval=nval+2
				   val=val+1.d0/work_Y(i,j,k)
     1					+1.d0/work_Y(i+1,j,k)
				endif
				if(k.ne.1) then
				   nval=nval+2
				   val=val+1.d0/work_Z(i,j,k-1)
     1					+1.d0/work_Z(i+1,j,k-1)
				endif
				if(k.ne.nz) then
				   nval=nval+2
				   val=val+1.d0/work_Z(i,j,k)
     1					+1.d0/work_Z(i+1,j,k)
				endif
				epsilon_X(i,j,k)=dble(nval)/val
1400			continue
1500		continue
1600	continue
c
	do 1900 k = 1,nz
		do 1800 j = 1,ny-1
			do 1700 i = 1,nx
				nval=1
				val=1.d0/work_Y(i,j,k)
				if(i.ne.1) then
				   nval=nval+2
				   val=val+1.d0/work_X(i-1,j,k)
     1					+1.d0/work_X(i-1,j+1,k)
				endif
				if(i.ne.nx) then
				   nval=nval+2
				   val=val+1.d0/work_X(i,j,k)
     1					+1.d0/work_X(i,j+1,k)
				endif
				if(k.ne.1) then
				   nval=nval+2
				   val=val+1.d0/work_Z(i,j,k-1)
     1					+1.d0/work_Z(i,j+1,k-1)
				endif
				if(k.ne.nz) then
				   nval=nval+2
				   val=val+1.d0/work_Z(i,j,k)
     1					+1.d0/work_Z(i,j+1,k)
				endif
				epsilon_Y(i,j,k)=dble(nval)/val
1700			continue
1800		continue
1900	continue
c
	do 2200 k = 1,nz-1
		do 2100 j = 1,ny
			do 2000 i = 1,nx
				nval=1
				val=1.d0/work_Z(i,j,k)
				if(i.ne.1) then
				   nval=nval+2
				   val=val+1.d0/work_X(i-1,j,k)
     1					+1.d0/work_X(i-1,j,k+1)
				endif
				if(i.ne.nx) then
				   nval=nval+2
				   val=val+1.d0/work_X(i,j,k)
     1					+1.d0/work_X(i,j,k+1)
				endif
				if(j.ne.1) then
				   nval=nval+2
				   val=val+1.d0/work_Y(i,j-1,k)
     1					+1.d0/work_Y(i,j-1,k+1)
				endif
				if(j.ne.ny) then
				   nval=nval+2
				   val=val+1.d0/work_Y(i,j,k)
     1					+1.d0/work_Y(i,j,k+1)
				endif
				epsilon_Z(i,j,k)=dble(nval)/val
2000			continue
2100		continue
2200	continue
c
        return
        end
c
c ==========================================================================================
c ==========================================================================================
c	set_boundary.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      8/13/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Set boundary conditions on the grid
c
c	Input:
c		- bnd_type : flag defining type of boundary condition considered
c				- 0	: set boundary values to 0
c				- 1	: dipolar. The boundary potential is approximated
c					  by the Debye-Huckel potential of the equivalent
c					  dipole to the charge distribution.
c				- 2	: Coulomb. The boundary potential is approximated
c					  by the sum of the Debye-Huckel potentials
c					  of the individual charges.
c				- 3	: focusing (not yet implemented)
c
c	Output:
c		- bndx	: value of potential for ix=1 and ix = nx
c		- bndy	: value of potential for iy=1 and iy = ny
c		- bndz	: value of potential for iz=1 and iz = nz
c
	subroutine set_boundary(bndx,bndy,bndz,hx,hy,hz,nx,ny,nz)
c
	include 'poisson.h'
c
c ==========================================================================================
c == Define all variables                        ===========================================
c ==========================================================================================
c	
c	Local variables
c ==========================================================================================
c
	integer	i,j,k
	integer	nx,ny,nz
	integer	ic
c
	integer	n_p,n_n
c
	real*8	Ion_strength,l_debinv,kappa2,kappa0
	real*8	crg_p,crg_n
	real*8	xdiffp1, xdiffn1, xdiffpN,xdiffnN
	real*8	ydiffp1, ydiffn1, ydiffpN,ydiffnN
	real*8	zdiffp1, zdiffn1, zdiffpN,zdiffnN
	real*8	distp,distn,val
	real*8	cg_p(3),cg_n(3),crg(3)
c
c	Boundary values
c ==========================================================================================
c
	real*8	bndx(ny,nz,*),bndy(nx,nz,*)
	real*8	bndz(nx,ny,*)
c
c	Parameter values
c ==========================================================================================
c
	integer	surf_type,flag_pymol,flag_box,flag_distrib,flag_eps
	integer	grid_type,bnd_type
	integer	ion_zone
c
	real*8	R_probe,R_ion
c
c       Grid info
c  =========================================================================================
c
        real*8  hx(*),hy(*),hz(*)
c
c	Constants
c ==========================================================================================
c
	integer	n_ion,n_zz
        integer method,nlevel,nlin,equation,ndiscrete,ijac,nsmooth
c
        real*8   tol
	real*8	l_Bjerrum,pi,Temp,KB,electron,Ang,beta,Avogadro,eps0
	real*8	C_charge,C_prop
	real*8	eps1,eps2,eps3,eps_bulk
	real*8	ion_conc(max_ion),ion_z(max_ion),kappa(max_ion)
	real*8	ion_rad(max_ion)
c
c	Information on molecule (charges)
c ==========================================================================================
c
	integer	natom,ncharge,iform
c
	real*8	atom_coord(*)
	real*8	atom_radius(*),atom_charge(*)
c
	character atom_label(*)*30
	character atom_label2(*)*12
c
c       Pointers
c ==========================================================================================
c
        pointer (ptr_coord,atom_coord)
        pointer (ptr_radius,atom_radius)
        pointer (ptr_charge,atom_charge)
        pointer (ptr_label,atom_label)
        pointer (ptr_label2,atom_label2)
c
c ==========================================================================================
c == All common blocks (common storage)          ===========================================
c ==========================================================================================
c	
	common /input_par/	R_probe,R_ion,flag_pymol,surf_type,
     1				grid_type,ion_zone,flag_box,
     2				flag_distrib,flag_eps,bnd_type
c
	common /constants/ 	l_Bjerrum,pi,Temp,KB,electron,Ang,
     1			   	Avogadro,eps0,beta,C_charge,C_prop
	common /elec_const/ 	eps1,eps2
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
	common /salts/		Ion_strength,kappa0
c
        common /charges/   ptr_coord,ptr_charge,ptr_radius,ptr_label,
     1                     ptr_label2,iform,natom,ncharge
c
	common /pdesolver/	tol,method,nlevel,nlin,equation,
     1                          ndiscrete,ijac,nsmooth
        common /bulk/           eps_bulk
c
        if(equation.le.3) then
                eps3 = eps2
        else
                eps3 = eps_bulk
        endif
c
c ==========================================================================================
c == Compute Debye length                        ===========================================
c ==========================================================================================
c
c	The Debye-Huckel parameter kappa is defined as:
c
c		kappa = ((8*pi*NA*ec^2)/(1000*kB*T*eps))^(1/2) * (Ion_strength)^(1/2)	
c
c	where NA is the Avogadro number, kB is the Boltzman constant, T the temperature,
c	ec the charge of the electron, and eps the dielectric constant of water.
c
c	The debye length is defined as:
c
c		l_debye = 1/kappa
c	and     l_debinv = kappa
c
	kappa2 = 8.48602807d0 * (298/Temp) * Ion_strength/eps3
c
	l_debinv = sqrt(kappa2)
c	if(Ion_strength.ne.0) then
c		l_debye = 1.d0/sqrt(kappa2)
c	else
c		l_debye = 0
c	endif
c
c ==========================================================================================
c == Dimensionless potential                     ===========================================
c ==========================================================================================
c
c	We solve the PB equation to find the dimensionless potential u defined as:
c
c	u(r) = (ec /kB T) phi(r)
c
c	where ec is the charge of the electron, kB is the Boltzman constant and T the temperature
c
c	ec / (kB T) is defined as C_prop
c
c ==========================================================================================
c == Generate boundary conditions                ===========================================
c ==========================================================================================
c
c	Initialize boundary potential values to 0
c
c	bndx(j,k,1) is the value at grid point 1,j,k
c	bndx(j,k,2) is the value at grid point nx,j,k
c
c	bndx(j,k,3) and bndx(j,k,4) are set to 0 (do not know what these values are,
c	seems to be used in MG ??)
c
	do 300 k=1,nz
		do 200 j = 1,ny
			bndx(j,k,1) = 0.d0
			bndx(j,k,2) = 0.d0
200		continue
300	continue
c
c	bndy(i,k,1) is the value at grid point i,1,k
c	bndy(i,k,2) is the value at grid point i,ny,k
c
c	bndy(i,k,3) and bndy(i,k,4) are set to 0 (do not know what these values are,
c	seems to be used in MG ??)
c
	do 500 k=1,nz
		do 400 i = 1,nx
			bndy(i,k,1) = 0.d0
			bndy(i,k,2) = 0.d0
400		continue
500	continue
c
c	bndz(i,j,1) is the value at grid point i,j,1
c	bndz(i,j,2) is the value at grid point i,j,nz
c
c	bndz(i,j,3) and bndz(i,j,4) are set to 0 (do not know what these values are,
c	seems to be used in MG ??)
c
	do 700 j=1,ny
		do 600 i = 1,nx
			bndz(i,j,1) = 0.d0
			bndz(i,j,2) = 0.d0
600		continue
700	continue
c
	if(bnd_type.eq.0) then
c
c		In this case, keep boundary potential values to 0
c
		return
c
	elseif(bnd_type.eq.1) then
c
c		In this case, the potential values at the boundaries are computed
c		based on the Debye potential of the dipole equivalent to all charges
c		of the solute.
c
c		Find equivalent dipole for all charges: find centers of mass C_n and
c		C_p for negative
c		and positive charges, and assign them the sum of negative (resp. positive) charges
c
		do 800 i = 1,3
			cg_p(i) = 0
			cg_n(i) = 0
800		continue
c
		n_p = 0
		n_n = 0
		crg_p = 0
		crg_n = 0
c
		do 1100 i = 1,ncharge
c
			if(atom_charge(i).lt.0) then
				n_n = n_n + 1
				crg_n = crg_n + atom_charge(i)
				do 900 j = 1,3
					cg_n(j)=cg_n(j)+
     1					atom_coord(3*(i-1)+j)
900				continue
			elseif(atom_charge(i).gt.0) then
				n_p = n_p + 1
				crg_p = crg_p + atom_charge(i)
				do 1000 j = 1,3
					cg_p(j)=cg_p(j)+
     1					atom_coord(3*(i-1)+j)
1000				continue
			endif
c
1100		continue
c
		do 1200 i = 1,3
			if(n_p.ne.0) cg_p(i) = cg_p(i)/n_p
			if(n_n.ne.0) cg_n(i) = cg_n(i)/n_n
1200		continue
c
c		Now compute potential at each boundary points
c
c		phi = crg_p  exp(-dist_p*l_debinv) / (eps_solv*dist_p)
c		    + crg_n  exp(-dist_n*l_debinv) / (eps_solv*dist_n)
c
c		bndx(j,k,1) is the value at grid point 1,j,k
c		bndx(j,k,2) is the value at grid point nx,j,k
c
		xdiffp1 = (hx(1)-cg_p(1))**2
		xdiffn1 = (hx(1)-cg_n(1))**2
		xdiffpN = (hx(nx)-cg_p(1))**2
		xdiffnN = (hx(nx)-cg_n(1))**2
c
		do 1400 k = 1,nz
			zdiffp1 = (hz(k)-cg_p(3))**2
			zdiffn1 = (hz(k)-cg_n(3))**2
			do 1300 j = 1,ny
				ydiffp1 = (hy(j)-cg_p(2))**2
				ydiffn1 = (hy(j)-cg_n(2))**2
				distp = xdiffp1+ydiffp1+zdiffp1
				distp=sqrt(distp)
				distn = xdiffn1+ydiffn1+zdiffn1
				distn=sqrt(distn)
				val=(crg_p*exp(-distp*l_debinv)
     1				/(eps3*distp)+crg_n*exp(-distn*
     2				l_debinv)/(eps3*distp))*C_prop
				bndx(j,k,1) = val
				distp = xdiffpN+ydiffp1+zdiffp1
				distp=sqrt(distp)
				distn = xdiffnN+ydiffn1+zdiffn1
				distn=sqrt(distn)
				val=(crg_p*exp(-distp*l_debinv)
     1				/(eps3*distp)+crg_n*exp(-distn*
     2				l_debinv)/(eps3*distp))*C_prop
				bndx(j,k,2) = val
1300			continue
1400		continue
c
c		bndy(i,k,1) is the value at grid point i,1,k
c		bndy(i,k,2) is the value at grid point i,ny,k
c
		ydiffp1 = (hy(1)-cg_p(2))**2
		ydiffn1 = (hy(1)-cg_n(2))**2
		ydiffpN = (hy(ny)-cg_p(2))**2
		ydiffnN = (hy(ny)-cg_n(2))**2
c
		do 1600 k = 1,nz
			zdiffp1 = (hz(k)-cg_p(3))**2
			zdiffn1 = (hz(k)-cg_n(3))**2
			do 1500 i = 1,nx
				xdiffp1 = (hx(i)-cg_p(1))**2
				xdiffn1 = (hx(i)-cg_n(1))**2
				distp = xdiffp1+ydiffp1+zdiffp1
				distp=sqrt(distp)
				distn = xdiffn1+ydiffn1+zdiffn1
				distn=sqrt(distn)
				val=(crg_p*exp(-distp*l_debinv)
     1				/(eps3*distp)+crg_n*exp(-distn*
     2				l_debinv)/(eps3*distp))*C_prop
				bndy(i,k,1) = val
				distp = xdiffp1+ydiffpN+zdiffp1
				distp=sqrt(distp)
				distn = xdiffn1+ydiffnN+zdiffn1
				distn=sqrt(distn)
				val=(crg_p*exp(-distp*l_debinv)
     1				/(eps3*distp)+crg_n*exp(-distn*
     2				l_debinv)/(eps3*distp))*C_prop
				bndy(i,k,2) = val
1500			continue
1600		continue
c
c		bndz(i,j,1) is the value at grid point i,j,1
c		bndz(i,j,2) is the value at grid point i,j,nz
c
		zdiffp1 = (hz(1)-cg_p(3))**2
		zdiffn1 = (hz(1)-cg_n(3))**2
		zdiffpN = (hz(nz)-cg_p(3))**2
		zdiffnN = (hz(nz)-cg_n(3))**2
c
		do 1800 j = 1,ny
			ydiffp1 = (hy(j)-cg_p(2))**2
			ydiffn1 = (hy(j)-cg_n(2))**2
			do 1700 i = 1,nx
 				xdiffp1 = (hx(i)-cg_p(1))**2
				xdiffn1 = (hx(i)-cg_n(1))**2
				distp = xdiffp1+ydiffp1+zdiffp1
				distp=sqrt(distp)
				distn = xdiffn1+ydiffn1+zdiffn1
				distn=sqrt(distn)
				val=(crg_p*exp(-distp*l_debinv)
     1				/(eps3*distp)+crg_n*exp(-distn*
     2				l_debinv)/(eps3*distp))*C_prop
				bndz(i,j,1) = val
				distp = xdiffp1+ydiffp1+zdiffpN
				distp=sqrt(distp)
				distn = xdiffn1+ydiffn1+zdiffnN
				distn=sqrt(distn)
				val=(crg_p*exp(-distp*l_debinv)
     1				/(eps3*distp)+crg_n*exp(-distn*
     2				l_debinv)/(eps3*distp))*C_prop
				bndz(i,j,2) = val
1700			continue
1800		continue
c
	elseif(bnd_type.eq.2) then
c
c		In this case, the potential values at the boundaries are computed
c		based on the Debye potential of all charges of the solute.
c
c		Now compute potential at each boundary points : 
c
c		phi = Sum_p crg_p  exp(-dist_p/l_debye) / (eps_solv*dist_p)
c
c		We loop over all charges, and accumulate the potential
c
		do 2600 ic = 1,ncharge
c
			if(atom_charge(ic).eq.0) goto 2600
c
			do 1900 j = 1,3
				crg(j) = atom_coord(3*(ic-1)+j)
1900			continue
			crg_p = atom_charge(ic)
c
c			bndx(j,k,1) is the value at grid point 1,j,k
c			bndx(j,k,2) is the value at grid point nx,j,k
c
			xdiffp1 = (hx(1)-crg(1))**2
			xdiffpN = (hx(nx)-crg(1))**2
c
			do 2100 k = 1,nz
			    zdiffp1 = (hz(k)-crg(3))**2
			    do 2000 j = 1,ny
				ydiffp1 = (hy(j)-crg(2))**2
				distp = xdiffp1+ydiffp1+zdiffp1
				distp=sqrt(distp)
				val=(crg_p*exp(-distp*l_debinv)
     1				/(eps3*distp))*C_prop
				bndx(j,k,1) = bndx(j,k,1)+val
				distp = xdiffpN+ydiffp1+zdiffp1
				distp=sqrt(distp)
				val=(crg_p*exp(-distp*l_debinv)
     1				/(eps3*distp))*C_prop
				bndx(j,k,2) = bndx(j,k,2)+val
2000			    continue
2100			continue
c
c			bndy(i,k,1) is the value at grid point i,1,k
c			bndy(i,k,2) is the value at grid point i,ny,k
c
			ydiffp1 = (hy(1)-crg(2))**2
			ydiffpN = (hy(ny)-crg(2))**2
c
			do 2300 k = 1,nz
			    zdiffp1 = (hz(k)-crg(3))**2
			    do 2200 i = 1,nx
				xdiffp1 = (hx(i)-crg(1))**2
				distp = xdiffp1+ydiffp1+zdiffp1
				distp=sqrt(distp)
				val=(crg_p*exp(-distp*l_debinv)
     1				/(eps3*distp))*C_prop
				bndy(i,k,1) = bndy(i,k,1)+val
				distp = xdiffp1+ydiffpN+zdiffp1
				distp=sqrt(distp)
				val=(crg_p*exp(-distp*l_debinv)
     1				/(eps3*distp))*C_prop
				bndy(i,k,2) = bndy(i,k,2)+val
2200			    continue
2300			continue
c
c			bndz(i,j,1) is the value at grid point i,j,1
c			bndz(i,j,2) is the value at grid point i,j,nz
c
			zdiffp1 = (hz(1)-crg(3))**2
			zdiffpN = (hz(nz)-crg(3))**2
c
			do 2500 j = 1,ny
			    ydiffp1 = (hy(j)-crg(2))**2
			    do 2400 i = 1,nx
				xdiffp1 = (hx(i)-crg(1))**2
				distp = xdiffp1+ydiffp1+zdiffp1
				distp=sqrt(distp)
				val=(crg_p*exp(-distp*l_debinv)
     1				/(eps3*distp))*C_prop
				bndz(i,j,1) = bndz(i,j,1)+val
				distp = xdiffp1+ydiffp1+zdiffpN
				distp=sqrt(distp)
				val=(crg_p*exp(-distp*l_debinv)
     1				/(eps3*distp))*C_prop
				bndz(i,j,2) = bndz(i,j,2)+val
2400			    continue
2500			continue
c
2600		continue
c
	endif
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	set_helmoltz.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      8/13/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Define region in which nonlinear term is not zero
c
c	(this is the solvent region, minus the ion excluded region)
c
	subroutine set_helmoltz(helmoltz_map,grid_partition,nx,ny,nz)
c
	include 'poisson.h'
c
c ==========================================================================================
c == Define all variables                        ===========================================
c ==========================================================================================
c	
c	Local variables
c ==========================================================================================
c
	integer	i,j,k
	integer	nx,ny,nz
c
c	Output
c ==========================================================================================
c
	real*8	helmoltz_map(nx,ny,nz)
c
c
c	Information on grid
c ==========================================================================================
c
	integer*1 grid_partition(nx,ny,nz)
c
c	Information on ions
c ==========================================================================================
c
	integer	n_ion,n_zz
c
	real*8	eps1,eps2
	real*8	Ion_Strength,kappa0
	real*8	ion_conc(max_ion),ion_z(max_ion),ion_rad(max_ion)
	real*8	kappa(max_ion)
c
c ==========================================================================================
c == Define all common blocks                    ===========================================
c ==========================================================================================
c	
	common /elec_const/ 	eps1,eps2
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
	common /salts/		Ion_Strength,kappa0
c
c ==========================================================================================
c == Set helmoltz map: 0 inside solute, 1 in solvent if ions are present ===================
c ==========================================================================================
c	
	do 300 k = 1,nz
		do 200 j = 1,ny
			do 100 i = 1,nx
                                if(ibits(grid_partition(i,j,k),1,1)
     1                          .ne.0) then
					helmoltz_map(i,j,k)=kappa0
				else
					helmoltz_map(i,j,k)=0
				endif
100			continue
200		continue
300	continue
c
	return
	end
