c       Copyright (C) 2007 Patrice Koehl
c
c       This library is free software; you can redistribute it and/or
c       modify it under the terms of the GNU Lesser General Public
c       License as published by the Free Software Foundation; either
c       version 2.1 of the License, or (at your option) any later version.
c
c       This library is distributed in the hope that it will be useful,
c       but WITHOUT ANY WARRANTY; without even the implied warranty of
c       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
c       Lesser General Public License for more details.
c
c       You should have received a copy of the GNU Lesser General Public
c       License along with this library; if not, write to the Free Software
c       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
c
c ==========================================================================================
c ==========================================================================================
c 	readmap_DX.f
c ==========================================================================================
c ==========================================================================================
c
c	This utility program reads in a map in DX format
c
	subroutine readmap_DX(Npointx,Npointy,Npointz,fname,map,ierr)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
	integer	i,j,k,nx,ny,nz
        integer Npointx,Npointy,Npointz
	integer	nval,ierr
        integer nline,ival
c
	real*8	map(Npointx,Npointy,Npointz)
c
	character*100	fname
	character*80	record
c
1	format(a)
2       format(1x,1pe15.6,1x,1pe15.6,1x,1pe15.6)
c
c ==========================================================================================
c	Read map
c ==========================================================================================
c
	ierr = 0
c
	open(unit=1,file=fname,status='old',err=600)
c
100	read(1,1,end=600) record
	if(record(1:14).eq.'object 1 class') then
		read(record(36:80),*) nx,ny,nz
	endif
	if(record(1:14).ne.'object 3 class') goto 100
c
        if(Npointx.ne.nx.or.Npointy.ne.ny.or.Npointz.ne.nz) then
                ierr = 1
                return
        endif
c
        read(1,2) (((map(i,j,k),k=1,Npointz),j=1,Npointy),i=1,Npointx)
c
	close(unit=1)
c
	return
c
600	ierr = 1
c
	return
        end
c ==========================================================================================
c ==========================================================================================
c 	readmap_EZD.f
c ==========================================================================================
c ==========================================================================================
c
c	This utility program reads in a map in EZD format
c
	subroutine readmap_EZD(Npointx,Npointy,Npointz,fname,map,ierr)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
        include 'poisson.h'
c
	integer	i,j,k,nx,ny,nz
        integer Npointx,Npointy,Npointz
	integer	nval,ierr
        integer nline,ival
c
	real*8	map(*)
c
	character*100	fname
	character*80	record
c
1	format(a)
c
c ==========================================================================================
c	Read map
c ==========================================================================================
c
	ierr = 0
c
	open(unit=1,file=fname,status='old',err=600)
c
100	read(1,1,end=600) record
	if(record(1:4).eq.'GRID') then
		read(record(5:80),*) nx,ny,nz
	endif
	if(record(1:8).ne.'MAP') goto 100
c
        if(Npointx.ne.nx.or.Npointy.ne.ny.or.Npointz.ne.nz) then
                ierr = 1
                return
        endif
c
	nval = nx*ny*nz
	nline = nval/7 + 1
	if(mod(nval,7).eq.0) nline = nline - 1
c
	ival = 0
	do 200 i = 1,nline -1
		read(1,*) (map(ival+j),j=1,7)
		ival = ival + 7
200	continue
	read(1,*) (map(i),i=ival+1,nval)
c
	close(unit=1)
c
	return
c
600	ierr = 1
c
	return
        end
c ==========================================================================================
c ==========================================================================================
c 	readmap_bin.f
c ==========================================================================================
c ==========================================================================================
c
c	This utility program reads in a map in binary format
c
	subroutine readmap_bin(Npointx,Npointy,Npointz,fname,map,ierr)
c
c ==========================================================================================
c	Declare all variables
c ==========================================================================================
c
        include 'poisson.h'
c
	integer	i,j,k,nx,ny,nz,n
        integer Npointx,Npointy,Npointz
	integer	nval,ierr
        integer nline,ival
c
	real*8	map(*)
        real*8  hx
c
	character*100	fname
	character*80	record
c
1	format(a)
c
c ==========================================================================================
c	Read map
c ==========================================================================================
c
	ierr = 0
c
	open(unit=1,file=fname,status='old',form='unformatted',err=600)
c
        read(1) nx,ny,nz
        n = min(nx,ny)
        n = min(n,nz)
        read(1) hx
        read(1) hx
        read(1) hx
c
        if(Npointx.ne.nx.or.Npointy.ne.ny.or.Npointz.ne.nz) then
                ierr = 1
                return
        endif
c
        read(1) (map(i),i=1,nx*ny*nz)
c
	close(unit=1)
c
	return
c
600	ierr = 1
c
	return
        end
