c ==========================================================================================
c ==========================================================================================
c	Read_Input.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	7/5/07
c	Author:		Patrice Koehl
c			in collaboration with Marc Delarue
c
c	Purpose:	Read in all information required for the run
c
	subroutine read_input(fname_input,l_log,ierr,
     1                  nx,ny,nz,sizex,sizey,sizez)
c
	include 'poisson.h'
c
c ==========================================================================================
c == Define all variables                  =================================================
c ==========================================================================================
c
c	General information
c ==========================================================================================
c
	integer	i,k,ival,l_log,nkeys,flag_pymol,surf_type,ierr
	integer method,equation,ndiscrete,nlevel,nlin,ion_zone
	integer	idx,idx2,npoint_sphere,ijac,ijac2,nsmooth
	integer flagcenter
c
c	Constants
c
	real*8	R_probe,R_ion
	real*8	l_Bjerrum,pi,Temp,KB,electron,Ang,beta,Avogadro,eps0
	real*8	C_charge,C_prop
	real*8	eps1,eps2
	real*8	tol
c
	character	keyword*20,value*50,value1*50
	character	mesh(2)*11,surf(2)*26,crg_distrib(4)*17
	character	boundary(4)*8
	character	solver(5)*10,epsval(3)*28
	character	equa_name(6)*45
        character       coarse(3)*30,smooth_name(3)*30
c
c	Information on grid
c ==========================================================================================
c
	integer	grid_type,flag_box,flag_distrib,bnd_type,flag_eps
        integer Npoint,N,nx,ny,nz
	real*8	sizex,sizey,sizez
c
c	Information on molecule
c ==========================================================================================
c
	integer	natom,ncharge,iform
c
	real*8	atom_coord(*)
	real*8	atom_radius(*),atom_charge(*)
	real*8	shiftx,shifty,shiftz
c
	character atom_label(*)*30
	character atom_label2(*)*12
c
c	Information on solvents and ions
c ==========================================================================================
c
	integer	n_ion,n_solv,n_zz
c
	real*8	a_lattice,v_lattice,zn_max,zp_max
	real*8	solv_conc(max_solv),solv_dipole(max_solv)
	real*8	solv_cvol(max_solv)
c
	real*8	ion_conc(max_ion),ion_z(max_ion)
	real*8	kappa(max_ion),ion_rad(max_ion)
c
c       Information on slab (if any)
c ==========================================================================================
c
        integer flag_slab
c
        real*8  z0,zh,eps_slab
c
c	Information for Yukawa potential in solvent (if any)
c ==========================================================================================
c
	real*8	v0,b,psi_bulk,lambda
c
c	Output maps
c ==========================================================================================
c
	integer		flag_phi,flag_edens,flag_nion,flag_dion
	integer		flag_dipole,flag_map_format,flag_psi
	character	fname_out_phi*100,fname_out_edens*100
	character	fname_out_nion*100,fname_out_dion*100
        character       fname_out_dipole*100,fname_out_psi*100
c
c	File names:
c ==========================================================================================
c
	character	fname_input*100
	character	fname_pdb*100,fname_log*100,fname_py*100
	character	fname_newpdb*100
	character	keys(100)*20
	character	line*100
	character	fpdbcenter*100
c
c ==========================================================================================
c == Pointers for atom arrays              =================================================
c ==========================================================================================
c
        pointer (ptr_coord,atom_coord)
        pointer (ptr_radius,atom_radius)
        pointer (ptr_charge,atom_charge)
        pointer (ptr_label,atom_label)
        pointer (ptr_label2,atom_label2)
c
c ==========================================================================================
c == Define all common blocks              =================================================
c ==========================================================================================
c
	common /input_file/	fname_pdb,fname_log,fname_py
	common /input_par/	R_probe,R_ion,flag_pymol,surf_type,
     1				grid_type,ion_zone,flag_box,
     2				flag_distrib,flag_eps,bnd_type
        common /charges/        ptr_coord,ptr_charge,ptr_radius,
     1                          ptr_label,ptr_label2,iform,natom,
     2                          ncharge
	common /constants/ 	l_Bjerrum,pi,Temp,KB,electron,Ang,
     1				Avogadro,eps0,beta,C_charge,C_prop
	common /elec_const/ 	eps1,eps2
	common /ions/		ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
        common /ion_max/        zp_max,zn_max
	common /solvent/	solv_conc,solv_dipole,solv_cvol,n_solv
	common /lattice/	a_lattice,v_lattice
        common /slab/           z0,zh,eps_slab,flag_slab
	common /pdesolver/	tol,method,nlevel,nlin,equation,
     1                          ndiscrete,ijac,nsmooth
	common /molshift/	shiftx,shifty,shiftz
	common /output_map/	fname_out_phi,fname_out_edens,
     1				fname_out_nion,fname_out_dion,
     2                          fname_out_dipole,fname_out_psi,
     3				flag_phi,flag_edens,flag_nion,flag_dion,
     4				flag_dipole,flag_map_format,flag_psi
        common /point_sphere/   npoint_sphere
	common /yukawa/		v0,b,psi_bulk
	common /damping/	lambda
	common /centering/	fpdbcenter
c
1	format(a)
2	format(70('='),/,2('='),'General information       ',40x,
     1	2('='),/,70('='),/)  
3	format('PDB file name                       : ',a50)
4	format('Log file name                       : ',a50)
5	format('Pymol CGO file for visualizing mesh : ',a50)
6	format('Mesh type                           : ',a11)
7	format('Surface type                        : ',a26)
8	format('Probe radius (angstroms)            : ',f8.3,
     1	' angstroms')
9	format('Charge distribution                 : ',a17)
10	format('Temperature for this run            : ',f8.3,' K')
11	format('Epsilon (inside molecule)           : ',f8.3)
12	format('Epsilon (inside solvent)            : ',f8.3)
13	format('Ion #                           ',i1,'   :',
     1	' valence : ',f6.2,' conc.  : ',f8.3,' mol/l',' size :',f8.3)
14	format('Solvent #                       ',i1,'   :',
     1	' dipole : ',f6.2,' Debye',' conc.  : ',f8.3,' mol/l')
15	format('Ion exclusion zone (Stern region)   : ',f8.3,
     1	' angstroms')
16	format('Type of boundary conditions         : ',a8)
17	format('Method used to solve equation       : ',a9)
18	format('Method to compute epsilon on grid   : ',a28)
19	format('Elliptic equation considered        : ',a45)
20	format('Number of levels for multigrid      : ',i4)
21	format('Output: Electrostatic potential map : ',a30)
22	format('Output: Energy density map          : ',a30)
23	format('Output: Ion density (number)        : ',a30)
24	format('Output: Ion density (charge)        : ',a30)
25	format('Output: PDB file for centered Mol   : ',a30)
26	format('Output: Water density (dipole)      : ',a30)
27      format('Method for discretization           : ',a30)
28      format('Smoothing technique                 : ',a30)
29	format('PDB centering file                  : ',a50)
c
c ==========================================================================================
c == Define Keywords for reading information  ==============================================
c ==========================================================================================
c
	data nkeys /49/
	data (keys(i),i=1,49) /
     &  		 'Npoint             :',
     &  		 'Npointx            :',
     &			 'Npointy            :',
     &			 'Npointz            :',
     &			 'Size               :',
     &			 'Sizex              :',
     &			 'Sizey              :',
     &			 'Sizez              :',
     &			 'Grid type          :',
     &			 'Surface type       :',
     &			 'Radius Probe       :',
     &			 'File Input Mol     :',
     &			 'File Input Format  :',
     &			 'File Output LOG    :',
     &			 'File Pymol Mesh    :',
     &			 'Show Solvent Box   :',
     &                   'Charge Distrib.    :',
     &                   'Temperature        :',
     &                   'Epsilon molecule   :',
     &                   'Epsilon solvent    :',
     &                   'Epsilon averaging  :',
     &                   'Ion                :',
     &			 'Boundary Condition :',
     &                   'PDE Solver         :',
     &                   'Tol (convergence)  :',
     &			 'Equation type      :',
     &			 'Nlevels (MG)       :',
     &                   'Shifts             :',
     &                   'Map format         :',
     &			 'File Output Phi    :',
     &			 'File Output Edens  :',
     &			 'File Output Ion    :',
     &			 'File Output IonC   :',
     &                   'Stern region size  :',
     &                   'Solvent            :',
     &			 'Lattice size       :',
     &			 'File Output Dipole :',
     &                   'z0: center of slab :',
     &                   'zh: 1/2 height     :',
     &                   'Epsilon slab       :',
     &                   'Point on sphere    :',
     &                   'Discretization     :',
     &                   'Equation           :',
     &                   'Stencil size       :',
     &                   'Smoothing method   :',
     &			 'Yukawa parameters  :',
     &			 'Damping parameter  :',
     &			 'File Output Psi    :',
     &			 'File Input Center  :'
     &			/
c
	data mesh    /'Cubic      ','Rectangular'/
	data surf    /'Solvent Accessible Surface','Molecular Surface'/
	data crg_distrib /'Trilinear','Spherical Uniform',
     1  'Antialiasing','Uniform surface'/
	data boundary /'Zero    ','Dipolar ','Coulomb ','Focusing'/
	data solver /'CGMG','Newton','Newton','Newton','SelfC'/
	data epsval /'Uniform','Mean (2 neighbours)',
     1			'Harmonic mean (9 neighbours)'/
	data equa_name/'Poisson','Linearized Poisson Boltzmann',
     1			'Non Linear Poisson Boltzmann',
     2			'Size modified Poisson Boltzmann',
     3			'Dipolar Poisson Boltzmann Langevin (PBL)',
     4                  'Yukawa-Poisson-Boltzmann Langevin  (YULP)'/
        data coarse /'Standard','Harmonic','Galerkin'/
        data smooth_name /'Red-Black Gauss-Seidel','Weighted Jacobi',
     1                    'SOR'/
c
c ==========================================================================================
c == Default values                           ==============================================
c ==========================================================================================
c
	nx      = 0
	ny      = 0
	nz      = 0
	Npoint  = 0
c
	sizex = 0.d0
	sizey = 0.d0
	sizez = 0.d0
c
	shiftx = 0.d0
	shifty = 0.d0
	shiftz = 0.d0
c
	fname_pdb = ' '
	fname_newpdb = ' '
	fname_out_dipole = ' '
	fname_log=' '
	fname_py=' '
	i = index(fname_input,'.')
	iform = 0
	fname_log(1:i+2) = fname_input(1:i-1)//'.log'
	fname_py(1:i+1)  = fname_input(1:i-1)//'.py'
c
	grid_type = 0
	bnd_type  = 0
	R_probe   = 1.4d0
c
	flag_pymol = 0
	flag_box   = 0
c
	flag_distrib = 0
c
	Temp = 300.d0
c
	eps1 = 2.d0
	eps2 = 80.d0
	flag_eps = 1
c
	n_ion = 0
        n_zz  = 0
c
	R_ion = 0.d0
	ion_zone = 0
c
	n_solv = 0
c
	a_lattice = 0
c
	nlin   = 1
	method = 1
        equation = 2
        ndiscrete = 2
        nsmooth = 1
        ijac = 0
        ijac2 = 7
	nlevel = 0
	tol = 1.d-8
c
	flag_map_format = 1
	flag_phi   = 0
	flag_edens = 0
	flag_nion  = 0
	flag_dion  = 0
c
        flag_slab = 0
        z0        = 0
        zh        = 0
        eps_slab  = 1.d0
c
	v0 = 0
	b = 1
	lambda = 0.5
c	
	flagcenter = 0
	fpdbcenter = ' '
c
c ==========================================================================================
c == Open file, and read line one at a time   ==============================================
c ==========================================================================================
c
	ierr = 0
	open(unit=99,file=fname_input,status='old')
c
100	read(99,1,end=400,err=1000) line
c
	if(line(1:1).eq.'#'.or.line(1:1).eq.' '.or.
     1		line(1:1).eq.'	'.or.line(1:1).eq.'!') goto 100
	keyword = line(1:20)
	value1   = line(21:70)
	call process_string(value1,value)
	if(value.eq.' ') goto 100
c
	do 200 i = 1,nkeys	
		if(keyword.eq.keys(i)) then
			goto 300
		endif
200	continue
c
	write(6,*)' '
	write(6,*)
     1	'WARNING :Problem reading input file: not in correct format'
	write(6,*)' '
	write(6,*)
     1	'	Please edit sample input file that came with program:'
	write(6,*) '	You can remove lines, change order, ...',
     1		'BUT: do not change keyword names !!'
	write(6,*)' '
	write(6,*) '	Offending keyword : "',keyword,'"'
	write(6,*)' '
c	write(6,*) '	Program exiting...please try again'
c	write(6,*)' '
c	stop
        goto 100
c
300	continue
c
	if(i.eq.1) then
		read(value,*) nx,ny,nz
		Npoint  = nx*ny*nz
	elseif(i.eq.2) then
		read(value,*) nx
	elseif(i.eq.3) then
		read(value,*) ny
	elseif(i.eq.4) then
		read(value,*) nz
	elseif(i.eq.5) then
		read(value,*) sizex,sizey,sizez
		if(sizex.eq.sizey.and.sizey.eq.sizez) grid_type=0
	elseif(i.eq.6) then
		read(value,*) sizex
	elseif(i.eq.7) then
		read(value,*) sizey
	elseif(i.eq.8) then
		read(value,*) sizez
	elseif(i.eq.9) then
		read(value,*) grid_type
	elseif(i.eq.10) then
		read(value,*) surf_type
	elseif(i.eq.11) then
		read(value,*) R_probe
	elseif(i.eq.12) then
		fname_pdb(1:50) = value
	elseif(i.eq.13) then
		if(value(1:3).eq.'PDB'.or.value(1:3).eq.'pdb') then
			iform = 0
		elseif(value(1:3).eq.'PQR'.or.value(1:3).eq.'pqr') then
			iform = 1
		elseif(value(1:3).eq.'FRE'.or.value(1:3).eq.'fre') then
			iform = 2
		else
			write(6,*) 
     1			'Input format for molecule not recognized:',
     2			value(1:3)
			stop
		endif
	elseif(i.eq.14) then
		fname_log(1:50) = value
	elseif(i.eq.15) then
		fname_py(1:50) = value
		flag_pymol = 1
	elseif(i.eq.16) then
		read(value,*) flag_box
	elseif(i.eq.17) then
		read(value,*) flag_distrib
	elseif(i.eq.18) then
		read(value,*) Temp
	elseif(i.eq.19) then
		read(value,*) eps1
	elseif(i.eq.20) then
		read(value,*) eps2
	elseif(i.eq.21) then
		read(value,*) flag_eps
	elseif(i.eq.22) then
		n_ion = n_ion + 1
		read(value,*) ion_z(n_ion),ion_conc(n_ion),
     1		ion_rad(n_ion)
	elseif(i.eq.23) then
		read(value,*) bnd_type
	elseif(i.eq.24) then
		read(value,*) method
	elseif(i.eq.25) then
		read(value,*) tol
	elseif(i.eq.26) then
		read(value,*) nlin
	elseif(i.eq.27) then
		read(value,*) nlevel
	elseif(i.eq.28) then
		read(value,*) shiftx,shifty,shiftz
	elseif(i.eq.29) then
		read(value,*) flag_map_format
	elseif(i.eq.30) then
		fname_out_phi(1:50) = value
		call truncate(fname_out_phi)
		flag_phi = 1
	elseif(i.eq.31) then
		fname_out_edens(1:50) = value
		call truncate(fname_out_edens)
		flag_edens = 1
	elseif(i.eq.32) then
		fname_out_nion(1:50) = value
		call truncate(fname_out_nion)
		flag_nion = 1
	elseif(i.eq.33) then
		fname_out_dion(1:50) = value
		call truncate(fname_out_dion)
		flag_dion = 1
	elseif(i.eq.34) then
		read(value,*) R_ion
		ion_zone = 1
	elseif(i.eq.35) then
		n_solv = n_solv + 1
		read(value,*) solv_dipole(n_solv),solv_conc(n_solv)
	elseif(i.eq.36) then
		read(value,*) a_lattice
	elseif(i.eq.37) then
		fname_out_dipole(1:50) = value
		call truncate(fname_out_dipole)
		flag_dipole = 1
        elseif(i.eq.38) then
                flag_slab = 1
                read(value,*) z0
        elseif(i.eq.39) then
                flag_slab = 1
                read(value,*) zh
        elseif(i.eq.40) then
                flag_slab = 1
                read(value,*) eps_slab
        elseif(i.eq.41) then
                read(value,*) npoint_sphere
        elseif(i.eq.42) then
                read(value,*) ndiscrete
        elseif(i.eq.43) then
                read(value,*) equation
        elseif(i.eq.44) then
                read(value,*) ijac2
        elseif(i.eq.45) then
                read(value,*) nsmooth
	elseif(i.eq.46) then
		read(value,*) v0,b
	elseif(i.eq.47) then
		read(value,*) lambda
	elseif(i.eq.48) then
		fname_out_psi(1:50) = value
		call truncate(fname_out_psi)
		flag_psi = 1
	elseif(i.eq.49) then
		flagcenter = 1
		fpdbcenter(1:50) = value
		call truncate(fpdbcenter)
	endif
c
	goto 100
c
400	continue
c
	if(equation.le.3.and.method.ge.4) method = 2
c
	Npoint  = nx*ny*nz
        N = max(nx,ny)
        N = max(N,nz)
c
c ==========================================================================================
c == Set number of coarse  grids considered   ==============================================
c ==========================================================================================
c
	if(nlevel.eq.0.or.nlevel.eq.-1) then
c
c               If nlevel is not defined in the input parameter file (or set to 0),
c               set it automatically such that the lowest level is 3,3,3
c
		ival = 1
		do 500 k = 1,10
			ival = 2*ival
			if(ival.eq.N-1) goto 600
500		continue
600		continue
		nlevel = k-1
		nlevel = min(nlevel,6)
c
	endif
c
c ==========================================================================================
c == Define if equation is linear or not      ==============================================
c ==========================================================================================
c
        if(equation.eq.0.or.equation.eq.1) then
                nlin = 0
        else
                nlin = 1
        endif
c
        if(ijac2.eq.7) then
                ijac = 0
        else
                ijac = 1
        endif
c
c ==========================================================================================
c ====  Special case: z:z electrolytes                ======================================
c ==========================================================================================
c
        zn_max = ion_z(1)
        zp_max = ion_z(1)
        do 625 i = 2,n_ion
                if(ion_z(i).gt.0) then
                        zp_max = max(zp_max,ion_z(i))
                else
                        zn_max = min(zn_max,ion_z(i))
                endif
625     continue
c
c ==========================================================================================
c       Make sure we have a Stern region if there are ions
c ==========================================================================================
c
        if(n_ion.ne.0) then
                do 650 i = 1,n_ion
                        R_ion = max(R_ion,ion_rad(i))
650             continue
c                if(R_ion.eq.0) R_ion = 1.5
                ion_zone = 1
        endif 
c
c ==========================================================================================
c       FIle name for PDB file containing centered molecule
c ==========================================================================================
c
	if(flagcenter.eq.0) then
	  fpdbcenter = fname_pdb
	endif
	idx = index(fname_pdb,'.') -1
	idx2 = index(fname_pdb,' ') -1
	fname_newpdb = fname_pdb(1:idx)//'_center'//
     1	fname_pdb(idx+1:idx2)
c
c ==========================================================================================
c ====  Miscallenii                 ========================================================
c ==========================================================================================
c
	l_log = 12
	write(6,*)  ' '
	write(6,*) 'Log file for this run : ',
     1		fname_log(1:index(fname_log,' ')-1)
	write(6,*)  ' '

	open(unit=l_log,file=fname_log,status='unknown',
     1		form='formatted')
c
	write(l_log,2)
	write(l_log,3) fname_pdb(1:50)
	write(l_log,29) fpdbcenter(1:50)
	write(l_log,4) fname_log(1:50)
	if(flag_pymol.eq.1) write(l_log,5) fname_py(1:50)
	write(l_log,6) mesh(grid_type+1)
	write(l_log,7) surf(surf_type+1)
	write(l_log,8) R_probe
	write(l_log,9) crg_distrib(flag_distrib+1)
	write(l_log,10) Temp
	write(l_log,11) eps1
	write(l_log,12) eps2
	write(l_log,18) epsval(flag_eps+1)
	do 700 i = 1,n_ion
		write(l_log,13) i,ion_z(i),ion_conc(i),ion_rad(i)
700	continue
	do 800 i = 1,n_solv
		write(l_log,14) i,solv_dipole(i),solv_conc(i)
800	continue
	write(l_log,15) R_ion
	write(l_log,16) boundary(bnd_type+1)
	write(l_log,19) equa_name(equation+1)
	write(l_log,17) solver(method+1)
        write(l_log,27) coarse(ndiscrete+1)
        write(l_log,28) smooth_name(nsmooth)
	if(method.ne.0) then 
		write(l_log,20) nlevel
	endif
	write(l_log,*) ' '
	if(flag_map_format.eq.3.or.flag_map_format.eq.4) then
		write(l_log,25) fname_newpdb
	endif
	if(flag_phi.ne.0) then
		write(l_log,21) fname_out_phi
	endif
	if(flag_edens.ne.0) then
		write(l_log,22) fname_out_edens
	endif
	if(flag_nion.ne.0) then
		write(l_log,23) fname_out_nion
	endif
	if(flag_dion.ne.0) then
		write(l_log,24) fname_out_dion
	endif
	if(flag_dipole.ne.0) then
		write(l_log,26) fname_out_dipole
	endif
c
	return
c
1000	ierr = 1
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	Set_Constants.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	7/5/07
c	Author:		Patrice Koehl
c			in collaboration with Marc Delarue
c
c	Purpose:	Defines all constants (universal, and those specific to
c			the run
c
	subroutine Set_constants
c
	include 'poisson.h'
c
c ==========================================================================================
c = Define all variables       =============================================================
c ==========================================================================================
c
c	Constants
c ==========================================================================================
c
	integer	i,n_ion,n_solv,n_zz
	integer	method,nlevel,nlin,equation,ndiscrete,ijac,nsmooth
c
	real*8	l_Bjerrum,pi,Temp,KB,electron,Ang,beta,Avogadro,eps0
	real*8	C_charge,C_prop,val
	real*8	eps1,eps2,eps_bulk
	real*8	ratio,phi0,coef_Z0
	real*8	tol
	real*8	vol_min,vol_max,d_min,d_max,a_lattice,v_lattice
	real*8	p0_e,coef_l
	real*8	Ion_Strength,kappa0
	real*8	solv_conc(max_solv)
	real*8	solv_cvol(max_solv),solv_dipole(max_solv)
	real*8	ion_conc(max_ion),ion_z(max_ion),ion_rad(max_ion)
	real*8	kappa(max_ion)
	real*8	ion_cvol(max_ion)
	real*8	l_ion(max_ion),l_dip(max_solv)
c
c	Common blocks
c ==========================================================================================
c
	common /constants/ 	l_Bjerrum,pi,Temp,KB,electron,Ang,
     1				Avogadro,eps0,beta,C_charge,C_prop
	common /elec_const/ 	eps1,eps2
	common /ions/		ion_conc,ion_z,ion_rad,	kappa,n_ion,n_zz
	common /solvent/	solv_conc,solv_dipole,solv_cvol,n_solv
	common /smpb/		ion_cvol,ratio,phi0,coef_Z0
	common /dipoles/	vol_max,p0_e,coef_l
	common /pdesolver/	tol,method,nlevel,nlin,equation,
     1                          ndiscrete,ijac,nsmooth
	common /salts/		Ion_strength,kappa0
	common /lattice/	a_lattice,v_lattice
        common /bulk/           eps_bulk
	common /fugacities/	l_dip,l_ion
c
c ==========================================================================================
c = Define constants           =============================================================
c ==========================================================================================
c
c	All units are defined in SI, except when
c	dealing with distance only maintain them in Angstrom)
c
c	KB is the Boltzmann constant (in J/K)
c
	KB = 1.3806505d-23
c
c	electron charge,  in international unit (C)
c
	electron = 1.602176462d-19 
c
c	pi set to arc cosine of -1
c
	pi = dacos(-1.d0)
c
c	Ang is a conversion factor from angstrom to m
c
	Ang = 1.d-10
c
c	Avogadro is the Avogadro Number
c
	Avogadro = 6.02214179d23
c
c	Eps0 is the electric constant
c
	eps0 = 8.854187817d-12
c
c ==========================================================================================
c == Set units depending on Temp              ==============================================
c ==========================================================================================
c
c
c	Set beta = 1./(KB*Temp)
c
	beta = 1.d0/(KB*Temp)
c
c	The Bjerrum length lB is the separation at which the electrostatic interaction 
c	between two elementary	charges is equal to kb T, i.e.: 
c		kb T = e^2/(4 pi eps lB)
c
	l_Bjerrum = 1.d10*beta*electron*electron/(4*pi*eps0)
c
c	write(6,*) 'l_Bjerrum in vacuo : ',l_Bjerrum
c	write(6,*) 'l_Bjerrum in water : ',l_Bjerrum/80.d0
c
c	The ionic strength of a solution is classically defined as:
c	Ion_strength = 0.5 * Sum_i (z_i)^2 Ci
c	where z_i is the valence of ion i, and Ci its concentration
c
	Ion_strength = 0
	do 100 i = 1,n_ion
		Ion_strength = Ion_strength + 0.5d0*ion_conc(i)*
     1				ion_z(i)*ion_z(i)
100	continue
c
c	kappa is the Debye-Huckel constant.
c
c	kappa = -4*pi*NA*(electron)^2 * ion_conc *ion_z / (kB T)
c	where NA is Avogadro number
c
c	Notes:
c		- since we use SI, we need to multiply by 1/(4*pi*eps)
c		- kappa is multiplied by a concentration, which is in mol/l,
c		  i.e. in mol/(dm^3); in SI, we need to multiply 1000
c		- finally, kappa is in unit m^-2; since we express distances
c		  in angstrom, we divide by 1.d20
c
	val = Avogadro*electron*electron*beta*1.d-17/eps0
c        write(6,*) 'val =',val
	kappa0 = val*2d0*Ion_strength
c
	do 200 i = 1,n_ion
		kappa(i) = -ion_conc(i) * ion_z(i)/(2.d0*Ion_strength)
200	continue
c
c	C_charge is the constant in front of charge, for dimensionless
c	Poisson-Boltzmann equation
c
c	C_charge = (4*pi*(electron)^2 / (kB *Temp)
c
c	Notes: 
c		- Since "electron" is expressed in C, we need to scale by 1/(4*pi*eps0)
c		- C_charge is in unit of m^-1; it is however easier to express distances
c		  in angstroms, in which case we multiply C_charge by 1.d10
c
	C_charge = 1.d10*electron*electron*beta/eps0
c        write(6,*) 'C_charge =',C_charge
c
c	C_prop is the coefficient to convert the Coulomb potential used for
c	boundary condition into a dimensionless potential value: 
c		C_prop = electron*electron/(kB*Temp)
c	Notes:
c		- Since "electron" is expressed in C, we need to scale by 1/(4*pi*eps0)
c		- since potential is in 1/r, and r is expressed in angstrom, we
c		  scale C_prop by 1.d10
c
c	C_prop = electron*electron*beta*1.d10/(4*pi*eps0)
	C_prop = C_charge/(4*pi)
c        write(6,*) 'C_prop =',C_prop
c
c	if(equation.le.2) return
c
c ==========================================================================================
c == Set constants for Size Modified Poisson Boltzmann equation ============================
c ==========================================================================================
c
c	Application of SMPB equation implies that we have one ion specie with a small
c	radius, and at least one other ion specie with a larger radius; we also
c	suppose that all larger radii are all equal.
c
c	For simplicity, we make sure that the smaller ion is in position 1, while
c	all other ions start at position 2.
c
	d_min = 2*ion_rad(1)
	do 300 i = 2,n_ion
		if(d_min.gt.2*ion_rad(i)) then
			d_min = ion_z(i)
			ion_z(i) = ion_z(1)
			ion_z(1) = d_min
			d_min = ion_conc(i)
			ion_conc(i) = ion_conc(1)
			ion_conc(1) = d_min
			d_min = kappa(i)
			kappa(i) = kappa(1)
			kappa(1) = d_min
			d_min = 2*ion_rad(i)
			ion_rad(i) = ion_rad(1)
			ion_rad(1) = d_min/2
		endif
300	continue
c
c	Make sure that radii of ions from 2 to n_ion are the same...
c
	d_max = 2.*ion_rad(2)
	do 400 i = 3,n_ion
	    if(d_max.lt.2.*ion_rad(i)) then
		write(6,*) 'SMPB can currently only handle 2 ion sizes'
		write(6,*) 'Please revise ion radii in input file'
		stop
	    endif
400	continue
c
	d_max = max(d_max,a_lattice)
c
c	Define vol_min, vol_max and ratio
c
	vol_min = d_min*d_min*d_min
	vol_max = d_max*d_max*d_max
c
	if(vol_min.ne.0) then
		ratio = vol_max / vol_min
	else
		ratio = 0.d0
	endif
c
c	For each ion specie, compute conc*vol, in number of ions
c
        phi0 = 0
        do 500 i = 1,n_ion
		ion_cvol(i) = vol_max*Avogadro*1.d-27*ion_conc(i)
		phi0 = phi0 + ion_cvol(i)
500     continue
c
	coef_Z0 = 1
	do 550 i = 2,n_ion
		coef_Z0 = coef_Z0 - ion_cvol(i)
550	continue
c
c	if(equation.le.3) return
c
c ==========================================================================================
c == Set constants for Generalized Langevin                     ============================
c ==========================================================================================
c
c	Define p0_e:
c
c	p0 is the solvent dipole moment. It is read in from the input file
c	in Debye, which we convert to C.m with factor 3.33 10^-30.
c	p0 will come in as : beta*p0*Efield.  Since we solve the equation
c	with a unitless potential (i.e. we find u = phi*electron/(kB*T) ),
c	beta*p0*Efield becomes p0 *Efield(unitless) / electron.
c	Since Efield is du/dx, and dx is in angstrom on the grid, we multiply
c	by 1.d10 (angstrom -> meter)
c
	p0_e = solv_dipole(1) * 3.33d-30 * 1.d10 / electron
c
c	Compute the fugacities of the ions and solvents
c
	do 600 i = 1,n_solv
		solv_cvol(i) = vol_max*Avogadro*1.d-27*solv_conc(i)
600	continue
c
	do 700 i = 1,n_solv
		phi0 = phi0 +solv_cvol(i)
700	continue
c
c	Define coef_l = 4*pi*lB*p0^2 / (a^3 * e^2)
c
	if(vol_max.ne.0) then
		coef_l = beta*electron*electron*p0_e*p0_e*1.d10/
     1			(vol_max*eps0)
	else
		coef_l = 0
	endif
c
        eps_bulk = coef_l * solv_cvol(1)/3
        eps_bulk = eps_bulk + 1
c
c       Define fugacities:
c
	do 800 i = 1,n_solv
		l_dip(i) = solv_cvol(i)/(1-phi0)
800	continue
c
	do 900 i = 1,n_ion
		l_ion(i) = ion_cvol(i)/(1-phi0)
900	continue
c
	v_lattice = vol_max
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	Process_string.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	7/5/07
c	Author:		Patrice Koehl
c			in collaboration with Marc Delarue
c
c	Purpose:	"cleans" a string: remove all comments after "!", remove trailing blanks, 
c			and extra characters
c
	subroutine process_string(string_in,string_out)
c
	integer	i,imax,imin
c
	character	string_in*50,string_out*50
c
	string_out = '                              '
c
c	Remove comments after "!"
c
	do 100 i = 1,50
		if(string_in(i:i).eq.'!') then
			imax = i-1
			goto 200
		endif
100	continue
	imax = 50
200	continue
c
c	Remove all tabs
c
	do 300 i = 1,imax
		if(string_in(i:i).eq.'	') string_in(i:i) = ' '
300	continue
c
c	Remove trailing blanks (right)
c
	do 400 i = imax,1,-1
		if(string_in(i:i).ne.' ') goto 500
400	continue
c
500	continue
c
	imax = i
c
c	Remove trailing blanks (left)
c
	do 600 i = 1,imax
		if(string_in(i:i).ne.' ') goto 700
600	continue
c
700	continue
c
	imin = i
c
	string_out(1:imax-imin+1) = string_in(imin:imax)
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c	Read_Charge.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	6/27/07
c	Author:		Patrice Koehl (based on Cyril Azuaras program)
c			in collaboration with Marc Delarue
c
c	Purpose:	Read in a PDB file containing the positions
c			of the atoms, as well as their partial charges
c			and radii
c			Once read, the molecule is centered on (0,0,0)
c			point 
c
c	Input:
c		- l_in		: name of the file containing the charges
c		- l_log		: name of log file for the run
c		- iform		: flag:
c					0 if file is in strict PDB format
c					1 if file is in "loose" PDB format
c					  (i.e. column separated, but not necessarily
c					   following Fortran format for PDB file)
c
c	Output:
c		- ierr		: flag:
c					0 : normal output
c					1 : problem while reading file
c		
c	Common blocks used:
c
c	    -charges 
c		- natom   	: number of atoms (charges)
c		- atom_coord	: coordinates of atoms (after recentering)
c		- atom_radius	: radii of atoms
c		- atom_charge	: charges of atoms
c		- atom_occup	: occupancies of atoms
c		- atom_label	: label in PDB file for each atom
c		(all these informations stored in common block CHARGES
c
c	    -centers
c		- cg(3)		: original center of molecule
c		- Center(3)	: center of the grid
c
c
	subroutine Read_Charge(fname_pdb,fpdbcenter,l_log,ierr)
c
	include 'poisson.h'
c
c ==========================================================================================
c = Define all variables       =============================================================
c ==========================================================================================
c
	integer	i,j,natom,ncharge,l_in,l_log,l_center
	integer ierr,iform
c
	real*8	x,crg_tot,crg_pos,crg_neg
	real*8	cg(3),c_grid(3),min_coord(3),max_coord(3)
	real*8	atom_coord(*)
	real*8	atom_radius(*),atom_charge(*)
c
	character record*80,fname_pdb*100
	character fpdbcenter*100
	character atom_label(*)*30,atom_label2(*)*12
c
c ==========================================================================================
c == Pointers for atom arrays              =================================================
c ==========================================================================================
c
        pointer (ptr_coord,atom_coord)
        pointer (ptr_radius,atom_radius)
        pointer (ptr_charge,atom_charge)
        pointer (ptr_label,atom_label)
        pointer (ptr_label2,atom_label2)
c
c ==========================================================================================
c == Common blocks                         =================================================
c ==========================================================================================
c
        common /charges/        ptr_coord,ptr_charge,ptr_radius,
     1                          ptr_label,ptr_label2,iform,natom,
     2                          ncharge
c
	common /centers/ cg,c_grid
c
c ==========================================================================================
c = Output format (to write in log file)  ==================================================
c ==========================================================================================
c
1	format(a)
2	format(3f8.3,2f6.2)
3	format(70('='),/,2('='),'Information on molecule(s)',40x,
     1	2('='),/,70('='),/)  
4	format('Number of atoms in molecule            : ',i6)
5	format('Total charge of the molecule           : ',f10.3, ' e')
6	format('Sum of negative charges in molecule    : ',f10.3, ' e')
7	format('Sum of positive charges in molecule    : ',f10.3, ' e')
8	format(a30)
9	format(a12)
10      format(3f8.3,1x,2f7.4)
c
	ierr = 0
	l_in = 11
	l_center = 103
c!!!!!! Next line changed 01/02/12
c	open(unit=l_in,file=fname_pdb,status='old',err=1000)
c	write(6,*) "Opening file: ",fpdbcenter," /"
	open(unit=l_center,file=fpdbcenter,status='old',err=1001)
c	write(6,*) "Opened: ",fpdbcenter
c
c ==========================================================================================
c = Read info from PDB         =============================================================
c ==========================================================================================
c
c	Read each line of input file: only keep lines that start
c	with "ATOM  ", and store coordinates and charges
c
c       First count number of atoms
c
	natom = 0
50	read(l_center,1,end=75,err=1001) record
c	write(6,*) record
	if(record(1:6).ne.'ATOM  ') goto 50
c
	natom = natom + 1
        goto 50
75      continue
c	write(6,*) "Number of atoms in ",fpdbcenter,"  = ",natom
c
c       Allocate space
c
        ptr_charge = malloc(8*natom)
        ptr_radius = malloc(8*natom)
        ptr_coord  = malloc(24*natom)
        ptr_label  = malloc(30*natom)
        ptr_label2 = malloc(12*natom)
c
        rewind(l_center)
c
	natom = 0
100     read(l_center,1,end=200,err=1001) record
	if(record(1:6).ne.'ATOM  ') goto 100
c
	natom = natom + 1
	read(record,8) atom_label(natom)
	read(record(55:66),9) atom_label2(natom)
	if(iform.eq.0) then
		read(record(31:80),2) (atom_coord(3*(natom-1)+i),i=1,3),
     1			atom_charge(natom),atom_radius(natom)
	elseif(iform.eq.1) then
		read(record(31:80),10)(atom_coord(3*(natom-1)+i),i=1,3),
     1			atom_charge(natom),atom_radius(natom)
	elseif(iform.eq.2) then
		read(record(31:80),*) (atom_coord(3*(natom-1)+i),i=1,3),
     1			atom_charge(natom),atom_radius(natom)
c		write(6,*) (atom_coord(3*(natom-1)+i),i=1,3),
c     1			atom_charge(natom),atom_radius(natom)
	endif
c	if(atom_charge(natom).eq.0.and.atom_radius(natom).eq.0) then
c		natom = natom -1
c	endif
	goto 100
c
200	continue
	close(unit=l_center)
	ncharge = natom
c	write(6,*) "Second pass: ",natom
c
c ==========================================================================================
c = Recenter molecule          =============================================================
c ==========================================================================================
c
c	A. Now that we have read all atoms, find center of mass
c
	do 300 i = 1,3
		cg(i) = atom_coord(i)
                min_coord(i) = cg(i)
                max_coord(i) = cg(i)
300	continue
c
	do 500 i = 2,natom
		do 400 j = 1,3
			x = atom_coord(3*(i-1)+j)
c			cg(j) = cg(j) + x
                        if(x.lt.min_coord(j)) min_coord(j) = x
                        if(x.gt.max_coord(j)) max_coord(j) = x
400		continue
500	continue
c
	do 600 i = 1,3
c		cg(i) = cg(i)/natom
                cg(i) = 0.5d0*(min_coord(i)+max_coord(i))
600	continue
	write(6,*) "Center of the box = ",cg(1),cg(2),cg(3)
c	
c ==========================================================================================
c = Now only keep molecules with non-zero radius and charge  ===============================
c ==========================================================================================
c	
	call Clean_Charges
c	
	open(unit=l_in,file=fname_pdb,status='old',err=1000)
c	
        natom = 0
101      read(l_in,1,end=102,err=1000) record
        if(record(1:6).ne.'ATOM  ') goto 101
c	
        natom = natom + 1
        goto 101
102      continue
c
c       Allocate space
c
        ptr_charge = malloc(8*natom)
        ptr_radius = malloc(8*natom)
        ptr_coord  = malloc(24*natom)
        ptr_label  = malloc(30*natom)
        ptr_label2 = malloc(12*natom)
c
        rewind(l_in)
c
        natom = 0
103     read(l_in,1,end=104,err=1000) record
        if(record(1:6).ne.'ATOM  ') goto 103
c
        natom = natom + 1
        read(record,8) atom_label(natom)
        read(record(55:66),9) atom_label2(natom)
        if(iform.eq.0) then
                read(record(31:80),2) (atom_coord(3*(natom-1)+i),i=1,3),
     1                  atom_charge(natom),atom_radius(natom)
        elseif(iform.eq.1) then
                read(record(31:80),10)(atom_coord(3*(natom-1)+i),i=1,3),
     1                  atom_charge(natom),atom_radius(natom)
        elseif(iform.eq.2) then
                read(record(31:80),*) (atom_coord(3*(natom-1)+i),i=1,3),
     1                  atom_charge(natom),atom_radius(natom)
c               write(6,*) (atom_coord(3*(natom-1)+i),i=1,3),
c     1                 atom_charge(natom),atom_radius(natom)
        endif
        if(atom_charge(natom).eq.0.and.atom_radius(natom).eq.0) then
                natom = natom -1
        endif
        goto 103
c
104     continue
        ncharge = natom
c	
c       B. Now recenter molecule such that it is centered on 0
c
        do 800 i = 1,natom
                do 700 j = 1,3
                        atom_coord(3*(i-1)+j) = atom_coord(3*(i-1)+j)
     1                          - cg(j)
700             continue
800     continue
c	
c ==========================================================================================
c = Compute stat on charges    =============================================================
c ==========================================================================================
c
	crg_tot = 0.d0
	crg_pos = 0.d0
	crg_neg = 0.d0
	do 900 i = 1,natom
		x = atom_charge(i)
		crg_tot=crg_tot + x
		if(atom_charge(i).gt.0) then
			crg_pos = crg_pos + x
		else
			crg_neg = crg_neg + x
		endif
900	continue
c
c ==========================================================================================
c = Write info in log file     =============================================================
c ==========================================================================================
c
	write(l_log,3)
	write(l_log,4) natom
	write(l_log,5) crg_tot 
	write(l_log,6) crg_neg 
	write(l_log,7) crg_pos 
	write(l_log,*) ' '
c
c ==========================================================================================
c = Return in calling program (normal exit) ================================================
c ==========================================================================================
c
	close(unit=l_in)
c
	return
c
1000	continue
c
c ==========================================================================================
c = Return in calling program (error reading file)  ========================================
c ==========================================================================================
c
	ierr = 1
c
	return
c	
1001	continue
c	
	ierr = 2
c	
	return
	end
c ==========================================================================================
c ==========================================================================================
c	Clean_Charges.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	7/5/07
c	Author:		Patrice Koehl
c			in collaboration with Marc Delarue
c
c	Purpose:	Free all arrays that stored charge information
c
        subroutine Clean_Charges
c
c	Information on molecule
c ==========================================================================================
c
	integer	natom,ncharge,iform
c
	real*8	atom_coord(*)
	real*8	atom_radius(*),atom_charge(*)
c
	character atom_label(*)*30
	character atom_label2(*)*12
c

c ==========================================================================================
c == Pointers for atom arrays              =================================================
c ==========================================================================================
c
        pointer (ptr_coord,atom_coord)
        pointer (ptr_radius,atom_radius)
        pointer (ptr_charge,atom_charge)
        pointer (ptr_label,atom_label)
        pointer (ptr_label2,atom_label2)
c
c ==========================================================================================
c == Common blocks                         =================================================
c ==========================================================================================
c
        common /charges/        ptr_coord,ptr_charge,ptr_radius,
     1                          ptr_label,ptr_label2,iform,natom,
     2                          ncharge
c
        call free(ptr_coord)
        call free(ptr_charge)
        call free(ptr_radius)
        call free(ptr_label)
        call free(ptr_label2)
c
        return
        end
c
	subroutine truncate(fname)
c
	integer	i
	character	fname*100,fname2*100
c
	do 100 i = 100,1,-1
		if(fname(i:i).ne.' ') goto 200
100	continue
c
200	continue
c
	fname2 = fname(1:i)
	fname  = ' '
	fname  = fname2
c
	return
	end
