c ==========================================================================================
c ==========================================================================================
c	IonSlab.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	7/04/07
c	Author:		Patrice Koehl 
c			in collaboration with Marc Delarue
c
c       Exclude ions inside the slab, except in pores inside the
c       molecule
c
c	Input:
c               z0      : center of the slab on Z axis
c               zh      : 1/2 height of the slab
c
	subroutine IonSlab(z0,zh,grid_partition,hz,nx,ny,nz)
c
c ==========================================================================================
c = Define all variables       =============================================================
c ==========================================================================================
c
c	Information on grid
c
        integer nx,ny,nz
        integer istop,istop2
c
	real*8	hz(nz)
c
c	Partition of the grid
c
	integer*1	grid_partition(nx,ny,nz)
c
c	Information on charges
c
c	Local variables
c
	integer	i,j,k
c
	real*8	z0,zh
c
        do 1000 k = 1,nz
c
           if(abs(hz(k)-z0).gt.zh) goto 1000
c
                do 900 j = 1,ny
c
c               Scan x-line: it should look like:
c
c               From 0 to first point protein point, reset
c               to 2
c               From nx back to first protein point, reset to 2
c
                        istop = 0
                        do 100 i = 1,nx
c
                                if(grid_partition(i,j,k).eq.1) then
                                        istop = i
                                        goto 200
                                endif
c
                                grid_partition(i,j,k) = 2
c
100                     continue
c
                        if(i.eq.nx) goto 900
c
200                     continue
c
                        istop2 = 0
                        do 300 i = nx,istop,-1
c
                                if(grid_partition(i,j,k).eq.1) then
                                        istop2 = i
                                        goto 400
                                endif
c
                                grid_partition(i,j,k) = 2
c
300                     continue
c
400                     continue
c
900             continue
c
1000    continue
c
        return
        end
c ==========================================================================================
c ==========================================================================================
c	EpsilonSlab.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	7/04/07
c	Author:		Patrice Koehl 
c			in collaboration with Marc Delarue
c
c       Set epsilon to eps_slab inside the slab, except in pores inside the
c       molecule
c
c	Input:
c               z0      : center of the slab on Z axis
c               zh      : 1/2 height of the slab
c
	subroutine EpsilonSlab(z0,zh,eps,eps_slab,grid_shift,ishift,
     1                          hz,nx,ny,nz)
c
c ==========================================================================================
c = Define all variables       =============================================================
c ==========================================================================================
c
c	Information on grid
c
        integer nx,ny,nz
        integer ishift,info
        integer istop,istop2
c
	real*8	hz(nz)
c
c	Partition of the grid
c
	integer*1	grid_shift(nx,ny,nz)
c
c	epsilon values        
c
        real*8  eps_slab
        real*8  eps(nx,ny,nz)
c
c	Local variables
c
	integer	i,j,k
c
	real*8	z0,zh
c
        do 1000 k = 1,nz
c
           if(abs(hz(k)-z0).gt.zh) goto 1000
c
                do 900 j = 1,ny
c
c               From 0 to first protein point, reset eps to eps_slab
c               From nx back to first protein point, reset eps to eps_slab
c
                        istop = 0
                        do 100 i = 1,nx
c
				info=ibits(grid_shift(i,j,k),ishift,1)
                                if(info.eq.0) then
                                        istop = i
                                        goto 200
                                endif
c
                                eps(i,j,k) = eps_slab
c
100                     continue
c
                        if(i.eq.nx) goto 900
c
200                     continue
c
                        istop2 = 0
                        do 300 i = nx,istop,-1
c
				info=ibits(grid_shift(i,j,k),ishift,1)
                                if(info.eq.0) then
                                        istop2 = i
                                        goto 400
                                endif
c
                                eps(i,j,k) = eps_slab
c
300                     continue
c
400                     continue
c
900             continue
c
1000    continue
c
        return
        end
