c       Copyright (C) 2007 Patrice Koehl
c
c       This library is free software; you can redistribute it and/or
c       modify it under the terms of the GNU Lesser General Public
c       License as published by the Free Software Foundation; either
c       version 2.1 of the License, or (at your option) any later version.
c
c       This library is distributed in the hope that it will be useful,
c       but WITHOUT ANY WARRANTY; without even the implied warranty of
c       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
c       Lesser General Public License for more details.
c
c       You should have received a copy of the GNU Lesser General Public
c       License along with this library; if not, write to the Free Software
c       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
c
c ==========================================================================================
c ==========================================================================================
c	residual_yukawa.f
c
c	This subroutines computes the residual for the non linear 
c       Yukawa equation
c ==========================================================================================
c ==========================================================================================
c
	subroutine residual_yukawa(nx,ny,nz,ipc,rpc,ac,grid,cc,fc,sinhu,
     1                          x,r,w1,hx,hy,hz)
c
c ==========================================================================================
c	Declaration 
c ==========================================================================================
c
	integer	nx,ny,nz,numdia
c
        integer*1 grid(*)
c
	integer	ipc(*)
c
	real*8	rpc(*),ac(nx*ny*nz,*),cc(*),fc(*),x(*),r(*),w1(*)
        real*8  sinhu(*),hx(*),hy(*),hz(*)
c
c ==========================================================================================
c	Call proper routine based on number of diagonals
c ==========================================================================================
c
	numdia = ipc(11)
c
	if(numdia.eq.7) then
		call residual7_yukawa(nx,ny,nz,ipc,rpc,ac(1,1),grid,
     1			cc,fc,sinhu,ac(1,2),ac(1,3),ac(1,4),
     2			x,r,w1,hx,hy,hz)
	elseif(numdia.eq.27) then
		call residual27_yukawa(nx,ny,nz,ipc,rpc,ac(1,1),grid,
     1		cc,fc,sinhu,ac(1,2),ac(1,3),ac(1,4),ac(1,5),ac(1,6),
     2		ac(1,7),ac(1,8),ac(1,9),ac(1,10),ac(1,11),ac(1,12),
     3		ac(1,13),ac(1,14),x,r,w1,hx,hy,hz)
	else
		write(6,*) 'Residual: bad stencil type...'
		stop
	endif
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	residual7_yukawa.f
c
c	This subroutines computes the residual when the matrix is stored with 
c	7-diagonal (banded) format
c ==========================================================================================
c ==========================================================================================
c
	subroutine residual7_yukawa(nx,ny,nz,ipc,rpc,oC,grid,cc,fc,
     1                  sinhu,oE,oN,uC,x,r,w1,hx,hy,hz)
c
c ==========================================================================================
c	Declaration 
c ==========================================================================================
c
	integer	nx,ny,nz,i,j,k
c
        integer*1 grid(*)
c
	integer	ipc(*)
c
	real*8	rpc(*)
	real*8	oE(nx,ny,nz),oN(nx,ny,nz),uC(nx,ny,nz)
	real*8	cc(nx,ny,nz),fc(nx,ny,nz),oC(nx,ny,nz)
	real*8	x(nx,ny,nz),r(nx,ny,nz),w1(nx,ny,nz)
        real*8  sinhu(nx,ny,nz)
	real*8	hx(*),hy(*),hz(*)
c
c ==========================================================================================
c	Compute residual
c ==========================================================================================
c
	call c_vec_yukawa(grid,cc,sinhu,x,w1,nx,ny,nz,hx,hy,hz)
c
	do 300 k = 2,nz-1
		do 200 j = 2,ny-1
			do 100 i = 2,nx-1
c
				r(i,j,k) = fc(i,j,k)
     1					+ oN(i,j,k)	*x(i,j+1,k)
     2					+ oN(i,j-1,k)	*x(i,j-1,k)
     3					+ oE(i,j,k)	*x(i+1,j,k)
     4					+ oE(i-1,j,k)	*x(i-1,j,k)
     5					+ uC(i,j,k)	*x(i,j,k+1)
     6					+ uC(i,j,k-1)	*x(i,j,k-1)
     7					- oC(i,j,k)*x(i,j,k) -w1(i,j,k)
c
100			continue
200		continue
300	continue
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	residual27.f
c
c	This subroutines computes the residual of the linear equation when A
c	is stored with 27-diagonal (banded) format
c ==========================================================================================
c ==========================================================================================
c
	subroutine residual27_yukawa(nx,ny,nz,ipc,rpc,oC,grid,cc,fc,
     1	sinhu,oE,oN,uC,oNE,oNW,uE,uW,uN,uS,uNE,uNW,uSE,uSW,x,r,w1,
     2	hx,hy,hz)
c
c ==========================================================================================
c	Declaration 
c ==========================================================================================
c
	integer	nx,ny,nz,i,j,k
c
        integer*1 grid(*)
c
	integer	ipc(*)
c
	real*8	tmp0,tmpU,tmpD
c
	real*8	rpc(*)
	real*8	hx(*),hy(*),hz(*)
	real*8	oE(nx,ny,nz),oN(nx,ny,nz),uC(nx,ny,nz)
	real*8	oNE(nx,ny,nz),oNW(nx,ny,nz),uE(nx,ny,nz)
	real*8	uW(nx,ny,nz),uN(nx,ny,nz),uS(nx,ny,nz)
	real*8	uNE(nx,ny,nz),uNW(nx,ny,nz),uSE(nx,ny,nz)
	real*8	uSW(nx,ny,nz)
	real*8	cc(nx,ny,nz),oC(nx,ny,nz),fc(nx,ny,nz)
	real*8	x(nx,ny,nz),r(nx,ny,nz),w1(nx,ny,nz)
        real*8  sinhu(nx,ny,nz)
c
c ==========================================================================================
c	Compute the residual
c ==========================================================================================
c
	call c_vec_yukawa(grid,cc,sinhu,x,w1,nx,ny,nz,hx,hy,hz)
c
	do 300 k=2,nz-1
		do 200 j = 2,ny-1
			do 100 i = 2,nx-1
c
				tmp0 =
     1				+ oN(i,j,k)	* x(i,j+1,k)
     2				+ oN(i,j-1,k)	* x(i,j-1,k)
     3				+ oE(i,j,k)	* x(i+1,j,k)
     4				+ oE(i-1,j,k)	* x(i-1,j,k)
     3				+ oNE(i,j,k)	* x(i+1,j+1,k)
     4				+ oNE(i-1,j-1,k)* x(i-1,j-1,k)
     3				+ oNW(i,j,k)	* x(i-1,j+1,k)
     4				+ oNW(i+1,j-1,k)* x(i+1,j-1,k)
c
				tmpU =
     1				+ uC(i,j,k)	* x(i,j,k+1)
     2				+ uN(i,j,k)	* x(i,j+1,k+1)
     3				+ uS(i,j,k)	* x(i,j-1,k+1)
     4				+ uE(i,j,k)	* x(i+1,j,k+1)
     3				+ uW(i,j,k)	* x(i-1,j,k+1)
     4				+ uNE(i,j,k)	* x(i+1,j+1,k+1)
     3				+ uNW(i,j,k)	* x(i-1,j+1,k+1)
     4				+ uSE(i,j,k)	* x(i+1,j-1,k+1)
     4				+ uSW(i,j,k)	* x(i-1,j-1,k+1)
c
				tmpD =
     1				+ uC(i,j,k-1)	* x(i,j,k-1)
     2				+ uS(i,j+1,k-1)	* x(i,j+1,k-1)
     3				+ uN(i,j-1,k-1)	* x(i,j-1,k-1)
     4				+ uW(i+1,j,k-1)	* x(i+1,j,k-1)
     3				+ uE(i-1,j,k-1)	* x(i-1,j,k-1)
     4				+ uSW(i+1,j+1,k-1)* x(i+1,j+1,k-1)
     3				+ uSE(i-1,j+1,k-1)* x(i-1,j+1,k-1)
     4				+ uNW(i+1,j-1,k-1)* x(i+1,j-1,k-1)
     4				+ uNE(i-1,j-1,k-1)* x(i-1,j-1,k-1)
c
				r(i,j,k) = fc(i,j,k)+tmp0 + tmpU + tmpD
     1				- oC(i,j,k)*x(i,j,k) -w1(i,j,k)
c
100			continue
200		continue
300	continue
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	C_vec_yukawa.f
c       ((non linear) function in Yukawa function
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      8/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	Defines the non linear term for the Yukawa equation
c
c	Input parameters:
c
c		u:	dimensionless Yukawa potential (in)
c		phi_ion Ion contribution (independent of u)
c		sinhu   Electric field information (independent of u)
c
c	Output:
c
c		c_nl: value of non linear term in PBE
c
	subroutine c_vec_yukawa(grid,phi_ion,sinhu,u,c_nl,nx,ny,nz,
     1				hx,hy,hz)
c
	include 'poisson.h'
c
c ==========================================================================================
c == Define all variables                  =================================================
c ==========================================================================================
c
c	Local variables
c===========================================================================================
c
	integer	i,j,k,i_ion
c
        real*8  val
	real*8	argument,argument0
	real*8	exp_min,exp_max
	parameter	(exp_min = -80.d0, exp_max = 80.d0)
c
c	Parameters (in and out)
c===========================================================================================
c
	integer nx,ny,nz
        integer info
c
        integer*1       grid(nx,ny,nz)
c
	real*8	v0,b,psi_bulk
	real*8	a_lattice,v_lattice
	real*8  l_ion(max_ion),l_dipole(max_solv)
	real*8	phi_ion(nx,ny,nz),sinhu(nx,ny,nz),u(nx,ny,nz)
	real*8	c_nl(nx,ny,nz)
c
c       Information on grid
c===========================================================================================
c
        real*8  x,y,z,dvol,g,d
        real*8  hx(nx),hy(ny),hz(nz)
c
	common /fugacities/	l_dipole,l_ion
	common /yukawa/		v0,b,psi_bulk
	common /lattice/	a_lattice,v_lattice
c
c ==========================================================================================
c == Compute non linear term               =================================================
c ==========================================================================================
c
	do 300 k = 1,nz
		do 200 j = 1,ny
			do 100 i = 1,nx
				c_nl(i,j,k) = 0.d0
100			continue
200		continue
300	continue
c
	do 600 k = 2,nz-1
        	z = hz(k+1)-hz(k-1)
		do 500 j = 2,ny-1
                	y = hy(j+1)-hy(j-1)
		   	do 400 i = 2,nx-1
c
                        	info = ibits(grid(i,j,k),0,2)
				if(info.eq.0) then
					c_nl(i,j,k) = 0.d0
					goto 400
				endif
                        	x = hx(i+1)-hx(i-1)
                        	dvol = x*y*z/8.d0
c
                        	argument = - u(i,j,k)+psi_bulk
				if(argument.gt.exp_max) argument=exp_max
				if(argument.lt.exp_min) argument=exp_min
c
				argument = exp(argument)
c
                        	val = l_dipole(1)*sinhu(i,j,k)*argument
                        	g = val*dvol/v_lattice
c
                        	d = 1 + val
                        	if(info.ne.1) then
                                	d = d + phi_ion(i,j,k)
                        	endif
c
                        	c_nl(i,j,k)=v0*g/d+(u(i,j,k)-psi_bulk)*
     1						dvol/(b*b)
c
400            		continue
500    		continue
600	continue
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c	DC_vec_yukawa.f
c       dervative of ((non linear) function in Yukawa equation
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      8/14/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	Defines the non linear term for the Yukawa equation
c
c	Input parameters:
c
c		u:	dimensionless Yukawa potential (in)
c		phi_ion Ion contribution (independent of u)
c		sinhu   Electric field information (independent of u)
c
c	Output:
c
c		dc_nl: value of derivatives of non linear term in Yukawa equation
c
	subroutine dc_vec_yukawa(grid,phi_ion,sinhu,u,dc_nl,nx,ny,nz,
     1				hx,hy,hz)
c
	include 'poisson.h'
c
c ==========================================================================================
c == Define all variables                  =================================================
c ==========================================================================================
c
c	Local variables
c===========================================================================================
c
	integer	i,j,k,i_ion
        integer info
c
        real*8  val
	real*8	argument,argument0
	real*8	exp_min,exp_max
	parameter	(exp_min = -80.d0, exp_max = 80.d0)
c
c	Parameters (in and out)
c===========================================================================================
c
	integer nx,ny,nz
c
        integer*1       grid(nx,ny,nz)
c
	real*8	v0,b,psi_bulk
	real*8	a_lattice,v_lattice
	real*8  l_ion(max_ion),l_dipole(max_solv)
	real*8	phi_ion(nx,ny,nz),sinhu(nx,ny,nz),u(nx,ny,nz)
	real*8	dc_nl(nx,ny,nz)
c
c       Information on grid
c===========================================================================================
c
        real*8  x,y,z,dvol,g,d,dg,dd
        real*8  hx(nx),hy(ny),hz(nz)
c
	common /fugacities/	l_dipole,l_ion
	common /yukawa/		v0,b
	common /lattice/	a_lattice,v_lattice
c
c ==========================================================================================
c == Compute non linear term               =================================================
c ==========================================================================================
c
	do 300 k = 1,nz
		do 200 j = 1,ny
			do 100 i = 1,nx
				dc_nl(i,j,k) = 0.d0
100			continue
200		continue
300	continue
c
	do 600 k = 2,nz-1
        	z = hz(k+1)-hz(k-1)
		do 500 j = 2,ny-1
                	y = hy(j+1)-hy(j-1)
		   	do 400 i = 2,nx-1
c
                        	info = ibits(grid(i,j,k),0,2)
				if(info.eq.0) then
					dc_nl(i,j,k) = 0.d0
					goto 400
				endif
                        	x = hx(i+1)-hx(i-1)
                        	dvol = x*y*z/8.d0
c
                        	argument = - u(i,j,k) + psi_bulk
				if(argument.gt.exp_max) argument=exp_max
				if(argument.lt.exp_min) argument=exp_min
c
				argument = exp(argument)
c
                        	val = l_dipole(1)*sinhu(i,j,k)*argument
                        	g = val*dvol/v_lattice
				dg = -g
c
                        	d = 1 + val
                        	if(info.ne.1) then
                                	d = d + phi_ion(i,j,k)
                        	endif
				dd = -val
c
                        	dc_nl(i,j,k) = v0*(dg*d-g*dd)/(d*d) +
     1                          dvol/(b*b)
c
400            		continue
500    		continue
600	continue
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c  	Vectcopy_all.f	
c
c	This subroutine copies a vector into another 
c ==========================================================================================
c ==========================================================================================
c
	subroutine vectcopy_all(nx,ny,nz,x,y)
c
c ==========================================================================================
c	Declaration
c ==========================================================================================
c
	integer	nx,ny,nz,i,j,k
c
	real*8	x(nx,ny,nz),y(nx,ny,nz)
c
c ==========================================================================================
c	copy vector
c ==========================================================================================
c
	do 300 k = 1,nz
		do 200 j = 1,ny
			do 100 i = 1,nx
				y(i,j,k) = x(i,j,k)
100			continue
200		continue
300	continue
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c  	xnorm1_diff.f	
c
c	This subroutine computes the 1-norm of the difference of two
c	vectors on a grid (excluding boundaries)
c ==========================================================================================
c ==========================================================================================
c
	function xnorm1_diff(nx,ny,nz,grid,x,y)
c
c ==========================================================================================
c	Declaration
c ==========================================================================================
c
	integer	nx,ny,nz,i,j,k,info
c
        integer*1       grid(nx,ny,nz)
c
	real*8	xnorm1_diff,xnorm1
	real*8	x(nx,ny,nz),y(nx,ny,nz)
c
c ==========================================================================================
c	Compute 1-norm
c ==========================================================================================
c
	xnorm1_diff = 0
        xnorm1      = 0
	do 300 k = 2,nz-1
		do 200 j = 2,ny-1
			do 100 i = 2,nx-1
                                info=ibits(grid(i,j,k),0,1)
                                if(info.ne.0) then
				        xnorm1_diff = xnorm1_diff + 
     1					dabs(x(i,j,k)-y(i,j,k))
				        xnorm1 = xnorm1 + 
     1					dabs(x(i,j,k))
                                endif
100			continue
200		continue
300	continue
c
        xnorm1_diff = xnorm1_diff/xnorm1
c
	return
	end
c
c ==========================================================================================
c ==========================================================================================
c  	xnorm2_diff.f	
c
c	This subroutine computes the 2-norm of the difference of two
c	vectors on a grid (excluding boundaries)
c ==========================================================================================
c ==========================================================================================
c
	function xnorm2_diff(nx,ny,nz,grid,x,y)
c
c ==========================================================================================
c	Declaration
c ==========================================================================================
c
	integer	nx,ny,nz,i,j,k
        integer info,nval
c
        integer*1 grid(nx,ny,nz)
	real*8	xnorm2_diff,dif
	real*8	x(nx,ny,nz),y(nx,ny,nz)
c
c ==========================================================================================
c	Compute 2-norm
c ==========================================================================================
c
	xnorm2_diff = 0
        nval = 0
	do 300 k = 2,nz-1
		do 200 j = 2,ny-1
			do 100 i = 2,nx-1
                                info=ibits(grid(i,j,k),0,1)
                                if(info.ne.0) then
                                        nval = nval + 1
				        dif = x(i,j,k)-y(i,j,k)
				        xnorm2_diff=xnorm2_diff+dif*dif
                                endif
100			continue
200		continue
300	continue
c
	xnorm2_diff = dsqrt(xnorm2_diff/nval)
c
	return
	end
c ==========================================================================================
c ==========================================================================================
c  	compute_ion_phi.f	
c
c       When solving the Yukawa equation, there is a term that depends on the
c       ions and electrostatic potential. Since this term is independent of
c       psi, we can pre-compute it!
c
c ==========================================================================================
c ==========================================================================================
c
	subroutine compute_ion_phi(nx,ny,nz,grid,phi,ion_phi)
c
c ==========================================================================================
c	Declaration
c ==========================================================================================
c
        include 'poisson.h'
c
        integer i,j,k,nx,ny,nz
        integer i_ion,info
        integer n_ion,n_zz
c
        integer*1 grid(nx,ny,nz)
c
        real*8  d,arg,exp_min,exp_max
	real*8  l_ion(max_ion),l_dipole(max_solv)
        real*8  ion_conc(max_ion),ion_z(max_ion),ion_rad(max_ion)
        real*8  kappa(max_ion)
c
        real*8  phi(nx,ny,nz),ion_phi(nx,ny,nz)
c
        parameter       (exp_min = -80.d0, exp_max=80.d0)
c
c       Common blocks
c ==========================================================================================
c
	common /fugacities/	l_dipole,l_ion
        common /ions/           ion_conc,ion_z,ion_rad,kappa,n_ion,n_zz
c
        call set_all_zeros(nx,ny,nz,ion_phi)
c
        do 400 k = 2,nz-1
            do 300 j = 2,ny-1
                do 200 i = 2,nx-1
c
                        info = ibits(grid(i,j,k),1,1)
                        if(info.ne.0) then
                            d = 0
                            do 100 i_ion = 1, n_ion
c
                                arg = -ion_z(i_ion)*phi(i,j,k)
                                if(arg.gt.exp_max) arg = exp_max
                                if(arg.lt.exp_min) arg = exp_min
                                arg = exp(arg)
                                d = d + l_ion(i_ion)*arg
c
100                         continue
                            ion_phi(i,j,k) = d
                        endif
c
200             continue
300         continue
400     continue
c
        return
        end
c ==========================================================================================
c ==========================================================================================
c  	compute_sinhu_phi.f	
c
c       When solving the Yukawa equation, there is a term that depends on the
c       dipoles and electric field. Since this term is independent of
c       psi, we can pre-compute it!
c
c ==========================================================================================
c ==========================================================================================
c
	subroutine compute_sinhu_phi(nx,ny,nz,hx,hy,hz,grid,phi,
     1                  sinhu_phi,vect2D)
c
c ==========================================================================================
c	Declaration
c ==========================================================================================
c
        integer i,j,k,nx,ny,nz
        integer i_ion,info
        integer n_ion,n_zz
c
        integer*1 grid(nx,ny,nz)
c
        real*8  u_val,F1_u,sinhc_u,exp_mu
        real*8  p0_e,vol,coef_l
        real*8  vect2D(nx,ny)
        real*8  hx(nx),hy(ny),hz(nz)
c
        real*8  phi(nx,ny,nz),sinhu_phi(nx,ny,nz)
c
        common /dipoles/        vol,p0_e,coef_l
c
c ==========================================================================================
c       Compute sinh_u/u, where u is p0*electron*Efield
c ==========================================================================================
c
        call set_all_zeros(nx,ny,nz,sinhu_phi)
c
        do 300 k = 2,nz-1
                call plane_efield(phi(1,1,k),phi(1,1,k-1),phi(1,1,k+1),
     1                  nx,ny,hx,hy,hz,k,vect2D)
                do 200 j = 2,ny-1
                        do 100 i = 2,nx-1
c
                                info = ibits(grid(i,j,k),0,1)
c
                                if(info.ne.0) then
                                        u_val = p0_e * vect2D(i,j)
                                        call compute_F1(u_val,F1_u,
     1                                  sinhc_u,exp_mu)
                                        sinhu_phi(i,j,k)=sinhc_u/exp_mu
                                endif
c
100                     continue
200             continue
300     continue
c
        return
        end
c ==========================================================================================
c ==========================================================================================
c	set_boundary_yukawa.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      8/13/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Set boundary conditions on the grid for the PSI potential of the Yukawa
c       equation
c
c	Output:
c		- bndx	: value of potential for ix=1 and ix = Npointx
c		- bndy	: value of potential for iy=1 and iy = Npointy
c		- bndz	: value of potential for iz=1 and iz = Npointz
c
	subroutine set_boundary_yukawa(psi_bulk,bndx,bndy,bndz,nx,ny,nz)
c
	include 'poisson.h'
c
c ==========================================================================================
c == Define all variables                        ===========================================
c ==========================================================================================
c	
c	Local variables
c ==========================================================================================
c
	integer	i,j,k
	integer	nx,ny,nz
c
        real*8  psi_bulk
c
c	Boundary values
c ==========================================================================================
c
	real*8	bndx(ny,nz,*),bndy(nx,nz,*)
	real*8	bndz(nx,ny,*)
c
c ==========================================================================================
c == Generate boundary conditions                ===========================================
c ==========================================================================================
c
c	Initialize boundary potential values to 0
c
c	bndx(j,k,1) is the value at grid point 1,j,k
c	bndx(j,k,2) is the value at grid point Npointx,j,k
c
c
	do 300 k=1,nz
		do 200 j = 1,ny
			bndx(j,k,1) = psi_bulk
			bndx(j,k,2) = psi_bulk
200		continue
300	continue
c
c	bndy(i,k,1) is the value at grid point i,1,k
c	bndy(i,k,2) is the value at grid point i,Npointy,k
c
	do 500 k=1,nz
		do 400 i = 1,nx
			bndy(i,k,1) = psi_bulk
			bndy(i,k,2) = psi_bulk
400		continue
500	continue
c
c	bndz(i,j,1) is the value at grid point i,j,1
c	bndz(i,j,2) is the value at grid point i,j,Npointz
c
	do 700 j=1,ny
		do 600 i = 1,nx
			bndz(i,j,1) = psi_bulk
			bndz(i,j,2) = psi_bulk
600             continue
700     continue
c
        return
        end
c ==========================================================================================
c ==========================================================================================
c	get_bulk_yukawa.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      8/13/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Get boundary value on the grid for the PSI potential of the Yukawa
c       equation
c
c	Output:
c               psi_bulk
c
	subroutine get_boundary_yukawa(l_log)
c
	include 'poisson.h'
c
c ==========================================================================================
c == Define all variables                        ===========================================
c ==========================================================================================
c	
c	Local variables
c ==========================================================================================
c
	integer	i,j,k
	integer	nx,ny,nz
        integer l_log
c
        real*8  accuracy
        real*8  fac,eps_bulk,get_psi_bulk
        real*8  eps_kirkwood,eps_onsager
        real*8  vn2,b,delta,eps_bulk_0
c
c	Constants
c ==========================================================================================
c
	real*8	l_Bjerrum,pi,Temp,KB,electron,Ang,beta,Avogadro,eps0
	real*8	C_charge,C_prop
        real*8  vol_max,p0_e,coef_l
        real*8  v0,b_yu,psi_bulk,test,test2
c
c       Information on solvent
c ==========================================================================================
c
        integer n_solv
        real*8  solv_conc(max_solv),solv_dipole(max_solv)
        real*8  solv_cvol(max_solv)
        real*8  l_dipole(max_solv),l_ion(max_ion)
c
c ==========================================================================================
c == All common blocks (common storage)          ===========================================
c ==========================================================================================
c	
	common /constants/ 	l_Bjerrum,pi,Temp,KB,electron,Ang,
     1			   	Avogadro,eps0,beta,C_charge,C_prop
        common /solvent/        solv_conc,solv_dipole,solv_cvol,n_solv
        common /dipoles/        vol_max,p0_e,coef_l
        common /fugacities/     l_dipole,l_ion
        common /yukawa/         v0,b_yu,psi_bulk
        common /bulk/           eps_bulk
c
c ==========================================================================================
c == Formats for log file                        ===========================================
c ==========================================================================================
c	
1       format(70('='),/,2('='),'Dielectric permittivity (bulk)',36x,
     1  2('='),/,70('='),/)
2       format('eps_bulk (Onsager)                     : ',f10.3)
3       format('eps_bulk (Kirkwood)                    : ',f10.3)
4       format('eps_bulk (GPBL)                        : ',f10.3)
5       format('eps_bulk (GPBL+Yukawa)                 : ',f10.3)
c
c ==========================================================================================
c       Get psi value for bulk
c ==========================================================================================
c
        psi_bulk = - solv_conc(1)*Avogadro*1.d-27*v0*b_yu*b_yu
c
c ==========================================================================================
c       Compute eps_bulk (and compare with Kirkwood and Onsager estimates)
c ==========================================================================================
c
        fac = beta*electron*electron*p0_e*p0_e*Avogadro*solv_conc(1)
     1          *1.d-17/eps0
        eps_bulk_0 = 1+fac/3
        eps_bulk = 1 + fac*exp(-psi_bulk)/3
        eps_kirkwood = (1+fac+sqrt((1+fac)**2+8))/4
        vn2 = 4./3.
        b = 2*vn2 - 1 + fac*((vn2+2)**2)/9
        delta = b**2 + 8*vn2
        eps_onsager = (b + sqrt(delta))/4
c
c        write(6,*) ' '
c        write(6,*) 'eps_bulk (Onsager)    : ',eps_onsager
c        write(6,*) 'eps_bulk (Kirkwood)   : ',eps_kirkwood
c        write(6,*) 'eps_bulk (no Yukawa)  : ',eps_bulk_0
c        write(6,*) 'eps_bulk (this model) : ',eps_bulk
c        write(6,*) ' '
c
c        write(l_log,1)
c        write(l_log,2) eps_onsager
c        write(l_log,3) eps_kirkwood
c        write(l_log,4) eps_bulk_0
c        write(l_log,5) eps_bulk
c        write(l_log,*) ' '
c
        return
        end
c ==========================================================================================
c ==========================================================================================
c	Set_psi_bulk.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	7/05/07
c	Author:		Patrice Koehl 
c			in collaboration with Marc Delarue
c
        subroutine set_psi_bulk(nx,ny,nz,psi_bulk,grid,psi)
c
        integer nx,ny,nz
        integer i,j,k
        integer info
c
        integer*1       grid(nx,ny,nz)
c
        real*8  psi_bulk
        real*8  psi(nx,ny,nz)
c
        do 300 k = 1,nz
                do 200 j = 1,ny
                        do 100 i = 1,nx
                                info=ibits(grid(i,j,k),0,1)
                                if(info.eq.0) then
                                        psi(i,j,k) = 0.d0
                                else
                                        psi(i,j,k) = psi_bulk
                                endif
100                     continue
200             continue
300     continue
c
        return
        end
c ==========================================================================================
c ==========================================================================================
c	set_epsilon_yukawa.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      7/13/07
c	Version 2:	8/28/07: uses GenSurf to generate epsilon map
c
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Set epsilon values on the grid: 
c	The surface of the protein separates two regions: the interior,
c	which is then set with epsilon = epsilon1, and the exterior where
c	the solvent is, set with epsilon = epsilon2
c
c	Output:
c		epsilon_X	: epsilon values on the grid (half point on along X axis
c		epsilon_Y	: epsilon values on the grid (half point on along Y axis
c		epsilon_Z	: epsilon values on the grid (half point on along Z axis
c
c
	subroutine set_epsilon_yukawa(epsilon_X,epsilon_Y,epsilon_Z,
     1          grid_shift,nx,ny,nz)
c
	include 'poisson.h'
c
c ==========================================================================================
c == Define all variables                        ===========================================
c ==========================================================================================
c	
c	Local variables
c ==========================================================================================
c
	integer	i,j,k
	integer	ipos,nval
	integer	nx,ny,nz
c
	real*8	eps_val(3)
c
c	Output
c ==========================================================================================
c
	real*8	epsilon_X(nx,ny,nz)
	real*8	epsilon_Y(nx,ny,nz)
	real*8	epsilon_Z(nx,ny,nz)
c
c	Information on shifted grids
c ==========================================================================================
c
	integer*1 grid_shift(nx,ny,nz)
c
	eps_val(1) = 1.d0
	eps_val(2) = 1.d0
	eps_val(3) = 1.d0
c
c	Set initial epsilon map based on molecule boundaries
c
c	X-shifted map
c
	do 800 k = 1,nz
	   do 700 j = 1,ny
		do 600 i = 1,nx
			ipos=ibits(grid_shift(i,j,k),2,1)+1
			epsilon_X(i,j,k) = eps_val(ipos)
600		continue
700	   continue
800	continue
c
c	Repeat for Y-shifted map
c
	do 1300 k = 1,nz
	   do 1200 j = 1,ny
		do 1100 i = 1,nx
			ipos=ibits(grid_shift(i,j,k),4,1)+1
			epsilon_Y(i,j,k) = eps_val(ipos)
1100		continue
1200	   continue
1300	continue
c
c	Repeat for Z-shifted map
c
	do 1700 k = 1,nz
	   do 1600 j = 1,ny
		do 1500 i = 1,nx
			ipos=ibits(grid_shift(i,j,k),6,1)+1
			epsilon_Z(i,j,k) = eps_val(ipos)
1500		continue
1600	   continue
1700	continue
c
        return
        end
c
        subroutine printline(nx,ny,nz,u)
c
        integer i,j,k
        integer nx,ny,nz
c
        real*8  u(nx,ny,nz)
c
        do 100 i = 1,nx
                write(6,*) 'i,psi(i,96,96) :',i,u(i,96,96)
100     continue
c
        return
        end
