c       Copyright (C) 2007 Patrice Koehl
c
c       This library is free software; you can redistribute it and/or
c       modify it under the terms of the GNU Lesser General Public
c       License as published by the Free Software Foundation; either
c       version 2.1 of the License, or (at your option) any later version.
c
c       This library is distributed in the hope that it will be useful,
c       but WITHOUT ANY WARRANTY; without even the implied warranty of
c       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
c       Lesser General Public License for more details.
c
c       You should have received a copy of the GNU Lesser General Public
c       License along with this library; if not, write to the Free Software
c       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
c
c ==========================================================================================
c ==========================================================================================
c	Solve_eq2.f
c ==========================================================================================
c ==========================================================================================
c
c       Version 1:      7/6/07
c       Author:         Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose:	provide a wrapper to Michael Holst PDE solvers
c
c	This is the wrapper to solve equation 2 in the YULP system of equations
c
	subroutine solve_eq2(grid_shift,eps_X,eps_Y,eps_Z,
     1		scalar_map,crg_map,gxcf,gycf,gzcf,xf,yf,zf,
     2		nx,ny,nz,u,rwork,iwork,rparam,iparam,sinhu)
c
	include 'poisson.h'
c
c ==========================================================================================
c == Define all variables                        ===========================================
c ==========================================================================================
c	
c	Local variables
c ==========================================================================================
c
	integer	i
	integer	nx,ny,nz
c
c	Variables for Michael Holst MG package
c ==========================================================================================
c
	integer	iparam(100),iwork(*)
c
	real*8	rwork(*)
	real*8	rparam(100)
c
	real*8	xf(*),yf(*),zf(*)
	real*8	gxcf(*),gycf(*),gzcf(*)
c
	integer*1 grid_shift(*)
c
c	Information on molecule
c ==========================================================================================
c
	integer	natom,ncharge,iform
c
	real*8	atom_coord(*)
	real*8	atom_radius(*),atom_charge(*)
c
	character atom_label(*)*30
	character atom_label2(*)*12
c
c	Information on maps over the grid:
c ==========================================================================================
c
c	phi_map:	electrostatics potential
c	crg_map:	charge map
c
	real*8	u(*)
	real*8	crg_map(*)
	real*8	scalar_map(*)
        real*8  sinhu(*)
c
c	epsilon_X defines the permittivity on the edges along the X axis
c	epsilon_Y defines the permittivity on the edges along the Y axis
c	epsilon_Z defines the permittivity on the edges along the Z axis
c
	real*8	eps_X(*),eps_Y(*),eps_Z(*)
c
c	Pointers
c ==========================================================================================
c
        pointer (ptr_coord,atom_coord)
        pointer (ptr_charge,atom_charge)
        pointer (ptr_radius,atom_radius)
        pointer (ptr_label,atom_label)
        pointer (ptr_label2,atom_label2)
c
c	Common blocks
c ==========================================================================================
c
        common /charges/        ptr_coord,ptr_charge,ptr_radius,
     1                          ptr_label,ptr_label2,iform,
     2                          natom,ncharge
c
	call Newton_Solver_yukawa(iparam,rparam,iwork,rwork,
     1		u,xf,yf,zf,gxcf,gycf,gzcf,eps_X,eps_Y,
     2		eps_Z,scalar_map,crg_map,sinhu,grid_shift)
c
	return
	end
