c       Copyright (C) 2007 Patrice Koehl
c
c       This library is free software; you can redistribute it and/or
c       modify it under the terms of the GNU Lesser General Public
c       License as published by the Free Software Foundation; either
c       version 2.1 of the License, or (at your option) any later version.
c
c       This library is distributed in the hope that it will be useful,
c       but WITHOUT ANY WARRANTY; without even the implied warranty of
c       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
c       Lesser General Public License for more details.
c
c       You should have received a copy of the GNU Lesser General Public
c       License along with this library; if not, write to the Free Software
c       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
c
c ==========================================================================================
c ==========================================================================================
c	Solve_sc.f
c ==========================================================================================
c ==========================================================================================
c
c	Version 1:	9/11/09
c	Author:		Patrice Koehl, in collaboration with Marc Delarue
c
c	Purpose: the is a generalized Poisson Boltzmann solver, using generic
c	PB solver with self consistent approach
c
        subroutine solve_sc(grid_shift,eps_X,eps_Y,eps_Z,helmoltz_map,
     1                          crg_map,gxcf,gycf,gzcf,xf,yf,zf,
     2                          nx,ny,nz,u,rwork,iwork,rparam,iparam,
     3                          hx,hy,hz,resid,u_old,space,free_density,
     4				free_charge,free_langevin)
c
c ==========================================================================================
c = poisson.h defines array sizes for AquaSol    ===========================================
c ==========================================================================================
c
	include 'poisson.h'
c
c ==========================================================================================
c = Declares all local/global arrays          ==============================================
c ==========================================================================================
c
c	Local variables
c ==========================================================================================
c
	integer	i,j
	integer	nx,ny,nz,flag
        integer nnode,nfudge,nsize,nsize_byte
        integer n_iter,niter_max,itmax
        integer method,equation,ndiscrete,nlevel,nlin,ijac,nsmooth
        integer zero,rwork_size2,imethod,info
        integer flag_free
c
        real*8  res,res_c,resnum,resnum_old,resden,rmem,tol,lambda
	real*8	xnorm1,val,free_charge,free_density,free_langevin
	real*8	free_ene_kT,free_ene
	real*8	l_Bjerrum,pi,Temp,KB,electron,Ang,Avogadro
	real*8	eps0,beta,C_charge,C_prop,fac
c
c	Information on molecule
c ==========================================================================================
c
        integer natom,ncharge,iform
c
        real*8  atom_coord(*),atom_radius(*)
        real*8  atom_charge(*)
c
        character  atom_label(*)*30,atom_label2(*)*12
c
c	Information on maps over the grid:
c ==========================================================================================
c
c	eps_X,eps_Y,eps_Z:	dielectric constants
c	crg_map:		charge map
c	gxcf,gycf,gzcf:		boundary conditions
c	helmoltz_map:		region in which helmoltz term is non zero
c
	real*8	eps_X(*)
	real*8	eps_Y(*)
	real*8	eps_Z(*)
	real*8	crg_map(*)
	real*8	helmoltz_map(*)
	real*8	gxcf(*)
	real*8	gycf(*)
	real*8	gzcf(*)
c
        real*8  resid(*)
        real*8  space(*)
        real*8  u_old(*)
c
        real*8  rwork(*)
        real*8  iwork(*)
c
c       solution of the equation
c ==========================================================================================
c
	real*8	u(*)
c
c	Information on grid
c ==========================================================================================
c
	real*8	hx(*),hy(*),hz(*)
        real*8  xf(*),yf(*),zf(*)
c
	integer*1 grid_shift(*)
c
        integer iparam(100)
        real*8  rparam(100)
c
c	Pointers
c ==========================================================================================
c
        pointer (ptr_coord,atom_coord)
        pointer (ptr_charge,atom_charge)
        pointer (ptr_radius,atom_radius)
        pointer (ptr_label,atom_label)
        pointer (ptr_label2,atom_label2)
c
c	Common blocks
c ==========================================================================================
c
        common /charges/        ptr_coord,ptr_charge,ptr_radius,
     1                          ptr_label,ptr_label2,iform,
     2                          natom,ncharge
        common /pdesolver/      tol,method,nlevel,nlin,equation,
     1                          ndiscrete,ijac,nsmooth
	common /residue/	resden,resnum_old
	common /constants/	l_Bjerrum,pi,Temp,KB,electron,Ang,
     1				Avogadro,eps0,beta,C_charge,C_prop
c
        niter_max = 100
        lambda    = 1.0
	equation  = 2
c
	fac       = 4.18/(KB*Temp*Avogadro*(1.d-3))
        flag_free = 0
        free_ene  = 0.d0
c
        itmax = 1
        info  = 0
c
        n_iter = 0
        call print_ini_free
        call print_residual_free(1,-99,resnum,resden,resnum_old,
     1          free_ene)
c
200     continue
c
c               Set Helmholtz map
c
                call set_helmoltz(helmoltz_map,grid_shift,
     1                  nx,ny,nz)
c
c               Set boundaries
c
                call set_boundary(gxcf,gycf,gzcf,hx,hy,hz,
     1                  nx,ny,nz)
c
c		Re-define epsilon maps, charge distrib
c		compute resid and update Helmholtz term
c
                call charge_distrib(crg_map,grid_shift,hx,
     1          hy,hz,nx,ny,nz)
c
                call langevin_new_eps(u,nx,ny,nz,hx,hy,hz,
     1          grid_shift,space,eps_X,eps_Y,eps_Z)
c
               	call langevin_resid(u,nx,ny,nz,hx,hy,hz,
     1          space,crg_map,helmoltz_map,resid,eps_X,eps_Y,eps_Z,
     2          free_ene,flag_free,grid_shift)
c
		call energy_fixedcharges(u,crg_map,free_charge,
     1		hx,hy,hz,nx,ny,nz)
		call energy_elecdensity0(u,hx,hy,hz,nx,ny,nz,
     1		free_density)
		call langevin_free_energy(u,free_langevin,grid_shift,
     1		hx,hy,hz,nx,ny,nz)
c
		free_ene = free_charge + free_density + free_langevin
c
c		Convert in kT:
c
		free_ene_kT = free_ene *fac
c
c               Compute residual and update Helmholtz term
c
                resnum = xnorm1(nx,ny,nz,resid)
c
                if(n_iter.eq.0) then
                        resden = resnum
                        resnum_old = resnum
                endif
c
                res = resnum/resden
                res_c = resnum/resnum_old
c
                call print_residual_free(1,n_iter,resnum,
     1                  resden,resnum_old,free_ene_kT)
c
                resnum_old = resnum
c
                if(res.lt.tol.or.n_iter.gt.niter_max) goto 400
c
	        call solve_pde(grid_shift,eps_X,eps_Y,eps_Z,
     1                  helmoltz_map,crg_map,gxcf,gycf,gzcf,
     2                  xf,yf,zf,nx,ny,nz,u,rwork,iwork,
     3                  rparam,iparam)
c
c               Update solution
c
                do 300 i = 1,nx*ny*nz
                        u(i) = lambda*u(i) + (1-lambda)*u_old(i)
                        u_old(i) = u(i)
300             continue
c
                n_iter = n_iter + 1
c
        goto 200
c
400     continue
c
        return
        end
